----------------------------------------------------------------------------------
-- Company: 		NTNU
-- Engineer: 		Kim Troennes
-- 
-- Create Date:	02:50:16 03/13/2014 
-- Design Name:	intra.vhd
-- Module Name:	intrawrap - Behavioral 
-- Project Name:  intra8x8final
-- Description:	8x8 luminance intra prediction for H.264 encoder
--						Computes neighbouring pixels for prediction module
----------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity intrawrap is
	generic(
		num : integer := 7;
		--measured in number of 8x8 blocks:
		frame_width: integer := 2;
		frame_height: integer := 2		
	);
	
	port(
		clk: in std_logic;
		rst: in std_logic;
		
		-- Original pixels, 4x4 matrix
		ycbcr0	    :	in std_logic_vector(num downto 0);	 
		ycbcr1 		:	in std_logic_vector(num downto 0);
		ycbcr2  	:	in std_logic_vector(num downto 0);
		ycbcr3		:	in std_logic_vector(num downto 0);
		ycbcr4		:	in std_logic_vector(num downto 0);
		ycbcr5		:	in std_logic_vector(num downto 0);
		ycbcr6		:	in std_logic_vector(num downto 0);
		ycbcr7		:	in std_logic_vector(num downto 0);
		ycbcr8		:	in std_logic_vector(num downto 0);
		ycbcr9		:	in std_logic_vector(num downto 0);
		ycbcr10		:	in std_logic_vector(num downto 0);
		ycbcr11		:	in std_logic_vector(num downto 0);
		ycbcr12		:	in std_logic_vector(num downto 0);			 
		ycbcr13		:	in std_logic_vector(num downto 0);
		ycbcr14		:	in std_logic_vector(num downto 0);
		ycbcr15		:	in std_logic_vector(num downto 0);
		
		-- Ready signals for pipeline support
		-- only rdy1 and idct is used for now
		rdy_idct   : in std_logic;	
		rdy1_in : in std_logic;	
--		rdy2_in : in std_logic;	
--		rdy3_in  : in std_logic;	
--		rdy4_in  : in std_logic;	
		
		-- Reconstructed residuals from previous 8x8 frame, 2x8 matrix
		-- | X| X| X| X| X| X| X| 7| 
		-- | 8| 9|10|11|12|13|14|15|
		ycbcr7_pr  :	in std_logic_vector(num+1 downto 0);
		ycbcr8_pr   :	in std_logic_vector(num+1 downto 0);
		ycbcr9_pr   :	in std_logic_vector(num+1 downto 0);
		ycbcr10_pr  :	in std_logic_vector(num+1 downto 0);
		ycbcr11_pr  :	in std_logic_vector(num+1 downto 0);
		ycbcr12_pr  :	in std_logic_vector(num+1 downto 0);
		ycbcr13_pr  :	in std_logic_vector(num+1 downto 0);		 
		ycbcr14_pr  :	in std_logic_vector(num+1 downto 0);
		ycbcr15_pr  :	in std_logic_vector(num+1 downto 0);	
		
		-- Residuals
		intra0  :	out std_logic_vector(num+1 downto 0);	  
		intra1  :	out std_logic_vector(num+1 downto 0);
		intra2  :	out std_logic_vector(num+1 downto 0);
		intra3  :	out std_logic_vector(num+1 downto 0);
		intra4  :	out std_logic_vector(num+1 downto 0);
		intra5  :	out std_logic_vector(num+1 downto 0);
		intra6  :	out std_logic_vector(num+1 downto 0);
		intra7  :	out std_logic_vector(num+1 downto 0);
		intra8  :	out std_logic_vector(num+1 downto 0);
		intra9  :	out std_logic_vector(num+1 downto 0);
		intra10  :	out std_logic_vector(num+1 downto 0);
		intra11  :	out std_logic_vector(num+1 downto 0);
		intra12  :	out std_logic_vector(num+1 downto 0);
		intra13  :	out std_logic_vector(num+1 downto 0);		 
		intra14  :	out std_logic_vector(num+1 downto 0);
		intra15  :	out std_logic_vector(num+1 downto 0);		   
		
		rdy_out   : out std_logic	 
	);

end intrawrap;

architecture Behavioral of intrawrap is

-- Input to pred module registers
signal ycbcr0_reg : std_logic_vector(num downto 0);
signal ycbcr1_reg : std_logic_vector(num downto 0);
signal ycbcr2_reg : std_logic_vector(num downto 0);
signal ycbcr3_reg : std_logic_vector(num downto 0);
signal ycbcr4_reg : std_logic_vector(num downto 0);
signal ycbcr5_reg : std_logic_vector(num downto 0);
signal ycbcr6_reg : std_logic_vector(num downto 0);
signal ycbcr7_reg : std_logic_vector(num downto 0);
signal ycbcr8_reg : std_logic_vector(num downto 0);
signal ycbcr9_reg : std_logic_vector(num downto 0);
signal ycbcr10_reg : std_logic_vector(num downto 0);
signal ycbcr11_reg : std_logic_vector(num downto 0);
signal ycbcr12_reg : std_logic_vector(num downto 0);
signal ycbcr13_reg : std_logic_vector(num downto 0);
signal ycbcr14_reg : std_logic_vector(num downto 0);
signal ycbcr15_reg : std_logic_vector(num downto 0);

signal ycbcr0_reg2 : std_logic_vector(num downto 0);
signal ycbcr1_reg2 : std_logic_vector(num downto 0);
signal ycbcr2_reg2 : std_logic_vector(num downto 0);
signal ycbcr3_reg2 : std_logic_vector(num downto 0);
signal ycbcr4_reg2 : std_logic_vector(num downto 0);
signal ycbcr5_reg2 : std_logic_vector(num downto 0);
signal ycbcr6_reg2 : std_logic_vector(num downto 0);
signal ycbcr7_reg2 : std_logic_vector(num downto 0);
signal ycbcr8_reg2 : std_logic_vector(num downto 0);
signal ycbcr9_reg2 : std_logic_vector(num downto 0);
signal ycbcr10_reg2 : std_logic_vector(num downto 0);
signal ycbcr11_reg2 : std_logic_vector(num downto 0);
signal ycbcr12_reg2 : std_logic_vector(num downto 0);
signal ycbcr13_reg2 : std_logic_vector(num downto 0);
signal ycbcr14_reg2 : std_logic_vector(num downto 0);
signal ycbcr15_reg2 : std_logic_vector(num downto 0);

signal rdy_pr_in: 		std_logic;
signal rdy_pr_in_reg : 	std_logic;
--signal rdy_pr_in_reg2 : std_logic;
--signal rdy_pr_in2: 		std_logic;
--signal rdy_pr_in2_reg:  std_logic;
--signal rdy_pr_in2_reg2: std_logic;
signal first_row : 		std_logic;
signal first_row_reg : 	std_logic;
--signal first_row_reg2 : std_logic;
signal first_col : 		std_logic;
signal first_col_reg : 	std_logic;
--signal first_col_reg2 : std_logic;
signal next_quart	: 		std_logic_vector (1 downto 0);
signal quart: 				std_logic_vector(1 downto 0);
signal quart_reg : 		std_logic_vector(1 downto 0);
--signal quart_reg2 : 		std_logic_vector(1 downto 0);

signal quart_pr : std_logic_vector(1 downto 0);
signal next_quart_pr: std_logic_vector(1 downto 0);

signal rdy_update_az : std_logic;

-- Delta signal for rdy_in, only high 1 clock cycle after rdy_in
signal rdy_in_start: std_logic;
signal rdy_in_counter: integer range 0 to 3 :=3;

-- Which 8x8 block in the macroblock are we working on:
signal internal_count: integer range 0 to 3 := 3;
-- Number of rows/cols transferred
signal row_count: integer range 0 to frame_height-1 := frame_height-1;
signal col_count: integer range 0 to frame_width-1 := frame_width-1;
signal next_row: integer range 0 to frame_width-1 := 0;
signal next_col: integer range 0 to frame_width-1 := 0;	 

-- Registers for storage of left side neighbours
-- a column of 16 pixels to be sent to q,r,s,t,u,v,w,x vals
signal ycbcr0_left : std_logic_vector(num downto 0);
signal ycbcr1_left : std_logic_vector(num downto 0);
signal ycbcr2_left : std_logic_vector(num downto 0);
signal ycbcr3_left : std_logic_vector(num downto 0);
signal ycbcr4_left : std_logic_vector(num downto 0);
signal ycbcr5_left : std_logic_vector(num downto 0);
signal ycbcr6_left : std_logic_vector(num downto 0);
signal ycbcr7_left : std_logic_vector(num downto 0);
signal ycbcr8_left : std_logic_vector(num downto 0);
signal ycbcr9_left : std_logic_vector(num downto 0);
signal ycbcr10_left : std_logic_vector(num downto 0);
signal ycbcr11_left : std_logic_vector(num downto 0);
signal ycbcr12_left : std_logic_vector(num downto 0);
signal ycbcr13_left : std_logic_vector(num downto 0);
signal ycbcr14_left : std_logic_vector(num downto 0);
signal ycbcr15_left : std_logic_vector(num downto 0);

-- memory for storing uppper horizontal values
-- 0 to (framewidth-1)+1 to make i-p zeros at end of frame
type ram_horizontal is array (0 to frame_width) of std_logic_vector(63 downto 0);
signal ram_h: ram_horizontal := (others=> (others=>'0')); 

-- read from ram_h, and write to abcdefghijklmnop values (16*8 bits)
signal ycbcr_upper : std_logic_vector(127 downto 0); 

-- Z value registers, z# means for which internal count it is updated on
signal ycbcr_z0 : std_logic_vector(num downto 0);
signal ycbcr_z1 : std_logic_vector(num downto 0);
signal ycbcr_z2 : std_logic_vector(num downto 0);
signal ycbcr_z3 : std_logic_vector(num downto 0);

-- Stored/delayed intra prediction signals prediction module
-- 9*8 bits, only use ycbcr_prev[7..15]
type reg_type is array (7 to 15) of std_logic_vector(num downto 0); 
signal intra_reg1	: reg_type;
signal intra_reg2	: reg_type;
signal intra_reg3	: reg_type;
signal intra_reg4	: reg_type;
signal intra_reg5	: reg_type;
signal intra_reg6	: reg_type;

-- Previous pixels = intra prediction + reconstructed residual
signal ycbcr7_prev  : std_logic_vector(num downto 0);
signal ycbcr8_prev  : std_logic_vector(num downto 0);
signal ycbcr9_prev  : std_logic_vector(num downto 0);
signal ycbcr10_prev : std_logic_vector(num downto 0);
signal ycbcr11_prev : std_logic_vector(num downto 0);
signal ycbcr12_prev : std_logic_vector(num downto 0);
signal ycbcr13_prev : std_logic_vector(num downto 0);		 
signal ycbcr14_prev : std_logic_vector(num downto 0);
signal ycbcr15_prev : std_logic_vector(num downto 0);

signal ycbcr7_prev_extra  : std_logic_vector(num+1 downto 0);
signal ycbcr8_prev_extra  : std_logic_vector(num+1 downto 0);
signal ycbcr9_prev_extra  : std_logic_vector(num+1 downto 0);
signal ycbcr10_prev_extra : std_logic_vector(num+1 downto 0);
signal ycbcr11_prev_extra : std_logic_vector(num+1 downto 0);
signal ycbcr12_prev_extra : std_logic_vector(num+1 downto 0);
signal ycbcr13_prev_extra : std_logic_vector(num+1 downto 0);		 
signal ycbcr14_prev_extra : std_logic_vector(num+1 downto 0);
signal ycbcr15_prev_extra : std_logic_vector(num+1 downto 0);

-- Output from prediction module: predicted pixels
signal intra7_pr1   :	 std_logic_vector(num downto 0);  
signal intra8_pr1   :	 std_logic_vector(num downto 0);  
signal intra9_pr1   :	 std_logic_vector(num downto 0);  
signal intra10_pr1  :	 std_logic_vector(num downto 0);  
signal intra11_pr1  :	 std_logic_vector(num downto 0);  
signal intra12_pr1  :	 std_logic_vector(num downto 0);  
signal intra13_pr1  :	 std_logic_vector(num downto 0);	
signal intra14_pr1  :	 std_logic_vector(num downto 0);  
signal intra15_pr1  :	 std_logic_vector(num downto 0);

-- Output from prediction module: residuals
signal intra0_res  :	std_logic_vector(num+1 downto 0);	
signal intra1_res  :	std_logic_vector(num+1 downto 0);   
signal intra2_res  :	std_logic_vector(num+1 downto 0);   
signal intra3_res  :	std_logic_vector(num+1 downto 0);   
signal intra4_res  :	std_logic_vector(num+1 downto 0);   
signal intra5_res  :	std_logic_vector(num+1 downto 0);   
signal intra6_res  :	std_logic_vector(num+1 downto 0);   
signal intra7_res  :	std_logic_vector(num+1 downto 0);   
signal intra8_res  :	std_logic_vector(num+1 downto 0);   
signal intra9_res  :	std_logic_vector(num+1 downto 0);   
signal intra10_res :	std_logic_vector(num+1 downto 0);   
signal intra11_res :	std_logic_vector(num+1 downto 0);   
signal intra12_res :	std_logic_vector(num+1 downto 0);   
signal intra13_res :	std_logic_vector(num+1 downto 0);	
signal intra14_res :	std_logic_vector(num+1 downto 0);   
signal intra15_res :	std_logic_vector(num+1 downto 0);

signal rdy_pr_out : std_logic;			   
signal rdy_out_reg: std_logic;	

-- A-Z values
signal a_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal b_val :	std_logic_vector(num downto 0):=(others=> '0');
signal c_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal d_val :	std_logic_vector(num downto 0):=(others=> '0');
signal e_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal f_val :	std_logic_vector(num downto 0):=(others=> '0');
signal g_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal h_val :	std_logic_vector(num downto 0):=(others=> '0');		  
signal i_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal j_val :	std_logic_vector(num downto 0):=(others=> '0');
signal k_val : 	std_logic_vector(num downto 0):=(others=> '0');
signal l_val :	std_logic_vector(num downto 0):=(others=> '0');
signal m_val :	std_logic_vector(num downto 0):=(others=> '0');
signal n_val :	std_logic_vector(num downto 0):=(others=> '0');
signal o_val :	std_logic_vector(num downto 0):=(others=> '0');
signal p_val :	std_logic_vector(num downto 0):=(others=> '0');
signal q_val :	std_logic_vector(num downto 0):=(others=> '0');
signal r_val :	std_logic_vector(num downto 0):=(others=> '0');
signal s_val :	std_logic_vector(num downto 0):=(others=> '0');
signal t_val :	std_logic_vector(num downto 0):=(others=> '0');
signal u_val :	std_logic_vector(num downto 0):=(others=> '0');
signal v_val :	std_logic_vector(num downto 0):=(others=> '0');
signal w_val :	std_logic_vector(num downto 0):=(others=> '0');
signal x_val :	std_logic_vector(num downto 0):=(others=> '0');
signal z_val :	std_logic_vector(num downto 0):=(others=> '0');

component intrapred is
	generic(
		num : integer := 7
	);
	port(	
		clk: in std_logic;
		rst: in std_logic;
		
		A	   :	in std_logic_vector(num downto 0);	 
		B 		:	in std_logic_vector(num downto 0);
		C  	:	in std_logic_vector(num downto 0);
		D		:	in std_logic_vector(num downto 0);
		E		:	in std_logic_vector(num downto 0);
		F		:	in std_logic_vector(num downto 0);
		G		:	in std_logic_vector(num downto 0);
		H		:	in std_logic_vector(num downto 0);
		I		:	in std_logic_vector(num downto 0);
		J		:	in std_logic_vector(num downto 0);
		K		:	in std_logic_vector(num downto 0);
		L		:	in std_logic_vector(num downto 0);
		M		:	in std_logic_vector(num downto 0);	
		N 		:	in std_logic_vector(num downto 0);
		O  	:	in std_logic_vector(num downto 0);
		P		:	in std_logic_vector(num downto 0);
		Q		:	in std_logic_vector(num downto 0);
		R		:	in std_logic_vector(num downto 0);
		S		:	in std_logic_vector(num downto 0);
		T		:	in std_logic_vector(num downto 0);
		U		:	in std_logic_vector(num downto 0);
		V		:	in std_logic_vector(num downto 0);
		W		:	in std_logic_vector(num downto 0);
		X		:	in std_logic_vector(num downto 0);
		Z		:	in std_logic_vector(num downto 0);	
		
		ycbcr0	   :	in std_logic_vector(num downto 0);	 
		ycbcr1 		:	in std_logic_vector(num downto 0);
		ycbcr2  		:	in std_logic_vector(num downto 0);
		ycbcr3		:	in std_logic_vector(num downto 0);
		ycbcr4		:	in std_logic_vector(num downto 0);
		ycbcr5		:	in std_logic_vector(num downto 0);
		ycbcr6		:	in std_logic_vector(num downto 0);
		ycbcr7		:	in std_logic_vector(num downto 0);
		ycbcr8		:	in std_logic_vector(num downto 0);
		ycbcr9		:	in std_logic_vector(num downto 0);
		ycbcr10		:	in std_logic_vector(num downto 0);
		ycbcr11		:	in std_logic_vector(num downto 0);
		ycbcr12		:	in std_logic_vector(num downto 0);			 
		ycbcr13		:	in std_logic_vector(num downto 0);
		ycbcr14		:	in std_logic_vector(num downto 0);
		ycbcr15		:	in std_logic_vector(num downto 0); 
		
		quart : in std_logic_vector(1 downto 0);
		
		first_row 		: in std_logic;
		first_col		: in std_logic;
		
		rdy_in_start: in std_logic;
		rdy_out     : out std_logic;		
			
		intra7   :	out std_logic_vector(num downto 0);
		intra8   :	out std_logic_vector(num downto 0);
		intra9   :	out std_logic_vector(num downto 0);
		intra10  :	out std_logic_vector(num downto 0);
		intra11  :	out std_logic_vector(num downto 0);
		intra12  :	out std_logic_vector(num downto 0);
		intra13  :	out std_logic_vector(num downto 0);		 
		intra14  :	out std_logic_vector(num downto 0);
		intra15  :	out std_logic_vector(num downto 0);	   
		
		Diff0   :	out std_logic_vector(num+1 downto 0);	  
		Diff1   :	out std_logic_vector(num+1 downto 0);
		Diff2   :	out std_logic_vector(num+1 downto 0);
		Diff3   :	out std_logic_vector(num+1 downto 0);
		Diff4   :	out std_logic_vector(num+1 downto 0);
		Diff5   :	out std_logic_vector(num+1 downto 0);
		Diff6   :	out std_logic_vector(num+1 downto 0);
		Diff7   :	out std_logic_vector(num+1 downto 0);
		Diff8   :	out std_logic_vector(num+1 downto 0);
		Diff9   :	out std_logic_vector(num+1 downto 0);
		Diff10  :	out std_logic_vector(num+1 downto 0);
		Diff11  :	out std_logic_vector(num+1 downto 0);
		Diff12  :	out std_logic_vector(num+1 downto 0);
		Diff13  :	out std_logic_vector(num+1 downto 0);		 
		Diff14  :	out std_logic_vector(num+1 downto 0);
		Diff15  :	out std_logic_vector(num+1 downto 0)
		
	);
end component;	

begin

Prediction: intrapred
generic map(
		num => num
	)
port map(
		clk => clk,
		rst => rst,
		
		A => a_val,
		B => b_val,
		C => c_val,
		D => d_val,
		E => e_val,
		F => f_val,
		G => g_val,
		H => h_val,
		I => i_val,
		J => j_val,
		K => k_val,
		L => l_val,
		M => m_val,
		N => n_val,
		O => o_val,
		P => p_val,
		Q => q_val,
		R => r_val,
		S => s_val,
		T => t_val,
		U => u_val,
		V => v_val,
		W => w_val,
		X => x_val,
		Z => z_val,
		
		ycbcr0 =>	ycbcr0_reg2, 
		ycbcr1 =>	ycbcr1_reg2,
		ycbcr2 =>	ycbcr2_reg2,
		ycbcr3 =>	ycbcr3_reg2, 
		ycbcr4 =>	ycbcr4_reg2, 
		ycbcr5 =>	ycbcr5_reg2, 
		ycbcr6 =>	ycbcr6_reg2, 
		ycbcr7 =>	ycbcr7_reg2, 
		ycbcr8 =>	ycbcr8_reg2, 
		ycbcr9 =>	ycbcr9_reg2, 
		ycbcr10 =>	ycbcr10_reg2, 
		ycbcr11 =>	ycbcr11_reg2, 
		ycbcr12 =>	ycbcr12_reg2, 	 
		ycbcr13 =>	ycbcr13_reg2, 
		ycbcr14 =>	ycbcr14_reg2, 
		ycbcr15 =>	ycbcr15_reg2,
		
		rdy_in_start => rdy_pr_in_reg,		
		first_row => first_row_reg,
		first_col => first_col_reg,		
		quart => quart_reg,
		
		intra7  =>	 intra7_pr1 	, 
		intra8  =>	 intra8_pr1 	, 
		intra9  =>	 intra9_pr1 	, 
		intra10 =>	 intra10_pr1	, 
		intra11 =>	 intra11_pr1	, 
		intra12 =>   intra12_pr1	, 
		intra13 =>   intra13_pr1	,  
		intra14 =>   intra14_pr1	, 
		intra15 =>   intra15_pr1	,    
		
		Diff0  =>	  intra0_res,
		Diff1  =>	  intra1_res,
		Diff2  =>	  intra2_res,
		Diff3  =>	  intra3_res,
		Diff4  =>	  intra4_res,
		Diff5  =>	  intra5_res,
		Diff6  =>	  intra6_res,
		Diff7  =>	  intra7_res,
		Diff8  =>	  intra8_res,
		Diff9  =>	  intra9_res,
		Diff10 =>	 intra10_res,
		Diff11 =>	 intra11_res,
		Diff12 =>	 intra12_res,
		Diff13 =>	 intra13_res,
		Diff14 =>	 intra14_res,
		Diff15 =>	 intra15_res,
		
		
		rdy_out   =>  rdy_pr_out 
	);

--#####################################
--##START
--#####################################


-- rdy_in_start is only high one cycle when rdy1_in is high for four cycles
process (clk, rst)
begin
   if rst='1' then   
      rdy_in_counter <= 3;
   elsif (clk'event and clk='1') then 
		if (rdy1_in = '1') then
			if (rdy_in_counter = 3) then
				rdy_in_counter <= 0;
			else
				rdy_in_counter <= rdy_in_counter+1;
			end if;
		end if;
	end if;	
end process;
rdy_pr_in <= rdy1_in when (rdy_in_counter = 0) else '0';

-- set quart input signal
process (clk, rst)
begin
	if (rst = '1') then
		next_quart <= "00";
	elsif (clk'event and clk = '1') then	
		if (rdy1_in = '1') then
			if (next_quart /= "11") then
				next_quart <= next_quart + '1';
			else
				next_quart <= "00";
			end if;
		end if;
	end if;
end process;

process (clk, rst)
begin
	if (rst = '1') then
		quart <= "00";
	elsif (clk'event and clk = '1') then
		quart <= next_quart;
	end if;
end process;

-- internal_count keeps track of which 8x8 block in a MB it is currently
-- based on rdy_in and rdyin_counter
process (clk, rst)
begin
   if rst='1' then 
		internal_count <= 3;
   elsif (clk'event and clk='1') then 	
		if (rdy1_in = '1' and rdy_in_counter = 3) then
			if (internal_count = 3) then
				internal_count <= 0;
			else
				internal_count <= internal_count+1;
			end if;
		end if;
	end if;	
end process;

-- row/col counts keep track of where we are
-- zigzag pattern within the macroblock
process (clk, rst)
begin
   if rst='1' then 
		next_row <= 0;
		next_col <= 0;
   elsif (clk'event and clk='1') then
		if (internal_count = 0) then
			next_row <= row_count;
			next_col <= col_count + 1;
		elsif (internal_count = 1) then	
			next_row <= row_count + 1;
			next_col <= col_count - 1;
		elsif (internal_count = 2) then	
			next_row <= row_count;
			next_col <= col_count + 1;
	  --optimize: from LESS THAN to NOT maybe..
		elsif (internal_count = 3 and col_count = frame_width-1 and row_count < frame_height-1) then	
			next_row <= row_count + 1;
			next_col <= 0;
		elsif (internal_count = 3 and col_count = frame_width-1 and row_count = frame_height-1) then
			next_row <= 0;
			next_col <= 0;
		else
			next_row <= row_count - 1;
			next_col <= col_count + 1;
		end if;	
	end if;	
end process;

process (clk, rst)
begin
   if rst='1' then 	
		col_count <= frame_width-1;
		row_count <= frame_height-1;
	elsif (clk'event and clk='1') then
		if (rdy1_in = '1' and rdy_in_counter = 3) then
			col_count <= next_col;
			row_count <= next_row;
		end if;
	end if;
end process;

first_row <= '1' when row_count = 0 else '0';
first_col <= '1' when col_count = 0 else '0';


-- Counter for output data (quart_pr)
-- only based on rdy_idct input and clk
process (clk, rst)
begin
	if (rst = '1') then
		next_quart_pr <= "00";
	elsif (clk'event and clk = '1') then	
		if (rdy_idct = '1') then
			if (next_quart_pr /= "11") then
				next_quart_pr <= next_quart_pr + '1';
			else
				next_quart_pr <= "00";
			end if;
		end if;
	end if;
end process;

process (clk, rst)
begin
	if (rst = '1') then
		quart_pr <= "00";
	elsif (clk'event and clk = '1') then
		quart_pr <= next_quart_pr;
	end if;
end process;

process (clk, rst)
begin
	if (rst = '1') then
		rdy_update_az <= '0';
	elsif (clk'event and clk = '1') then
        if (quart_pr = "11") then
            rdy_update_az <= '1';
        else
            rdy_update_az <= '0';
		end if;
	end if;
end process;

-- read and store left side neighbours
-- they come as 2x8 where quart_pr and internal_count separates them
process (clk, rst)
begin
	if (rst = '1') then
		ycbcr0_left <= (others=>'0'); ycbcr1_left <= (others=>'0');
		ycbcr2_left <= (others=>'0'); ycbcr3_left <= (others=>'0');
		ycbcr4_left <= (others=>'0'); ycbcr5_left <= (others=>'0');
		ycbcr6_left <= (others=>'0'); ycbcr7_left <= (others=>'0');
		ycbcr8_left <= (others=>'0'); ycbcr9_left <= (others=>'0');
		ycbcr10_left <= (others=>'0'); ycbcr11_left <= (others=>'0');
		ycbcr12_left <= (others=>'0'); ycbcr13_left <= (others=>'0');
		ycbcr14_left <= (others=>'0'); ycbcr15_left <= (others=>'0');
	elsif (clk'event and clk = '1') then
		if (internal_count = 0 or internal_count = 1) then
			if (rdy_idct = '1' and quart_pr = "00") then
				ycbcr0_left <= ycbcr7_prev;
				ycbcr1_left <= ycbcr15_prev;
			elsif (quart_pr = "01") then
				ycbcr2_left <= ycbcr7_prev;
				ycbcr3_left <= ycbcr15_prev;
			elsif (quart_pr = "10") then
				ycbcr4_left <= ycbcr7_prev;
				ycbcr5_left <= ycbcr15_prev;
			elsif (quart_pr = "11") then
				ycbcr6_left <= ycbcr7_prev;
				ycbcr7_left <= ycbcr15_prev;
			end if;
		elsif	(internal_count = 2 or internal_count = 3) then
			if (rdy_idct = '1' and quart_pr = "00") then
				ycbcr8_left <= ycbcr7_prev;
				ycbcr9_left <= ycbcr15_prev;
			elsif (quart_pr = "01") then
				ycbcr10_left <= ycbcr7_prev;
				ycbcr11_left <= ycbcr15_prev;
			elsif (quart_pr = "10") then
				ycbcr12_left <= ycbcr7_prev;
				ycbcr13_left <= ycbcr15_prev;
			elsif (quart_pr = "11") then
				ycbcr14_left <= ycbcr7_prev;
				ycbcr15_left <= ycbcr15_prev;
			end if;
		end if;
	end if;
end process;

-- update qrstuvwx vals based on internal count at rdy_update_az
process (clk, rst)
begin	
	if (rst = '1') then
		q_val <= (others=>'0'); r_val <= (others=>'0');
		s_val <= (others=>'0');	t_val <= (others=>'0');
		u_val <= (others=>'0');	v_val <= (others=>'0');
		w_val <= (others=>'0');	x_val <= (others=>'0');
	elsif (clk'event and clk = '1') then
        if (rdy_update_az = '1') then
            if (internal_count = 0 or (internal_count = 3 and col_count < frame_width-1)) then
                q_val <= ycbcr0_left;
                r_val <= ycbcr1_left;
                s_val <= ycbcr2_left;	
                t_val <= ycbcr3_left;
                u_val <= ycbcr4_left;	
                v_val <= ycbcr5_left;
                w_val <= ycbcr6_left;	
                x_val <= ycbcr7_left;
            elsif	(internal_count = 1 or internal_count = 2) then
                q_val <= ycbcr8_left;
                r_val <= ycbcr9_left;
                s_val <= ycbcr10_left;	
                t_val <= ycbcr11_left;
                u_val <= ycbcr12_left;	
                v_val <= ycbcr13_left;
                w_val <= ycbcr14_left;	
                x_val <= ycbcr15_left;	
    --		else	
            elsif (internal_count = 3 and col_count = frame_width-1) then
                q_val <= (others=>'0'); r_val <= (others=>'0');
                s_val <= (others=>'0');	t_val <= (others=>'0');
                u_val <= (others=>'0');	v_val <= (others=>'0');
                w_val <= (others=>'0');	x_val <= (others=>'0');					
            end if;
        end if;
	end if;
end process;	


-- store prev pixels in horizontal memory
process (clk, rst)
begin
	if (rst='1') then
		ram_h <= (others=>(others=>'0'));
	elsif (clk'event and clk = '1') then
		if (quart_pr = "11") then
			ram_h(col_count) <= ycbcr8_prev & ycbcr9_prev & ycbcr10_prev & ycbcr11_prev & ycbcr12_prev & ycbcr13_prev & ycbcr14_prev & ycbcr15_prev;
		end if;	
	end if;
end process;	

-- read from ram_h to get abcdefghijklmnop values for next 8x8 block
process (clk, rst)
begin
	if (rst='1') then
		ycbcr_upper <= (others=>'0');
	elsif (clk'event and clk = '1') then
        if (rdy_update_az = '1') then
            if (internal_count = 0) then
                ycbcr_upper <= ram_h(col_count+1) & ram_h(col_count+2);
            elsif	(internal_count = 1) then
                ycbcr_upper <= ram_h(col_count-1) & ram_h(col_count);
            elsif	(internal_count = 2) then
                ycbcr_upper <= ram_h(col_count+1) & 
                    ram_h(col_count+1)(7 downto 0) & ram_h(col_count+1)(7 downto 0) &
                    ram_h(col_count+1)(7 downto 0) & ram_h(col_count+1)(7 downto 0) &
                    ram_h(col_count+1)(7 downto 0) & ram_h(col_count+1)(7 downto 0) &
                    ram_h(col_count+1)(7 downto 0) & ram_h(col_count+1)(7 downto 0);
            elsif (internal_count = 3 and col_count = frame_width-1) then
                ycbcr_upper <= ram_h(0) & ram_h(1);
            else --internal_count = 3 and col_count < frame_width-1) then
                ycbcr_upper <= ram_h(col_count+1) & ram_h(col_count+2);
            end if;
        end if;
	end if;
end process;		

--update abcdefghijklmnop
a_val <= ycbcr_upper(127 downto 120);
b_val <= ycbcr_upper(119 downto 112); 
c_val <= ycbcr_upper(111 downto 104); 
d_val <= ycbcr_upper(103 downto 96); 
e_val <= ycbcr_upper(95 downto 88); 
f_val <= ycbcr_upper(87 downto 80); 
g_val <= ycbcr_upper(79 downto 72); 
h_val <= ycbcr_upper(71 downto 64); 
i_val <= ycbcr_upper(63 downto 56); 
j_val <= ycbcr_upper(55 downto 48); 
k_val <= ycbcr_upper(47 downto 40); 
l_val <= ycbcr_upper(39 downto 32); 
m_val <= ycbcr_upper(31 downto 24); 
n_val <= ycbcr_upper(23 downto 16); 
o_val <= ycbcr_upper(15 downto 8); 
p_val <= ycbcr_upper(7 downto 0); 

--z_val fetch before neighbouur registers update
process (clk, rst)
begin
	if (rst='1') then
		ycbcr_z0 <= (others=>'0');
		ycbcr_z1 <= (others=>'0');
		ycbcr_z2 <= (others=>'0');
		ycbcr_z3 <= (others=>'0');
	elsif (clk'event and clk = '1') then
		if (internal_count = 0) then
			if (quart_pr = "11") then
				ycbcr_z3 <= ycbcr15_prev;
			--must be stored before x_val changes	
			elsif (rdy_idct = '1' and quart_pr = "00") then
				ycbcr_z2 <= x_val;
				ycbcr_z1 <= h_val;
			end if;
		elsif (internal_count = 1) then
			if (rdy_idct = '1' and quart_pr = "00") then
				ycbcr_z0 <= h_val;
			end if;	
		end if;		
	end if;
end process;

--update z_val before new data comes in, at rdy_update_az = 1
process (clk, rst)
begin	
	if (rst='1') then
		z_val <= (others=>'0');
	elsif (clk'event and clk = '1') then
		if (rdy_update_az = '1') then
			if (internal_count = 0) then
				z_val <= ycbcr_z1;
			elsif (internal_count = 1) then
				z_val <= ycbcr_z2;
			elsif (internal_count = 2) then
				z_val <= ycbcr_z3;
			elsif (internal_count = 3 and col_count < frame_width-1) then
				z_val <= ycbcr_z0;
			else --new row, and its zero
				z_val <= (others=>'0');
			end if;
		end if;	
	end if;
end process;
	
--#######################################################
-- Delay/Buffer Registers for module: intrapred
--########################################################


process (clk, rst)
begin
   if rst='1' then   
		ycbcr0_reg <= (others=>'0');
		ycbcr1_reg <= (others=>'0');
		ycbcr2_reg <= (others=>'0');
		ycbcr3_reg <= (others=>'0');
		ycbcr4_reg <= (others=>'0');
		ycbcr5_reg <= (others=>'0');
		ycbcr6_reg <= (others=>'0');
		ycbcr7_reg <= (others=>'0');
		ycbcr8_reg <= (others=>'0');
		ycbcr9_reg <= (others=>'0');
		ycbcr10_reg <= (others=>'0');
		ycbcr11_reg <= (others=>'0');
		ycbcr12_reg <= (others=>'0');
		ycbcr13_reg <= (others=>'0');
		ycbcr14_reg <= (others=>'0');
		ycbcr15_reg <= (others=>'0');
		ycbcr0_reg2 <= (others=>'0');
		ycbcr1_reg2 <= (others=>'0');
		ycbcr2_reg2 <= (others=>'0');
		ycbcr3_reg2 <= (others=>'0');
		ycbcr4_reg2 <= (others=>'0');
		ycbcr5_reg2 <= (others=>'0');
		ycbcr6_reg2 <= (others=>'0');
		ycbcr7_reg2 <= (others=>'0');
		ycbcr8_reg2 <= (others=>'0');
		ycbcr9_reg2 <= (others=>'0');
		ycbcr10_reg2 <= (others=>'0');
		ycbcr11_reg2 <= (others=>'0');
		ycbcr12_reg2 <= (others=>'0');
		ycbcr13_reg2 <= (others=>'0');
		ycbcr14_reg2 <= (others=>'0');
		ycbcr15_reg2 <= (others=>'0');

		rdy_pr_in_reg <= '0';
		first_row_reg <= '0';
		first_col_reg <= '0';
		quart_reg 	<= (others=>'0');	
		
   elsif (clk'event and clk='1') then 
	
		ycbcr0_reg <= ycbcr0;
		ycbcr1_reg <= ycbcr1;
		ycbcr2_reg <= ycbcr2;
		ycbcr3_reg <= ycbcr3;
		ycbcr4_reg <= ycbcr4;
		ycbcr5_reg <= ycbcr5;
		ycbcr6_reg <= ycbcr6;
		ycbcr7_reg <= ycbcr7;
		ycbcr8_reg <= ycbcr8;
		ycbcr9_reg <= ycbcr9;
		ycbcr10_reg <= ycbcr10;
		ycbcr11_reg <= ycbcr11;
		ycbcr12_reg <= ycbcr12;
		ycbcr13_reg <= ycbcr13;
		ycbcr14_reg <= ycbcr14;
		ycbcr15_reg <= ycbcr15;
		
	elsif (clk'event and clk='0') then
		ycbcr0_reg2 <= ycbcr0_reg;
		ycbcr1_reg2 <= ycbcr1_reg;
		ycbcr2_reg2 <= ycbcr2_reg;
		ycbcr3_reg2 <= ycbcr3_reg;
		ycbcr4_reg2 <= ycbcr4_reg;
		ycbcr5_reg2 <= ycbcr5_reg;
		ycbcr6_reg2 <= ycbcr6_reg;
		ycbcr7_reg2 <= ycbcr7_reg;
		ycbcr8_reg2 <= ycbcr8_reg;
		ycbcr9_reg2 <= ycbcr9_reg;
		ycbcr10_reg2 <= ycbcr10_reg;
		ycbcr11_reg2 <= ycbcr11_reg;
		ycbcr12_reg2 <= ycbcr12_reg;
		ycbcr13_reg2 <= ycbcr13_reg;
		ycbcr14_reg2 <= ycbcr14_reg;
		ycbcr15_reg2 <= ycbcr15_reg;
		
		rdy_pr_in_reg <= rdy_pr_in;
		first_row_reg <= first_row;
		first_col_reg <= first_col;
		quart_reg 	  <= quart;			
   end if;
end process;

-- Assuming 6 clk cycles delay from reconstruction loop
-- Delaying predicted pixels from intrapred while residuals are being reconstructed
process (clk, rst)
begin
	if (rst = '1') then
		intra_reg1 <= (others=> (others=>'0'));
		intra_reg2 <= (others=> (others=>'0'));
		intra_reg3 <= (others=> (others=>'0'));
		intra_reg4 <= (others=> (others=>'0'));
		intra_reg5 <= (others=> (others=>'0'));
		intra_reg6 <= (others=> (others=>'0'));
		
	elsif (clk'event and clk = '1') then
		--store output from intrapred module 
		intra_reg1(7)  <= intra7_pr1;	
		intra_reg1(8)  <= intra8_pr1;  intra_reg1(9)  <= intra9_pr1;	 
		intra_reg1(10) <= intra10_pr1; intra_reg1(11) <= intra11_pr1;
		intra_reg1(12) <= intra12_pr1; intra_reg1(13) <= intra13_pr1; 
		intra_reg1(14) <= intra14_pr1; intra_reg1(15) <= intra15_pr1;

		intra_reg2 <= intra_reg1;
		intra_reg3 <= intra_reg2;
		intra_reg4 <= intra_reg3;
		intra_reg5 <= intra_reg4;
		intra_reg6 <= intra_reg5;
	end if;	
end process;

-- after reconstruction loop (reconstructed residual + delayed intra prediction best mode)
-- make delay signals after finding reconstruction wait time
-- these are 2x8, not 4x4           
ycbcr7_prev_extra  <= ycbcr7_pr(num+1 downto 0)  + ('0' & intra_reg6(7));             
ycbcr8_prev_extra  <= ycbcr8_pr(num+1 downto 0)  + ('0' & intra_reg6(8));          
ycbcr9_prev_extra  <= ycbcr9_pr(num+1 downto 0)  + ('0' & intra_reg6(9));          
ycbcr10_prev_extra <= ycbcr10_pr(num+1 downto 0) + ('0' & intra_reg6(10));              
ycbcr11_prev_extra <= ycbcr11_pr(num+1 downto 0) + ('0' & intra_reg6(11));              
ycbcr12_prev_extra <= ycbcr12_pr(num+1 downto 0) + ('0' & intra_reg6(12));              
ycbcr13_prev_extra <= ycbcr13_pr(num+1 downto 0) + ('0' & intra_reg6(13));              
ycbcr14_prev_extra <= ycbcr14_pr(num+1 downto 0) + ('0' & intra_reg6(14));              
ycbcr15_prev_extra <= ycbcr15_pr(num+1 downto 0) + ('0' & intra_reg6(15));

ycbcr7_prev  <= ycbcr7_prev_extra(num downto 0) when ycbcr7_prev_extra(num+1)='0'  else (not(ycbcr7_prev_extra(num downto 0))+1);
ycbcr8_prev  <= ycbcr8_prev_extra(num downto 0) when ycbcr8_prev_extra(num+1)='0'  else (not(ycbcr8_prev_extra(num downto 0))+1);
ycbcr9_prev  <= ycbcr9_prev_extra(num downto 0) when ycbcr9_prev_extra(num+1)='0'  else (not(ycbcr9_prev_extra(num downto 0))+1);
ycbcr10_prev <= ycbcr10_prev_extra(num downto 0)when ycbcr10_prev_extra(num+1)='0' else (not(ycbcr10_prev_extra(num downto 0))+1);
ycbcr11_prev <= ycbcr11_prev_extra(num downto 0)when ycbcr11_prev_extra(num+1)='0' else (not(ycbcr11_prev_extra(num downto 0))+1);
ycbcr12_prev <= ycbcr12_prev_extra(num downto 0)when ycbcr12_prev_extra(num+1)='0' else (not(ycbcr12_prev_extra(num downto 0))+1);
ycbcr13_prev <= ycbcr13_prev_extra(num downto 0)when ycbcr13_prev_extra(num+1)='0' else (not(ycbcr13_prev_extra(num downto 0))+1);
ycbcr14_prev <= ycbcr14_prev_extra(num downto 0)when ycbcr14_prev_extra(num+1)='0' else (not(ycbcr14_prev_extra(num downto 0))+1);
ycbcr15_prev <= ycbcr15_prev_extra(num downto 0)when ycbcr15_prev_extra(num+1)='0' else (not(ycbcr15_prev_extra(num downto 0))+1);


--outputs directly from intra prediction module
intra0  <= intra0_res;
intra1  <= intra1_res;
intra2  <= intra2_res;
intra3  <= intra3_res;
intra4  <= intra4_res;
intra5  <= intra5_res;
intra6  <= intra6_res;
intra7  <= intra7_res;
intra8  <= intra8_res;
intra9  <= intra9_res;
intra10 <= intra10_res;
intra11 <= intra11_res;
intra12 <= intra12_res;
intra13 <= intra13_res;
intra14 <= intra14_res;
intra15 <= intra15_res;
		
rdy_out <= rdy_pr_out;	


end Behavioral;

