----------------------------------------------------------------------------------
-- Company: 		NTNU
-- Engineer: 		Kim Troennes
-- 
-- Create Date:	02:48:16 03/13/2014 
-- Design Name:	pred.vhd
-- Module Name:	intrapred - Behavioral  
-- Project Name:  intra8x8final
-- Description:	8x8 luminance intra prediction for H.264 encoder
--						Computes residuals and best mode predictions
----------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

entity intrapred is
	generic(
		num : integer := 7
	);
	
	port(
		clk: in std_logic;
		rst: in std_logic;
		
		--a to z must be zeros when firstrow or firstcol  = 1
		A	   :	in std_logic_vector(num downto 0);	 
		B 		:	in std_logic_vector(num downto 0);
		C  	:	in std_logic_vector(num downto 0);
		D		:	in std_logic_vector(num downto 0);
		E		:	in std_logic_vector(num downto 0);
		F		:	in std_logic_vector(num downto 0);
		G		:	in std_logic_vector(num downto 0);
		H		:	in std_logic_vector(num downto 0);
		I		:	in std_logic_vector(num downto 0);
		J		:	in std_logic_vector(num downto 0);
		K		:	in std_logic_vector(num downto 0);
		L		:	in std_logic_vector(num downto 0);
		M		:	in std_logic_vector(num downto 0);	
		N 		:	in std_logic_vector(num downto 0);
		O  	:	in std_logic_vector(num downto 0);
		P		:	in std_logic_vector(num downto 0);
		Q		:	in std_logic_vector(num downto 0);
		R		:	in std_logic_vector(num downto 0);
		S		:	in std_logic_vector(num downto 0);
		T		:	in std_logic_vector(num downto 0);
		U		:	in std_logic_vector(num downto 0);
		V		:	in std_logic_vector(num downto 0);
		W		:	in std_logic_vector(num downto 0);
		X		:	in std_logic_vector(num downto 0);
		Z		:	in std_logic_vector(num downto 0);	
		
		ycbcr0	   :	in std_logic_vector(num downto 0);	 
		ycbcr1 		:	in std_logic_vector(num downto 0);
		ycbcr2  		:	in std_logic_vector(num downto 0);
		ycbcr3		:	in std_logic_vector(num downto 0);
		ycbcr4		:	in std_logic_vector(num downto 0);
		ycbcr5		:	in std_logic_vector(num downto 0);
		ycbcr6		:	in std_logic_vector(num downto 0);
		ycbcr7		:	in std_logic_vector(num downto 0);
		ycbcr8		:	in std_logic_vector(num downto 0);
		ycbcr9		:	in std_logic_vector(num downto 0);
		ycbcr10		:	in std_logic_vector(num downto 0);
		ycbcr11		:	in std_logic_vector(num downto 0);
		ycbcr12		:	in std_logic_vector(num downto 0);			 
		ycbcr13		:	in std_logic_vector(num downto 0);
		ycbcr14		:	in std_logic_vector(num downto 0);
		ycbcr15		:	in std_logic_vector(num downto 0); 

		rdy_in_start : in std_logic;
		
		--which of the 4x4's in the 8x8 are we working on?
		quart : in std_logic_vector(1 downto 0);
		
		first_row 		: in std_logic;
		first_col		: in std_logic;		

		-- Predicted pixels used in calculating next A-Z, 
		-- 8x2 matrix where X is not used
		-- |X |X | X| X| X| X| X| 7|
		-- |8 |9 |10|11|12|13|14|15|		

		intra7   :	out std_logic_vector(num downto 0);
		intra8   :	out std_logic_vector(num downto 0);
		intra9   :	out std_logic_vector(num downto 0);
		intra10  :	out std_logic_vector(num downto 0);
		intra11  :	out std_logic_vector(num downto 0);
		intra12  :	out std_logic_vector(num downto 0);
		intra13  :	out std_logic_vector(num downto 0);		 
		intra14  :	out std_logic_vector(num downto 0);
		intra15  :	out std_logic_vector(num downto 0);	   
		
		-- Residuals: 8x2 matrix, not 4x4
		Diff0   :	out std_logic_vector(num+1 downto 0);	  
		Diff1   :	out std_logic_vector(num+1 downto 0);
		Diff2   :	out std_logic_vector(num+1 downto 0);
		Diff3   :	out std_logic_vector(num+1 downto 0);
		Diff4   :	out std_logic_vector(num+1 downto 0);
		Diff5   :	out std_logic_vector(num+1 downto 0);
		Diff6   :	out std_logic_vector(num+1 downto 0);
		Diff7   :	out std_logic_vector(num+1 downto 0);
		Diff8   :	out std_logic_vector(num+1 downto 0);
		Diff9   :	out std_logic_vector(num+1 downto 0);
		Diff10  :	out std_logic_vector(num+1 downto 0);
		Diff11  :	out std_logic_vector(num+1 downto 0);
		Diff12  :	out std_logic_vector(num+1 downto 0);
		Diff13  :	out std_logic_vector(num+1 downto 0);		 
		Diff14  :	out std_logic_vector(num+1 downto 0);
		Diff15  :	out std_logic_vector(num+1 downto 0);
		
		rdy_out   : out std_logic
	
	);
end intrapred;

architecture Behavioral of intrapred is

-- Timing signals
signal rdy_in_1 	: std_logic;
signal rdy_in_2	: std_logic;
signal rdy_in_3	: std_logic;
signal rdy_in_4	: std_logic;
signal rdy_in_5	: std_logic;
signal rdy_in_6	: std_logic;
signal rdy_in_7	: std_logic;
signal rdy_in_8	: std_logic;
signal rdy_in_9	: std_logic;
signal rdy_in_10	: std_logic;

signal first_row_reg			: std_logic;
signal first_row_reg2		: std_logic;

signal first_col_reg			: std_logic;
signal first_col_reg2		: std_logic;

signal	ycbcr0_reg	   :	 std_logic_vector(num downto 0);	 
signal	ycbcr1_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr2_reg  	:	 std_logic_vector(num downto 0);
signal	ycbcr3_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr4_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr5_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr6_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr7_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr8_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr9_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr10_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr11_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr12_reg		:	 std_logic_vector(num downto 0);			 
signal	ycbcr13_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr14_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr15_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr16_reg	   :	 std_logic_vector(num downto 0);	 
signal	ycbcr17_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr18_reg  	:	 std_logic_vector(num downto 0);
signal	ycbcr19_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr20_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr21_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr22_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr23_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr24_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr25_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr26_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr27_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr28_reg		:	 std_logic_vector(num downto 0);			 
signal	ycbcr29_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr30_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr31_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr32_reg	   :	 std_logic_vector(num downto 0);	 
signal	ycbcr33_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr34_reg  	:	 std_logic_vector(num downto 0);
signal	ycbcr35_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr36_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr37_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr38_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr39_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr40_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr41_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr42_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr43_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr44_reg		:	 std_logic_vector(num downto 0);			 
--signal	ycbcr45_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr46_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr47_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr48_reg	   :	 std_logic_vector(num downto 0);	 
signal	ycbcr49_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr50_reg  	:	 std_logic_vector(num downto 0);
signal	ycbcr51_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr52_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr53_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr54_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr55_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr56_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr57_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr58_reg		:	 std_logic_vector(num downto 0);
signal	ycbcr59_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr60_reg		:	 std_logic_vector(num downto 0);			 
--signal	ycbcr61_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr62_reg		:	 std_logic_vector(num downto 0);
--signal	ycbcr63_reg		:	 std_logic_vector(num downto 0);

-- Vertical prediction
signal V_0				:    std_logic_vector(num downto 0);               
signal V_1				:    std_logic_vector(num downto 0);  
signal V_2				:    std_logic_vector(num downto 0);  
signal V_3				:    std_logic_vector(num downto 0);  
signal V_4				:    std_logic_vector(num downto 0);  
signal V_5				:    std_logic_vector(num downto 0);  	 
signal V_6				:    std_logic_vector(num downto 0);               
signal V_7				:    std_logic_vector(num downto 0);  
signal V_8				:    std_logic_vector(num downto 0);  
signal V_9				:    std_logic_vector(num downto 0);  
signal V_10				:    std_logic_vector(num downto 0);  
signal V_11				:    std_logic_vector(num downto 0);  
signal V_12				:    std_logic_vector(num downto 0);  
signal V_13				:    std_logic_vector(num downto 0);  
signal V_14				:    std_logic_vector(num downto 0);  
signal V_15				:    std_logic_vector(num downto 0);
signal V_16				:    std_logic_vector(num downto 0);               
signal V_17				:    std_logic_vector(num downto 0);  
signal V_18				:    std_logic_vector(num downto 0);  
signal V_19				:    std_logic_vector(num downto 0);  
signal V_20				:    std_logic_vector(num downto 0);  
signal V_21				:    std_logic_vector(num downto 0);  	 
signal V_22				:    std_logic_vector(num downto 0);               
signal V_23				:    std_logic_vector(num downto 0);  
signal V_24				:    std_logic_vector(num downto 0);  
signal V_25				:    std_logic_vector(num downto 0);  
signal V_26				:    std_logic_vector(num downto 0);  
signal V_27				:    std_logic_vector(num downto 0);  
signal V_28				:    std_logic_vector(num downto 0);  
signal V_29				:    std_logic_vector(num downto 0);  
signal V_30				:    std_logic_vector(num downto 0);  
signal V_31				:    std_logic_vector(num downto 0); 
signal V_32				:    std_logic_vector(num downto 0);               
signal V_33				:    std_logic_vector(num downto 0);  
signal V_34				:    std_logic_vector(num downto 0);  
signal V_35				:    std_logic_vector(num downto 0);  
signal V_36				:    std_logic_vector(num downto 0);  
signal V_37				:    std_logic_vector(num downto 0);  	 
signal V_38				:    std_logic_vector(num downto 0);               
signal V_39				:    std_logic_vector(num downto 0);  
signal V_40				:    std_logic_vector(num downto 0);  
signal V_41				:    std_logic_vector(num downto 0);  
signal V_42				:    std_logic_vector(num downto 0);  
signal V_43				:    std_logic_vector(num downto 0);  
signal V_44				:    std_logic_vector(num downto 0);  
signal V_45				:    std_logic_vector(num downto 0);  
signal V_46				:    std_logic_vector(num downto 0);  
signal V_47				:    std_logic_vector(num downto 0); 
signal V_48				:    std_logic_vector(num downto 0);               
signal V_49				:    std_logic_vector(num downto 0);  
signal V_50				:    std_logic_vector(num downto 0);  
signal V_51				:    std_logic_vector(num downto 0);  
signal V_52				:    std_logic_vector(num downto 0);  
signal V_53				:    std_logic_vector(num downto 0);  	 
signal V_54				:    std_logic_vector(num downto 0);               
signal V_55				:    std_logic_vector(num downto 0);  
signal V_56				:    std_logic_vector(num downto 0);  
signal V_57				:    std_logic_vector(num downto 0);  
signal V_58				:    std_logic_vector(num downto 0);  
signal V_59				:    std_logic_vector(num downto 0);  
signal V_60				:    std_logic_vector(num downto 0);  
signal V_61				:    std_logic_vector(num downto 0);  
signal V_62				:    std_logic_vector(num downto 0);  
signal V_63				:    std_logic_vector(num downto 0);  

-- Horizontal prediction
signal H_0				:    std_logic_vector(num downto 0);               
signal H_1				:    std_logic_vector(num downto 0);  
signal H_2				:    std_logic_vector(num downto 0);  
signal H_3				:    std_logic_vector(num downto 0);  
signal H_4				:    std_logic_vector(num downto 0);  
signal H_5				:    std_logic_vector(num downto 0);  	 
signal H_6				:    std_logic_vector(num downto 0);               
signal H_7				:    std_logic_vector(num downto 0);  
signal H_8				:    std_logic_vector(num downto 0);  
signal H_9				:    std_logic_vector(num downto 0);  
signal H_10				:    std_logic_vector(num downto 0);  
signal H_11				:    std_logic_vector(num downto 0);  
signal H_12				:    std_logic_vector(num downto 0);  
signal H_13				:    std_logic_vector(num downto 0);  
signal H_14				:    std_logic_vector(num downto 0);  
signal H_15				:    std_logic_vector(num downto 0);
signal H_16				:    std_logic_vector(num downto 0);               
signal H_17				:    std_logic_vector(num downto 0);  
signal H_18				:    std_logic_vector(num downto 0);  
signal H_19				:    std_logic_vector(num downto 0);  
signal H_20				:    std_logic_vector(num downto 0);  
signal H_21				:    std_logic_vector(num downto 0);  	 
signal H_22				:    std_logic_vector(num downto 0);               
signal H_23				:    std_logic_vector(num downto 0);  
signal H_24				:    std_logic_vector(num downto 0);  
signal H_25				:    std_logic_vector(num downto 0);  
signal H_26				:    std_logic_vector(num downto 0);  
signal H_27				:    std_logic_vector(num downto 0);  
signal H_28				:    std_logic_vector(num downto 0);  
signal H_29				:    std_logic_vector(num downto 0);  
signal H_30				:    std_logic_vector(num downto 0);  
signal H_31				:    std_logic_vector(num downto 0); 
signal H_32				:    std_logic_vector(num downto 0);               
signal H_33				:    std_logic_vector(num downto 0);  
signal H_34				:    std_logic_vector(num downto 0);  
signal H_35				:    std_logic_vector(num downto 0);  
signal H_36				:    std_logic_vector(num downto 0);  
signal H_37				:    std_logic_vector(num downto 0);  	 
signal H_38				:    std_logic_vector(num downto 0);               
signal H_39				:    std_logic_vector(num downto 0);  
signal H_40				:    std_logic_vector(num downto 0);  
signal H_41				:    std_logic_vector(num downto 0);  
signal H_42				:    std_logic_vector(num downto 0);  
signal H_43				:    std_logic_vector(num downto 0);  
signal H_44				:    std_logic_vector(num downto 0);  
signal H_45				:    std_logic_vector(num downto 0);  
signal H_46				:    std_logic_vector(num downto 0);  
signal H_47				:    std_logic_vector(num downto 0); 
signal H_48				:    std_logic_vector(num downto 0);               
signal H_49				:    std_logic_vector(num downto 0);  
signal H_50				:    std_logic_vector(num downto 0);  
signal H_51				:    std_logic_vector(num downto 0);  
signal H_52				:    std_logic_vector(num downto 0);  
signal H_53				:    std_logic_vector(num downto 0);  	 
signal H_54				:    std_logic_vector(num downto 0);               
signal H_55				:    std_logic_vector(num downto 0);  
signal H_56				:    std_logic_vector(num downto 0);  
signal H_57				:    std_logic_vector(num downto 0);  
signal H_58				:    std_logic_vector(num downto 0);  
signal H_59				:    std_logic_vector(num downto 0);  
signal H_60				:    std_logic_vector(num downto 0);  
signal H_61				:    std_logic_vector(num downto 0);  
signal H_62				:    std_logic_vector(num downto 0);  
signal H_63				:    std_logic_vector(num downto 0);

-- DC prediction
signal DC				: 	  std_logic_vector(num downto 0); 
signal sum_DC			:    std_logic_vector(num+4 downto 0);
signal sum_DC_TOTAL	:    std_logic_vector(num+4 downto 0);

-- DDL
signal DDL_0				:    std_logic_vector(num downto 0);               
signal DDL_1				:    std_logic_vector(num downto 0);  
signal DDL_2				:    std_logic_vector(num downto 0);  
signal DDL_3				:    std_logic_vector(num downto 0);  
signal DDL_4				:    std_logic_vector(num downto 0);  
signal DDL_5				:    std_logic_vector(num downto 0);  	 
signal DDL_6				:    std_logic_vector(num downto 0);               
signal DDL_7				:    std_logic_vector(num downto 0);  
signal DDL_8				:    std_logic_vector(num downto 0);  
signal DDL_9				:    std_logic_vector(num downto 0);  
signal DDL_10				:    std_logic_vector(num downto 0);  
signal DDL_11				:    std_logic_vector(num downto 0);  
signal DDL_12				:    std_logic_vector(num downto 0);  
signal DDL_13				:    std_logic_vector(num downto 0);  
signal DDL_14				:    std_logic_vector(num downto 0);  
signal DDL_15				:    std_logic_vector(num downto 0);
signal DDL_16				:    std_logic_vector(num downto 0);               
signal DDL_17				:    std_logic_vector(num downto 0);  
signal DDL_18				:    std_logic_vector(num downto 0);  
signal DDL_19				:    std_logic_vector(num downto 0);  
signal DDL_20				:    std_logic_vector(num downto 0);  
signal DDL_21				:    std_logic_vector(num downto 0);  	 
signal DDL_22				:    std_logic_vector(num downto 0);               
signal DDL_23				:    std_logic_vector(num downto 0);  
signal DDL_24				:    std_logic_vector(num downto 0);  
signal DDL_25				:    std_logic_vector(num downto 0);  
signal DDL_26				:    std_logic_vector(num downto 0);  
signal DDL_27				:    std_logic_vector(num downto 0);  
signal DDL_28				:    std_logic_vector(num downto 0);  
signal DDL_29				:    std_logic_vector(num downto 0);  
signal DDL_30				:    std_logic_vector(num downto 0);  
signal DDL_31				:    std_logic_vector(num downto 0); 
signal DDL_32				:    std_logic_vector(num downto 0);               
signal DDL_33				:    std_logic_vector(num downto 0);  
signal DDL_34				:    std_logic_vector(num downto 0);  
signal DDL_35				:    std_logic_vector(num downto 0);  
signal DDL_36				:    std_logic_vector(num downto 0);  
signal DDL_37				:    std_logic_vector(num downto 0);  	 
signal DDL_38				:    std_logic_vector(num downto 0);               
signal DDL_39				:    std_logic_vector(num downto 0);  
signal DDL_40				:    std_logic_vector(num downto 0);  
signal DDL_41				:    std_logic_vector(num downto 0);  
signal DDL_42				:    std_logic_vector(num downto 0);  
signal DDL_43				:    std_logic_vector(num downto 0);  
signal DDL_44				:    std_logic_vector(num downto 0);  
signal DDL_45				:    std_logic_vector(num downto 0);  
signal DDL_46				:    std_logic_vector(num downto 0);  
signal DDL_47				:    std_logic_vector(num downto 0); 
signal DDL_48				:    std_logic_vector(num downto 0);               
signal DDL_49				:    std_logic_vector(num downto 0);  
signal DDL_50				:    std_logic_vector(num downto 0);  
signal DDL_51				:    std_logic_vector(num downto 0);  
signal DDL_52				:    std_logic_vector(num downto 0);  
signal DDL_53				:    std_logic_vector(num downto 0);  	 
signal DDL_54				:    std_logic_vector(num downto 0);               
signal DDL_55				:    std_logic_vector(num downto 0);  
signal DDL_56				:    std_logic_vector(num downto 0);  
signal DDL_57				:    std_logic_vector(num downto 0);  
signal DDL_58				:    std_logic_vector(num downto 0);  
signal DDL_59				:    std_logic_vector(num downto 0);  
signal DDL_60				:    std_logic_vector(num downto 0);  
signal DDL_61				:    std_logic_vector(num downto 0);  
signal DDL_62				:    std_logic_vector(num downto 0);  
signal DDL_63				:    std_logic_vector(num downto 0);
 
-- DDR
signal DDR_0				:    std_logic_vector(num downto 0);               
signal DDR_1				:    std_logic_vector(num downto 0);  
signal DDR_2				:    std_logic_vector(num downto 0);  
signal DDR_3				:    std_logic_vector(num downto 0);  
signal DDR_4				:    std_logic_vector(num downto 0);  
signal DDR_5				:    std_logic_vector(num downto 0);  	 
signal DDR_6				:    std_logic_vector(num downto 0);               
signal DDR_7				:    std_logic_vector(num downto 0);  
signal DDR_8				:    std_logic_vector(num downto 0);  
signal DDR_9				:    std_logic_vector(num downto 0);  
signal DDR_10				:    std_logic_vector(num downto 0);  
signal DDR_11				:    std_logic_vector(num downto 0);  
signal DDR_12				:    std_logic_vector(num downto 0);  
signal DDR_13				:    std_logic_vector(num downto 0);  
signal DDR_14				:    std_logic_vector(num downto 0);  
signal DDR_15				:    std_logic_vector(num downto 0);
signal DDR_16				:    std_logic_vector(num downto 0);               
signal DDR_17				:    std_logic_vector(num downto 0);  
signal DDR_18				:    std_logic_vector(num downto 0);  
signal DDR_19				:    std_logic_vector(num downto 0);  
signal DDR_20				:    std_logic_vector(num downto 0);  
signal DDR_21				:    std_logic_vector(num downto 0);  	 
signal DDR_22				:    std_logic_vector(num downto 0);               
signal DDR_23				:    std_logic_vector(num downto 0);  
signal DDR_24				:    std_logic_vector(num downto 0);  
signal DDR_25				:    std_logic_vector(num downto 0);  
signal DDR_26				:    std_logic_vector(num downto 0);  
signal DDR_27				:    std_logic_vector(num downto 0);  
signal DDR_28				:    std_logic_vector(num downto 0);  
signal DDR_29				:    std_logic_vector(num downto 0);  
signal DDR_30				:    std_logic_vector(num downto 0);  
signal DDR_31				:    std_logic_vector(num downto 0); 
signal DDR_32				:    std_logic_vector(num downto 0);               
signal DDR_33				:    std_logic_vector(num downto 0);  
signal DDR_34				:    std_logic_vector(num downto 0);  
signal DDR_35				:    std_logic_vector(num downto 0);  
signal DDR_36				:    std_logic_vector(num downto 0);  
signal DDR_37				:    std_logic_vector(num downto 0);  	 
signal DDR_38				:    std_logic_vector(num downto 0);               
signal DDR_39				:    std_logic_vector(num downto 0);  
signal DDR_40				:    std_logic_vector(num downto 0);  
signal DDR_41				:    std_logic_vector(num downto 0);  
signal DDR_42				:    std_logic_vector(num downto 0);  
signal DDR_43				:    std_logic_vector(num downto 0);  
signal DDR_44				:    std_logic_vector(num downto 0);  
signal DDR_45				:    std_logic_vector(num downto 0);  
signal DDR_46				:    std_logic_vector(num downto 0);  
signal DDR_47				:    std_logic_vector(num downto 0); 
signal DDR_48				:    std_logic_vector(num downto 0);               
signal DDR_49				:    std_logic_vector(num downto 0);  
signal DDR_50				:    std_logic_vector(num downto 0);  
signal DDR_51				:    std_logic_vector(num downto 0);  
signal DDR_52				:    std_logic_vector(num downto 0);  
signal DDR_53				:    std_logic_vector(num downto 0);  	 
signal DDR_54				:    std_logic_vector(num downto 0);               
signal DDR_55				:    std_logic_vector(num downto 0);  
signal DDR_56				:    std_logic_vector(num downto 0);  
signal DDR_57				:    std_logic_vector(num downto 0);  
signal DDR_58				:    std_logic_vector(num downto 0);  
signal DDR_59				:    std_logic_vector(num downto 0);  
signal DDR_60				:    std_logic_vector(num downto 0);  
signal DDR_61				:    std_logic_vector(num downto 0);  
signal DDR_62				:    std_logic_vector(num downto 0);  
signal DDR_63				:    std_logic_vector(num downto 0); 

-- VR
signal VR_0					:    std_logic_vector(num downto 0);               
signal VR_1					:    std_logic_vector(num downto 0);  
signal VR_2					:    std_logic_vector(num downto 0);  
signal VR_3					:    std_logic_vector(num downto 0);  
signal VR_4					:    std_logic_vector(num downto 0);  
signal VR_5					:    std_logic_vector(num downto 0);  	 
signal VR_6					:    std_logic_vector(num downto 0);               
signal VR_7					:    std_logic_vector(num downto 0);  
signal VR_8					:    std_logic_vector(num downto 0);  
signal VR_9					:    std_logic_vector(num downto 0);  
signal VR_10				:    std_logic_vector(num downto 0);  
signal VR_11				:    std_logic_vector(num downto 0);  
signal VR_12				:    std_logic_vector(num downto 0);  
signal VR_13				:    std_logic_vector(num downto 0);  
signal VR_14				:    std_logic_vector(num downto 0);  
signal VR_15				:    std_logic_vector(num downto 0);
signal VR_16				:    std_logic_vector(num downto 0);               
signal VR_17				:    std_logic_vector(num downto 0);  
signal VR_18				:    std_logic_vector(num downto 0);  
signal VR_19				:    std_logic_vector(num downto 0);  
signal VR_20				:    std_logic_vector(num downto 0);  
signal VR_21				:    std_logic_vector(num downto 0);  	 
signal VR_22				:    std_logic_vector(num downto 0);               
signal VR_23				:    std_logic_vector(num downto 0);  
signal VR_24				:    std_logic_vector(num downto 0);  
signal VR_25				:    std_logic_vector(num downto 0);  
signal VR_26				:    std_logic_vector(num downto 0);  
signal VR_27				:    std_logic_vector(num downto 0);  
signal VR_28				:    std_logic_vector(num downto 0);  
signal VR_29				:    std_logic_vector(num downto 0);  
signal VR_30				:    std_logic_vector(num downto 0);  
signal VR_31				:    std_logic_vector(num downto 0); 
signal VR_32				:    std_logic_vector(num downto 0);               
signal VR_33				:    std_logic_vector(num downto 0);  
signal VR_34				:    std_logic_vector(num downto 0);  
signal VR_35				:    std_logic_vector(num downto 0);  
signal VR_36				:    std_logic_vector(num downto 0);  
signal VR_37				:    std_logic_vector(num downto 0);  	 
signal VR_38				:    std_logic_vector(num downto 0);               
signal VR_39				:    std_logic_vector(num downto 0);  
signal VR_40				:    std_logic_vector(num downto 0);  
signal VR_41				:    std_logic_vector(num downto 0);  
signal VR_42				:    std_logic_vector(num downto 0);  
signal VR_43				:    std_logic_vector(num downto 0);  
signal VR_44				:    std_logic_vector(num downto 0);  
signal VR_45				:    std_logic_vector(num downto 0);  
signal VR_46				:    std_logic_vector(num downto 0);  
signal VR_47				:    std_logic_vector(num downto 0); 
signal VR_48				:    std_logic_vector(num downto 0);               
signal VR_49				:    std_logic_vector(num downto 0);  
signal VR_50				:    std_logic_vector(num downto 0);  
signal VR_51				:    std_logic_vector(num downto 0);  
signal VR_52				:    std_logic_vector(num downto 0);  
signal VR_53				:    std_logic_vector(num downto 0);  	 
signal VR_54				:    std_logic_vector(num downto 0);               
signal VR_55				:    std_logic_vector(num downto 0);  
signal VR_56				:    std_logic_vector(num downto 0);  
signal VR_57				:    std_logic_vector(num downto 0);  
signal VR_58				:    std_logic_vector(num downto 0);  
signal VR_59				:    std_logic_vector(num downto 0);  
signal VR_60				:    std_logic_vector(num downto 0);  
signal VR_61				:    std_logic_vector(num downto 0);  
signal VR_62				:    std_logic_vector(num downto 0);  
signal VR_63				:    std_logic_vector(num downto 0);
 
-- HD
signal HD_0					:    std_logic_vector(num downto 0);               
signal HD_1					:    std_logic_vector(num downto 0);  
signal HD_2					:    std_logic_vector(num downto 0);  
signal HD_3					:    std_logic_vector(num downto 0);  
signal HD_4					:    std_logic_vector(num downto 0);  
signal HD_5					:    std_logic_vector(num downto 0);  	 
signal HD_6					:    std_logic_vector(num downto 0);               
signal HD_7					:    std_logic_vector(num downto 0);  
signal HD_8					:    std_logic_vector(num downto 0);  
signal HD_9					:    std_logic_vector(num downto 0);  
signal HD_10				:    std_logic_vector(num downto 0);  
signal HD_11				:    std_logic_vector(num downto 0);  
signal HD_12				:    std_logic_vector(num downto 0);  
signal HD_13				:    std_logic_vector(num downto 0);  
signal HD_14				:    std_logic_vector(num downto 0);  
signal HD_15				:    std_logic_vector(num downto 0);
signal HD_16				:    std_logic_vector(num downto 0);               
signal HD_17				:    std_logic_vector(num downto 0);  
signal HD_18				:    std_logic_vector(num downto 0);  
signal HD_19				:    std_logic_vector(num downto 0);  
signal HD_20				:    std_logic_vector(num downto 0);  
signal HD_21				:    std_logic_vector(num downto 0);  	 
signal HD_22				:    std_logic_vector(num downto 0);               
signal HD_23				:    std_logic_vector(num downto 0);  
signal HD_24				:    std_logic_vector(num downto 0);  
signal HD_25				:    std_logic_vector(num downto 0);  
signal HD_26				:    std_logic_vector(num downto 0);  
signal HD_27				:    std_logic_vector(num downto 0);  
signal HD_28				:    std_logic_vector(num downto 0);  
signal HD_29				:    std_logic_vector(num downto 0);  
signal HD_30				:    std_logic_vector(num downto 0);  
signal HD_31				:    std_logic_vector(num downto 0); 
signal HD_32				:    std_logic_vector(num downto 0);               
signal HD_33				:    std_logic_vector(num downto 0);  
signal HD_34				:    std_logic_vector(num downto 0);  
signal HD_35				:    std_logic_vector(num downto 0);  
signal HD_36				:    std_logic_vector(num downto 0);  
signal HD_37				:    std_logic_vector(num downto 0);  	 
signal HD_38				:    std_logic_vector(num downto 0);               
signal HD_39				:    std_logic_vector(num downto 0);  
signal HD_40				:    std_logic_vector(num downto 0);  
signal HD_41				:    std_logic_vector(num downto 0);  
signal HD_42				:    std_logic_vector(num downto 0);  
signal HD_43				:    std_logic_vector(num downto 0);  
signal HD_44				:    std_logic_vector(num downto 0);  
signal HD_45				:    std_logic_vector(num downto 0);  
signal HD_46				:    std_logic_vector(num downto 0);  
signal HD_47				:    std_logic_vector(num downto 0); 
signal HD_48				:    std_logic_vector(num downto 0);               
signal HD_49				:    std_logic_vector(num downto 0);  
signal HD_50				:    std_logic_vector(num downto 0);  
signal HD_51				:    std_logic_vector(num downto 0);  
signal HD_52				:    std_logic_vector(num downto 0);  
signal HD_53				:    std_logic_vector(num downto 0);  	 
signal HD_54				:    std_logic_vector(num downto 0);               
signal HD_55				:    std_logic_vector(num downto 0);  
signal HD_56				:    std_logic_vector(num downto 0);  
signal HD_57				:    std_logic_vector(num downto 0);  
signal HD_58				:    std_logic_vector(num downto 0);  
signal HD_59				:    std_logic_vector(num downto 0);  
signal HD_60				:    std_logic_vector(num downto 0);  
signal HD_61				:    std_logic_vector(num downto 0);  
signal HD_62				:    std_logic_vector(num downto 0);  
signal HD_63				:    std_logic_vector(num downto 0); 

-- VL
signal VL_0					:    std_logic_vector(num downto 0);               
signal VL_1					:    std_logic_vector(num downto 0);  
signal VL_2					:    std_logic_vector(num downto 0);  
signal VL_3					:    std_logic_vector(num downto 0);  
signal VL_4					:    std_logic_vector(num downto 0);  
signal VL_5					:    std_logic_vector(num downto 0);  	 
signal VL_6					:    std_logic_vector(num downto 0);               
signal VL_7					:    std_logic_vector(num downto 0);  
signal VL_8					:    std_logic_vector(num downto 0);  
signal VL_9					:    std_logic_vector(num downto 0);  
signal VL_10				:    std_logic_vector(num downto 0);  
signal VL_11				:    std_logic_vector(num downto 0);  
signal VL_12				:    std_logic_vector(num downto 0);  
signal VL_13				:    std_logic_vector(num downto 0);  
signal VL_14				:    std_logic_vector(num downto 0);  
signal VL_15				:    std_logic_vector(num downto 0);
signal VL_16				:    std_logic_vector(num downto 0);               
signal VL_17				:    std_logic_vector(num downto 0);  
signal VL_18				:    std_logic_vector(num downto 0);  
signal VL_19				:    std_logic_vector(num downto 0);  
signal VL_20				:    std_logic_vector(num downto 0);  
signal VL_21				:    std_logic_vector(num downto 0);  	 
signal VL_22				:    std_logic_vector(num downto 0);               
signal VL_23				:    std_logic_vector(num downto 0);  
signal VL_24				:    std_logic_vector(num downto 0);  
signal VL_25				:    std_logic_vector(num downto 0);  
signal VL_26				:    std_logic_vector(num downto 0);  
signal VL_27				:    std_logic_vector(num downto 0);  
signal VL_28				:    std_logic_vector(num downto 0);  
signal VL_29				:    std_logic_vector(num downto 0);  
signal VL_30				:    std_logic_vector(num downto 0);  
signal VL_31				:    std_logic_vector(num downto 0); 
signal VL_32				:    std_logic_vector(num downto 0);               
signal VL_33				:    std_logic_vector(num downto 0);  
signal VL_34				:    std_logic_vector(num downto 0);  
signal VL_35				:    std_logic_vector(num downto 0);  
signal VL_36				:    std_logic_vector(num downto 0);  
signal VL_37				:    std_logic_vector(num downto 0);  	 
signal VL_38				:    std_logic_vector(num downto 0);               
signal VL_39				:    std_logic_vector(num downto 0);  
signal VL_40				:    std_logic_vector(num downto 0);  
signal VL_41				:    std_logic_vector(num downto 0);  
signal VL_42				:    std_logic_vector(num downto 0);  
signal VL_43				:    std_logic_vector(num downto 0);  
signal VL_44				:    std_logic_vector(num downto 0);  
signal VL_45				:    std_logic_vector(num downto 0);  
signal VL_46				:    std_logic_vector(num downto 0);  
signal VL_47				:    std_logic_vector(num downto 0); 
signal VL_48				:    std_logic_vector(num downto 0);               
signal VL_49				:    std_logic_vector(num downto 0);  
signal VL_50				:    std_logic_vector(num downto 0);  
signal VL_51				:    std_logic_vector(num downto 0);  
signal VL_52				:    std_logic_vector(num downto 0);  
signal VL_53				:    std_logic_vector(num downto 0);  	 
signal VL_54				:    std_logic_vector(num downto 0);               
signal VL_55				:    std_logic_vector(num downto 0);  
signal VL_56				:    std_logic_vector(num downto 0);  
signal VL_57				:    std_logic_vector(num downto 0);  
signal VL_58				:    std_logic_vector(num downto 0);  
signal VL_59				:    std_logic_vector(num downto 0);  
signal VL_60				:    std_logic_vector(num downto 0);  
signal VL_61				:    std_logic_vector(num downto 0);  
signal VL_62				:    std_logic_vector(num downto 0);  
signal VL_63				:    std_logic_vector(num downto 0); 

-- HU
signal HU_0					:    std_logic_vector(num downto 0);               
signal HU_1					:    std_logic_vector(num downto 0);  
signal HU_2					:    std_logic_vector(num downto 0);  
signal HU_3					:    std_logic_vector(num downto 0);  
signal HU_4					:    std_logic_vector(num downto 0);  
signal HU_5					:    std_logic_vector(num downto 0);  	 
signal HU_6					:    std_logic_vector(num downto 0);               
signal HU_7					:    std_logic_vector(num downto 0);  
signal HU_8					:    std_logic_vector(num downto 0);  
signal HU_9					:    std_logic_vector(num downto 0);  
signal HU_10				:    std_logic_vector(num downto 0);  
signal HU_11				:    std_logic_vector(num downto 0);  
signal HU_12				:    std_logic_vector(num downto 0);  
signal HU_13				:    std_logic_vector(num downto 0);  
signal HU_14				:    std_logic_vector(num downto 0);  
signal HU_15				:    std_logic_vector(num downto 0);
signal HU_16				:    std_logic_vector(num downto 0);               
signal HU_17				:    std_logic_vector(num downto 0);  
signal HU_18				:    std_logic_vector(num downto 0);  
signal HU_19				:    std_logic_vector(num downto 0);  
signal HU_20				:    std_logic_vector(num downto 0);  
signal HU_21				:    std_logic_vector(num downto 0);  	 
signal HU_22				:    std_logic_vector(num downto 0);               
signal HU_23				:    std_logic_vector(num downto 0);  
signal HU_24				:    std_logic_vector(num downto 0);  
signal HU_25				:    std_logic_vector(num downto 0);  
signal HU_26				:    std_logic_vector(num downto 0);  
signal HU_27				:    std_logic_vector(num downto 0);  
signal HU_28				:    std_logic_vector(num downto 0);  
signal HU_29				:    std_logic_vector(num downto 0);  
signal HU_30				:    std_logic_vector(num downto 0);  
signal HU_31				:    std_logic_vector(num downto 0); 
signal HU_32				:    std_logic_vector(num downto 0);               
signal HU_33				:    std_logic_vector(num downto 0);  
signal HU_34				:    std_logic_vector(num downto 0);  
signal HU_35				:    std_logic_vector(num downto 0);  
signal HU_36				:    std_logic_vector(num downto 0);  
signal HU_37				:    std_logic_vector(num downto 0);  	 
signal HU_38				:    std_logic_vector(num downto 0);               
signal HU_39				:    std_logic_vector(num downto 0);  
signal HU_40				:    std_logic_vector(num downto 0);  
signal HU_41				:    std_logic_vector(num downto 0);  
signal HU_42				:    std_logic_vector(num downto 0);  
signal HU_43				:    std_logic_vector(num downto 0);  
signal HU_44				:    std_logic_vector(num downto 0);  
signal HU_45				:    std_logic_vector(num downto 0);  
signal HU_46				:    std_logic_vector(num downto 0);  
signal HU_47				:    std_logic_vector(num downto 0); 
signal HU_48				:    std_logic_vector(num downto 0);               
signal HU_49				:    std_logic_vector(num downto 0);  
signal HU_50				:    std_logic_vector(num downto 0);  
signal HU_51				:    std_logic_vector(num downto 0);  
signal HU_52				:    std_logic_vector(num downto 0);  
signal HU_53				:    std_logic_vector(num downto 0);  	 
signal HU_54				:    std_logic_vector(num downto 0);               
signal HU_55				:    std_logic_vector(num downto 0);  
signal HU_56				:    std_logic_vector(num downto 0);  
signal HU_57				:    std_logic_vector(num downto 0);  
signal HU_58				:    std_logic_vector(num downto 0);  
signal HU_59				:    std_logic_vector(num downto 0);  
signal HU_60				:    std_logic_vector(num downto 0);  
signal HU_61				:    std_logic_vector(num downto 0);  
signal HU_62				:    std_logic_vector(num downto 0);  
signal HU_63				:    std_logic_vector(num downto 0); 

-- Equations
signal eABC       : std_logic_vector(num+2 downto 0);
signal eBCD       : std_logic_vector(num+2 downto 0);
signal eCDE       : std_logic_vector(num+2 downto 0);
signal eDEF       : std_logic_vector(num+2 downto 0);
signal eEFG       : std_logic_vector(num+2 downto 0);
signal eFGH       : std_logic_vector(num+2 downto 0);
signal eGHI       : std_logic_vector(num+2 downto 0);
signal eHIJ       : std_logic_vector(num+2 downto 0);
signal eIJK       : std_logic_vector(num+2 downto 0);
signal eJKL       : std_logic_vector(num+2 downto 0);
signal eKLM       : std_logic_vector(num+2 downto 0);
signal eLMN       : std_logic_vector(num+2 downto 0);	
signal eMNO       : std_logic_vector(num+2 downto 0); 
signal eNOP       : std_logic_vector(num+2 downto 0);	
signal eO3P       : std_logic_vector(num+2 downto 0); 

signal eZAB       : std_logic_vector(num+2 downto 0);
signal eQZA       : std_logic_vector(num+2 downto 0);
signal eRQZ       : std_logic_vector(num+2 downto 0);
signal eSRQ       : std_logic_vector(num+2 downto 0);
signal eTSR       : std_logic_vector(num+2 downto 0);
signal eUTS       : std_logic_vector(num+2 downto 0);
signal eVUT       : std_logic_vector(num+2 downto 0);	
signal eWVU       : std_logic_vector(num+2 downto 0); 
signal eXWV       : std_logic_vector(num+2 downto 0);	
signal e3XW       : std_logic_vector(num+2 downto 0); 

signal eAB      : std_logic_vector(num+1 downto 0);
signal eBC      : std_logic_vector(num+1 downto 0);
signal eCD      : std_logic_vector(num+1 downto 0);
signal eDE      : std_logic_vector(num+1 downto 0);
signal eEF      : std_logic_vector(num+1 downto 0);
signal eFG      : std_logic_vector(num+1 downto 0);
signal eGH      : std_logic_vector(num+1 downto 0);
signal eHI      : std_logic_vector(num+1 downto 0);
signal eIJ      : std_logic_vector(num+1 downto 0);
signal eJK      : std_logic_vector(num+1 downto 0);
signal eKL      : std_logic_vector(num+1 downto 0);

signal eZA      : std_logic_vector(num+1 downto 0);
signal eQZ      : std_logic_vector(num+1 downto 0);
signal eRQ      : std_logic_vector(num+1 downto 0);
signal eSR      : std_logic_vector(num+1 downto 0);
signal eTS      : std_logic_vector(num+1 downto 0);
signal eUT      : std_logic_vector(num+1 downto 0);
signal eVU      : std_logic_vector(num+1 downto 0);
signal eWV      : std_logic_vector(num+1 downto 0);
signal eXW      : std_logic_vector(num+1 downto 0);

signal eLM      : std_logic_vector(num+1 downto 0);
signal eMN      : std_logic_vector(num+1 downto 0);
signal eNO      : std_logic_vector(num+1 downto 0);
signal eOP      : std_logic_vector(num+1 downto 0);

--DIFF SAD REG START
signal Diff0_V    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_V    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_V   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_V   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_V   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_V   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_V   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_V    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_V    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_V   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_V   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_V   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_V   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_V   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_V_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_V_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_V_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_V_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_V_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_V_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_V_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_H    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_H    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_H   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_H   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_H   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_H   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_H   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_H    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_H    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_H   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_H   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_H   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_H   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_H   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_H_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_H_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_H_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_H_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_H_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_H_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_H_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_DC    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_DC   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_DC    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DC    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DC   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DC   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DC   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DC   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DC   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DC_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DC_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_DDL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_DDL   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_DDL    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DDL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DDL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DDL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DDL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DDL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DDL   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DDL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DDL_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_DDR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_DDR   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_DDR    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DDR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DDR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DDR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DDR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DDR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DDR   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_DDR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_DDR_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_VR    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_VR   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_VR    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_VR    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_VR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_VR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_VR   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_VR   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_VR   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_VR_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_VR_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_HU    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_HU   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_HU    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_HU    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_HU   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_HU   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_HU   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_HU   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_HU   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_HU_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_HU_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_VL    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_VL   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_VL    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_VL    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_VL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_VL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_VL   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_VL   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_VL   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_VL_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_VL_reg   : std_logic_vector(num downto 0):=(others=>'0');

signal Diff0_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff1_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff2_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff3_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff4_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff5_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff6_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff7_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff8_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff9_HD    : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff10_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff11_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff12_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff13_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff14_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff15_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff16_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff17_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff18_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff19_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff20_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff21_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff22_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff23_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff24_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff25_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff26_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff27_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff28_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff29_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff30_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff31_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff32_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff33_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff34_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff35_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff36_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff37_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff38_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff39_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff40_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff41_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff42_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff43_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff44_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff45_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff46_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff47_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff48_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');	  
signal Diff49_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff50_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff51_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff52_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff53_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff54_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff55_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff56_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff57_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff58_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff59_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff60_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff61_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff62_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');
signal Diff63_HD   : std_logic_vector(num+1 downto 0):=(others=>'0');

signal SAD0_HD    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_HD    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_HD   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_HD   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_HD   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_HD   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_HD   : std_logic_vector(num downto 0):=(others=>'0');	   

signal SAD0_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD1_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD2_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD3_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD4_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD5_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD6_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD7_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD8_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD9_HD_reg    : std_logic_vector(num downto 0):=(others=>'0');
signal SAD10_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD11_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD12_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD13_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD14_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD15_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD16_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD17_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD18_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD19_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD20_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD21_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD22_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD23_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD24_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD25_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD26_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD27_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD28_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD29_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD30_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD31_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD32_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD33_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD34_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD35_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD36_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD37_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD38_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD39_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD40_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD41_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD42_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD43_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD44_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD45_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD46_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD47_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD48_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');	  
signal SAD49_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD50_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD51_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD52_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD53_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD54_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD55_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD56_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD57_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD58_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD59_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD60_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD61_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD62_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
signal SAD63_HD_reg   : std_logic_vector(num downto 0):=(others=>'0');
--DIFF SAD REG END

signal SAD_V_quart0        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart0        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart0      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart0      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart0       :	std_logic_vector(num+4 downto 0):=(others=>'0');		   
														
signal SAD_V_quart_reg0        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart_reg0        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart_reg0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart_reg0      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart_reg0      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart_reg0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart_reg0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart_reg0       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart_reg0       :	std_logic_vector(num+4 downto 0):=(others=>'0');

signal SAD_V_quart1        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart1        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart1      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart1      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart1       :	std_logic_vector(num+4 downto 0):=(others=>'0');		   
														
signal SAD_V_quart_reg1        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart_reg1        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart_reg1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart_reg1      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart_reg1      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart_reg1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart_reg1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart_reg1       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart_reg1       :	std_logic_vector(num+4 downto 0):=(others=>'0');

signal SAD_V_quart2        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart2        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart2      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart2      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart2       :	std_logic_vector(num+4 downto 0):=(others=>'0');		   
														
signal SAD_V_quart_reg2        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart_reg2        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart_reg2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart_reg2      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart_reg2      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart_reg2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart_reg2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart_reg2       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart_reg2       :	std_logic_vector(num+4 downto 0):=(others=>'0');

signal SAD_V_quart3        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart3        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart3      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart3      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart3       :	std_logic_vector(num+4 downto 0):=(others=>'0');		   
														
signal SAD_V_quart_reg3        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_H_quart_reg3        :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DC_quart_reg3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDL_quart_reg3      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_DDR_quart_reg3      :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VR_quart_reg3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HD_quart_reg3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_VL_quart_reg3       :	std_logic_vector(num+4 downto 0):=(others=>'0');
signal SAD_HU_quart_reg3       :	std_logic_vector(num+4 downto 0):=(others=>'0');

signal SAD_V		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_H		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DC		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DDL		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DDR		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_VR		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_HD		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_VL		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_HU		: std_logic_vector(num+6 downto 0) := (others => '0');

signal SAD_V_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_H_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DC_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DDL_reg	: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_DDR_reg	: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_VR_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_HD_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_VL_reg		: std_logic_vector(num+6 downto 0) := (others => '0');
signal SAD_HU_reg		: std_logic_vector(num+6 downto 0) := (others => '0');

SIGNAL COMP0            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL COMP1            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL COMP2            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL COMP3            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL COMP4            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL COMP5            :  std_logic_vector(num+6 downto 0):=(others=>'0');		
SIGNAL COMP6            :  std_logic_vector(num+6 downto 0):=(others=>'0');
SIGNAL MINIMUM          :  std_logic_vector(num+6 downto 0):=(others=>'0');	


type reg_type is array (0 to 63) of std_logic_vector(num downto 0); 
type reg_type2 is array (0 to 63) of std_logic_vector(num+1 downto 0); 
--9*4*8 bit for intra output
type reg_type3 is array (0 to 35) of std_logic_vector(num downto 0);

signal V_reg	: reg_type;
signal H_reg	: reg_type;
signal DC_reg	: std_logic_vector(num downto 0);
signal DDL_reg	: reg_type;
signal DDR_reg	: reg_type;
signal VL_reg	: reg_type;
signal HD_reg	: reg_type;
signal VR_reg	: reg_type;
signal HU_reg	: reg_type;


signal Diff_regV 		: reg_type2;
signal Diff_regH 		: reg_type2;
signal Diff_regDC 	: reg_type2;
signal Diff_regDDL 	: reg_type2;
signal Diff_regDDR 	: reg_type2;
signal Diff_regVL 	: reg_type2;
signal Diff_regHD 	: reg_type2;
signal Diff_regVR 	: reg_type2;
signal Diff_regHU 	: reg_type2;


signal Intra_out_reg : reg_type; --64*8 bit
signal Diff_out_reg  : reg_type2; --64*9 bit

-------------------------------------------------------------------------------------------------------------------------
--#####################################################################################################################--
--# Begin (finally..)
--#####################################################################################################################--
-------------------------------------------------------------------------------------------------------------------------


begin

-- store input original pixels in correct registers based on quart (0-3)
process (clk, rst)
begin
	if (rst = '1') then
		ycbcr0_reg  <= (others=>'0'); ycbcr1_reg  <= (others=>'0'); ycbcr2_reg  <= (others=>'0'); ycbcr3_reg  <= (others=>'0');
		ycbcr4_reg  <= (others=>'0'); ycbcr5_reg  <= (others=>'0'); ycbcr6_reg  <= (others=>'0'); ycbcr7_reg  <= (others=>'0');
		ycbcr8_reg  <= (others=>'0'); ycbcr9_reg  <= (others=>'0'); ycbcr10_reg <= (others=>'0'); ycbcr11_reg <= (others=>'0');
		ycbcr12_reg <= (others=>'0'); ycbcr13_reg <= (others=>'0'); ycbcr14_reg <= (others=>'0'); ycbcr15_reg <= (others=>'0');
		ycbcr16_reg <= (others=>'0'); ycbcr17_reg <= (others=>'0'); ycbcr18_reg <= (others=>'0'); ycbcr19_reg <= (others=>'0');
		ycbcr20_reg <= (others=>'0'); ycbcr21_reg <= (others=>'0'); ycbcr22_reg <= (others=>'0'); ycbcr23_reg <= (others=>'0');
		ycbcr24_reg <= (others=>'0'); ycbcr25_reg <= (others=>'0'); ycbcr26_reg <= (others=>'0'); ycbcr27_reg <= (others=>'0');
		ycbcr28_reg <= (others=>'0'); ycbcr29_reg <= (others=>'0'); ycbcr30_reg <= (others=>'0'); ycbcr31_reg <= (others=>'0');
		ycbcr32_reg <= (others=>'0'); ycbcr33_reg <= (others=>'0'); ycbcr34_reg <= (others=>'0'); ycbcr35_reg <= (others=>'0');
		-- gap registers not needed: 36-39, 44-47, 52-55, 60-63
		ycbcr40_reg <= (others=>'0'); ycbcr41_reg <= (others=>'0'); ycbcr42_reg <= (others=>'0'); ycbcr43_reg <= (others=>'0');		
		ycbcr48_reg <= (others=>'0'); ycbcr49_reg <= (others=>'0'); ycbcr50_reg <= (others=>'0'); ycbcr51_reg <= (others=>'0');
		ycbcr56_reg <= (others=>'0'); ycbcr57_reg <= (others=>'0'); ycbcr58_reg <= (others=>'0'); ycbcr59_reg <= (others=>'0');	
		
	elsif (clk'event and clk = '1') then
	-- 4x4 input, to 8x8 storage	
		if (quart = "00") and (rdy_in_start = '1') then		
			ycbcr0_reg  <=  ycbcr0;   ycbcr1_reg  <=  ycbcr1;	ycbcr2_reg  <=  ycbcr2;   ycbcr3_reg  <=  ycbcr3;    	  
			ycbcr8_reg  <=  ycbcr4;   ycbcr9_reg  <=  ycbcr5;  ycbcr10_reg <=  ycbcr6;   ycbcr11_reg <=  ycbcr7;  
			ycbcr16_reg <=  ycbcr8;   ycbcr17_reg <=  ycbcr9;  ycbcr18_reg <= ycbcr10;   ycbcr19_reg <= ycbcr11;  
			ycbcr24_reg <= ycbcr12;   ycbcr25_reg <= ycbcr13;  ycbcr26_reg <= ycbcr14;   ycbcr27_reg <= ycbcr15; 
		
		elsif (quart = "01") then		
			ycbcr4_reg  <=  ycbcr0;   ycbcr5_reg  <=  ycbcr1;	ycbcr6_reg  <=  ycbcr2;   ycbcr7_reg  <=  ycbcr3;    	  
			ycbcr12_reg <=  ycbcr4;   ycbcr13_reg <=  ycbcr5;  ycbcr14_reg <=  ycbcr6;   ycbcr15_reg <=  ycbcr7;  
			ycbcr20_reg <=  ycbcr8;   ycbcr21_reg <=  ycbcr9;  ycbcr22_reg <= ycbcr10;   ycbcr23_reg <= ycbcr11;  
			ycbcr28_reg <= ycbcr12;   ycbcr29_reg <= ycbcr13;  ycbcr30_reg <= ycbcr14;   ycbcr31_reg <= ycbcr15; 	
	
		elsif (quart = "10") then	
			ycbcr32_reg <=  ycbcr0;   ycbcr33_reg <=  ycbcr1;	ycbcr34_reg <=  ycbcr2;   ycbcr35_reg <=  ycbcr3;    	  
			ycbcr40_reg <=  ycbcr4;   ycbcr41_reg <=  ycbcr5;  ycbcr42_reg <=  ycbcr6;   ycbcr43_reg <=  ycbcr7;  
			ycbcr48_reg <=  ycbcr8;   ycbcr49_reg <=  ycbcr9;  ycbcr50_reg <= ycbcr10;   ycbcr51_reg <= ycbcr11;  
			ycbcr56_reg <= ycbcr12;   ycbcr57_reg <= ycbcr13;  ycbcr58_reg <= ycbcr14;   ycbcr59_reg <= ycbcr15; 
	
		end if;
	end if;
end process;		

-- Timing logic process
process (clk, rst)
begin
	if (rst = '1') then
		rdy_in_1 <= '0';
		rdy_in_2 <= '0';
		rdy_in_3 <= '0';
		rdy_in_4 <= '0';
		rdy_in_5 <= '0';
		rdy_in_6 <= '0';
		rdy_in_7 <= '0';
		rdy_in_8 <= '0';
		rdy_in_9 <= '0';
		rdy_in_10 <= '0';
	
	elsif (clk'event and clk = '1') then
		rdy_in_1 <= rdy_in_start;
		rdy_in_2 <= rdy_in_1;
		rdy_in_3 <= rdy_in_2;
		rdy_in_4 <= rdy_in_3;
		rdy_in_5 <= rdy_in_4;
		rdy_in_6 <= rdy_in_5;
		rdy_in_7 <= rdy_in_6;
		rdy_in_8 <= rdy_in_7;
		rdy_in_9 <= rdy_in_8;
		rdy_in_10 <= rdy_in_9;		
	end if;
end process;

-- Registers to store first row/col information
process (clk, rst)
begin
	if (rst = '1') then
		first_row_reg <= '0';
		first_row_reg2 <= '0';	
		first_col_reg <= '0';
		first_col_reg2 <= '0';
		
	elsif (clk'event and clk = '1') then
		if (rdy_in_start = '1') then
			first_row_reg <= first_row;
			first_col_reg <= first_col;
		elsif (rdy_in_3 = '1') then
			first_row_reg2 <= first_row_reg;
			first_col_reg2 <= first_col_reg;
		end if;	
	end if;
end process;	

-- Assign equations
	-- for these, use eAB(num+1 downto 1) to shift left by 1/divide by 2
eABC <= (("0" & eAB(num+1 downto 0)) + ("0" & eBC(num+1 downto 0)));
eBCD <= (("0" & eBC(num+1 downto 0)) + ("0" & eCD(num+1 downto 0)));
eCDE <= (("0" & eCD(num+1 downto 0)) + ("0" & eDE(num+1 downto 0)));
eDEF <= (("0" & eDE(num+1 downto 0)) + ("0" & eEF(num+1 downto 0)));
eEFG <= (("0" & eEF(num+1 downto 0)) + ("0" & eFG(num+1 downto 0)));
eFGH <= (("0" & eFG(num+1 downto 0)) + ("0" & eGH(num+1 downto 0)));
eGHI <= (("0" & eGH(num+1 downto 0)) + ("0" & eHI(num+1 downto 0)));
eHIJ <= (("0" & eHI(num+1 downto 0)) + ("0" & eIJ(num+1 downto 0)));
eIJK <= (("0" & eIJ(num+1 downto 0)) + ("0" & eJK(num+1 downto 0)));
eJKL <= (("0" & eJK(num+1 downto 0)) + ("0" & eKL(num+1 downto 0)));
eKLM <= (("0" & eKL(num+1 downto 0)) + ("0" & eLM(num+1 downto 0)));
eLMN <= (("0" & eLM(num+1 downto 0)) + ("0" & eMN(num+1 downto 0)));
eMNO <= (("0" & eMN(num+1 downto 0)) + ("0" & eNO(num+1 downto 0)));
eNOP <= (("0" & eNO(num+1 downto 0)) + ("0" & eOP(num+1 downto 0)));
eO3P <= (("0" & P(num downto 0)&"0") + ("0" & eOP(num+1 downto 0))+1);

eZAB <= (("0" & eZA(num+1 downto 0)) + ("0" & eAB(num+1 downto 0)));
eQZA <= (("0" & eQZ(num+1 downto 0)) + ("0" & eZA(num+1 downto 0)));
eRQZ <= (("0" & eRQ(num+1 downto 0)) + ("0" & eQZ(num+1 downto 0)));
eSRQ <= (("0" & eSR(num+1 downto 0)) + ("0" & eRQ(num+1 downto 0)));
eTSR <= (("0" & eTS(num+1 downto 0)) + ("0" & eSR(num+1 downto 0)));
eUTS <= (("0" & eUT(num+1 downto 0)) + ("0" & eTS(num+1 downto 0)));
eVUT <= (("0" & eVU(num+1 downto 0)) + ("0" & eUT(num+1 downto 0)));
eWVU <= (("0" & eWV(num+1 downto 0)) + ("0" & eVU(num+1 downto 0)));
eXWV <= (("0" & eXW(num+1 downto 0)) + ("0" & eWV(num+1 downto 0)));
e3XW <= (("0" & X(num downto 0)&"0") + ("0" & eXW(num+1 downto 0))+1);

eAB <= (("0" & A(num downto 0)) + ("0" & B(num downto 0)) + 1); 
eBC <= (("0" & B(num downto 0)) + ("0" & C(num downto 0)) + 1);
eCD <= (("0" & C(num downto 0)) + ("0" & D(num downto 0)) + 1);
eDE <= (("0" & D(num downto 0)) + ("0" & E(num downto 0)) + 1);
eEF <= (("0" & E(num downto 0)) + ("0" & F(num downto 0)) + 1);
eFG <= (("0" & F(num downto 0)) + ("0" & G(num downto 0)) + 1);
eGH <= (("0" & G(num downto 0)) + ("0" & H(num downto 0)) + 1);
eHI <= (("0" & H(num downto 0)) + ("0" & I(num downto 0)) + 1);
eIJ <= (("0" & I(num downto 0)) + ("0" & J(num downto 0)) + 1);
eJK <= (("0" & J(num downto 0)) + ("0" & K(num downto 0)) + 1);
eKL <= (("0" & K(num downto 0)) + ("0" & L(num downto 0)) + 1);

eZA <= (("0" & Z(num downto 0)) + ("0" & A(num downto 0)) + 1);
eQZ <= (("0" & Q(num downto 0)) + ("0" & Z(num downto 0)) + 1);
eRQ <= (("0" & R(num downto 0)) + ("0" & Q(num downto 0)) + 1);
eSR <= (("0" & S(num downto 0)) + ("0" & R(num downto 0)) + 1);
eTS <= (("0" & T(num downto 0)) + ("0" & S(num downto 0)) + 1);
eUT <= (("0" & U(num downto 0)) + ("0" & T(num downto 0)) + 1);
eVU <= (("0" & V(num downto 0)) + ("0" & U(num downto 0)) + 1);
eWV <= (("0" & W(num downto 0)) + ("0" & V(num downto 0)) + 1);
eXW <= (("0" & X(num downto 0)) + ("0" & W(num downto 0)) + 1);

-- these 4 are only needed to make 3 tap filter equations..
eLM <= (("0" & L(num downto 0)) + ("0" & M(num downto 0)) + 1);
eMN <= (("0" & M(num downto 0)) + ("0" & N(num downto 0)) + 1);
eNO <= (("0" & N(num downto 0)) + ("0" & O(num downto 0)) + 1);
eOP <= (("0" & O(num downto 0)) + ("0" & P(num downto 0)) + 1);


-------------------------------------------------------------------------------------
--###################################################################################
--# Make Predictions
--###################################################################################
-------------------------------------------------------------------------------------


-- Assign vertical predictions
V_0 <= A; V_8  <= A; V_16 <= A; V_24 <= A; V_32 <= A; V_40 <= A; V_48 <= A; V_56 <= A; 
V_1 <= B; V_9  <= B; V_17 <= B; V_25 <= B; V_33 <= B; V_41 <= B; V_49 <= B; V_57 <= B; 
V_2 <= C; V_10 <= C; V_18 <= C; V_26 <= C; V_34 <= C; V_42 <= C; V_50 <= C; V_58 <= C; 
V_3 <= D; V_11 <= D; V_19 <= D; V_27 <= D; V_35 <= D; V_43 <= D; V_51 <= D; V_59 <= D; 
V_4 <= E; V_12 <= E; V_20 <= E; V_28 <= E; V_36 <= E; V_44 <= E; V_52 <= E; V_60 <= E; 
V_5 <= F; V_13 <= F; V_21 <= F; V_29 <= F; V_37 <= F; V_45 <= F; V_53 <= F; V_61 <= F; 
V_6 <= G; V_14 <= G; V_22 <= G; V_30 <= G; V_38 <= G; V_46 <= G; V_54 <= G; V_62 <= G; 
V_7 <= H; V_15 <= H; V_23 <= H; V_31 <= H; V_39 <= H; V_47 <= H; V_55 <= H; V_63 <= H; 

-- Assign horizontal predictions
H_0  <= Q; H_1  <= Q; H_2  <= Q; H_3  <= Q; H_4  <= Q; H_5  <= Q; H_6  <= Q; H_7  <= Q; 
H_8  <= R; H_9  <= R; H_10 <= R; H_11 <= R; H_12 <= R; H_13 <= R; H_14 <= R; H_15 <= R; 
H_16 <= S; H_17 <= S; H_18 <= S; H_19 <= S; H_20 <= S; H_21 <= S; H_22 <= S; H_23 <= S;
H_24 <= T; H_25 <= T; H_26 <= T; H_27 <= T; H_28 <= T; H_29 <= T; H_30 <= T; H_31 <= T;
H_32 <= U; H_33 <= U; H_34 <= U; H_35 <= U; H_36 <= U; H_37 <= U; H_38 <= U; H_39 <= U;   
H_40 <= V; H_41 <= V; H_42 <= V; H_43 <= V; H_44 <= V; H_45 <= V; H_46 <= V; H_47 <= V; 
H_48 <= W; H_49 <= W; H_50 <= W; H_51 <= W; H_52 <= W; H_53 <= W; H_54 <= W; H_55 <= W; 
H_56 <= X; H_57 <= X; H_58 <= X; H_59 <= X; H_60 <= X; H_61 <= X; H_62 <= X; H_63 <= X; 

-- Assign DC
sum_DC_TOTAL <=  (("0000" & A(num downto 0)) + ("0000" & Q(num downto 0)) +
						("0000" & B(num downto 0)) + ("0000" & R(num downto 0)) +
						("0000" & C(num downto 0)) + ("0000" & S(num downto 0)) +
						("0000" & D(num downto 0)) + ("0000" & T(num downto 0)) +
						("0000" & E(num downto 0)) + ("0000" & U(num downto 0)) +
						("0000" & F(num downto 0)) + ("0000" & V(num downto 0)) +
						("0000" & G(num downto 0)) + ("0000" & W(num downto 0)) +
						("0000" & H(num downto 0)) + ("0000" & X(num downto 0)) );
						
						
sum_DC <= sum_DC_TOTAL(num+4 downto 0) + "000000000010" when  
				((first_row='0' and first_col='1') or (first_row='1' and first_col='0')) else 
			 sum_DC_TOTAL(num+4 downto 0) + "000000000100" when  
				(first_row='0' and first_col='0') else 
			 (others=>'1');
DC <= sum_DC(num+3 downto 3) when  
			((first_row='0' and first_col='1') or (first_row='1' and first_col='0')) else 
		sum_DC(num+4 downto 4) when  
			(first_row='0' and first_col='0') else 
		(others=>'0');	
						
-- Assign DDL
DDL_0  <= eABC(num+2 downto 2);
DDL_1  <= eBCD(num+2 downto 2); DDL_8  <= eBCD(num+2 downto 2);
DDL_2  <= eCDE(num+2 downto 2); DDL_9  <= eCDE(num+2 downto 2); DDL_16 <= eCDE(num+2 downto 2); 
DDL_3  <= eDEF(num+2 downto 2); DDL_10 <= eDEF(num+2 downto 2); DDL_17 <= eDEF(num+2 downto 2); DDL_24 <= eDEF(num+2 downto 2);
DDL_4  <= eEFG(num+2 downto 2); DDL_11 <= eEFG(num+2 downto 2); DDL_18 <= eEFG(num+2 downto 2); DDL_25 <= eEFG(num+2 downto 2); DDL_32 <= eEFG(num+2 downto 2); 
DDL_5  <= eFGH(num+2 downto 2); DDL_12 <= eFGH(num+2 downto 2); DDL_19 <= eFGH(num+2 downto 2); DDL_26 <= eFGH(num+2 downto 2); DDL_33 <= eFGH(num+2 downto 2); DDL_40 <= eFGH(num+2 downto 2);
DDL_6  <= eGHI(num+2 downto 2); DDL_13 <= eGHI(num+2 downto 2); DDL_20 <= eGHI(num+2 downto 2); DDL_27 <= eGHI(num+2 downto 2); DDL_34 <= eGHI(num+2 downto 2); DDL_41 <= eGHI(num+2 downto 2); DDL_48 <= eGHI(num+2 downto 2); 
DDL_7  <= eHIJ(num+2 downto 2); DDL_14 <= eHIJ(num+2 downto 2); DDL_21 <= eHIJ(num+2 downto 2); DDL_28 <= eHIJ(num+2 downto 2); DDL_35 <= eHIJ(num+2 downto 2); DDL_42 <= eHIJ(num+2 downto 2); DDL_49 <= eHIJ(num+2 downto 2); DDL_56 <= eHIJ(num+2 downto 2); 
DDL_15 <= eIJK(num+2 downto 2); DDL_22 <= eIJK(num+2 downto 2); DDL_29 <= eIJK(num+2 downto 2); DDL_36 <= eIJK(num+2 downto 2); DDL_43 <= eIJK(num+2 downto 2); DDL_50 <= eIJK(num+2 downto 2); DDL_57 <= eIJK(num+2 downto 2);
DDL_23 <= eJKL(num+2 downto 2); DDL_30 <= eJKL(num+2 downto 2); DDL_37 <= eJKL(num+2 downto 2); DDL_44 <= eJKL(num+2 downto 2); DDL_51 <= eJKL(num+2 downto 2); DDL_58 <= eJKL(num+2 downto 2);
DDL_31 <= eKLM(num+2 downto 2); DDL_38 <= eKLM(num+2 downto 2); DDL_45 <= eKLM(num+2 downto 2); DDL_52 <= eKLM(num+2 downto 2); DDL_59 <= eKLM(num+2 downto 2);
DDL_39 <= eLMN(num+2 downto 2); DDL_46 <= eLMN(num+2 downto 2); DDL_53 <= eLMN(num+2 downto 2); DDL_60 <= eLMN(num+2 downto 2);
DDL_47 <= eMNO(num+2 downto 2); DDL_54 <= eMNO(num+2 downto 2); DDL_61 <= eMNO(num+2 downto 2);
DDL_55 <= eNOP(num+2 downto 2); DDL_62 <= eNOP(num+2 downto 2);
DDL_63 <= eO3P(num+2 downto 2);

--Assign DDR
-- Mode 4 - Diagonal Down Right
DDR_7  <= eFGH(num+2 downto 2);
DDR_6  <= eEFG(num+2 downto 2); DDR_15 <= eEFG(num+2 downto 2);
DDR_5  <= eDEF(num+2 downto 2); DDR_14 <= eDEF(num+2 downto 2); DDR_23 <= eDEF(num+2 downto 2); 
DDR_4  <= eCDE(num+2 downto 2); DDR_13 <= eCDE(num+2 downto 2); DDR_22 <= eCDE(num+2 downto 2); DDR_31 <= eCDE(num+2 downto 2);
DDR_3  <= eBCD(num+2 downto 2); DDR_12 <= eBCD(num+2 downto 2); DDR_21 <= eBCD(num+2 downto 2); DDR_30 <= eBCD(num+2 downto 2); DDR_39 <= eBCD(num+2 downto 2); 
DDR_2  <= eABC(num+2 downto 2); DDR_11 <= eABC(num+2 downto 2); DDR_20 <= eABC(num+2 downto 2); DDR_29 <= eABC(num+2 downto 2); DDR_38 <= eABC(num+2 downto 2); DDR_47 <= eABC(num+2 downto 2);
DDR_1  <= eZAB(num+2 downto 2); DDR_10 <= eZAB(num+2 downto 2); DDR_19 <= eZAB(num+2 downto 2); DDR_28 <= eZAB(num+2 downto 2); DDR_37 <= eZAB(num+2 downto 2); DDR_46 <= eZAB(num+2 downto 2); DDR_55 <= eZAB(num+2 downto 2); 
DDR_0  <= eQZA(num+2 downto 2); DDR_9  <= eQZA(num+2 downto 2); DDR_18 <= eQZA(num+2 downto 2); DDR_27 <= eQZA(num+2 downto 2); DDR_36 <= eQZA(num+2 downto 2); DDR_45 <= eQZA(num+2 downto 2); DDR_54 <= eQZA(num+2 downto 2); DDR_63 <= eQZA(num+2 downto 2); 
DDR_8  <= eRQZ(num+2 downto 2); DDR_17 <= eRQZ(num+2 downto 2); DDR_26 <= eRQZ(num+2 downto 2); DDR_35 <= eRQZ(num+2 downto 2); DDR_44 <= eRQZ(num+2 downto 2); DDR_53 <= eRQZ(num+2 downto 2); DDR_62 <= eRQZ(num+2 downto 2);
DDR_16 <= eSRQ(num+2 downto 2); DDR_25 <= eSRQ(num+2 downto 2); DDR_34 <= eSRQ(num+2 downto 2); DDR_43 <= eSRQ(num+2 downto 2); DDR_52 <= eSRQ(num+2 downto 2); DDR_61 <= eSRQ(num+2 downto 2);
DDR_24 <= eTSR(num+2 downto 2); DDR_33 <= eTSR(num+2 downto 2); DDR_42 <= eTSR(num+2 downto 2); DDR_51 <= eTSR(num+2 downto 2); DDR_60 <= eTSR(num+2 downto 2);
DDR_32 <= eUTS(num+2 downto 2); DDR_41 <= eUTS(num+2 downto 2); DDR_50 <= eUTS(num+2 downto 2); DDR_59 <= eUTS(num+2 downto 2);
DDR_40 <= eVUT(num+2 downto 2); DDR_49 <= eVUT(num+2 downto 2); DDR_58 <= eVUT(num+2 downto 2);
DDR_48 <= eWVU(num+2 downto 2); DDR_57 <= eWVU(num+2 downto 2);
DDR_56 <= eXWV(num+2 downto 2);

-- Mode 5 - Vertical Right
VR_0  <= eZA(num+1 downto 1);  VR_1  <= eAB(num+1 downto 1);  VR_2  <= eBC(num+1 downto 1);  VR_3  <= eCD(num+1 downto 1);  VR_4  <= eDE(num+1 downto 1);  VR_5  <= eEF(num+1 downto 1);  VR_6  <= eFG(num+1 downto 1);  VR_7  <= eGH(num+1 downto 1);
VR_8  <= eQZA(num+2 downto 2); VR_9  <= eZAB(num+2 downto 2); VR_10 <= eABC(num+2 downto 2); VR_11 <= eBCD(num+2 downto 2); VR_12 <= eCDE(num+2 downto 2); VR_13 <= eDEF(num+2 downto 2); VR_14 <= eEFG(num+2 downto 2); VR_15 <= eFGH(num+2 downto 2); 
VR_16 <= eRQZ(num+2 downto 2); VR_17 <= eZA(num+1 downto 1);  VR_18 <= eAB(num+1 downto 1);  VR_19 <= eBC(num+1 downto 1);  VR_20 <= eCD(num+1 downto 1);  VR_21 <= eDE(num+1 downto 1);  VR_22 <= eEF(num+1 downto 1);  VR_23 <= eFG(num+1 downto 1);
VR_24 <= eSRQ(num+2 downto 2); VR_25 <= eQZA(num+2 downto 2); VR_26 <= eZAB(num+2 downto 2); VR_27 <= eABC(num+2 downto 2); VR_28 <= eBCD(num+2 downto 2); VR_29 <= eCDE(num+2 downto 2); VR_30 <= eDEF(num+2 downto 2); VR_31 <= eEFG(num+2 downto 2);
VR_32 <= eTSR(num+2 downto 2); VR_33 <= eRQZ(num+2 downto 2); VR_34 <= eZA(num+1 downto 1);  VR_35 <= eAB(num+1 downto 1);  VR_36 <= eBC(num+1 downto 1);  VR_37 <= eCD(num+1 downto 1);  VR_38 <= eDE(num+1 downto 1);  VR_39 <= eEF(num+1 downto 1);
VR_40 <= eUTS(num+2 downto 2); VR_41 <= eSRQ(num+2 downto 2); VR_42 <= eQZA(num+2 downto 2); VR_43 <= eZAB(num+2 downto 2); VR_44 <= eABC(num+2 downto 2); VR_45 <= eBCD(num+2 downto 2); VR_46 <= eCDE(num+2 downto 2); VR_47 <= eDEF(num+2 downto 2);
VR_48 <= eVUT(num+2 downto 2); VR_49 <= eTSR(num+2 downto 2); VR_50 <= eRQZ(num+2 downto 2); VR_51 <= eZA(num+1 downto 1);  VR_52 <= eAB(num+1 downto 1);  VR_53 <= eBC(num+1 downto 1);  VR_54 <= eCD(num+1 downto 1);  VR_55 <= eDE(num+1 downto 1);
VR_56 <= eWVU(num+2 downto 2); VR_57 <= eUTS(num+2 downto 2); VR_58 <= eSRQ(num+2 downto 2); VR_59 <= eQZA(num+2 downto 2); VR_60 <= eZAB(num+2 downto 2); VR_61 <= eABC(num+2 downto 2); VR_62 <= eBCD(num+2 downto 2); VR_63 <= eCDE(num+2 downto 2); 

-- Mode 6 - Horizontal Down
HD_0  <= eQZ(num+1 downto 1); HD_1  <= eQZA(num+2 downto 2); HD_2  <= eZAB(num+2 downto 2);HD_3  <= eABC(num+2 downto 2); HD_4  <= eBCD(num+2 downto 2);HD_5  <= eCDE(num+2 downto 2); HD_6  <= eDEF(num+2 downto 2);HD_7  <= eEFG(num+2 downto 2);
HD_8  <= eRQ(num+1 downto 1); HD_9  <= eRQZ(num+2 downto 2); HD_10 <= eQZ(num+1 downto 1); HD_11 <= eQZA(num+2 downto 2); HD_12 <= eZAB(num+2 downto 2);HD_13 <= eABC(num+2 downto 2); HD_14 <= eBCD(num+2 downto 2);HD_15 <= eCDE(num+2 downto 2);
HD_16 <= eSR(num+1 downto 1); HD_17 <= eSRQ(num+2 downto 2); HD_18 <= eRQ(num+1 downto 1); HD_19 <= eRQZ(num+2 downto 2); HD_20 <= eQZ(num+1 downto 1); HD_21 <= eQZA(num+2 downto 2); HD_22 <= eZAB(num+2 downto 2);HD_23 <= eABC(num+2 downto 2);
HD_24 <= eTS(num+1 downto 1); HD_25 <= eTSR(num+2 downto 2); HD_26 <= eSR(num+1 downto 1); HD_27 <= eSRQ(num+2 downto 2); HD_28 <= eRQ(num+1 downto 1); HD_29 <= eRQZ(num+2 downto 2); HD_30 <= eQZ(num+1 downto 1); HD_31 <= eQZA(num+2 downto 2);
HD_32 <= eUT(num+1 downto 1); HD_33 <= eUTS(num+2 downto 2); HD_34 <= eTS(num+1 downto 1); HD_35 <= eTSR(num+2 downto 2); HD_36 <= eSR(num+1 downto 1); HD_37 <= eSRQ(num+2 downto 2); HD_38 <= eRQ(num+1 downto 1); HD_39 <= eRQZ(num+2 downto 2);
HD_40 <= eVU(num+1 downto 1); HD_41 <= eVUT(num+2 downto 2); HD_42 <= eUT(num+1 downto 1); HD_43 <= eUTS(num+2 downto 2); HD_44 <= eTS(num+1 downto 1); HD_45 <= eTSR(num+2 downto 2); HD_46 <= eSR(num+1 downto 1); HD_47 <= eSRQ(num+2 downto 2);
HD_48 <= eWV(num+1 downto 1); HD_49 <= eWVU(num+2 downto 2); HD_50 <= eVU(num+1 downto 1); HD_51 <= eVUT(num+2 downto 2); HD_52 <= eUT(num+1 downto 1); HD_53 <= eUTS(num+2 downto 2); HD_54 <= eTS(num+1 downto 1); HD_55 <= eTSR(num+2 downto 2);
HD_56 <= eXW(num+1 downto 1); HD_57 <= eXWV(num+2 downto 2); HD_58 <= eWV(num+1 downto 1); HD_59 <= eWVU(num+2 downto 2); HD_60 <= eVU(num+1 downto 1); HD_61 <= eVUT(num+2 downto 2); HD_62 <= eUT(num+1 downto 1); HD_63 <= eUTS(num+2 downto 2);

-- Mode 7 -Vertical Left
VL_0  <= eAB(num+1 downto 1);  VL_1  <= eBC(num+1 downto 1);  VL_2  <= eCD(num+1 downto 1);  VL_3  <= eDE(num+1 downto 1);  VL_4  <= eEF(num+1 downto 1);  VL_5  <= eFG(num+1 downto 1);  VL_6  <= eGH(num+1 downto 1);  VL_7  <= eHI(num+1 downto 1);
VL_8  <= eABC(num+2 downto 2); VL_9  <= eBCD(num+2 downto 2); VL_10 <= eCDE(num+2 downto 2); VL_11 <= eDEF(num+2 downto 2); VL_12 <= eEFG(num+2 downto 2); VL_13 <= eFGH(num+2 downto 2); VL_14 <= eGHI(num+2 downto 2); VL_15 <= eHIJ(num+2 downto 2);
VL_16 <= eBC(num+1 downto 1);  VL_17 <= eCD(num+1 downto 1);  VL_18 <= eDE(num+1 downto 1);  VL_19 <= eEF(num+1 downto 1);  VL_20 <= eFG(num+1 downto 1);  VL_21 <= eGH(num+1 downto 1);  VL_22 <= eHI(num+1 downto 1);  VL_23 <= eIJ(num+1 downto 1);
VL_24 <= eBCD(num+2 downto 2); VL_25 <= eCDE(num+2 downto 2); VL_26 <= eDEF(num+2 downto 2); VL_27 <= eEFG(num+2 downto 2); VL_28 <= eFGH(num+2 downto 2); VL_29 <= eGHI(num+2 downto 2); VL_30 <= eHIJ(num+2 downto 2); VL_31 <= eIJK(num+2 downto 2);
VL_32 <= eCD(num+1 downto 1);  VL_33 <= eDE(num+1 downto 1);  VL_34 <= eEF(num+1 downto 1);  VL_35 <= eFG(num+1 downto 1);  VL_36 <= eGH(num+1 downto 1);  VL_37 <= eHI(num+1 downto 1);  VL_38 <= eIJ(num+1 downto 1);  VL_39 <= eJK(num+1 downto 1);
VL_40 <= eCDE(num+2 downto 2); VL_41 <= eDEF(num+2 downto 2); VL_42 <= eEFG(num+2 downto 2); VL_43 <= eFGH(num+2 downto 2); VL_44 <= eGHI(num+2 downto 2); VL_45 <= eHIJ(num+2 downto 2); VL_46 <= eIJK(num+2 downto 2); VL_47 <= eJKL(num+2 downto 2);
VL_48 <= eDE(num+1 downto 1);  VL_49 <= eEF(num+1 downto 1);  VL_50 <= eFG(num+1 downto 1);  VL_51 <= eGH(num+1 downto 1);  VL_52 <= eHI(num+1 downto 1);  VL_53 <= eIJ(num+1 downto 1);  VL_54 <= eJK(num+1 downto 1);  VL_55 <= eKL(num+1 downto 1);
VL_56 <= eDEF(num+2 downto 2); VL_57 <= eEFG(num+2 downto 2); VL_58 <= eFGH(num+2 downto 2); VL_59 <= eGHI(num+2 downto 2); VL_60 <= eHIJ(num+2 downto 2); VL_61 <= eIJK(num+2 downto 2); VL_62 <= eJKL(num+2 downto 2); VL_63 <= eKLM(num+2 downto 2);

-- Mode 8 - Horizontal Up
HU_0  <= eRQ(num+1 downto 1); HU_1  <= eSRQ(num+2 downto 2); HU_2  <= eSR(num+1 downto 1); HU_3  <= eTSR(num+2 downto 2); HU_4  <= eTS(num+1 downto 1); HU_5  <= eUTS(num+2 downto 2); HU_6  <= eUT(num+1 downto 1); HU_7  <= eVUT(num+2 downto 2);
HU_8  <= eSR(num+1 downto 1); HU_9  <= eTSR(num+2 downto 2); HU_10 <= eTS(num+1 downto 1); HU_11 <= eUTS(num+2 downto 2); HU_12 <= eUT(num+1 downto 1); HU_13 <= eVUT(num+2 downto 2); HU_14 <= eVU(num+1 downto 1); HU_15 <= eWVU(num+2 downto 2);
HU_16 <= eTS(num+1 downto 1); HU_17 <= eUTS(num+2 downto 2); HU_18 <= eUT(num+1 downto 1); HU_19 <= eVUT(num+2 downto 2); HU_20 <= eVU(num+1 downto 1); HU_21 <= eWVU(num+2 downto 2); HU_22 <= eWV(num+1 downto 1); HU_23 <= eXWV(num+2 downto 2);
HU_24 <= eUT(num+1 downto 1); HU_25 <= eVUT(num+2 downto 2); HU_26 <= eVU(num+1 downto 1); HU_27 <= eWVU(num+2 downto 2); HU_28 <= eWV(num+1 downto 1); HU_29 <= eXWV(num+2 downto 2); HU_30 <= eXW(num+1 downto 1); HU_31 <= e3XW(num+2 downto 2);
HU_32 <= eVU(num+1 downto 1); HU_33 <= eWVU(num+2 downto 2); HU_34 <= eWV(num+1 downto 1); HU_35 <= eXWV(num+2 downto 2); HU_36 <= eXW(num+1 downto 1); HU_37 <= e3XW(num+2 downto 2); HU_38 <= X;   HU_39 <= X;
HU_40 <= eWV(num+1 downto 1); HU_41 <= eXWV(num+2 downto 2); HU_42 <= eXW(num+1 downto 1); HU_43 <= e3XW(num+2 downto 2); HU_44 <= X;   HU_45 <= X;    HU_46 <= X;   HU_47 <= X;
HU_48 <= eXW(num+1 downto 1); HU_49 <= e3XW(num+2 downto 2); HU_50 <= X;   					 HU_51 <= X;    					  HU_52 <= X;   HU_53 <= X;    HU_54 <= X;   HU_55 <= X;
HU_56 <= X; 						HU_57 <= X; 						 HU_58 <= X;   					 HU_59 <= X;    					  HU_60 <= X;   HU_61 <= X;    HU_62 <= X;   HU_63 <= X;


-------------------------------------------------------------------------------------------------------------------------
--#####################################################################################################################--
--# Diff stuff
--#####################################################################################################################--
-------------------------------------------------------------------------------------------------------------------------

-- For all predicted pixels, Diff = Original - Predicted
-- use 3*16 ycbcr registers and current input ycbcr pixels, when quart is 3
-- 600 lines or so

-- Vertical
Diff0_V   <= (("0"& ycbcr0_reg)  -  ("0"&V_0 )) when (quart="11") else (others=>'0');
Diff1_V   <= (("0"& ycbcr1_reg)  -  ("0"&V_1 )) when (quart="11") else (others=>'0');
Diff2_V   <= (("0"& ycbcr2_reg)  -  ("0"&V_2 )) when (quart="11") else (others=>'0');
Diff3_V   <= (("0"& ycbcr3_reg)  -  ("0"&V_3 )) when (quart="11") else (others=>'0');
Diff4_V   <= (("0"& ycbcr4_reg)  -  ("0"&V_4 )) when (quart="11") else (others=>'0');
Diff5_V   <= (("0"& ycbcr5_reg)  -  ("0"&V_5 )) when (quart="11") else (others=>'0');
Diff6_V   <= (("0"& ycbcr6_reg)  -  ("0"&V_6 )) when (quart="11") else (others=>'0');
Diff7_V   <= (("0"& ycbcr7_reg)  -  ("0"&V_7 )) when (quart="11") else (others=>'0');
Diff8_V   <= (("0"& ycbcr8_reg)  -  ("0"&V_8 )) when (quart="11") else (others=>'0');
Diff9_V   <= (("0"& ycbcr9_reg)  -  ("0"&V_9 )) when (quart="11") else (others=>'0');
Diff10_V  <= (("0"& ycbcr10_reg)  - ("0"&V_10)) when (quart="11") else (others=>'0');
Diff11_V  <= (("0"& ycbcr11_reg)  - ("0"&V_11)) when (quart="11") else (others=>'0');	 
Diff12_V  <= (("0"& ycbcr12_reg)  - ("0"&V_12)) when (quart="11") else (others=>'0');
Diff13_V  <= (("0"& ycbcr13_reg)  - ("0"&V_13)) when (quart="11") else (others=>'0');
Diff14_V  <= (("0"& ycbcr14_reg)  - ("0"&V_14)) when (quart="11") else (others=>'0');  
Diff15_V  <= (("0"& ycbcr15_reg)  - ("0"&V_15)) when (quart="11") else (others=>'0');
Diff16_V  <= (("0"& ycbcr16_reg)  - ("0"&V_16)) when (quart="11") else (others=>'0');
Diff17_V  <= (("0"& ycbcr17_reg)  - ("0"&V_17)) when (quart="11") else (others=>'0');
Diff18_V  <= (("0"& ycbcr18_reg)  - ("0"&V_18)) when (quart="11") else (others=>'0');
Diff19_V  <= (("0"& ycbcr19_reg)  - ("0"&V_19)) when (quart="11") else (others=>'0');
Diff20_V  <= (("0"& ycbcr20_reg)  - ("0"&V_20)) when (quart="11") else (others=>'0');
Diff21_V  <= (("0"& ycbcr21_reg)  - ("0"&V_21)) when (quart="11") else (others=>'0');
Diff22_V  <= (("0"& ycbcr22_reg)  - ("0"&V_22)) when (quart="11") else (others=>'0');
Diff23_V  <= (("0"& ycbcr23_reg)  - ("0"&V_23)) when (quart="11") else (others=>'0');
Diff24_V  <= (("0"& ycbcr24_reg)  - ("0"&V_24)) when (quart="11") else (others=>'0');
Diff25_V  <= (("0"& ycbcr25_reg)  - ("0"&V_25)) when (quart="11") else (others=>'0');
Diff26_V  <= (("0"& ycbcr26_reg)  - ("0"&V_26)) when (quart="11") else (others=>'0');
Diff27_V  <= (("0"& ycbcr27_reg)  - ("0"&V_27)) when (quart="11") else (others=>'0');	 
Diff28_V  <= (("0"& ycbcr28_reg)  - ("0"&V_28)) when (quart="11") else (others=>'0');
Diff29_V  <= (("0"& ycbcr29_reg)  - ("0"&V_29)) when (quart="11") else (others=>'0');
Diff30_V  <= (("0"& ycbcr30_reg)  - ("0"&V_30)) when (quart="11") else (others=>'0');  
Diff31_V  <= (("0"& ycbcr31_reg)  - ("0"&V_31)) when (quart="11") else (others=>'0'); 
Diff32_V  <= (("0"& ycbcr32_reg)  - ("0"&V_32)) when (quart="11") else (others=>'0');
Diff33_V  <= (("0"& ycbcr33_reg)  - ("0"&V_33)) when (quart="11") else (others=>'0');
Diff34_V  <= (("0"& ycbcr34_reg)  - ("0"&V_34)) when (quart="11") else (others=>'0');
Diff35_V  <= (("0"& ycbcr35_reg)  - ("0"&V_35)) when (quart="11") else (others=>'0');
Diff36_V  <= (("0"& ycbcr0)       - ("0"&V_36)) when (quart="11") else (others=>'0');
Diff37_V  <= (("0"& ycbcr1)       - ("0"&V_37)) when (quart="11") else (others=>'0');
Diff38_V  <= (("0"& ycbcr2)       - ("0"&V_38)) when (quart="11") else (others=>'0');
Diff39_V  <= (("0"& ycbcr3)       - ("0"&V_39)) when (quart="11") else (others=>'0');
Diff40_V  <= (("0"& ycbcr40_reg)  - ("0"&V_40)) when (quart="11") else (others=>'0');
Diff41_V  <= (("0"& ycbcr41_reg)  - ("0"&V_41)) when (quart="11") else (others=>'0');
Diff42_V  <= (("0"& ycbcr42_reg)  - ("0"&V_42)) when (quart="11") else (others=>'0');
Diff43_V  <= (("0"& ycbcr43_reg)  - ("0"&V_43)) when (quart="11") else (others=>'0');	 
Diff44_V  <= (("0"& ycbcr4)       - ("0"&V_44)) when (quart="11") else (others=>'0');
Diff45_V  <= (("0"& ycbcr5)       - ("0"&V_45)) when (quart="11") else (others=>'0');
Diff46_V  <= (("0"& ycbcr6)       - ("0"&V_46)) when (quart="11") else (others=>'0');  
Diff47_V  <= (("0"& ycbcr7)       - ("0"&V_47)) when (quart="11") else (others=>'0');
Diff48_V  <= (("0"& ycbcr48_reg)  - ("0"&V_48)) when (quart="11") else (others=>'0');
Diff49_V  <= (("0"& ycbcr49_reg)  - ("0"&V_49)) when (quart="11") else (others=>'0');
Diff50_V  <= (("0"& ycbcr50_reg)  - ("0"&V_50)) when (quart="11") else (others=>'0');
Diff51_V  <= (("0"& ycbcr51_reg)  - ("0"&V_51)) when (quart="11") else (others=>'0');
Diff52_V  <= (("0"& ycbcr8)       - ("0"&V_52)) when (quart="11") else (others=>'0');
Diff53_V  <= (("0"& ycbcr9)       - ("0"&V_53)) when (quart="11") else (others=>'0');
Diff54_V  <= (("0"& ycbcr10)      - ("0"&V_54)) when (quart="11") else (others=>'0');
Diff55_V  <= (("0"& ycbcr11)      - ("0"&V_55)) when (quart="11") else (others=>'0');
Diff56_V  <= (("0"& ycbcr56_reg)  - ("0"&V_56)) when (quart="11") else (others=>'0');
Diff57_V  <= (("0"& ycbcr57_reg)  - ("0"&V_57)) when (quart="11") else (others=>'0');
Diff58_V  <= (("0"& ycbcr58_reg)  - ("0"&V_58)) when (quart="11") else (others=>'0');
Diff59_V  <= (("0"& ycbcr59_reg)  - ("0"&V_59)) when (quart="11") else (others=>'0');	 
Diff60_V  <= (("0"& ycbcr12)      - ("0"&V_60)) when (quart="11") else (others=>'0');
Diff61_V  <= (("0"& ycbcr13)      - ("0"&V_61)) when (quart="11") else (others=>'0');
Diff62_V  <= (("0"& ycbcr14)      - ("0"&V_62)) when (quart="11") else (others=>'0');  
Diff63_V  <= (("0"& ycbcr15)      - ("0"&V_63)) when (quart="11") else (others=>'0'); 

-- Horizontal
Diff0_H   <= (("0"& ycbcr0_reg)  -  ("0"&H_0 )) when (quart="11") else (others=>'0');
Diff1_H   <= (("0"& ycbcr1_reg)  -  ("0"&H_1 )) when (quart="11") else (others=>'0');
Diff2_H   <= (("0"& ycbcr2_reg)  -  ("0"&H_2 )) when (quart="11") else (others=>'0');
Diff3_H   <= (("0"& ycbcr3_reg)  -  ("0"&H_3 )) when (quart="11") else (others=>'0');
Diff4_H   <= (("0"& ycbcr4_reg)  -  ("0"&H_4 )) when (quart="11") else (others=>'0');
Diff5_H   <= (("0"& ycbcr5_reg)  -  ("0"&H_5 )) when (quart="11") else (others=>'0');
Diff6_H   <= (("0"& ycbcr6_reg)  -  ("0"&H_6 )) when (quart="11") else (others=>'0');
Diff7_H   <= (("0"& ycbcr7_reg)  -  ("0"&H_7 )) when (quart="11") else (others=>'0');
Diff8_H   <= (("0"& ycbcr8_reg)  -  ("0"&H_8 )) when (quart="11") else (others=>'0');
Diff9_H   <= (("0"& ycbcr9_reg)  -  ("0"&H_9 )) when (quart="11") else (others=>'0');
Diff10_H  <= (("0"& ycbcr10_reg)  - ("0"&H_10)) when (quart="11") else (others=>'0');
Diff11_H  <= (("0"& ycbcr11_reg)  - ("0"&H_11)) when (quart="11") else (others=>'0');	 
Diff12_H  <= (("0"& ycbcr12_reg)  - ("0"&H_12)) when (quart="11") else (others=>'0');
Diff13_H  <= (("0"& ycbcr13_reg)  - ("0"&H_13)) when (quart="11") else (others=>'0');
Diff14_H  <= (("0"& ycbcr14_reg)  - ("0"&H_14)) when (quart="11") else (others=>'0');  
Diff15_H  <= (("0"& ycbcr15_reg)  - ("0"&H_15)) when (quart="11") else (others=>'0');
Diff16_H  <= (("0"& ycbcr16_reg)  - ("0"&H_16)) when (quart="11") else (others=>'0');
Diff17_H  <= (("0"& ycbcr17_reg)  - ("0"&H_17)) when (quart="11") else (others=>'0');
Diff18_H  <= (("0"& ycbcr18_reg)  - ("0"&H_18)) when (quart="11") else (others=>'0');
Diff19_H  <= (("0"& ycbcr19_reg)  - ("0"&H_19)) when (quart="11") else (others=>'0');
Diff20_H  <= (("0"& ycbcr20_reg)  - ("0"&H_20)) when (quart="11") else (others=>'0');
Diff21_H  <= (("0"& ycbcr21_reg)  - ("0"&H_21)) when (quart="11") else (others=>'0');
Diff22_H  <= (("0"& ycbcr22_reg)  - ("0"&H_22)) when (quart="11") else (others=>'0');
Diff23_H  <= (("0"& ycbcr23_reg)  - ("0"&H_23)) when (quart="11") else (others=>'0');
Diff24_H  <= (("0"& ycbcr24_reg)  - ("0"&H_24)) when (quart="11") else (others=>'0');
Diff25_H  <= (("0"& ycbcr25_reg)  - ("0"&H_25)) when (quart="11") else (others=>'0');
Diff26_H  <= (("0"& ycbcr26_reg)  - ("0"&H_26)) when (quart="11") else (others=>'0');
Diff27_H  <= (("0"& ycbcr27_reg)  - ("0"&H_27)) when (quart="11") else (others=>'0');	 
Diff28_H  <= (("0"& ycbcr28_reg)  - ("0"&H_28)) when (quart="11") else (others=>'0');
Diff29_H  <= (("0"& ycbcr29_reg)  - ("0"&H_29)) when (quart="11") else (others=>'0');
Diff30_H  <= (("0"& ycbcr30_reg)  - ("0"&H_30)) when (quart="11") else (others=>'0');  
Diff31_H  <= (("0"& ycbcr31_reg)  - ("0"&H_31)) when (quart="11") else (others=>'0'); 
Diff32_H  <= (("0"& ycbcr32_reg)  - ("0"&H_32)) when (quart="11") else (others=>'0');
Diff33_H  <= (("0"& ycbcr33_reg)  - ("0"&H_33)) when (quart="11") else (others=>'0');
Diff34_H  <= (("0"& ycbcr34_reg)  - ("0"&H_34)) when (quart="11") else (others=>'0');
Diff35_H  <= (("0"& ycbcr35_reg)  - ("0"&H_35)) when (quart="11") else (others=>'0');
Diff36_H  <= (("0"& ycbcr0)       - ("0"&H_36)) when (quart="11") else (others=>'0');
Diff37_H  <= (("0"& ycbcr1)       - ("0"&H_37)) when (quart="11") else (others=>'0');
Diff38_H  <= (("0"& ycbcr2)       - ("0"&H_38)) when (quart="11") else (others=>'0');
Diff39_H  <= (("0"& ycbcr3)       - ("0"&H_39)) when (quart="11") else (others=>'0');
Diff40_H  <= (("0"& ycbcr40_reg)  - ("0"&H_40)) when (quart="11") else (others=>'0');
Diff41_H  <= (("0"& ycbcr41_reg)  - ("0"&H_41)) when (quart="11") else (others=>'0');
Diff42_H  <= (("0"& ycbcr42_reg)  - ("0"&H_42)) when (quart="11") else (others=>'0');
Diff43_H  <= (("0"& ycbcr43_reg)  - ("0"&H_43)) when (quart="11") else (others=>'0');	 
Diff44_H  <= (("0"& ycbcr4)       - ("0"&H_44)) when (quart="11") else (others=>'0');
Diff45_H  <= (("0"& ycbcr5)       - ("0"&H_45)) when (quart="11") else (others=>'0');
Diff46_H  <= (("0"& ycbcr6)       - ("0"&H_46)) when (quart="11") else (others=>'0');  
Diff47_H  <= (("0"& ycbcr7)       - ("0"&H_47)) when (quart="11") else (others=>'0');
Diff48_H  <= (("0"& ycbcr48_reg)  - ("0"&H_48)) when (quart="11") else (others=>'0');
Diff49_H  <= (("0"& ycbcr49_reg)  - ("0"&H_49)) when (quart="11") else (others=>'0');
Diff50_H  <= (("0"& ycbcr50_reg)  - ("0"&H_50)) when (quart="11") else (others=>'0');
Diff51_H  <= (("0"& ycbcr51_reg)  - ("0"&H_51)) when (quart="11") else (others=>'0');
Diff52_H  <= (("0"& ycbcr8)       - ("0"&H_52)) when (quart="11") else (others=>'0');
Diff53_H  <= (("0"& ycbcr9)       - ("0"&H_53)) when (quart="11") else (others=>'0');
Diff54_H  <= (("0"& ycbcr10)      - ("0"&H_54)) when (quart="11") else (others=>'0');
Diff55_H  <= (("0"& ycbcr11)      - ("0"&H_55)) when (quart="11") else (others=>'0');
Diff56_H  <= (("0"& ycbcr56_reg)  - ("0"&H_56)) when (quart="11") else (others=>'0');
Diff57_H  <= (("0"& ycbcr57_reg)  - ("0"&H_57)) when (quart="11") else (others=>'0');
Diff58_H  <= (("0"& ycbcr58_reg)  - ("0"&H_58)) when (quart="11") else (others=>'0');
Diff59_H  <= (("0"& ycbcr59_reg)  - ("0"&H_59)) when (quart="11") else (others=>'0');	 
Diff60_H  <= (("0"& ycbcr12)      - ("0"&H_60)) when (quart="11") else (others=>'0');
Diff61_H  <= (("0"& ycbcr13)      - ("0"&H_61)) when (quart="11") else (others=>'0');
Diff62_H  <= (("0"& ycbcr14)      - ("0"&H_62)) when (quart="11") else (others=>'0');  
Diff63_H  <= (("0"& ycbcr15)      - ("0"&H_63)) when (quart="11") else (others=>'0');

--DC
Diff0_DC   <= (("0"& ycbcr0_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff1_DC   <= (("0"& ycbcr1_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff2_DC   <= (("0"& ycbcr2_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff3_DC   <= (("0"& ycbcr3_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff4_DC   <= (("0"& ycbcr4_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff5_DC   <= (("0"& ycbcr5_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff6_DC   <= (("0"& ycbcr6_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff7_DC   <= (("0"& ycbcr7_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff8_DC   <= (("0"& ycbcr8_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff9_DC   <= (("0"& ycbcr9_reg)  -  ("0"&DC)) when (quart="11") else (others=>'0');
Diff10_DC  <= (("0"& ycbcr10_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff11_DC  <= (("0"& ycbcr11_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');	 
Diff12_DC  <= (("0"& ycbcr12_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff13_DC  <= (("0"& ycbcr13_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff14_DC  <= (("0"& ycbcr14_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');  
Diff15_DC  <= (("0"& ycbcr15_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff16_DC  <= (("0"& ycbcr16_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff17_DC  <= (("0"& ycbcr17_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff18_DC  <= (("0"& ycbcr18_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff19_DC  <= (("0"& ycbcr19_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff20_DC  <= (("0"& ycbcr20_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff21_DC  <= (("0"& ycbcr21_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff22_DC  <= (("0"& ycbcr22_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff23_DC  <= (("0"& ycbcr23_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff24_DC  <= (("0"& ycbcr24_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff25_DC  <= (("0"& ycbcr25_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff26_DC  <= (("0"& ycbcr26_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff27_DC  <= (("0"& ycbcr27_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');	 
Diff28_DC  <= (("0"& ycbcr28_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff29_DC  <= (("0"& ycbcr29_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff30_DC  <= (("0"& ycbcr30_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');  
Diff31_DC  <= (("0"& ycbcr31_reg)  - ("0"&DC)) when (quart="11") else (others=>'0'); 
Diff32_DC  <= (("0"& ycbcr32_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff33_DC  <= (("0"& ycbcr33_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff34_DC  <= (("0"& ycbcr34_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff35_DC  <= (("0"& ycbcr35_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff36_DC  <= (("0"& ycbcr0)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff37_DC  <= (("0"& ycbcr1)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff38_DC  <= (("0"& ycbcr2)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff39_DC  <= (("0"& ycbcr3)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff40_DC  <= (("0"& ycbcr40_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff41_DC  <= (("0"& ycbcr41_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff42_DC  <= (("0"& ycbcr42_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff43_DC  <= (("0"& ycbcr43_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');	 
Diff44_DC  <= (("0"& ycbcr4)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff45_DC  <= (("0"& ycbcr5)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff46_DC  <= (("0"& ycbcr6)       - ("0"&DC)) when (quart="11") else (others=>'0');  
Diff47_DC  <= (("0"& ycbcr7)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff48_DC  <= (("0"& ycbcr48_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff49_DC  <= (("0"& ycbcr49_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff50_DC  <= (("0"& ycbcr50_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff51_DC  <= (("0"& ycbcr51_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff52_DC  <= (("0"& ycbcr8)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff53_DC  <= (("0"& ycbcr9)       - ("0"&DC)) when (quart="11") else (others=>'0');
Diff54_DC  <= (("0"& ycbcr10)      - ("0"&DC)) when (quart="11") else (others=>'0');
Diff55_DC  <= (("0"& ycbcr11)      - ("0"&DC)) when (quart="11") else (others=>'0');
Diff56_DC  <= (("0"& ycbcr56_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff57_DC  <= (("0"& ycbcr57_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff58_DC  <= (("0"& ycbcr58_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');
Diff59_DC  <= (("0"& ycbcr59_reg)  - ("0"&DC)) when (quart="11") else (others=>'0');	 
Diff60_DC  <= (("0"& ycbcr12)      - ("0"&DC)) when (quart="11") else (others=>'0');
Diff61_DC  <= (("0"& ycbcr13)      - ("0"&DC)) when (quart="11") else (others=>'0');
Diff62_DC  <= (("0"& ycbcr14)      - ("0"&DC)) when (quart="11") else (others=>'0');  
Diff63_DC  <= (("0"& ycbcr15)      - ("0"&DC)) when (quart="11") else (others=>'0');
--DDL
Diff0_DDL   <= (("0"& ycbcr0_reg)  -  ("0"&DDL_0 )) when (quart="11") else (others=>'0');
Diff1_DDL   <= (("0"& ycbcr1_reg)  -  ("0"&DDL_1 )) when (quart="11") else (others=>'0');
Diff2_DDL   <= (("0"& ycbcr2_reg)  -  ("0"&DDL_2 )) when (quart="11") else (others=>'0');
Diff3_DDL   <= (("0"& ycbcr3_reg)  -  ("0"&DDL_3 )) when (quart="11") else (others=>'0');
Diff4_DDL   <= (("0"& ycbcr4_reg)  -  ("0"&DDL_4 )) when (quart="11") else (others=>'0');
Diff5_DDL   <= (("0"& ycbcr5_reg)  -  ("0"&DDL_5 )) when (quart="11") else (others=>'0');
Diff6_DDL   <= (("0"& ycbcr6_reg)  -  ("0"&DDL_6 )) when (quart="11") else (others=>'0');
Diff7_DDL   <= (("0"& ycbcr7_reg)  -  ("0"&DDL_7 )) when (quart="11") else (others=>'0');
Diff8_DDL   <= (("0"& ycbcr8_reg)  -  ("0"&DDL_8 )) when (quart="11") else (others=>'0');
Diff9_DDL   <= (("0"& ycbcr9_reg)  -  ("0"&DDL_9 )) when (quart="11") else (others=>'0');
Diff10_DDL  <= (("0"& ycbcr10_reg)  - ("0"&DDL_10)) when (quart="11") else (others=>'0');
Diff11_DDL  <= (("0"& ycbcr11_reg)  - ("0"&DDL_11)) when (quart="11") else (others=>'0');	 
Diff12_DDL  <= (("0"& ycbcr12_reg)  - ("0"&DDL_12)) when (quart="11") else (others=>'0');
Diff13_DDL  <= (("0"& ycbcr13_reg)  - ("0"&DDL_13)) when (quart="11") else (others=>'0');
Diff14_DDL  <= (("0"& ycbcr14_reg)  - ("0"&DDL_14)) when (quart="11") else (others=>'0');  
Diff15_DDL  <= (("0"& ycbcr15_reg)  - ("0"&DDL_15)) when (quart="11") else (others=>'0');
Diff16_DDL  <= (("0"& ycbcr16_reg)  - ("0"&DDL_16)) when (quart="11") else (others=>'0');
Diff17_DDL  <= (("0"& ycbcr17_reg)  - ("0"&DDL_17)) when (quart="11") else (others=>'0');
Diff18_DDL  <= (("0"& ycbcr18_reg)  - ("0"&DDL_18)) when (quart="11") else (others=>'0');
Diff19_DDL  <= (("0"& ycbcr19_reg)  - ("0"&DDL_19)) when (quart="11") else (others=>'0');
Diff20_DDL  <= (("0"& ycbcr20_reg)  - ("0"&DDL_20)) when (quart="11") else (others=>'0');
Diff21_DDL  <= (("0"& ycbcr21_reg)  - ("0"&DDL_21)) when (quart="11") else (others=>'0');
Diff22_DDL  <= (("0"& ycbcr22_reg)  - ("0"&DDL_22)) when (quart="11") else (others=>'0');
Diff23_DDL  <= (("0"& ycbcr23_reg)  - ("0"&DDL_23)) when (quart="11") else (others=>'0');
Diff24_DDL  <= (("0"& ycbcr24_reg)  - ("0"&DDL_24)) when (quart="11") else (others=>'0');
Diff25_DDL  <= (("0"& ycbcr25_reg)  - ("0"&DDL_25)) when (quart="11") else (others=>'0');
Diff26_DDL  <= (("0"& ycbcr26_reg)  - ("0"&DDL_26)) when (quart="11") else (others=>'0');
Diff27_DDL  <= (("0"& ycbcr27_reg)  - ("0"&DDL_27)) when (quart="11") else (others=>'0');	 
Diff28_DDL  <= (("0"& ycbcr28_reg)  - ("0"&DDL_28)) when (quart="11") else (others=>'0');
Diff29_DDL  <= (("0"& ycbcr29_reg)  - ("0"&DDL_29)) when (quart="11") else (others=>'0');
Diff30_DDL  <= (("0"& ycbcr30_reg)  - ("0"&DDL_30)) when (quart="11") else (others=>'0');  
Diff31_DDL  <= (("0"& ycbcr31_reg)  - ("0"&DDL_31)) when (quart="11") else (others=>'0'); 
Diff32_DDL  <= (("0"& ycbcr32_reg)  - ("0"&DDL_32)) when (quart="11") else (others=>'0');
Diff33_DDL  <= (("0"& ycbcr33_reg)  - ("0"&DDL_33)) when (quart="11") else (others=>'0');
Diff34_DDL  <= (("0"& ycbcr34_reg)  - ("0"&DDL_34)) when (quart="11") else (others=>'0');
Diff35_DDL  <= (("0"& ycbcr35_reg)  - ("0"&DDL_35)) when (quart="11") else (others=>'0');
Diff36_DDL  <= (("0"& ycbcr0)       - ("0"&DDL_36)) when (quart="11") else (others=>'0');
Diff37_DDL  <= (("0"& ycbcr1)       - ("0"&DDL_37)) when (quart="11") else (others=>'0');
Diff38_DDL  <= (("0"& ycbcr2)       - ("0"&DDL_38)) when (quart="11") else (others=>'0');
Diff39_DDL  <= (("0"& ycbcr3)       - ("0"&DDL_39)) when (quart="11") else (others=>'0');
Diff40_DDL  <= (("0"& ycbcr40_reg)  - ("0"&DDL_40)) when (quart="11") else (others=>'0');
Diff41_DDL  <= (("0"& ycbcr41_reg)  - ("0"&DDL_41)) when (quart="11") else (others=>'0');
Diff42_DDL  <= (("0"& ycbcr42_reg)  - ("0"&DDL_42)) when (quart="11") else (others=>'0');
Diff43_DDL  <= (("0"& ycbcr43_reg)  - ("0"&DDL_43)) when (quart="11") else (others=>'0');	 
Diff44_DDL  <= (("0"& ycbcr4)       - ("0"&DDL_44)) when (quart="11") else (others=>'0');
Diff45_DDL  <= (("0"& ycbcr5)       - ("0"&DDL_45)) when (quart="11") else (others=>'0');
Diff46_DDL  <= (("0"& ycbcr6)       - ("0"&DDL_46)) when (quart="11") else (others=>'0');  
Diff47_DDL  <= (("0"& ycbcr7)       - ("0"&DDL_47)) when (quart="11") else (others=>'0');
Diff48_DDL  <= (("0"& ycbcr48_reg)  - ("0"&DDL_48)) when (quart="11") else (others=>'0');
Diff49_DDL  <= (("0"& ycbcr49_reg)  - ("0"&DDL_49)) when (quart="11") else (others=>'0');
Diff50_DDL  <= (("0"& ycbcr50_reg)  - ("0"&DDL_50)) when (quart="11") else (others=>'0');
Diff51_DDL  <= (("0"& ycbcr51_reg)  - ("0"&DDL_51)) when (quart="11") else (others=>'0');
Diff52_DDL  <= (("0"& ycbcr8)       - ("0"&DDL_52)) when (quart="11") else (others=>'0');
Diff53_DDL  <= (("0"& ycbcr9)       - ("0"&DDL_53)) when (quart="11") else (others=>'0');
Diff54_DDL  <= (("0"& ycbcr10)      - ("0"&DDL_54)) when (quart="11") else (others=>'0');
Diff55_DDL  <= (("0"& ycbcr11)      - ("0"&DDL_55)) when (quart="11") else (others=>'0');
Diff56_DDL  <= (("0"& ycbcr56_reg)  - ("0"&DDL_56)) when (quart="11") else (others=>'0');
Diff57_DDL  <= (("0"& ycbcr57_reg)  - ("0"&DDL_57)) when (quart="11") else (others=>'0');
Diff58_DDL  <= (("0"& ycbcr58_reg)  - ("0"&DDL_58)) when (quart="11") else (others=>'0');
Diff59_DDL  <= (("0"& ycbcr59_reg)  - ("0"&DDL_59)) when (quart="11") else (others=>'0');	 
Diff60_DDL  <= (("0"& ycbcr12)      - ("0"&DDL_60)) when (quart="11") else (others=>'0');
Diff61_DDL  <= (("0"& ycbcr13)      - ("0"&DDL_61)) when (quart="11") else (others=>'0');
Diff62_DDL  <= (("0"& ycbcr14)      - ("0"&DDL_62)) when (quart="11") else (others=>'0');  
Diff63_DDL  <= (("0"& ycbcr15)      - ("0"&DDL_63)) when (quart="11") else (others=>'0');
--DDR

Diff0_DDR   <= (("0"& ycbcr0_reg)  -  ("0"&DDR_0 )) when (quart="11") else (others=>'0');
Diff1_DDR   <= (("0"& ycbcr1_reg)  -  ("0"&DDR_1 )) when (quart="11") else (others=>'0');
Diff2_DDR   <= (("0"& ycbcr2_reg)  -  ("0"&DDR_2 )) when (quart="11") else (others=>'0');
Diff3_DDR   <= (("0"& ycbcr3_reg)  -  ("0"&DDR_3 )) when (quart="11") else (others=>'0');
Diff4_DDR   <= (("0"& ycbcr4_reg)  -  ("0"&DDR_4 )) when (quart="11") else (others=>'0');
Diff5_DDR   <= (("0"& ycbcr5_reg)  -  ("0"&DDR_5 )) when (quart="11") else (others=>'0');
Diff6_DDR   <= (("0"& ycbcr6_reg)  -  ("0"&DDR_6 )) when (quart="11") else (others=>'0');
Diff7_DDR   <= (("0"& ycbcr7_reg)  -  ("0"&DDR_7 )) when (quart="11") else (others=>'0');
Diff8_DDR   <= (("0"& ycbcr8_reg)  -  ("0"&DDR_8 )) when (quart="11") else (others=>'0');
Diff9_DDR   <= (("0"& ycbcr9_reg)  -  ("0"&DDR_9 )) when (quart="11") else (others=>'0');
Diff10_DDR  <= (("0"& ycbcr10_reg)  - ("0"&DDR_10)) when (quart="11") else (others=>'0');
Diff11_DDR  <= (("0"& ycbcr11_reg)  - ("0"&DDR_11)) when (quart="11") else (others=>'0');	 
Diff12_DDR  <= (("0"& ycbcr12_reg)  - ("0"&DDR_12)) when (quart="11") else (others=>'0');
Diff13_DDR  <= (("0"& ycbcr13_reg)  - ("0"&DDR_13)) when (quart="11") else (others=>'0');
Diff14_DDR  <= (("0"& ycbcr14_reg)  - ("0"&DDR_14)) when (quart="11") else (others=>'0');  
Diff15_DDR  <= (("0"& ycbcr15_reg)  - ("0"&DDR_15)) when (quart="11") else (others=>'0');
Diff16_DDR  <= (("0"& ycbcr16_reg)  - ("0"&DDR_16)) when (quart="11") else (others=>'0');
Diff17_DDR  <= (("0"& ycbcr17_reg)  - ("0"&DDR_17)) when (quart="11") else (others=>'0');
Diff18_DDR  <= (("0"& ycbcr18_reg)  - ("0"&DDR_18)) when (quart="11") else (others=>'0');
Diff19_DDR  <= (("0"& ycbcr19_reg)  - ("0"&DDR_19)) when (quart="11") else (others=>'0');
Diff20_DDR  <= (("0"& ycbcr20_reg)  - ("0"&DDR_20)) when (quart="11") else (others=>'0');
Diff21_DDR  <= (("0"& ycbcr21_reg)  - ("0"&DDR_21)) when (quart="11") else (others=>'0');
Diff22_DDR  <= (("0"& ycbcr22_reg)  - ("0"&DDR_22)) when (quart="11") else (others=>'0');
Diff23_DDR  <= (("0"& ycbcr23_reg)  - ("0"&DDR_23)) when (quart="11") else (others=>'0');
Diff24_DDR  <= (("0"& ycbcr24_reg)  - ("0"&DDR_24)) when (quart="11") else (others=>'0');
Diff25_DDR  <= (("0"& ycbcr25_reg)  - ("0"&DDR_25)) when (quart="11") else (others=>'0');
Diff26_DDR  <= (("0"& ycbcr26_reg)  - ("0"&DDR_26)) when (quart="11") else (others=>'0');
Diff27_DDR  <= (("0"& ycbcr27_reg)  - ("0"&DDR_27)) when (quart="11") else (others=>'0');	 
Diff28_DDR  <= (("0"& ycbcr28_reg)  - ("0"&DDR_28)) when (quart="11") else (others=>'0');
Diff29_DDR  <= (("0"& ycbcr29_reg)  - ("0"&DDR_29)) when (quart="11") else (others=>'0');
Diff30_DDR  <= (("0"& ycbcr30_reg)  - ("0"&DDR_30)) when (quart="11") else (others=>'0');  
Diff31_DDR  <= (("0"& ycbcr31_reg)  - ("0"&DDR_31)) when (quart="11") else (others=>'0'); 
Diff32_DDR  <= (("0"& ycbcr32_reg)  - ("0"&DDR_32)) when (quart="11") else (others=>'0');
Diff33_DDR  <= (("0"& ycbcr33_reg)  - ("0"&DDR_33)) when (quart="11") else (others=>'0');
Diff34_DDR  <= (("0"& ycbcr34_reg)  - ("0"&DDR_34)) when (quart="11") else (others=>'0');
Diff35_DDR  <= (("0"& ycbcr35_reg)  - ("0"&DDR_35)) when (quart="11") else (others=>'0');
Diff36_DDR  <= (("0"& ycbcr0)       - ("0"&DDR_36)) when (quart="11") else (others=>'0');
Diff37_DDR  <= (("0"& ycbcr1)       - ("0"&DDR_37)) when (quart="11") else (others=>'0');
Diff38_DDR  <= (("0"& ycbcr2)       - ("0"&DDR_38)) when (quart="11") else (others=>'0');
Diff39_DDR  <= (("0"& ycbcr3)       - ("0"&DDR_39)) when (quart="11") else (others=>'0');
Diff40_DDR  <= (("0"& ycbcr40_reg)  - ("0"&DDR_40)) when (quart="11") else (others=>'0');
Diff41_DDR  <= (("0"& ycbcr41_reg)  - ("0"&DDR_41)) when (quart="11") else (others=>'0');
Diff42_DDR  <= (("0"& ycbcr42_reg)  - ("0"&DDR_42)) when (quart="11") else (others=>'0');
Diff43_DDR  <= (("0"& ycbcr43_reg)  - ("0"&DDR_43)) when (quart="11") else (others=>'0');	 
Diff44_DDR  <= (("0"& ycbcr4)       - ("0"&DDR_44)) when (quart="11") else (others=>'0');
Diff45_DDR  <= (("0"& ycbcr5)       - ("0"&DDR_45)) when (quart="11") else (others=>'0');
Diff46_DDR  <= (("0"& ycbcr6)       - ("0"&DDR_46)) when (quart="11") else (others=>'0');  
Diff47_DDR  <= (("0"& ycbcr7)       - ("0"&DDR_47)) when (quart="11") else (others=>'0');
Diff48_DDR  <= (("0"& ycbcr48_reg)  - ("0"&DDR_48)) when (quart="11") else (others=>'0');
Diff49_DDR  <= (("0"& ycbcr49_reg)  - ("0"&DDR_49)) when (quart="11") else (others=>'0');
Diff50_DDR  <= (("0"& ycbcr50_reg)  - ("0"&DDR_50)) when (quart="11") else (others=>'0');
Diff51_DDR  <= (("0"& ycbcr51_reg)  - ("0"&DDR_51)) when (quart="11") else (others=>'0');
Diff52_DDR  <= (("0"& ycbcr8)       - ("0"&DDR_52)) when (quart="11") else (others=>'0');
Diff53_DDR  <= (("0"& ycbcr9)       - ("0"&DDR_53)) when (quart="11") else (others=>'0');
Diff54_DDR  <= (("0"& ycbcr10)      - ("0"&DDR_54)) when (quart="11") else (others=>'0');
Diff55_DDR  <= (("0"& ycbcr11)      - ("0"&DDR_55)) when (quart="11") else (others=>'0');
Diff56_DDR  <= (("0"& ycbcr56_reg)  - ("0"&DDR_56)) when (quart="11") else (others=>'0');
Diff57_DDR  <= (("0"& ycbcr57_reg)  - ("0"&DDR_57)) when (quart="11") else (others=>'0');
Diff58_DDR  <= (("0"& ycbcr58_reg)  - ("0"&DDR_58)) when (quart="11") else (others=>'0');
Diff59_DDR  <= (("0"& ycbcr59_reg)  - ("0"&DDR_59)) when (quart="11") else (others=>'0');	 
Diff60_DDR  <= (("0"& ycbcr12)      - ("0"&DDR_60)) when (quart="11") else (others=>'0');
Diff61_DDR  <= (("0"& ycbcr13)      - ("0"&DDR_61)) when (quart="11") else (others=>'0');
Diff62_DDR  <= (("0"& ycbcr14)      - ("0"&DDR_62)) when (quart="11") else (others=>'0');  
Diff63_DDR  <= (("0"& ycbcr15)      - ("0"&DDR_63)) when (quart="11") else (others=>'0');

--VR
Diff0_VR   <= (("0"& ycbcr0_reg)  -  ("0"&VR_0 )) when (quart="11") else (others=>'0');
Diff1_VR   <= (("0"& ycbcr1_reg)  -  ("0"&VR_1 )) when (quart="11") else (others=>'0');
Diff2_VR   <= (("0"& ycbcr2_reg)  -  ("0"&VR_2 )) when (quart="11") else (others=>'0');
Diff3_VR   <= (("0"& ycbcr3_reg)  -  ("0"&VR_3 )) when (quart="11") else (others=>'0');
Diff4_VR   <= (("0"& ycbcr4_reg)  -  ("0"&VR_4 )) when (quart="11") else (others=>'0');
Diff5_VR   <= (("0"& ycbcr5_reg)  -  ("0"&VR_5 )) when (quart="11") else (others=>'0');
Diff6_VR   <= (("0"& ycbcr6_reg)  -  ("0"&VR_6 )) when (quart="11") else (others=>'0');
Diff7_VR   <= (("0"& ycbcr7_reg)  -  ("0"&VR_7 )) when (quart="11") else (others=>'0');
Diff8_VR   <= (("0"& ycbcr8_reg)  -  ("0"&VR_8 )) when (quart="11") else (others=>'0');
Diff9_VR   <= (("0"& ycbcr9_reg)  -  ("0"&VR_9 )) when (quart="11") else (others=>'0');
Diff10_VR  <= (("0"& ycbcr10_reg)  - ("0"&VR_10)) when (quart="11") else (others=>'0');
Diff11_VR  <= (("0"& ycbcr11_reg)  - ("0"&VR_11)) when (quart="11") else (others=>'0');	 
Diff12_VR  <= (("0"& ycbcr12_reg)  - ("0"&VR_12)) when (quart="11") else (others=>'0');
Diff13_VR  <= (("0"& ycbcr13_reg)  - ("0"&VR_13)) when (quart="11") else (others=>'0');
Diff14_VR  <= (("0"& ycbcr14_reg)  - ("0"&VR_14)) when (quart="11") else (others=>'0');  
Diff15_VR  <= (("0"& ycbcr15_reg)  - ("0"&VR_15)) when (quart="11") else (others=>'0');
Diff16_VR  <= (("0"& ycbcr16_reg)  - ("0"&VR_16)) when (quart="11") else (others=>'0');
Diff17_VR  <= (("0"& ycbcr17_reg)  - ("0"&VR_17)) when (quart="11") else (others=>'0');
Diff18_VR  <= (("0"& ycbcr18_reg)  - ("0"&VR_18)) when (quart="11") else (others=>'0');
Diff19_VR  <= (("0"& ycbcr19_reg)  - ("0"&VR_19)) when (quart="11") else (others=>'0');
Diff20_VR  <= (("0"& ycbcr20_reg)  - ("0"&VR_20)) when (quart="11") else (others=>'0');
Diff21_VR  <= (("0"& ycbcr21_reg)  - ("0"&VR_21)) when (quart="11") else (others=>'0');
Diff22_VR  <= (("0"& ycbcr22_reg)  - ("0"&VR_22)) when (quart="11") else (others=>'0');
Diff23_VR  <= (("0"& ycbcr23_reg)  - ("0"&VR_23)) when (quart="11") else (others=>'0');
Diff24_VR  <= (("0"& ycbcr24_reg)  - ("0"&VR_24)) when (quart="11") else (others=>'0');
Diff25_VR  <= (("0"& ycbcr25_reg)  - ("0"&VR_25)) when (quart="11") else (others=>'0');
Diff26_VR  <= (("0"& ycbcr26_reg)  - ("0"&VR_26)) when (quart="11") else (others=>'0');
Diff27_VR  <= (("0"& ycbcr27_reg)  - ("0"&VR_27)) when (quart="11") else (others=>'0');	 
Diff28_VR  <= (("0"& ycbcr28_reg)  - ("0"&VR_28)) when (quart="11") else (others=>'0');
Diff29_VR  <= (("0"& ycbcr29_reg)  - ("0"&VR_29)) when (quart="11") else (others=>'0');
Diff30_VR  <= (("0"& ycbcr30_reg)  - ("0"&VR_30)) when (quart="11") else (others=>'0');  
Diff31_VR  <= (("0"& ycbcr31_reg)  - ("0"&VR_31)) when (quart="11") else (others=>'0'); 
Diff32_VR  <= (("0"& ycbcr32_reg)  - ("0"&VR_32)) when (quart="11") else (others=>'0');
Diff33_VR  <= (("0"& ycbcr33_reg)  - ("0"&VR_33)) when (quart="11") else (others=>'0');
Diff34_VR  <= (("0"& ycbcr34_reg)  - ("0"&VR_34)) when (quart="11") else (others=>'0');
Diff35_VR  <= (("0"& ycbcr35_reg)  - ("0"&VR_35)) when (quart="11") else (others=>'0');
Diff36_VR  <= (("0"& ycbcr0)       - ("0"&VR_36)) when (quart="11") else (others=>'0');
Diff37_VR  <= (("0"& ycbcr1)       - ("0"&VR_37)) when (quart="11") else (others=>'0');
Diff38_VR  <= (("0"& ycbcr2)       - ("0"&VR_38)) when (quart="11") else (others=>'0');
Diff39_VR  <= (("0"& ycbcr3)       - ("0"&VR_39)) when (quart="11") else (others=>'0');
Diff40_VR  <= (("0"& ycbcr40_reg)  - ("0"&VR_40)) when (quart="11") else (others=>'0');
Diff41_VR  <= (("0"& ycbcr41_reg)  - ("0"&VR_41)) when (quart="11") else (others=>'0');
Diff42_VR  <= (("0"& ycbcr42_reg)  - ("0"&VR_42)) when (quart="11") else (others=>'0');
Diff43_VR  <= (("0"& ycbcr43_reg)  - ("0"&VR_43)) when (quart="11") else (others=>'0');	 
Diff44_VR  <= (("0"& ycbcr4)       - ("0"&VR_44)) when (quart="11") else (others=>'0');
Diff45_VR  <= (("0"& ycbcr5)       - ("0"&VR_45)) when (quart="11") else (others=>'0');
Diff46_VR  <= (("0"& ycbcr6)       - ("0"&VR_46)) when (quart="11") else (others=>'0');  
Diff47_VR  <= (("0"& ycbcr7)       - ("0"&VR_47)) when (quart="11") else (others=>'0');
Diff48_VR  <= (("0"& ycbcr48_reg)  - ("0"&VR_48)) when (quart="11") else (others=>'0');
Diff49_VR  <= (("0"& ycbcr49_reg)  - ("0"&VR_49)) when (quart="11") else (others=>'0');
Diff50_VR  <= (("0"& ycbcr50_reg)  - ("0"&VR_50)) when (quart="11") else (others=>'0');
Diff51_VR  <= (("0"& ycbcr51_reg)  - ("0"&VR_51)) when (quart="11") else (others=>'0');
Diff52_VR  <= (("0"& ycbcr8)       - ("0"&VR_52)) when (quart="11") else (others=>'0');
Diff53_VR  <= (("0"& ycbcr9)       - ("0"&VR_53)) when (quart="11") else (others=>'0');
Diff54_VR  <= (("0"& ycbcr10)      - ("0"&VR_54)) when (quart="11") else (others=>'0');
Diff55_VR  <= (("0"& ycbcr11)      - ("0"&VR_55)) when (quart="11") else (others=>'0');
Diff56_VR  <= (("0"& ycbcr56_reg)  - ("0"&VR_56)) when (quart="11") else (others=>'0');
Diff57_VR  <= (("0"& ycbcr57_reg)  - ("0"&VR_57)) when (quart="11") else (others=>'0');
Diff58_VR  <= (("0"& ycbcr58_reg)  - ("0"&VR_58)) when (quart="11") else (others=>'0');
Diff59_VR  <= (("0"& ycbcr59_reg)  - ("0"&VR_59)) when (quart="11") else (others=>'0');	 
Diff60_VR  <= (("0"& ycbcr12)      - ("0"&VR_60)) when (quart="11") else (others=>'0');
Diff61_VR  <= (("0"& ycbcr13)      - ("0"&VR_61)) when (quart="11") else (others=>'0');
Diff62_VR  <= (("0"& ycbcr14)      - ("0"&VR_62)) when (quart="11") else (others=>'0');  
Diff63_VR  <= (("0"& ycbcr15)      - ("0"&VR_63)) when (quart="11") else (others=>'0');
--HD
Diff0_HD   <= (("0"& ycbcr0_reg)  -  ("0"&HD_0 )) when (quart="11") else (others=>'0');
Diff1_HD   <= (("0"& ycbcr1_reg)  -  ("0"&HD_1 )) when (quart="11") else (others=>'0');
Diff2_HD   <= (("0"& ycbcr2_reg)  -  ("0"&HD_2 )) when (quart="11") else (others=>'0');
Diff3_HD   <= (("0"& ycbcr3_reg)  -  ("0"&HD_3 )) when (quart="11") else (others=>'0');
Diff4_HD   <= (("0"& ycbcr4_reg)  -  ("0"&HD_4 )) when (quart="11") else (others=>'0');
Diff5_HD   <= (("0"& ycbcr5_reg)  -  ("0"&HD_5 )) when (quart="11") else (others=>'0');
Diff6_HD   <= (("0"& ycbcr6_reg)  -  ("0"&HD_6 )) when (quart="11") else (others=>'0');
Diff7_HD   <= (("0"& ycbcr7_reg)  -  ("0"&HD_7 )) when (quart="11") else (others=>'0');
Diff8_HD   <= (("0"& ycbcr8_reg)  -  ("0"&HD_8 )) when (quart="11") else (others=>'0');
Diff9_HD   <= (("0"& ycbcr9_reg)  -  ("0"&HD_9 )) when (quart="11") else (others=>'0');
Diff10_HD  <= (("0"& ycbcr10_reg)  - ("0"&HD_10)) when (quart="11") else (others=>'0');
Diff11_HD  <= (("0"& ycbcr11_reg)  - ("0"&HD_11)) when (quart="11") else (others=>'0');	 
Diff12_HD  <= (("0"& ycbcr12_reg)  - ("0"&HD_12)) when (quart="11") else (others=>'0');
Diff13_HD  <= (("0"& ycbcr13_reg)  - ("0"&HD_13)) when (quart="11") else (others=>'0');
Diff14_HD  <= (("0"& ycbcr14_reg)  - ("0"&HD_14)) when (quart="11") else (others=>'0');  
Diff15_HD  <= (("0"& ycbcr15_reg)  - ("0"&HD_15)) when (quart="11") else (others=>'0');
Diff16_HD  <= (("0"& ycbcr16_reg)  - ("0"&HD_16)) when (quart="11") else (others=>'0');
Diff17_HD  <= (("0"& ycbcr17_reg)  - ("0"&HD_17)) when (quart="11") else (others=>'0');
Diff18_HD  <= (("0"& ycbcr18_reg)  - ("0"&HD_18)) when (quart="11") else (others=>'0');
Diff19_HD  <= (("0"& ycbcr19_reg)  - ("0"&HD_19)) when (quart="11") else (others=>'0');
Diff20_HD  <= (("0"& ycbcr20_reg)  - ("0"&HD_20)) when (quart="11") else (others=>'0');
Diff21_HD  <= (("0"& ycbcr21_reg)  - ("0"&HD_21)) when (quart="11") else (others=>'0');
Diff22_HD  <= (("0"& ycbcr22_reg)  - ("0"&HD_22)) when (quart="11") else (others=>'0');
Diff23_HD  <= (("0"& ycbcr23_reg)  - ("0"&HD_23)) when (quart="11") else (others=>'0');
Diff24_HD  <= (("0"& ycbcr24_reg)  - ("0"&HD_24)) when (quart="11") else (others=>'0');
Diff25_HD  <= (("0"& ycbcr25_reg)  - ("0"&HD_25)) when (quart="11") else (others=>'0');
Diff26_HD  <= (("0"& ycbcr26_reg)  - ("0"&HD_26)) when (quart="11") else (others=>'0');
Diff27_HD  <= (("0"& ycbcr27_reg)  - ("0"&HD_27)) when (quart="11") else (others=>'0');	 
Diff28_HD  <= (("0"& ycbcr28_reg)  - ("0"&HD_28)) when (quart="11") else (others=>'0');
Diff29_HD  <= (("0"& ycbcr29_reg)  - ("0"&HD_29)) when (quart="11") else (others=>'0');
Diff30_HD  <= (("0"& ycbcr30_reg)  - ("0"&HD_30)) when (quart="11") else (others=>'0');  
Diff31_HD  <= (("0"& ycbcr31_reg)  - ("0"&HD_31)) when (quart="11") else (others=>'0'); 
Diff32_HD  <= (("0"& ycbcr32_reg)  - ("0"&HD_32)) when (quart="11") else (others=>'0');
Diff33_HD  <= (("0"& ycbcr33_reg)  - ("0"&HD_33)) when (quart="11") else (others=>'0');
Diff34_HD  <= (("0"& ycbcr34_reg)  - ("0"&HD_34)) when (quart="11") else (others=>'0');
Diff35_HD  <= (("0"& ycbcr35_reg)  - ("0"&HD_35)) when (quart="11") else (others=>'0');
Diff36_HD  <= (("0"& ycbcr0)       - ("0"&HD_36)) when (quart="11") else (others=>'0');
Diff37_HD  <= (("0"& ycbcr1)       - ("0"&HD_37)) when (quart="11") else (others=>'0');
Diff38_HD  <= (("0"& ycbcr2)       - ("0"&HD_38)) when (quart="11") else (others=>'0');
Diff39_HD  <= (("0"& ycbcr3)       - ("0"&HD_39)) when (quart="11") else (others=>'0');
Diff40_HD  <= (("0"& ycbcr40_reg)  - ("0"&HD_40)) when (quart="11") else (others=>'0');
Diff41_HD  <= (("0"& ycbcr41_reg)  - ("0"&HD_41)) when (quart="11") else (others=>'0');
Diff42_HD  <= (("0"& ycbcr42_reg)  - ("0"&HD_42)) when (quart="11") else (others=>'0');
Diff43_HD  <= (("0"& ycbcr43_reg)  - ("0"&HD_43)) when (quart="11") else (others=>'0');	 
Diff44_HD  <= (("0"& ycbcr4)       - ("0"&HD_44)) when (quart="11") else (others=>'0');
Diff45_HD  <= (("0"& ycbcr5)       - ("0"&HD_45)) when (quart="11") else (others=>'0');
Diff46_HD  <= (("0"& ycbcr6)       - ("0"&HD_46)) when (quart="11") else (others=>'0');  
Diff47_HD  <= (("0"& ycbcr7)       - ("0"&HD_47)) when (quart="11") else (others=>'0');
Diff48_HD  <= (("0"& ycbcr48_reg)  - ("0"&HD_48)) when (quart="11") else (others=>'0');
Diff49_HD  <= (("0"& ycbcr49_reg)  - ("0"&HD_49)) when (quart="11") else (others=>'0');
Diff50_HD  <= (("0"& ycbcr50_reg)  - ("0"&HD_50)) when (quart="11") else (others=>'0');
Diff51_HD  <= (("0"& ycbcr51_reg)  - ("0"&HD_51)) when (quart="11") else (others=>'0');
Diff52_HD  <= (("0"& ycbcr8)       - ("0"&HD_52)) when (quart="11") else (others=>'0');
Diff53_HD  <= (("0"& ycbcr9)       - ("0"&HD_53)) when (quart="11") else (others=>'0');
Diff54_HD  <= (("0"& ycbcr10)      - ("0"&HD_54)) when (quart="11") else (others=>'0');
Diff55_HD  <= (("0"& ycbcr11)      - ("0"&HD_55)) when (quart="11") else (others=>'0');
Diff56_HD  <= (("0"& ycbcr56_reg)  - ("0"&HD_56)) when (quart="11") else (others=>'0');
Diff57_HD  <= (("0"& ycbcr57_reg)  - ("0"&HD_57)) when (quart="11") else (others=>'0');
Diff58_HD  <= (("0"& ycbcr58_reg)  - ("0"&HD_58)) when (quart="11") else (others=>'0');
Diff59_HD  <= (("0"& ycbcr59_reg)  - ("0"&HD_59)) when (quart="11") else (others=>'0');	 
Diff60_HD  <= (("0"& ycbcr12)      - ("0"&HD_60)) when (quart="11") else (others=>'0');
Diff61_HD  <= (("0"& ycbcr13)      - ("0"&HD_61)) when (quart="11") else (others=>'0');
Diff62_HD  <= (("0"& ycbcr14)      - ("0"&HD_62)) when (quart="11") else (others=>'0');  
Diff63_HD  <= (("0"& ycbcr15)      - ("0"&HD_63)) when (quart="11") else (others=>'0');
--VL
Diff0_VL   <= (("0"& ycbcr0_reg)  -  ("0"&VL_0 )) when (quart="11") else (others=>'0');
Diff1_VL   <= (("0"& ycbcr1_reg)  -  ("0"&VL_1 )) when (quart="11") else (others=>'0');
Diff2_VL   <= (("0"& ycbcr2_reg)  -  ("0"&VL_2 )) when (quart="11") else (others=>'0');
Diff3_VL   <= (("0"& ycbcr3_reg)  -  ("0"&VL_3 )) when (quart="11") else (others=>'0');
Diff4_VL   <= (("0"& ycbcr4_reg)  -  ("0"&VL_4 )) when (quart="11") else (others=>'0');
Diff5_VL   <= (("0"& ycbcr5_reg)  -  ("0"&VL_5 )) when (quart="11") else (others=>'0');
Diff6_VL   <= (("0"& ycbcr6_reg)  -  ("0"&VL_6 )) when (quart="11") else (others=>'0');
Diff7_VL   <= (("0"& ycbcr7_reg)  -  ("0"&VL_7 )) when (quart="11") else (others=>'0');
Diff8_VL   <= (("0"& ycbcr8_reg)  -  ("0"&VL_8 )) when (quart="11") else (others=>'0');
Diff9_VL   <= (("0"& ycbcr9_reg)  -  ("0"&VL_9 )) when (quart="11") else (others=>'0');
Diff10_VL  <= (("0"& ycbcr10_reg)  - ("0"&VL_10)) when (quart="11") else (others=>'0');
Diff11_VL  <= (("0"& ycbcr11_reg)  - ("0"&VL_11)) when (quart="11") else (others=>'0');	 
Diff12_VL  <= (("0"& ycbcr12_reg)  - ("0"&VL_12)) when (quart="11") else (others=>'0');
Diff13_VL  <= (("0"& ycbcr13_reg)  - ("0"&VL_13)) when (quart="11") else (others=>'0');
Diff14_VL  <= (("0"& ycbcr14_reg)  - ("0"&VL_14)) when (quart="11") else (others=>'0');  
Diff15_VL  <= (("0"& ycbcr15_reg)  - ("0"&VL_15)) when (quart="11") else (others=>'0');
Diff16_VL  <= (("0"& ycbcr16_reg)  - ("0"&VL_16)) when (quart="11") else (others=>'0');
Diff17_VL  <= (("0"& ycbcr17_reg)  - ("0"&VL_17)) when (quart="11") else (others=>'0');
Diff18_VL  <= (("0"& ycbcr18_reg)  - ("0"&VL_18)) when (quart="11") else (others=>'0');
Diff19_VL  <= (("0"& ycbcr19_reg)  - ("0"&VL_19)) when (quart="11") else (others=>'0');
Diff20_VL  <= (("0"& ycbcr20_reg)  - ("0"&VL_20)) when (quart="11") else (others=>'0');
Diff21_VL  <= (("0"& ycbcr21_reg)  - ("0"&VL_21)) when (quart="11") else (others=>'0');
Diff22_VL  <= (("0"& ycbcr22_reg)  - ("0"&VL_22)) when (quart="11") else (others=>'0');
Diff23_VL  <= (("0"& ycbcr23_reg)  - ("0"&VL_23)) when (quart="11") else (others=>'0');
Diff24_VL  <= (("0"& ycbcr24_reg)  - ("0"&VL_24)) when (quart="11") else (others=>'0');
Diff25_VL  <= (("0"& ycbcr25_reg)  - ("0"&VL_25)) when (quart="11") else (others=>'0');
Diff26_VL  <= (("0"& ycbcr26_reg)  - ("0"&VL_26)) when (quart="11") else (others=>'0');
Diff27_VL  <= (("0"& ycbcr27_reg)  - ("0"&VL_27)) when (quart="11") else (others=>'0');	 
Diff28_VL  <= (("0"& ycbcr28_reg)  - ("0"&VL_28)) when (quart="11") else (others=>'0');
Diff29_VL  <= (("0"& ycbcr29_reg)  - ("0"&VL_29)) when (quart="11") else (others=>'0');
Diff30_VL  <= (("0"& ycbcr30_reg)  - ("0"&VL_30)) when (quart="11") else (others=>'0');  
Diff31_VL  <= (("0"& ycbcr31_reg)  - ("0"&VL_31)) when (quart="11") else (others=>'0'); 
Diff32_VL  <= (("0"& ycbcr32_reg)  - ("0"&VL_32)) when (quart="11") else (others=>'0');
Diff33_VL  <= (("0"& ycbcr33_reg)  - ("0"&VL_33)) when (quart="11") else (others=>'0');
Diff34_VL  <= (("0"& ycbcr34_reg)  - ("0"&VL_34)) when (quart="11") else (others=>'0');
Diff35_VL  <= (("0"& ycbcr35_reg)  - ("0"&VL_35)) when (quart="11") else (others=>'0');
Diff36_VL  <= (("0"& ycbcr0)       - ("0"&VL_36)) when (quart="11") else (others=>'0');
Diff37_VL  <= (("0"& ycbcr1)       - ("0"&VL_37)) when (quart="11") else (others=>'0');
Diff38_VL  <= (("0"& ycbcr2)       - ("0"&VL_38)) when (quart="11") else (others=>'0');
Diff39_VL  <= (("0"& ycbcr3)       - ("0"&VL_39)) when (quart="11") else (others=>'0');
Diff40_VL  <= (("0"& ycbcr40_reg)  - ("0"&VL_40)) when (quart="11") else (others=>'0');
Diff41_VL  <= (("0"& ycbcr41_reg)  - ("0"&VL_41)) when (quart="11") else (others=>'0');
Diff42_VL  <= (("0"& ycbcr42_reg)  - ("0"&VL_42)) when (quart="11") else (others=>'0');
Diff43_VL  <= (("0"& ycbcr43_reg)  - ("0"&VL_43)) when (quart="11") else (others=>'0');	 
Diff44_VL  <= (("0"& ycbcr4)       - ("0"&VL_44)) when (quart="11") else (others=>'0');
Diff45_VL  <= (("0"& ycbcr5)       - ("0"&VL_45)) when (quart="11") else (others=>'0');
Diff46_VL  <= (("0"& ycbcr6)       - ("0"&VL_46)) when (quart="11") else (others=>'0');  
Diff47_VL  <= (("0"& ycbcr7)       - ("0"&VL_47)) when (quart="11") else (others=>'0');
Diff48_VL  <= (("0"& ycbcr48_reg)  - ("0"&VL_48)) when (quart="11") else (others=>'0');
Diff49_VL  <= (("0"& ycbcr49_reg)  - ("0"&VL_49)) when (quart="11") else (others=>'0');
Diff50_VL  <= (("0"& ycbcr50_reg)  - ("0"&VL_50)) when (quart="11") else (others=>'0');
Diff51_VL  <= (("0"& ycbcr51_reg)  - ("0"&VL_51)) when (quart="11") else (others=>'0');
Diff52_VL  <= (("0"& ycbcr8)       - ("0"&VL_52)) when (quart="11") else (others=>'0');
Diff53_VL  <= (("0"& ycbcr9)       - ("0"&VL_53)) when (quart="11") else (others=>'0');
Diff54_VL  <= (("0"& ycbcr10)      - ("0"&VL_54)) when (quart="11") else (others=>'0');
Diff55_VL  <= (("0"& ycbcr11)      - ("0"&VL_55)) when (quart="11") else (others=>'0');
Diff56_VL  <= (("0"& ycbcr56_reg)  - ("0"&VL_56)) when (quart="11") else (others=>'0');
Diff57_VL  <= (("0"& ycbcr57_reg)  - ("0"&VL_57)) when (quart="11") else (others=>'0');
Diff58_VL  <= (("0"& ycbcr58_reg)  - ("0"&VL_58)) when (quart="11") else (others=>'0');
Diff59_VL  <= (("0"& ycbcr59_reg)  - ("0"&VL_59)) when (quart="11") else (others=>'0');	 
Diff60_VL  <= (("0"& ycbcr12)      - ("0"&VL_60)) when (quart="11") else (others=>'0');
Diff61_VL  <= (("0"& ycbcr13)      - ("0"&VL_61)) when (quart="11") else (others=>'0');
Diff62_VL  <= (("0"& ycbcr14)      - ("0"&VL_62)) when (quart="11") else (others=>'0');  
Diff63_VL  <= (("0"& ycbcr15)      - ("0"&VL_63)) when (quart="11") else (others=>'0');
--HU
Diff0_HU   <= (("0"& ycbcr0_reg)  -  ("0"&HU_0 )) when (quart="11") else (others=>'0');
Diff1_HU   <= (("0"& ycbcr1_reg)  -  ("0"&HU_1 )) when (quart="11") else (others=>'0');
Diff2_HU   <= (("0"& ycbcr2_reg)  -  ("0"&HU_2 )) when (quart="11") else (others=>'0');
Diff3_HU   <= (("0"& ycbcr3_reg)  -  ("0"&HU_3 )) when (quart="11") else (others=>'0');
Diff4_HU   <= (("0"& ycbcr4_reg)  -  ("0"&HU_4 )) when (quart="11") else (others=>'0');
Diff5_HU   <= (("0"& ycbcr5_reg)  -  ("0"&HU_5 )) when (quart="11") else (others=>'0');
Diff6_HU   <= (("0"& ycbcr6_reg)  -  ("0"&HU_6 )) when (quart="11") else (others=>'0');
Diff7_HU   <= (("0"& ycbcr7_reg)  -  ("0"&HU_7 )) when (quart="11") else (others=>'0');
Diff8_HU   <= (("0"& ycbcr8_reg)  -  ("0"&HU_8 )) when (quart="11") else (others=>'0');
Diff9_HU   <= (("0"& ycbcr9_reg)  -  ("0"&HU_9 )) when (quart="11") else (others=>'0');
Diff10_HU  <= (("0"& ycbcr10_reg)  - ("0"&HU_10)) when (quart="11") else (others=>'0');
Diff11_HU  <= (("0"& ycbcr11_reg)  - ("0"&HU_11)) when (quart="11") else (others=>'0');	 
Diff12_HU  <= (("0"& ycbcr12_reg)  - ("0"&HU_12)) when (quart="11") else (others=>'0');
Diff13_HU  <= (("0"& ycbcr13_reg)  - ("0"&HU_13)) when (quart="11") else (others=>'0');
Diff14_HU  <= (("0"& ycbcr14_reg)  - ("0"&HU_14)) when (quart="11") else (others=>'0');  
Diff15_HU  <= (("0"& ycbcr15_reg)  - ("0"&HU_15)) when (quart="11") else (others=>'0');
Diff16_HU  <= (("0"& ycbcr16_reg)  - ("0"&HU_16)) when (quart="11") else (others=>'0');
Diff17_HU  <= (("0"& ycbcr17_reg)  - ("0"&HU_17)) when (quart="11") else (others=>'0');
Diff18_HU  <= (("0"& ycbcr18_reg)  - ("0"&HU_18)) when (quart="11") else (others=>'0');
Diff19_HU  <= (("0"& ycbcr19_reg)  - ("0"&HU_19)) when (quart="11") else (others=>'0');
Diff20_HU  <= (("0"& ycbcr20_reg)  - ("0"&HU_20)) when (quart="11") else (others=>'0');
Diff21_HU  <= (("0"& ycbcr21_reg)  - ("0"&HU_21)) when (quart="11") else (others=>'0');
Diff22_HU  <= (("0"& ycbcr22_reg)  - ("0"&HU_22)) when (quart="11") else (others=>'0');
Diff23_HU  <= (("0"& ycbcr23_reg)  - ("0"&HU_23)) when (quart="11") else (others=>'0');
Diff24_HU  <= (("0"& ycbcr24_reg)  - ("0"&HU_24)) when (quart="11") else (others=>'0');
Diff25_HU  <= (("0"& ycbcr25_reg)  - ("0"&HU_25)) when (quart="11") else (others=>'0');
Diff26_HU  <= (("0"& ycbcr26_reg)  - ("0"&HU_26)) when (quart="11") else (others=>'0');
Diff27_HU  <= (("0"& ycbcr27_reg)  - ("0"&HU_27)) when (quart="11") else (others=>'0');	 
Diff28_HU  <= (("0"& ycbcr28_reg)  - ("0"&HU_28)) when (quart="11") else (others=>'0');
Diff29_HU  <= (("0"& ycbcr29_reg)  - ("0"&HU_29)) when (quart="11") else (others=>'0');
Diff30_HU  <= (("0"& ycbcr30_reg)  - ("0"&HU_30)) when (quart="11") else (others=>'0');  
Diff31_HU  <= (("0"& ycbcr31_reg)  - ("0"&HU_31)) when (quart="11") else (others=>'0'); 
Diff32_HU  <= (("0"& ycbcr32_reg)  - ("0"&HU_32)) when (quart="11") else (others=>'0');
Diff33_HU  <= (("0"& ycbcr33_reg)  - ("0"&HU_33)) when (quart="11") else (others=>'0');
Diff34_HU  <= (("0"& ycbcr34_reg)  - ("0"&HU_34)) when (quart="11") else (others=>'0');
Diff35_HU  <= (("0"& ycbcr35_reg)  - ("0"&HU_35)) when (quart="11") else (others=>'0');
Diff36_HU  <= (("0"& ycbcr0)       - ("0"&HU_36)) when (quart="11") else (others=>'0');
Diff37_HU  <= (("0"& ycbcr1)       - ("0"&HU_37)) when (quart="11") else (others=>'0');
Diff38_HU  <= (("0"& ycbcr2)       - ("0"&HU_38)) when (quart="11") else (others=>'0');
Diff39_HU  <= (("0"& ycbcr3)       - ("0"&HU_39)) when (quart="11") else (others=>'0');
Diff40_HU  <= (("0"& ycbcr40_reg)  - ("0"&HU_40)) when (quart="11") else (others=>'0');
Diff41_HU  <= (("0"& ycbcr41_reg)  - ("0"&HU_41)) when (quart="11") else (others=>'0');
Diff42_HU  <= (("0"& ycbcr42_reg)  - ("0"&HU_42)) when (quart="11") else (others=>'0');
Diff43_HU  <= (("0"& ycbcr43_reg)  - ("0"&HU_43)) when (quart="11") else (others=>'0');	 
Diff44_HU  <= (("0"& ycbcr4)       - ("0"&HU_44)) when (quart="11") else (others=>'0');
Diff45_HU  <= (("0"& ycbcr5)       - ("0"&HU_45)) when (quart="11") else (others=>'0');
Diff46_HU  <= (("0"& ycbcr6)       - ("0"&HU_46)) when (quart="11") else (others=>'0');  
Diff47_HU  <= (("0"& ycbcr7)       - ("0"&HU_47)) when (quart="11") else (others=>'0');
Diff48_HU  <= (("0"& ycbcr48_reg)  - ("0"&HU_48)) when (quart="11") else (others=>'0');
Diff49_HU  <= (("0"& ycbcr49_reg)  - ("0"&HU_49)) when (quart="11") else (others=>'0');
Diff50_HU  <= (("0"& ycbcr50_reg)  - ("0"&HU_50)) when (quart="11") else (others=>'0');
Diff51_HU  <= (("0"& ycbcr51_reg)  - ("0"&HU_51)) when (quart="11") else (others=>'0');
Diff52_HU  <= (("0"& ycbcr8)       - ("0"&HU_52)) when (quart="11") else (others=>'0');
Diff53_HU  <= (("0"& ycbcr9)       - ("0"&HU_53)) when (quart="11") else (others=>'0');
Diff54_HU  <= (("0"& ycbcr10)      - ("0"&HU_54)) when (quart="11") else (others=>'0');
Diff55_HU  <= (("0"& ycbcr11)      - ("0"&HU_55)) when (quart="11") else (others=>'0');
Diff56_HU  <= (("0"& ycbcr56_reg)  - ("0"&HU_56)) when (quart="11") else (others=>'0');
Diff57_HU  <= (("0"& ycbcr57_reg)  - ("0"&HU_57)) when (quart="11") else (others=>'0');
Diff58_HU  <= (("0"& ycbcr58_reg)  - ("0"&HU_58)) when (quart="11") else (others=>'0');
Diff59_HU  <= (("0"& ycbcr59_reg)  - ("0"&HU_59)) when (quart="11") else (others=>'0');	 
Diff60_HU  <= (("0"& ycbcr12)      - ("0"&HU_60)) when (quart="11") else (others=>'0');
Diff61_HU  <= (("0"& ycbcr13)      - ("0"&HU_61)) when (quart="11") else (others=>'0');
Diff62_HU  <= (("0"& ycbcr14)      - ("0"&HU_62)) when (quart="11") else (others=>'0');  
Diff63_HU  <= (("0"& ycbcr15)      - ("0"&HU_63)) when (quart="11") else (others=>'0'); 

-------------------------------------------------------------------------------------------------------------------------
--#####################################################################################################################--
--# SAD stuff :(
--#####################################################################################################################--
-------------------------------------------------------------------------------------------------------------------------

-- Get the absolute value of all the diffs. These shall be summed later
-- approx 600 lines 
SAD0_V     <= Diff0_V(num downto 0) when Diff0_V(num+1) = '0' else (not(Diff0_V(num downto 0))+1);
SAD1_V     <= Diff1_V(num downto 0) when Diff1_V(num+1) = '0' else (not(Diff1_V(num downto 0))+1);
SAD2_V     <= Diff2_V(num downto 0) when Diff2_V(num+1) = '0' else (not(Diff2_V(num downto 0))+1);
SAD3_V     <= Diff3_V(num downto 0) when Diff3_V(num+1) = '0' else (not(Diff3_V(num downto 0))+1);
SAD4_V     <= Diff4_V(num downto 0) when Diff4_V(num+1) = '0' else (not(Diff4_V(num downto 0))+1);
SAD5_V     <= Diff5_V(num downto 0) when Diff5_V(num+1) = '0' else (not(Diff5_V(num downto 0))+1);
SAD6_V     <= Diff6_V(num downto 0) when Diff6_V(num+1) = '0' else (not(Diff6_V(num downto 0))+1);
SAD7_V     <= Diff7_V(num downto 0) when Diff7_V(num+1) = '0' else (not(Diff7_V(num downto 0))+1);
SAD8_V     <= Diff8_V(num downto 0) when Diff8_V(num+1) = '0' else (not(Diff8_V(num downto 0))+1);
SAD9_V     <= Diff9_V(num downto 0) when Diff9_V(num+1) = '0' else (not(Diff9_V(num downto 0))+1);
SAD10_V    <= Diff10_V(num downto 0) when Diff10_V(num+1) = '0' else (not(Diff10_V(num downto 0))+1);
SAD11_V    <= Diff11_V(num downto 0) when Diff11_V(num+1) = '0' else (not(Diff11_V(num downto 0))+1);
SAD12_V    <= Diff12_V(num downto 0) when Diff12_V(num+1) = '0' else (not(Diff12_V(num downto 0))+1);
SAD13_V    <= Diff13_V(num downto 0) when Diff13_V(num+1) = '0' else (not(Diff13_V(num downto 0))+1);
SAD14_V    <= Diff14_V(num downto 0) when Diff14_V(num+1) = '0' else (not(Diff14_V(num downto 0))+1);		
SAD15_V    <= Diff15_V(num downto 0) when Diff15_V(num+1) = '0' else (not(Diff15_V(num downto 0))+1);
SAD16_V    <= Diff16_V(num downto 0) when Diff16_V(num+1) = '0' else (not(Diff16_V(num downto 0))+1);
SAD17_V    <= Diff17_V(num downto 0) when Diff17_V(num+1) = '0' else (not(Diff17_V(num downto 0))+1);
SAD18_V    <= Diff18_V(num downto 0) when Diff18_V(num+1) = '0' else (not(Diff18_V(num downto 0))+1);
SAD19_V    <= Diff19_V(num downto 0) when Diff19_V(num+1) = '0' else (not(Diff19_V(num downto 0))+1);
SAD20_V    <= Diff20_V(num downto 0) when Diff20_V(num+1) = '0' else (not(Diff20_V(num downto 0))+1);
SAD21_V    <= Diff21_V(num downto 0) when Diff21_V(num+1) = '0' else (not(Diff21_V(num downto 0))+1);
SAD22_V    <= Diff22_V(num downto 0) when Diff22_V(num+1) = '0' else (not(Diff22_V(num downto 0))+1);
SAD23_V    <= Diff23_V(num downto 0) when Diff23_V(num+1) = '0' else (not(Diff23_V(num downto 0))+1);
SAD24_V    <= Diff24_V(num downto 0) when Diff24_V(num+1) = '0' else (not(Diff24_V(num downto 0))+1);
SAD25_V    <= Diff25_V(num downto 0) when Diff25_V(num+1) = '0' else (not(Diff25_V(num downto 0))+1);
SAD26_V    <= Diff26_V(num downto 0) when Diff26_V(num+1) = '0' else (not(Diff26_V(num downto 0))+1);
SAD27_V    <= Diff27_V(num downto 0) when Diff27_V(num+1) = '0' else (not(Diff27_V(num downto 0))+1);
SAD28_V    <= Diff28_V(num downto 0) when Diff28_V(num+1) = '0' else (not(Diff28_V(num downto 0))+1);
SAD29_V    <= Diff29_V(num downto 0) when Diff29_V(num+1) = '0' else (not(Diff29_V(num downto 0))+1);
SAD30_V    <= Diff30_V(num downto 0) when Diff30_V(num+1) = '0' else (not(Diff30_V(num downto 0))+1);		
SAD31_V    <= Diff31_V(num downto 0) when Diff31_V(num+1) = '0' else (not(Diff31_V(num downto 0))+1);
SAD32_V    <= Diff32_V(num downto 0) when Diff32_V(num+1) = '0' else (not(Diff32_V(num downto 0))+1);
SAD33_V    <= Diff33_V(num downto 0) when Diff33_V(num+1) = '0' else (not(Diff33_V(num downto 0))+1);
SAD34_V    <= Diff34_V(num downto 0) when Diff34_V(num+1) = '0' else (not(Diff34_V(num downto 0))+1);
SAD35_V    <= Diff35_V(num downto 0) when Diff35_V(num+1) = '0' else (not(Diff35_V(num downto 0))+1);
SAD36_V    <= Diff36_V(num downto 0) when Diff36_V(num+1) = '0' else (not(Diff36_V(num downto 0))+1);
SAD37_V    <= Diff37_V(num downto 0) when Diff37_V(num+1) = '0' else (not(Diff37_V(num downto 0))+1);
SAD38_V    <= Diff38_V(num downto 0) when Diff38_V(num+1) = '0' else (not(Diff38_V(num downto 0))+1);
SAD39_V    <= Diff39_V(num downto 0) when Diff39_V(num+1) = '0' else (not(Diff39_V(num downto 0))+1);
SAD40_V    <= Diff40_V(num downto 0) when Diff40_V(num+1) = '0' else (not(Diff40_V(num downto 0))+1);
SAD41_V    <= Diff41_V(num downto 0) when Diff41_V(num+1) = '0' else (not(Diff41_V(num downto 0))+1);
SAD42_V    <= Diff42_V(num downto 0) when Diff42_V(num+1) = '0' else (not(Diff42_V(num downto 0))+1);
SAD43_V    <= Diff43_V(num downto 0) when Diff43_V(num+1) = '0' else (not(Diff43_V(num downto 0))+1);
SAD44_V    <= Diff44_V(num downto 0) when Diff44_V(num+1) = '0' else (not(Diff44_V(num downto 0))+1);
SAD45_V    <= Diff45_V(num downto 0) when Diff45_V(num+1) = '0' else (not(Diff45_V(num downto 0))+1);
SAD46_V    <= Diff46_V(num downto 0) when Diff46_V(num+1) = '0' else (not(Diff46_V(num downto 0))+1);		
SAD47_V    <= Diff47_V(num downto 0) when Diff47_V(num+1) = '0' else (not(Diff47_V(num downto 0))+1);
SAD48_V    <= Diff48_V(num downto 0) when Diff48_V(num+1) = '0' else (not(Diff48_V(num downto 0))+1);
SAD49_V    <= Diff49_V(num downto 0) when Diff49_V(num+1) = '0' else (not(Diff49_V(num downto 0))+1);
SAD50_V    <= Diff50_V(num downto 0) when Diff50_V(num+1) = '0' else (not(Diff50_V(num downto 0))+1);
SAD51_V    <= Diff51_V(num downto 0) when Diff51_V(num+1) = '0' else (not(Diff51_V(num downto 0))+1);
SAD52_V    <= Diff52_V(num downto 0) when Diff52_V(num+1) = '0' else (not(Diff52_V(num downto 0))+1);
SAD53_V    <= Diff53_V(num downto 0) when Diff53_V(num+1) = '0' else (not(Diff53_V(num downto 0))+1);
SAD54_V    <= Diff54_V(num downto 0) when Diff54_V(num+1) = '0' else (not(Diff54_V(num downto 0))+1);
SAD55_V    <= Diff55_V(num downto 0) when Diff55_V(num+1) = '0' else (not(Diff55_V(num downto 0))+1);
SAD56_V    <= Diff56_V(num downto 0) when Diff56_V(num+1) = '0' else (not(Diff56_V(num downto 0))+1);
SAD57_V    <= Diff57_V(num downto 0) when Diff57_V(num+1) = '0' else (not(Diff57_V(num downto 0))+1);
SAD58_V    <= Diff58_V(num downto 0) when Diff58_V(num+1) = '0' else (not(Diff58_V(num downto 0))+1);
SAD59_V    <= Diff59_V(num downto 0) when Diff59_V(num+1) = '0' else (not(Diff59_V(num downto 0))+1);
SAD60_V    <= Diff60_V(num downto 0) when Diff60_V(num+1) = '0' else (not(Diff60_V(num downto 0))+1);
SAD61_V    <= Diff61_V(num downto 0) when Diff61_V(num+1) = '0' else (not(Diff61_V(num downto 0))+1);
SAD62_V    <= Diff62_V(num downto 0) when Diff62_V(num+1) = '0' else (not(Diff62_V(num downto 0))+1);		
SAD63_V    <= Diff63_V(num downto 0) when Diff63_V(num+1) = '0' else (not(Diff63_V(num downto 0))+1);
--Horizontal
SAD0_H     <= Diff0_H(num downto 0) when Diff0_H(num+1) = '0' else (not(Diff0_H(num downto 0))+1);
SAD1_H     <= Diff1_H(num downto 0) when Diff1_H(num+1) = '0' else (not(Diff1_H(num downto 0))+1);
SAD2_H     <= Diff2_H(num downto 0) when Diff2_H(num+1) = '0' else (not(Diff2_H(num downto 0))+1);
SAD3_H     <= Diff3_H(num downto 0) when Diff3_H(num+1) = '0' else (not(Diff3_H(num downto 0))+1);
SAD4_H     <= Diff4_H(num downto 0) when Diff4_H(num+1) = '0' else (not(Diff4_H(num downto 0))+1);
SAD5_H     <= Diff5_H(num downto 0) when Diff5_H(num+1) = '0' else (not(Diff5_H(num downto 0))+1);
SAD6_H     <= Diff6_H(num downto 0) when Diff6_H(num+1) = '0' else (not(Diff6_H(num downto 0))+1);
SAD7_H     <= Diff7_H(num downto 0) when Diff7_H(num+1) = '0' else (not(Diff7_H(num downto 0))+1);
SAD8_H     <= Diff8_H(num downto 0) when Diff8_H(num+1) = '0' else (not(Diff8_H(num downto 0))+1);
SAD9_H     <= Diff9_H(num downto 0) when Diff9_H(num+1) = '0' else (not(Diff9_H(num downto 0))+1);
SAD10_H    <= Diff10_H(num downto 0) when Diff10_H(num+1) = '0' else (not(Diff10_H(num downto 0))+1);
SAD11_H    <= Diff11_H(num downto 0) when Diff11_H(num+1) = '0' else (not(Diff11_H(num downto 0))+1);
SAD12_H    <= Diff12_H(num downto 0) when Diff12_H(num+1) = '0' else (not(Diff12_H(num downto 0))+1);
SAD13_H    <= Diff13_H(num downto 0) when Diff13_H(num+1) = '0' else (not(Diff13_H(num downto 0))+1);
SAD14_H    <= Diff14_H(num downto 0) when Diff14_H(num+1) = '0' else (not(Diff14_H(num downto 0))+1);		
SAD15_H    <= Diff15_H(num downto 0) when Diff15_H(num+1) = '0' else (not(Diff15_H(num downto 0))+1);
SAD16_H    <= Diff16_H(num downto 0) when Diff16_H(num+1) = '0' else (not(Diff16_H(num downto 0))+1);
SAD17_H    <= Diff17_H(num downto 0) when Diff17_H(num+1) = '0' else (not(Diff17_H(num downto 0))+1);
SAD18_H    <= Diff18_H(num downto 0) when Diff18_H(num+1) = '0' else (not(Diff18_H(num downto 0))+1);
SAD19_H    <= Diff19_H(num downto 0) when Diff19_H(num+1) = '0' else (not(Diff19_H(num downto 0))+1);
SAD20_H    <= Diff20_H(num downto 0) when Diff20_H(num+1) = '0' else (not(Diff20_H(num downto 0))+1);
SAD21_H    <= Diff21_H(num downto 0) when Diff21_H(num+1) = '0' else (not(Diff21_H(num downto 0))+1);
SAD22_H    <= Diff22_H(num downto 0) when Diff22_H(num+1) = '0' else (not(Diff22_H(num downto 0))+1);
SAD23_H    <= Diff23_H(num downto 0) when Diff23_H(num+1) = '0' else (not(Diff23_H(num downto 0))+1);
SAD24_H    <= Diff24_H(num downto 0) when Diff24_H(num+1) = '0' else (not(Diff24_H(num downto 0))+1);
SAD25_H    <= Diff25_H(num downto 0) when Diff25_H(num+1) = '0' else (not(Diff25_H(num downto 0))+1);
SAD26_H    <= Diff26_H(num downto 0) when Diff26_H(num+1) = '0' else (not(Diff26_H(num downto 0))+1);
SAD27_H    <= Diff27_H(num downto 0) when Diff27_H(num+1) = '0' else (not(Diff27_H(num downto 0))+1);
SAD28_H    <= Diff28_H(num downto 0) when Diff28_H(num+1) = '0' else (not(Diff28_H(num downto 0))+1);
SAD29_H    <= Diff29_H(num downto 0) when Diff29_H(num+1) = '0' else (not(Diff29_H(num downto 0))+1);
SAD30_H    <= Diff30_H(num downto 0) when Diff30_H(num+1) = '0' else (not(Diff30_H(num downto 0))+1);		
SAD31_H    <= Diff31_H(num downto 0) when Diff31_H(num+1) = '0' else (not(Diff31_H(num downto 0))+1);
SAD32_H    <= Diff32_H(num downto 0) when Diff32_H(num+1) = '0' else (not(Diff32_H(num downto 0))+1);
SAD33_H    <= Diff33_H(num downto 0) when Diff33_H(num+1) = '0' else (not(Diff33_H(num downto 0))+1);
SAD34_H    <= Diff34_H(num downto 0) when Diff34_H(num+1) = '0' else (not(Diff34_H(num downto 0))+1);
SAD35_H    <= Diff35_H(num downto 0) when Diff35_H(num+1) = '0' else (not(Diff35_H(num downto 0))+1);
SAD36_H    <= Diff36_H(num downto 0) when Diff36_H(num+1) = '0' else (not(Diff36_H(num downto 0))+1);
SAD37_H    <= Diff37_H(num downto 0) when Diff37_H(num+1) = '0' else (not(Diff37_H(num downto 0))+1);
SAD38_H    <= Diff38_H(num downto 0) when Diff38_H(num+1) = '0' else (not(Diff38_H(num downto 0))+1);
SAD39_H    <= Diff39_H(num downto 0) when Diff39_H(num+1) = '0' else (not(Diff39_H(num downto 0))+1);
SAD40_H    <= Diff40_H(num downto 0) when Diff40_H(num+1) = '0' else (not(Diff40_H(num downto 0))+1);
SAD41_H    <= Diff41_H(num downto 0) when Diff41_H(num+1) = '0' else (not(Diff41_H(num downto 0))+1);
SAD42_H    <= Diff42_H(num downto 0) when Diff42_H(num+1) = '0' else (not(Diff42_H(num downto 0))+1);
SAD43_H    <= Diff43_H(num downto 0) when Diff43_H(num+1) = '0' else (not(Diff43_H(num downto 0))+1);
SAD44_H    <= Diff44_H(num downto 0) when Diff44_H(num+1) = '0' else (not(Diff44_H(num downto 0))+1);
SAD45_H    <= Diff45_H(num downto 0) when Diff45_H(num+1) = '0' else (not(Diff45_H(num downto 0))+1);
SAD46_H    <= Diff46_H(num downto 0) when Diff46_H(num+1) = '0' else (not(Diff46_H(num downto 0))+1);		
SAD47_H    <= Diff47_H(num downto 0) when Diff47_H(num+1) = '0' else (not(Diff47_H(num downto 0))+1);
SAD48_H    <= Diff48_H(num downto 0) when Diff48_H(num+1) = '0' else (not(Diff48_H(num downto 0))+1);
SAD49_H    <= Diff49_H(num downto 0) when Diff49_H(num+1) = '0' else (not(Diff49_H(num downto 0))+1);
SAD50_H    <= Diff50_H(num downto 0) when Diff50_H(num+1) = '0' else (not(Diff50_H(num downto 0))+1);
SAD51_H    <= Diff51_H(num downto 0) when Diff51_H(num+1) = '0' else (not(Diff51_H(num downto 0))+1);
SAD52_H    <= Diff52_H(num downto 0) when Diff52_H(num+1) = '0' else (not(Diff52_H(num downto 0))+1);
SAD53_H    <= Diff53_H(num downto 0) when Diff53_H(num+1) = '0' else (not(Diff53_H(num downto 0))+1);
SAD54_H    <= Diff54_H(num downto 0) when Diff54_H(num+1) = '0' else (not(Diff54_H(num downto 0))+1);
SAD55_H    <= Diff55_H(num downto 0) when Diff55_H(num+1) = '0' else (not(Diff55_H(num downto 0))+1);
SAD56_H    <= Diff56_H(num downto 0) when Diff56_H(num+1) = '0' else (not(Diff56_H(num downto 0))+1);
SAD57_H    <= Diff57_H(num downto 0) when Diff57_H(num+1) = '0' else (not(Diff57_H(num downto 0))+1);
SAD58_H    <= Diff58_H(num downto 0) when Diff58_H(num+1) = '0' else (not(Diff58_H(num downto 0))+1);
SAD59_H    <= Diff59_H(num downto 0) when Diff59_H(num+1) = '0' else (not(Diff59_H(num downto 0))+1);
SAD60_H    <= Diff60_H(num downto 0) when Diff60_H(num+1) = '0' else (not(Diff60_H(num downto 0))+1);
SAD61_H    <= Diff61_H(num downto 0) when Diff61_H(num+1) = '0' else (not(Diff61_H(num downto 0))+1);
SAD62_H    <= Diff62_H(num downto 0) when Diff62_H(num+1) = '0' else (not(Diff62_H(num downto 0))+1);		
SAD63_H    <= Diff63_H(num downto 0) when Diff63_H(num+1) = '0' else (not(Diff63_H(num downto 0))+1);
--DC
SAD0_DC     <= Diff0_DC(num downto 0) when Diff0_DC(num+1) = '0' else (not(Diff0_DC(num downto 0))+1);
SAD1_DC     <= Diff1_DC(num downto 0) when Diff1_DC(num+1) = '0' else (not(Diff1_DC(num downto 0))+1);
SAD2_DC     <= Diff2_DC(num downto 0) when Diff2_DC(num+1) = '0' else (not(Diff2_DC(num downto 0))+1);
SAD3_DC     <= Diff3_DC(num downto 0) when Diff3_DC(num+1) = '0' else (not(Diff3_DC(num downto 0))+1);
SAD4_DC     <= Diff4_DC(num downto 0) when Diff4_DC(num+1) = '0' else (not(Diff4_DC(num downto 0))+1);
SAD5_DC     <= Diff5_DC(num downto 0) when Diff5_DC(num+1) = '0' else (not(Diff5_DC(num downto 0))+1);
SAD6_DC     <= Diff6_DC(num downto 0) when Diff6_DC(num+1) = '0' else (not(Diff6_DC(num downto 0))+1);
SAD7_DC     <= Diff7_DC(num downto 0) when Diff7_DC(num+1) = '0' else (not(Diff7_DC(num downto 0))+1);
SAD8_DC     <= Diff8_DC(num downto 0) when Diff8_DC(num+1) = '0' else (not(Diff8_DC(num downto 0))+1);
SAD9_DC     <= Diff9_DC(num downto 0) when Diff9_DC(num+1) = '0' else (not(Diff9_DC(num downto 0))+1);
SAD10_DC    <= Diff10_DC(num downto 0) when Diff10_DC(num+1) = '0' else (not(Diff10_DC(num downto 0))+1);
SAD11_DC    <= Diff11_DC(num downto 0) when Diff11_DC(num+1) = '0' else (not(Diff11_DC(num downto 0))+1);
SAD12_DC    <= Diff12_DC(num downto 0) when Diff12_DC(num+1) = '0' else (not(Diff12_DC(num downto 0))+1);
SAD13_DC    <= Diff13_DC(num downto 0) when Diff13_DC(num+1) = '0' else (not(Diff13_DC(num downto 0))+1);
SAD14_DC    <= Diff14_DC(num downto 0) when Diff14_DC(num+1) = '0' else (not(Diff14_DC(num downto 0))+1);		
SAD15_DC    <= Diff15_DC(num downto 0) when Diff15_DC(num+1) = '0' else (not(Diff15_DC(num downto 0))+1);
SAD16_DC    <= Diff16_DC(num downto 0) when Diff16_DC(num+1) = '0' else (not(Diff16_DC(num downto 0))+1);
SAD17_DC    <= Diff17_DC(num downto 0) when Diff17_DC(num+1) = '0' else (not(Diff17_DC(num downto 0))+1);
SAD18_DC    <= Diff18_DC(num downto 0) when Diff18_DC(num+1) = '0' else (not(Diff18_DC(num downto 0))+1);
SAD19_DC    <= Diff19_DC(num downto 0) when Diff19_DC(num+1) = '0' else (not(Diff19_DC(num downto 0))+1);
SAD20_DC    <= Diff20_DC(num downto 0) when Diff20_DC(num+1) = '0' else (not(Diff20_DC(num downto 0))+1);
SAD21_DC    <= Diff21_DC(num downto 0) when Diff21_DC(num+1) = '0' else (not(Diff21_DC(num downto 0))+1);
SAD22_DC    <= Diff22_DC(num downto 0) when Diff22_DC(num+1) = '0' else (not(Diff22_DC(num downto 0))+1);
SAD23_DC    <= Diff23_DC(num downto 0) when Diff23_DC(num+1) = '0' else (not(Diff23_DC(num downto 0))+1);
SAD24_DC    <= Diff24_DC(num downto 0) when Diff24_DC(num+1) = '0' else (not(Diff24_DC(num downto 0))+1);
SAD25_DC    <= Diff25_DC(num downto 0) when Diff25_DC(num+1) = '0' else (not(Diff25_DC(num downto 0))+1);
SAD26_DC    <= Diff26_DC(num downto 0) when Diff26_DC(num+1) = '0' else (not(Diff26_DC(num downto 0))+1);
SAD27_DC    <= Diff27_DC(num downto 0) when Diff27_DC(num+1) = '0' else (not(Diff27_DC(num downto 0))+1);
SAD28_DC    <= Diff28_DC(num downto 0) when Diff28_DC(num+1) = '0' else (not(Diff28_DC(num downto 0))+1);
SAD29_DC    <= Diff29_DC(num downto 0) when Diff29_DC(num+1) = '0' else (not(Diff29_DC(num downto 0))+1);
SAD30_DC    <= Diff30_DC(num downto 0) when Diff30_DC(num+1) = '0' else (not(Diff30_DC(num downto 0))+1);		
SAD31_DC    <= Diff31_DC(num downto 0) when Diff31_DC(num+1) = '0' else (not(Diff31_DC(num downto 0))+1);
SAD32_DC    <= Diff32_DC(num downto 0) when Diff32_DC(num+1) = '0' else (not(Diff32_DC(num downto 0))+1);
SAD33_DC    <= Diff33_DC(num downto 0) when Diff33_DC(num+1) = '0' else (not(Diff33_DC(num downto 0))+1);
SAD34_DC    <= Diff34_DC(num downto 0) when Diff34_DC(num+1) = '0' else (not(Diff34_DC(num downto 0))+1);
SAD35_DC    <= Diff35_DC(num downto 0) when Diff35_DC(num+1) = '0' else (not(Diff35_DC(num downto 0))+1);
SAD36_DC    <= Diff36_DC(num downto 0) when Diff36_DC(num+1) = '0' else (not(Diff36_DC(num downto 0))+1);
SAD37_DC    <= Diff37_DC(num downto 0) when Diff37_DC(num+1) = '0' else (not(Diff37_DC(num downto 0))+1);
SAD38_DC    <= Diff38_DC(num downto 0) when Diff38_DC(num+1) = '0' else (not(Diff38_DC(num downto 0))+1);
SAD39_DC    <= Diff39_DC(num downto 0) when Diff39_DC(num+1) = '0' else (not(Diff39_DC(num downto 0))+1);
SAD40_DC    <= Diff40_DC(num downto 0) when Diff40_DC(num+1) = '0' else (not(Diff40_DC(num downto 0))+1);
SAD41_DC    <= Diff41_DC(num downto 0) when Diff41_DC(num+1) = '0' else (not(Diff41_DC(num downto 0))+1);
SAD42_DC    <= Diff42_DC(num downto 0) when Diff42_DC(num+1) = '0' else (not(Diff42_DC(num downto 0))+1);
SAD43_DC    <= Diff43_DC(num downto 0) when Diff43_DC(num+1) = '0' else (not(Diff43_DC(num downto 0))+1);
SAD44_DC    <= Diff44_DC(num downto 0) when Diff44_DC(num+1) = '0' else (not(Diff44_DC(num downto 0))+1);
SAD45_DC    <= Diff45_DC(num downto 0) when Diff45_DC(num+1) = '0' else (not(Diff45_DC(num downto 0))+1);
SAD46_DC    <= Diff46_DC(num downto 0) when Diff46_DC(num+1) = '0' else (not(Diff46_DC(num downto 0))+1);		
SAD47_DC    <= Diff47_DC(num downto 0) when Diff47_DC(num+1) = '0' else (not(Diff47_DC(num downto 0))+1);
SAD48_DC    <= Diff48_DC(num downto 0) when Diff48_DC(num+1) = '0' else (not(Diff48_DC(num downto 0))+1);
SAD49_DC    <= Diff49_DC(num downto 0) when Diff49_DC(num+1) = '0' else (not(Diff49_DC(num downto 0))+1);
SAD50_DC    <= Diff50_DC(num downto 0) when Diff50_DC(num+1) = '0' else (not(Diff50_DC(num downto 0))+1);
SAD51_DC    <= Diff51_DC(num downto 0) when Diff51_DC(num+1) = '0' else (not(Diff51_DC(num downto 0))+1);
SAD52_DC    <= Diff52_DC(num downto 0) when Diff52_DC(num+1) = '0' else (not(Diff52_DC(num downto 0))+1);
SAD53_DC    <= Diff53_DC(num downto 0) when Diff53_DC(num+1) = '0' else (not(Diff53_DC(num downto 0))+1);
SAD54_DC    <= Diff54_DC(num downto 0) when Diff54_DC(num+1) = '0' else (not(Diff54_DC(num downto 0))+1);
SAD55_DC    <= Diff55_DC(num downto 0) when Diff55_DC(num+1) = '0' else (not(Diff55_DC(num downto 0))+1);
SAD56_DC    <= Diff56_DC(num downto 0) when Diff56_DC(num+1) = '0' else (not(Diff56_DC(num downto 0))+1);
SAD57_DC    <= Diff57_DC(num downto 0) when Diff57_DC(num+1) = '0' else (not(Diff57_DC(num downto 0))+1);
SAD58_DC    <= Diff58_DC(num downto 0) when Diff58_DC(num+1) = '0' else (not(Diff58_DC(num downto 0))+1);
SAD59_DC    <= Diff59_DC(num downto 0) when Diff59_DC(num+1) = '0' else (not(Diff59_DC(num downto 0))+1);
SAD60_DC    <= Diff60_DC(num downto 0) when Diff60_DC(num+1) = '0' else (not(Diff60_DC(num downto 0))+1);
SAD61_DC    <= Diff61_DC(num downto 0) when Diff61_DC(num+1) = '0' else (not(Diff61_DC(num downto 0))+1);
SAD62_DC    <= Diff62_DC(num downto 0) when Diff62_DC(num+1) = '0' else (not(Diff62_DC(num downto 0))+1);		
SAD63_DC    <= Diff63_DC(num downto 0) when Diff63_DC(num+1) = '0' else (not(Diff63_DC(num downto 0))+1);
--DDL
SAD0_DDL     <= Diff0_DDL(num downto 0) when Diff0_DDL(num+1) = '0' else (not(Diff0_DDL(num downto 0))+1);
SAD1_DDL     <= Diff1_DDL(num downto 0) when Diff1_DDL(num+1) = '0' else (not(Diff1_DDL(num downto 0))+1);
SAD2_DDL     <= Diff2_DDL(num downto 0) when Diff2_DDL(num+1) = '0' else (not(Diff2_DDL(num downto 0))+1);
SAD3_DDL     <= Diff3_DDL(num downto 0) when Diff3_DDL(num+1) = '0' else (not(Diff3_DDL(num downto 0))+1);
SAD4_DDL     <= Diff4_DDL(num downto 0) when Diff4_DDL(num+1) = '0' else (not(Diff4_DDL(num downto 0))+1);
SAD5_DDL     <= Diff5_DDL(num downto 0) when Diff5_DDL(num+1) = '0' else (not(Diff5_DDL(num downto 0))+1);
SAD6_DDL     <= Diff6_DDL(num downto 0) when Diff6_DDL(num+1) = '0' else (not(Diff6_DDL(num downto 0))+1);
SAD7_DDL     <= Diff7_DDL(num downto 0) when Diff7_DDL(num+1) = '0' else (not(Diff7_DDL(num downto 0))+1);
SAD8_DDL     <= Diff8_DDL(num downto 0) when Diff8_DDL(num+1) = '0' else (not(Diff8_DDL(num downto 0))+1);
SAD9_DDL     <= Diff9_DDL(num downto 0) when Diff9_DDL(num+1) = '0' else (not(Diff9_DDL(num downto 0))+1);
SAD10_DDL    <= Diff10_DDL(num downto 0) when Diff10_DDL(num+1) = '0' else (not(Diff10_DDL(num downto 0))+1);
SAD11_DDL    <= Diff11_DDL(num downto 0) when Diff11_DDL(num+1) = '0' else (not(Diff11_DDL(num downto 0))+1);
SAD12_DDL    <= Diff12_DDL(num downto 0) when Diff12_DDL(num+1) = '0' else (not(Diff12_DDL(num downto 0))+1);
SAD13_DDL    <= Diff13_DDL(num downto 0) when Diff13_DDL(num+1) = '0' else (not(Diff13_DDL(num downto 0))+1);
SAD14_DDL    <= Diff14_DDL(num downto 0) when Diff14_DDL(num+1) = '0' else (not(Diff14_DDL(num downto 0))+1);		
SAD15_DDL    <= Diff15_DDL(num downto 0) when Diff15_DDL(num+1) = '0' else (not(Diff15_DDL(num downto 0))+1);
SAD16_DDL    <= Diff16_DDL(num downto 0) when Diff16_DDL(num+1) = '0' else (not(Diff16_DDL(num downto 0))+1);
SAD17_DDL    <= Diff17_DDL(num downto 0) when Diff17_DDL(num+1) = '0' else (not(Diff17_DDL(num downto 0))+1);
SAD18_DDL    <= Diff18_DDL(num downto 0) when Diff18_DDL(num+1) = '0' else (not(Diff18_DDL(num downto 0))+1);
SAD19_DDL    <= Diff19_DDL(num downto 0) when Diff19_DDL(num+1) = '0' else (not(Diff19_DDL(num downto 0))+1);
SAD20_DDL    <= Diff20_DDL(num downto 0) when Diff20_DDL(num+1) = '0' else (not(Diff20_DDL(num downto 0))+1);
SAD21_DDL    <= Diff21_DDL(num downto 0) when Diff21_DDL(num+1) = '0' else (not(Diff21_DDL(num downto 0))+1);
SAD22_DDL    <= Diff22_DDL(num downto 0) when Diff22_DDL(num+1) = '0' else (not(Diff22_DDL(num downto 0))+1);
SAD23_DDL    <= Diff23_DDL(num downto 0) when Diff23_DDL(num+1) = '0' else (not(Diff23_DDL(num downto 0))+1);
SAD24_DDL    <= Diff24_DDL(num downto 0) when Diff24_DDL(num+1) = '0' else (not(Diff24_DDL(num downto 0))+1);
SAD25_DDL    <= Diff25_DDL(num downto 0) when Diff25_DDL(num+1) = '0' else (not(Diff25_DDL(num downto 0))+1);
SAD26_DDL    <= Diff26_DDL(num downto 0) when Diff26_DDL(num+1) = '0' else (not(Diff26_DDL(num downto 0))+1);
SAD27_DDL    <= Diff27_DDL(num downto 0) when Diff27_DDL(num+1) = '0' else (not(Diff27_DDL(num downto 0))+1);
SAD28_DDL    <= Diff28_DDL(num downto 0) when Diff28_DDL(num+1) = '0' else (not(Diff28_DDL(num downto 0))+1);
SAD29_DDL    <= Diff29_DDL(num downto 0) when Diff29_DDL(num+1) = '0' else (not(Diff29_DDL(num downto 0))+1);
SAD30_DDL    <= Diff30_DDL(num downto 0) when Diff30_DDL(num+1) = '0' else (not(Diff30_DDL(num downto 0))+1);		
SAD31_DDL    <= Diff31_DDL(num downto 0) when Diff31_DDL(num+1) = '0' else (not(Diff31_DDL(num downto 0))+1);
SAD32_DDL    <= Diff32_DDL(num downto 0) when Diff32_DDL(num+1) = '0' else (not(Diff32_DDL(num downto 0))+1);
SAD33_DDL    <= Diff33_DDL(num downto 0) when Diff33_DDL(num+1) = '0' else (not(Diff33_DDL(num downto 0))+1);
SAD34_DDL    <= Diff34_DDL(num downto 0) when Diff34_DDL(num+1) = '0' else (not(Diff34_DDL(num downto 0))+1);
SAD35_DDL    <= Diff35_DDL(num downto 0) when Diff35_DDL(num+1) = '0' else (not(Diff35_DDL(num downto 0))+1);
SAD36_DDL    <= Diff36_DDL(num downto 0) when Diff36_DDL(num+1) = '0' else (not(Diff36_DDL(num downto 0))+1);
SAD37_DDL    <= Diff37_DDL(num downto 0) when Diff37_DDL(num+1) = '0' else (not(Diff37_DDL(num downto 0))+1);
SAD38_DDL    <= Diff38_DDL(num downto 0) when Diff38_DDL(num+1) = '0' else (not(Diff38_DDL(num downto 0))+1);
SAD39_DDL    <= Diff39_DDL(num downto 0) when Diff39_DDL(num+1) = '0' else (not(Diff39_DDL(num downto 0))+1);
SAD40_DDL    <= Diff40_DDL(num downto 0) when Diff40_DDL(num+1) = '0' else (not(Diff40_DDL(num downto 0))+1);
SAD41_DDL    <= Diff41_DDL(num downto 0) when Diff41_DDL(num+1) = '0' else (not(Diff41_DDL(num downto 0))+1);
SAD42_DDL    <= Diff42_DDL(num downto 0) when Diff42_DDL(num+1) = '0' else (not(Diff42_DDL(num downto 0))+1);
SAD43_DDL    <= Diff43_DDL(num downto 0) when Diff43_DDL(num+1) = '0' else (not(Diff43_DDL(num downto 0))+1);
SAD44_DDL    <= Diff44_DDL(num downto 0) when Diff44_DDL(num+1) = '0' else (not(Diff44_DDL(num downto 0))+1);
SAD45_DDL    <= Diff45_DDL(num downto 0) when Diff45_DDL(num+1) = '0' else (not(Diff45_DDL(num downto 0))+1);
SAD46_DDL    <= Diff46_DDL(num downto 0) when Diff46_DDL(num+1) = '0' else (not(Diff46_DDL(num downto 0))+1);		
SAD47_DDL    <= Diff47_DDL(num downto 0) when Diff47_DDL(num+1) = '0' else (not(Diff47_DDL(num downto 0))+1);
SAD48_DDL    <= Diff48_DDL(num downto 0) when Diff48_DDL(num+1) = '0' else (not(Diff48_DDL(num downto 0))+1);
SAD49_DDL    <= Diff49_DDL(num downto 0) when Diff49_DDL(num+1) = '0' else (not(Diff49_DDL(num downto 0))+1);
SAD50_DDL    <= Diff50_DDL(num downto 0) when Diff50_DDL(num+1) = '0' else (not(Diff50_DDL(num downto 0))+1);
SAD51_DDL    <= Diff51_DDL(num downto 0) when Diff51_DDL(num+1) = '0' else (not(Diff51_DDL(num downto 0))+1);
SAD52_DDL    <= Diff52_DDL(num downto 0) when Diff52_DDL(num+1) = '0' else (not(Diff52_DDL(num downto 0))+1);
SAD53_DDL    <= Diff53_DDL(num downto 0) when Diff53_DDL(num+1) = '0' else (not(Diff53_DDL(num downto 0))+1);
SAD54_DDL    <= Diff54_DDL(num downto 0) when Diff54_DDL(num+1) = '0' else (not(Diff54_DDL(num downto 0))+1);
SAD55_DDL    <= Diff55_DDL(num downto 0) when Diff55_DDL(num+1) = '0' else (not(Diff55_DDL(num downto 0))+1);
SAD56_DDL    <= Diff56_DDL(num downto 0) when Diff56_DDL(num+1) = '0' else (not(Diff56_DDL(num downto 0))+1);
SAD57_DDL    <= Diff57_DDL(num downto 0) when Diff57_DDL(num+1) = '0' else (not(Diff57_DDL(num downto 0))+1);
SAD58_DDL    <= Diff58_DDL(num downto 0) when Diff58_DDL(num+1) = '0' else (not(Diff58_DDL(num downto 0))+1);
SAD59_DDL    <= Diff59_DDL(num downto 0) when Diff59_DDL(num+1) = '0' else (not(Diff59_DDL(num downto 0))+1);
SAD60_DDL    <= Diff60_DDL(num downto 0) when Diff60_DDL(num+1) = '0' else (not(Diff60_DDL(num downto 0))+1);
SAD61_DDL    <= Diff61_DDL(num downto 0) when Diff61_DDL(num+1) = '0' else (not(Diff61_DDL(num downto 0))+1);
SAD62_DDL    <= Diff62_DDL(num downto 0) when Diff62_DDL(num+1) = '0' else (not(Diff62_DDL(num downto 0))+1);		
SAD63_DDL    <= Diff63_DDL(num downto 0) when Diff63_DDL(num+1) = '0' else (not(Diff63_DDL(num downto 0))+1);
--DDR
SAD0_DDR     <= Diff0_DDR(num downto 0) when Diff0_DDR(num+1) = '0' else (not(Diff0_DDR(num downto 0))+1);
SAD1_DDR     <= Diff1_DDR(num downto 0) when Diff1_DDR(num+1) = '0' else (not(Diff1_DDR(num downto 0))+1);
SAD2_DDR     <= Diff2_DDR(num downto 0) when Diff2_DDR(num+1) = '0' else (not(Diff2_DDR(num downto 0))+1);
SAD3_DDR     <= Diff3_DDR(num downto 0) when Diff3_DDR(num+1) = '0' else (not(Diff3_DDR(num downto 0))+1);
SAD4_DDR     <= Diff4_DDR(num downto 0) when Diff4_DDR(num+1) = '0' else (not(Diff4_DDR(num downto 0))+1);
SAD5_DDR     <= Diff5_DDR(num downto 0) when Diff5_DDR(num+1) = '0' else (not(Diff5_DDR(num downto 0))+1);
SAD6_DDR     <= Diff6_DDR(num downto 0) when Diff6_DDR(num+1) = '0' else (not(Diff6_DDR(num downto 0))+1);
SAD7_DDR     <= Diff7_DDR(num downto 0) when Diff7_DDR(num+1) = '0' else (not(Diff7_DDR(num downto 0))+1);
SAD8_DDR     <= Diff8_DDR(num downto 0) when Diff8_DDR(num+1) = '0' else (not(Diff8_DDR(num downto 0))+1);
SAD9_DDR     <= Diff9_DDR(num downto 0) when Diff9_DDR(num+1) = '0' else (not(Diff9_DDR(num downto 0))+1);
SAD10_DDR    <= Diff10_DDR(num downto 0) when Diff10_DDR(num+1) = '0' else (not(Diff10_DDR(num downto 0))+1);
SAD11_DDR    <= Diff11_DDR(num downto 0) when Diff11_DDR(num+1) = '0' else (not(Diff11_DDR(num downto 0))+1);
SAD12_DDR    <= Diff12_DDR(num downto 0) when Diff12_DDR(num+1) = '0' else (not(Diff12_DDR(num downto 0))+1);
SAD13_DDR    <= Diff13_DDR(num downto 0) when Diff13_DDR(num+1) = '0' else (not(Diff13_DDR(num downto 0))+1);
SAD14_DDR    <= Diff14_DDR(num downto 0) when Diff14_DDR(num+1) = '0' else (not(Diff14_DDR(num downto 0))+1);		
SAD15_DDR    <= Diff15_DDR(num downto 0) when Diff15_DDR(num+1) = '0' else (not(Diff15_DDR(num downto 0))+1);
SAD16_DDR    <= Diff16_DDR(num downto 0) when Diff16_DDR(num+1) = '0' else (not(Diff16_DDR(num downto 0))+1);
SAD17_DDR    <= Diff17_DDR(num downto 0) when Diff17_DDR(num+1) = '0' else (not(Diff17_DDR(num downto 0))+1);
SAD18_DDR    <= Diff18_DDR(num downto 0) when Diff18_DDR(num+1) = '0' else (not(Diff18_DDR(num downto 0))+1);
SAD19_DDR    <= Diff19_DDR(num downto 0) when Diff19_DDR(num+1) = '0' else (not(Diff19_DDR(num downto 0))+1);
SAD20_DDR    <= Diff20_DDR(num downto 0) when Diff20_DDR(num+1) = '0' else (not(Diff20_DDR(num downto 0))+1);
SAD21_DDR    <= Diff21_DDR(num downto 0) when Diff21_DDR(num+1) = '0' else (not(Diff21_DDR(num downto 0))+1);
SAD22_DDR    <= Diff22_DDR(num downto 0) when Diff22_DDR(num+1) = '0' else (not(Diff22_DDR(num downto 0))+1);
SAD23_DDR    <= Diff23_DDR(num downto 0) when Diff23_DDR(num+1) = '0' else (not(Diff23_DDR(num downto 0))+1);
SAD24_DDR    <= Diff24_DDR(num downto 0) when Diff24_DDR(num+1) = '0' else (not(Diff24_DDR(num downto 0))+1);
SAD25_DDR    <= Diff25_DDR(num downto 0) when Diff25_DDR(num+1) = '0' else (not(Diff25_DDR(num downto 0))+1);
SAD26_DDR    <= Diff26_DDR(num downto 0) when Diff26_DDR(num+1) = '0' else (not(Diff26_DDR(num downto 0))+1);
SAD27_DDR    <= Diff27_DDR(num downto 0) when Diff27_DDR(num+1) = '0' else (not(Diff27_DDR(num downto 0))+1);
SAD28_DDR    <= Diff28_DDR(num downto 0) when Diff28_DDR(num+1) = '0' else (not(Diff28_DDR(num downto 0))+1);
SAD29_DDR    <= Diff29_DDR(num downto 0) when Diff29_DDR(num+1) = '0' else (not(Diff29_DDR(num downto 0))+1);
SAD30_DDR    <= Diff30_DDR(num downto 0) when Diff30_DDR(num+1) = '0' else (not(Diff30_DDR(num downto 0))+1);		
SAD31_DDR    <= Diff31_DDR(num downto 0) when Diff31_DDR(num+1) = '0' else (not(Diff31_DDR(num downto 0))+1);
SAD32_DDR    <= Diff32_DDR(num downto 0) when Diff32_DDR(num+1) = '0' else (not(Diff32_DDR(num downto 0))+1);
SAD33_DDR    <= Diff33_DDR(num downto 0) when Diff33_DDR(num+1) = '0' else (not(Diff33_DDR(num downto 0))+1);
SAD34_DDR    <= Diff34_DDR(num downto 0) when Diff34_DDR(num+1) = '0' else (not(Diff34_DDR(num downto 0))+1);
SAD35_DDR    <= Diff35_DDR(num downto 0) when Diff35_DDR(num+1) = '0' else (not(Diff35_DDR(num downto 0))+1);
SAD36_DDR    <= Diff36_DDR(num downto 0) when Diff36_DDR(num+1) = '0' else (not(Diff36_DDR(num downto 0))+1);
SAD37_DDR    <= Diff37_DDR(num downto 0) when Diff37_DDR(num+1) = '0' else (not(Diff37_DDR(num downto 0))+1);
SAD38_DDR    <= Diff38_DDR(num downto 0) when Diff38_DDR(num+1) = '0' else (not(Diff38_DDR(num downto 0))+1);
SAD39_DDR    <= Diff39_DDR(num downto 0) when Diff39_DDR(num+1) = '0' else (not(Diff39_DDR(num downto 0))+1);
SAD40_DDR    <= Diff40_DDR(num downto 0) when Diff40_DDR(num+1) = '0' else (not(Diff40_DDR(num downto 0))+1);
SAD41_DDR    <= Diff41_DDR(num downto 0) when Diff41_DDR(num+1) = '0' else (not(Diff41_DDR(num downto 0))+1);
SAD42_DDR    <= Diff42_DDR(num downto 0) when Diff42_DDR(num+1) = '0' else (not(Diff42_DDR(num downto 0))+1);
SAD43_DDR    <= Diff43_DDR(num downto 0) when Diff43_DDR(num+1) = '0' else (not(Diff43_DDR(num downto 0))+1);
SAD44_DDR    <= Diff44_DDR(num downto 0) when Diff44_DDR(num+1) = '0' else (not(Diff44_DDR(num downto 0))+1);
SAD45_DDR    <= Diff45_DDR(num downto 0) when Diff45_DDR(num+1) = '0' else (not(Diff45_DDR(num downto 0))+1);
SAD46_DDR    <= Diff46_DDR(num downto 0) when Diff46_DDR(num+1) = '0' else (not(Diff46_DDR(num downto 0))+1);		
SAD47_DDR    <= Diff47_DDR(num downto 0) when Diff47_DDR(num+1) = '0' else (not(Diff47_DDR(num downto 0))+1);
SAD48_DDR    <= Diff48_DDR(num downto 0) when Diff48_DDR(num+1) = '0' else (not(Diff48_DDR(num downto 0))+1);
SAD49_DDR    <= Diff49_DDR(num downto 0) when Diff49_DDR(num+1) = '0' else (not(Diff49_DDR(num downto 0))+1);
SAD50_DDR    <= Diff50_DDR(num downto 0) when Diff50_DDR(num+1) = '0' else (not(Diff50_DDR(num downto 0))+1);
SAD51_DDR    <= Diff51_DDR(num downto 0) when Diff51_DDR(num+1) = '0' else (not(Diff51_DDR(num downto 0))+1);
SAD52_DDR    <= Diff52_DDR(num downto 0) when Diff52_DDR(num+1) = '0' else (not(Diff52_DDR(num downto 0))+1);
SAD53_DDR    <= Diff53_DDR(num downto 0) when Diff53_DDR(num+1) = '0' else (not(Diff53_DDR(num downto 0))+1);
SAD54_DDR    <= Diff54_DDR(num downto 0) when Diff54_DDR(num+1) = '0' else (not(Diff54_DDR(num downto 0))+1);
SAD55_DDR    <= Diff55_DDR(num downto 0) when Diff55_DDR(num+1) = '0' else (not(Diff55_DDR(num downto 0))+1);
SAD56_DDR    <= Diff56_DDR(num downto 0) when Diff56_DDR(num+1) = '0' else (not(Diff56_DDR(num downto 0))+1);
SAD57_DDR    <= Diff57_DDR(num downto 0) when Diff57_DDR(num+1) = '0' else (not(Diff57_DDR(num downto 0))+1);
SAD58_DDR    <= Diff58_DDR(num downto 0) when Diff58_DDR(num+1) = '0' else (not(Diff58_DDR(num downto 0))+1);
SAD59_DDR    <= Diff59_DDR(num downto 0) when Diff59_DDR(num+1) = '0' else (not(Diff59_DDR(num downto 0))+1);
SAD60_DDR    <= Diff60_DDR(num downto 0) when Diff60_DDR(num+1) = '0' else (not(Diff60_DDR(num downto 0))+1);
SAD61_DDR    <= Diff61_DDR(num downto 0) when Diff61_DDR(num+1) = '0' else (not(Diff61_DDR(num downto 0))+1);
SAD62_DDR    <= Diff62_DDR(num downto 0) when Diff62_DDR(num+1) = '0' else (not(Diff62_DDR(num downto 0))+1);		
SAD63_DDR    <= Diff63_DDR(num downto 0) when Diff63_DDR(num+1) = '0' else (not(Diff63_DDR(num downto 0))+1);
--VR
SAD0_VR     <= Diff0_VR(num downto 0) when Diff0_VR(num+1) = '0' else (not(Diff0_VR(num downto 0))+1);
SAD1_VR     <= Diff1_VR(num downto 0) when Diff1_VR(num+1) = '0' else (not(Diff1_VR(num downto 0))+1);
SAD2_VR     <= Diff2_VR(num downto 0) when Diff2_VR(num+1) = '0' else (not(Diff2_VR(num downto 0))+1);
SAD3_VR     <= Diff3_VR(num downto 0) when Diff3_VR(num+1) = '0' else (not(Diff3_VR(num downto 0))+1);
SAD4_VR     <= Diff4_VR(num downto 0) when Diff4_VR(num+1) = '0' else (not(Diff4_VR(num downto 0))+1);
SAD5_VR     <= Diff5_VR(num downto 0) when Diff5_VR(num+1) = '0' else (not(Diff5_VR(num downto 0))+1);
SAD6_VR     <= Diff6_VR(num downto 0) when Diff6_VR(num+1) = '0' else (not(Diff6_VR(num downto 0))+1);
SAD7_VR     <= Diff7_VR(num downto 0) when Diff7_VR(num+1) = '0' else (not(Diff7_VR(num downto 0))+1);
SAD8_VR     <= Diff8_VR(num downto 0) when Diff8_VR(num+1) = '0' else (not(Diff8_VR(num downto 0))+1);
SAD9_VR     <= Diff9_VR(num downto 0) when Diff9_VR(num+1) = '0' else (not(Diff9_VR(num downto 0))+1);
SAD10_VR    <= Diff10_VR(num downto 0) when Diff10_VR(num+1) = '0' else (not(Diff10_VR(num downto 0))+1);
SAD11_VR    <= Diff11_VR(num downto 0) when Diff11_VR(num+1) = '0' else (not(Diff11_VR(num downto 0))+1);
SAD12_VR    <= Diff12_VR(num downto 0) when Diff12_VR(num+1) = '0' else (not(Diff12_VR(num downto 0))+1);
SAD13_VR    <= Diff13_VR(num downto 0) when Diff13_VR(num+1) = '0' else (not(Diff13_VR(num downto 0))+1);
SAD14_VR    <= Diff14_VR(num downto 0) when Diff14_VR(num+1) = '0' else (not(Diff14_VR(num downto 0))+1);		
SAD15_VR    <= Diff15_VR(num downto 0) when Diff15_VR(num+1) = '0' else (not(Diff15_VR(num downto 0))+1);
SAD16_VR    <= Diff16_VR(num downto 0) when Diff16_VR(num+1) = '0' else (not(Diff16_VR(num downto 0))+1);
SAD17_VR    <= Diff17_VR(num downto 0) when Diff17_VR(num+1) = '0' else (not(Diff17_VR(num downto 0))+1);
SAD18_VR    <= Diff18_VR(num downto 0) when Diff18_VR(num+1) = '0' else (not(Diff18_VR(num downto 0))+1);
SAD19_VR    <= Diff19_VR(num downto 0) when Diff19_VR(num+1) = '0' else (not(Diff19_VR(num downto 0))+1);
SAD20_VR    <= Diff20_VR(num downto 0) when Diff20_VR(num+1) = '0' else (not(Diff20_VR(num downto 0))+1);
SAD21_VR    <= Diff21_VR(num downto 0) when Diff21_VR(num+1) = '0' else (not(Diff21_VR(num downto 0))+1);
SAD22_VR    <= Diff22_VR(num downto 0) when Diff22_VR(num+1) = '0' else (not(Diff22_VR(num downto 0))+1);
SAD23_VR    <= Diff23_VR(num downto 0) when Diff23_VR(num+1) = '0' else (not(Diff23_VR(num downto 0))+1);
SAD24_VR    <= Diff24_VR(num downto 0) when Diff24_VR(num+1) = '0' else (not(Diff24_VR(num downto 0))+1);
SAD25_VR    <= Diff25_VR(num downto 0) when Diff25_VR(num+1) = '0' else (not(Diff25_VR(num downto 0))+1);
SAD26_VR    <= Diff26_VR(num downto 0) when Diff26_VR(num+1) = '0' else (not(Diff26_VR(num downto 0))+1);
SAD27_VR    <= Diff27_VR(num downto 0) when Diff27_VR(num+1) = '0' else (not(Diff27_VR(num downto 0))+1);
SAD28_VR    <= Diff28_VR(num downto 0) when Diff28_VR(num+1) = '0' else (not(Diff28_VR(num downto 0))+1);
SAD29_VR    <= Diff29_VR(num downto 0) when Diff29_VR(num+1) = '0' else (not(Diff29_VR(num downto 0))+1);
SAD30_VR    <= Diff30_VR(num downto 0) when Diff30_VR(num+1) = '0' else (not(Diff30_VR(num downto 0))+1);		
SAD31_VR    <= Diff31_VR(num downto 0) when Diff31_VR(num+1) = '0' else (not(Diff31_VR(num downto 0))+1);
SAD32_VR    <= Diff32_VR(num downto 0) when Diff32_VR(num+1) = '0' else (not(Diff32_VR(num downto 0))+1);
SAD33_VR    <= Diff33_VR(num downto 0) when Diff33_VR(num+1) = '0' else (not(Diff33_VR(num downto 0))+1);
SAD34_VR    <= Diff34_VR(num downto 0) when Diff34_VR(num+1) = '0' else (not(Diff34_VR(num downto 0))+1);
SAD35_VR    <= Diff35_VR(num downto 0) when Diff35_VR(num+1) = '0' else (not(Diff35_VR(num downto 0))+1);
SAD36_VR    <= Diff36_VR(num downto 0) when Diff36_VR(num+1) = '0' else (not(Diff36_VR(num downto 0))+1);
SAD37_VR    <= Diff37_VR(num downto 0) when Diff37_VR(num+1) = '0' else (not(Diff37_VR(num downto 0))+1);
SAD38_VR    <= Diff38_VR(num downto 0) when Diff38_VR(num+1) = '0' else (not(Diff38_VR(num downto 0))+1);
SAD39_VR    <= Diff39_VR(num downto 0) when Diff39_VR(num+1) = '0' else (not(Diff39_VR(num downto 0))+1);
SAD40_VR    <= Diff40_VR(num downto 0) when Diff40_VR(num+1) = '0' else (not(Diff40_VR(num downto 0))+1);
SAD41_VR    <= Diff41_VR(num downto 0) when Diff41_VR(num+1) = '0' else (not(Diff41_VR(num downto 0))+1);
SAD42_VR    <= Diff42_VR(num downto 0) when Diff42_VR(num+1) = '0' else (not(Diff42_VR(num downto 0))+1);
SAD43_VR    <= Diff43_VR(num downto 0) when Diff43_VR(num+1) = '0' else (not(Diff43_VR(num downto 0))+1);
SAD44_VR    <= Diff44_VR(num downto 0) when Diff44_VR(num+1) = '0' else (not(Diff44_VR(num downto 0))+1);
SAD45_VR    <= Diff45_VR(num downto 0) when Diff45_VR(num+1) = '0' else (not(Diff45_VR(num downto 0))+1);
SAD46_VR    <= Diff46_VR(num downto 0) when Diff46_VR(num+1) = '0' else (not(Diff46_VR(num downto 0))+1);		
SAD47_VR    <= Diff47_VR(num downto 0) when Diff47_VR(num+1) = '0' else (not(Diff47_VR(num downto 0))+1);
SAD48_VR    <= Diff48_VR(num downto 0) when Diff48_VR(num+1) = '0' else (not(Diff48_VR(num downto 0))+1);
SAD49_VR    <= Diff49_VR(num downto 0) when Diff49_VR(num+1) = '0' else (not(Diff49_VR(num downto 0))+1);
SAD50_VR    <= Diff50_VR(num downto 0) when Diff50_VR(num+1) = '0' else (not(Diff50_VR(num downto 0))+1);
SAD51_VR    <= Diff51_VR(num downto 0) when Diff51_VR(num+1) = '0' else (not(Diff51_VR(num downto 0))+1);
SAD52_VR    <= Diff52_VR(num downto 0) when Diff52_VR(num+1) = '0' else (not(Diff52_VR(num downto 0))+1);
SAD53_VR    <= Diff53_VR(num downto 0) when Diff53_VR(num+1) = '0' else (not(Diff53_VR(num downto 0))+1);
SAD54_VR    <= Diff54_VR(num downto 0) when Diff54_VR(num+1) = '0' else (not(Diff54_VR(num downto 0))+1);
SAD55_VR    <= Diff55_VR(num downto 0) when Diff55_VR(num+1) = '0' else (not(Diff55_VR(num downto 0))+1);
SAD56_VR    <= Diff56_VR(num downto 0) when Diff56_VR(num+1) = '0' else (not(Diff56_VR(num downto 0))+1);
SAD57_VR    <= Diff57_VR(num downto 0) when Diff57_VR(num+1) = '0' else (not(Diff57_VR(num downto 0))+1);
SAD58_VR    <= Diff58_VR(num downto 0) when Diff58_VR(num+1) = '0' else (not(Diff58_VR(num downto 0))+1);
SAD59_VR    <= Diff59_VR(num downto 0) when Diff59_VR(num+1) = '0' else (not(Diff59_VR(num downto 0))+1);
SAD60_VR    <= Diff60_VR(num downto 0) when Diff60_VR(num+1) = '0' else (not(Diff60_VR(num downto 0))+1);
SAD61_VR    <= Diff61_VR(num downto 0) when Diff61_VR(num+1) = '0' else (not(Diff61_VR(num downto 0))+1);
SAD62_VR    <= Diff62_VR(num downto 0) when Diff62_VR(num+1) = '0' else (not(Diff62_VR(num downto 0))+1);		
SAD63_VR    <= Diff63_VR(num downto 0) when Diff63_VR(num+1) = '0' else (not(Diff63_VR(num downto 0))+1);
--HD
SAD0_HD     <= Diff0_HD(num downto 0) when Diff0_HD(num+1) = '0' else (not(Diff0_HD(num downto 0))+1);
SAD1_HD     <= Diff1_HD(num downto 0) when Diff1_HD(num+1) = '0' else (not(Diff1_HD(num downto 0))+1);
SAD2_HD     <= Diff2_HD(num downto 0) when Diff2_HD(num+1) = '0' else (not(Diff2_HD(num downto 0))+1);
SAD3_HD     <= Diff3_HD(num downto 0) when Diff3_HD(num+1) = '0' else (not(Diff3_HD(num downto 0))+1);
SAD4_HD     <= Diff4_HD(num downto 0) when Diff4_HD(num+1) = '0' else (not(Diff4_HD(num downto 0))+1);
SAD5_HD     <= Diff5_HD(num downto 0) when Diff5_HD(num+1) = '0' else (not(Diff5_HD(num downto 0))+1);
SAD6_HD     <= Diff6_HD(num downto 0) when Diff6_HD(num+1) = '0' else (not(Diff6_HD(num downto 0))+1);
SAD7_HD     <= Diff7_HD(num downto 0) when Diff7_HD(num+1) = '0' else (not(Diff7_HD(num downto 0))+1);
SAD8_HD     <= Diff8_HD(num downto 0) when Diff8_HD(num+1) = '0' else (not(Diff8_HD(num downto 0))+1);
SAD9_HD     <= Diff9_HD(num downto 0) when Diff9_HD(num+1) = '0' else (not(Diff9_HD(num downto 0))+1);
SAD10_HD    <= Diff10_HD(num downto 0) when Diff10_HD(num+1) = '0' else (not(Diff10_HD(num downto 0))+1);
SAD11_HD    <= Diff11_HD(num downto 0) when Diff11_HD(num+1) = '0' else (not(Diff11_HD(num downto 0))+1);
SAD12_HD    <= Diff12_HD(num downto 0) when Diff12_HD(num+1) = '0' else (not(Diff12_HD(num downto 0))+1);
SAD13_HD    <= Diff13_HD(num downto 0) when Diff13_HD(num+1) = '0' else (not(Diff13_HD(num downto 0))+1);
SAD14_HD    <= Diff14_HD(num downto 0) when Diff14_HD(num+1) = '0' else (not(Diff14_HD(num downto 0))+1);		
SAD15_HD    <= Diff15_HD(num downto 0) when Diff15_HD(num+1) = '0' else (not(Diff15_HD(num downto 0))+1);
SAD16_HD    <= Diff16_HD(num downto 0) when Diff16_HD(num+1) = '0' else (not(Diff16_HD(num downto 0))+1);
SAD17_HD    <= Diff17_HD(num downto 0) when Diff17_HD(num+1) = '0' else (not(Diff17_HD(num downto 0))+1);
SAD18_HD    <= Diff18_HD(num downto 0) when Diff18_HD(num+1) = '0' else (not(Diff18_HD(num downto 0))+1);
SAD19_HD    <= Diff19_HD(num downto 0) when Diff19_HD(num+1) = '0' else (not(Diff19_HD(num downto 0))+1);
SAD20_HD    <= Diff20_HD(num downto 0) when Diff20_HD(num+1) = '0' else (not(Diff20_HD(num downto 0))+1);
SAD21_HD    <= Diff21_HD(num downto 0) when Diff21_HD(num+1) = '0' else (not(Diff21_HD(num downto 0))+1);
SAD22_HD    <= Diff22_HD(num downto 0) when Diff22_HD(num+1) = '0' else (not(Diff22_HD(num downto 0))+1);
SAD23_HD    <= Diff23_HD(num downto 0) when Diff23_HD(num+1) = '0' else (not(Diff23_HD(num downto 0))+1);
SAD24_HD    <= Diff24_HD(num downto 0) when Diff24_HD(num+1) = '0' else (not(Diff24_HD(num downto 0))+1);
SAD25_HD    <= Diff25_HD(num downto 0) when Diff25_HD(num+1) = '0' else (not(Diff25_HD(num downto 0))+1);
SAD26_HD    <= Diff26_HD(num downto 0) when Diff26_HD(num+1) = '0' else (not(Diff26_HD(num downto 0))+1);
SAD27_HD    <= Diff27_HD(num downto 0) when Diff27_HD(num+1) = '0' else (not(Diff27_HD(num downto 0))+1);
SAD28_HD    <= Diff28_HD(num downto 0) when Diff28_HD(num+1) = '0' else (not(Diff28_HD(num downto 0))+1);
SAD29_HD    <= Diff29_HD(num downto 0) when Diff29_HD(num+1) = '0' else (not(Diff29_HD(num downto 0))+1);
SAD30_HD    <= Diff30_HD(num downto 0) when Diff30_HD(num+1) = '0' else (not(Diff30_HD(num downto 0))+1);		
SAD31_HD    <= Diff31_HD(num downto 0) when Diff31_HD(num+1) = '0' else (not(Diff31_HD(num downto 0))+1);
SAD32_HD    <= Diff32_HD(num downto 0) when Diff32_HD(num+1) = '0' else (not(Diff32_HD(num downto 0))+1);
SAD33_HD    <= Diff33_HD(num downto 0) when Diff33_HD(num+1) = '0' else (not(Diff33_HD(num downto 0))+1);
SAD34_HD    <= Diff34_HD(num downto 0) when Diff34_HD(num+1) = '0' else (not(Diff34_HD(num downto 0))+1);
SAD35_HD    <= Diff35_HD(num downto 0) when Diff35_HD(num+1) = '0' else (not(Diff35_HD(num downto 0))+1);
SAD36_HD    <= Diff36_HD(num downto 0) when Diff36_HD(num+1) = '0' else (not(Diff36_HD(num downto 0))+1);
SAD37_HD    <= Diff37_HD(num downto 0) when Diff37_HD(num+1) = '0' else (not(Diff37_HD(num downto 0))+1);
SAD38_HD    <= Diff38_HD(num downto 0) when Diff38_HD(num+1) = '0' else (not(Diff38_HD(num downto 0))+1);
SAD39_HD    <= Diff39_HD(num downto 0) when Diff39_HD(num+1) = '0' else (not(Diff39_HD(num downto 0))+1);
SAD40_HD    <= Diff40_HD(num downto 0) when Diff40_HD(num+1) = '0' else (not(Diff40_HD(num downto 0))+1);
SAD41_HD    <= Diff41_HD(num downto 0) when Diff41_HD(num+1) = '0' else (not(Diff41_HD(num downto 0))+1);
SAD42_HD    <= Diff42_HD(num downto 0) when Diff42_HD(num+1) = '0' else (not(Diff42_HD(num downto 0))+1);
SAD43_HD    <= Diff43_HD(num downto 0) when Diff43_HD(num+1) = '0' else (not(Diff43_HD(num downto 0))+1);
SAD44_HD    <= Diff44_HD(num downto 0) when Diff44_HD(num+1) = '0' else (not(Diff44_HD(num downto 0))+1);
SAD45_HD    <= Diff45_HD(num downto 0) when Diff45_HD(num+1) = '0' else (not(Diff45_HD(num downto 0))+1);
SAD46_HD    <= Diff46_HD(num downto 0) when Diff46_HD(num+1) = '0' else (not(Diff46_HD(num downto 0))+1);		
SAD47_HD    <= Diff47_HD(num downto 0) when Diff47_HD(num+1) = '0' else (not(Diff47_HD(num downto 0))+1);
SAD48_HD    <= Diff48_HD(num downto 0) when Diff48_HD(num+1) = '0' else (not(Diff48_HD(num downto 0))+1);
SAD49_HD    <= Diff49_HD(num downto 0) when Diff49_HD(num+1) = '0' else (not(Diff49_HD(num downto 0))+1);
SAD50_HD    <= Diff50_HD(num downto 0) when Diff50_HD(num+1) = '0' else (not(Diff50_HD(num downto 0))+1);
SAD51_HD    <= Diff51_HD(num downto 0) when Diff51_HD(num+1) = '0' else (not(Diff51_HD(num downto 0))+1);
SAD52_HD    <= Diff52_HD(num downto 0) when Diff52_HD(num+1) = '0' else (not(Diff52_HD(num downto 0))+1);
SAD53_HD    <= Diff53_HD(num downto 0) when Diff53_HD(num+1) = '0' else (not(Diff53_HD(num downto 0))+1);
SAD54_HD    <= Diff54_HD(num downto 0) when Diff54_HD(num+1) = '0' else (not(Diff54_HD(num downto 0))+1);
SAD55_HD    <= Diff55_HD(num downto 0) when Diff55_HD(num+1) = '0' else (not(Diff55_HD(num downto 0))+1);
SAD56_HD    <= Diff56_HD(num downto 0) when Diff56_HD(num+1) = '0' else (not(Diff56_HD(num downto 0))+1);
SAD57_HD    <= Diff57_HD(num downto 0) when Diff57_HD(num+1) = '0' else (not(Diff57_HD(num downto 0))+1);
SAD58_HD    <= Diff58_HD(num downto 0) when Diff58_HD(num+1) = '0' else (not(Diff58_HD(num downto 0))+1);
SAD59_HD    <= Diff59_HD(num downto 0) when Diff59_HD(num+1) = '0' else (not(Diff59_HD(num downto 0))+1);
SAD60_HD    <= Diff60_HD(num downto 0) when Diff60_HD(num+1) = '0' else (not(Diff60_HD(num downto 0))+1);
SAD61_HD    <= Diff61_HD(num downto 0) when Diff61_HD(num+1) = '0' else (not(Diff61_HD(num downto 0))+1);
SAD62_HD    <= Diff62_HD(num downto 0) when Diff62_HD(num+1) = '0' else (not(Diff62_HD(num downto 0))+1);		
SAD63_HD    <= Diff63_HD(num downto 0) when Diff63_HD(num+1) = '0' else (not(Diff63_HD(num downto 0))+1);
--VL
SAD0_VL     <= Diff0_VL(num downto 0) when Diff0_VL(num+1) = '0' else (not(Diff0_VL(num downto 0))+1);
SAD1_VL     <= Diff1_VL(num downto 0) when Diff1_VL(num+1) = '0' else (not(Diff1_VL(num downto 0))+1);
SAD2_VL     <= Diff2_VL(num downto 0) when Diff2_VL(num+1) = '0' else (not(Diff2_VL(num downto 0))+1);
SAD3_VL     <= Diff3_VL(num downto 0) when Diff3_VL(num+1) = '0' else (not(Diff3_VL(num downto 0))+1);
SAD4_VL     <= Diff4_VL(num downto 0) when Diff4_VL(num+1) = '0' else (not(Diff4_VL(num downto 0))+1);
SAD5_VL     <= Diff5_VL(num downto 0) when Diff5_VL(num+1) = '0' else (not(Diff5_VL(num downto 0))+1);
SAD6_VL     <= Diff6_VL(num downto 0) when Diff6_VL(num+1) = '0' else (not(Diff6_VL(num downto 0))+1);
SAD7_VL     <= Diff7_VL(num downto 0) when Diff7_VL(num+1) = '0' else (not(Diff7_VL(num downto 0))+1);
SAD8_VL     <= Diff8_VL(num downto 0) when Diff8_VL(num+1) = '0' else (not(Diff8_VL(num downto 0))+1);
SAD9_VL     <= Diff9_VL(num downto 0) when Diff9_VL(num+1) = '0' else (not(Diff9_VL(num downto 0))+1);
SAD10_VL    <= Diff10_VL(num downto 0) when Diff10_VL(num+1) = '0' else (not(Diff10_VL(num downto 0))+1);
SAD11_VL    <= Diff11_VL(num downto 0) when Diff11_VL(num+1) = '0' else (not(Diff11_VL(num downto 0))+1);
SAD12_VL    <= Diff12_VL(num downto 0) when Diff12_VL(num+1) = '0' else (not(Diff12_VL(num downto 0))+1);
SAD13_VL    <= Diff13_VL(num downto 0) when Diff13_VL(num+1) = '0' else (not(Diff13_VL(num downto 0))+1);
SAD14_VL    <= Diff14_VL(num downto 0) when Diff14_VL(num+1) = '0' else (not(Diff14_VL(num downto 0))+1);		
SAD15_VL    <= Diff15_VL(num downto 0) when Diff15_VL(num+1) = '0' else (not(Diff15_VL(num downto 0))+1);
SAD16_VL    <= Diff16_VL(num downto 0) when Diff16_VL(num+1) = '0' else (not(Diff16_VL(num downto 0))+1);
SAD17_VL    <= Diff17_VL(num downto 0) when Diff17_VL(num+1) = '0' else (not(Diff17_VL(num downto 0))+1);
SAD18_VL    <= Diff18_VL(num downto 0) when Diff18_VL(num+1) = '0' else (not(Diff18_VL(num downto 0))+1);
SAD19_VL    <= Diff19_VL(num downto 0) when Diff19_VL(num+1) = '0' else (not(Diff19_VL(num downto 0))+1);
SAD20_VL    <= Diff20_VL(num downto 0) when Diff20_VL(num+1) = '0' else (not(Diff20_VL(num downto 0))+1);
SAD21_VL    <= Diff21_VL(num downto 0) when Diff21_VL(num+1) = '0' else (not(Diff21_VL(num downto 0))+1);
SAD22_VL    <= Diff22_VL(num downto 0) when Diff22_VL(num+1) = '0' else (not(Diff22_VL(num downto 0))+1);
SAD23_VL    <= Diff23_VL(num downto 0) when Diff23_VL(num+1) = '0' else (not(Diff23_VL(num downto 0))+1);
SAD24_VL    <= Diff24_VL(num downto 0) when Diff24_VL(num+1) = '0' else (not(Diff24_VL(num downto 0))+1);
SAD25_VL    <= Diff25_VL(num downto 0) when Diff25_VL(num+1) = '0' else (not(Diff25_VL(num downto 0))+1);
SAD26_VL    <= Diff26_VL(num downto 0) when Diff26_VL(num+1) = '0' else (not(Diff26_VL(num downto 0))+1);
SAD27_VL    <= Diff27_VL(num downto 0) when Diff27_VL(num+1) = '0' else (not(Diff27_VL(num downto 0))+1);
SAD28_VL    <= Diff28_VL(num downto 0) when Diff28_VL(num+1) = '0' else (not(Diff28_VL(num downto 0))+1);
SAD29_VL    <= Diff29_VL(num downto 0) when Diff29_VL(num+1) = '0' else (not(Diff29_VL(num downto 0))+1);
SAD30_VL    <= Diff30_VL(num downto 0) when Diff30_VL(num+1) = '0' else (not(Diff30_VL(num downto 0))+1);		
SAD31_VL    <= Diff31_VL(num downto 0) when Diff31_VL(num+1) = '0' else (not(Diff31_VL(num downto 0))+1);
SAD32_VL    <= Diff32_VL(num downto 0) when Diff32_VL(num+1) = '0' else (not(Diff32_VL(num downto 0))+1);
SAD33_VL    <= Diff33_VL(num downto 0) when Diff33_VL(num+1) = '0' else (not(Diff33_VL(num downto 0))+1);
SAD34_VL    <= Diff34_VL(num downto 0) when Diff34_VL(num+1) = '0' else (not(Diff34_VL(num downto 0))+1);
SAD35_VL    <= Diff35_VL(num downto 0) when Diff35_VL(num+1) = '0' else (not(Diff35_VL(num downto 0))+1);
SAD36_VL    <= Diff36_VL(num downto 0) when Diff36_VL(num+1) = '0' else (not(Diff36_VL(num downto 0))+1);
SAD37_VL    <= Diff37_VL(num downto 0) when Diff37_VL(num+1) = '0' else (not(Diff37_VL(num downto 0))+1);
SAD38_VL    <= Diff38_VL(num downto 0) when Diff38_VL(num+1) = '0' else (not(Diff38_VL(num downto 0))+1);
SAD39_VL    <= Diff39_VL(num downto 0) when Diff39_VL(num+1) = '0' else (not(Diff39_VL(num downto 0))+1);
SAD40_VL    <= Diff40_VL(num downto 0) when Diff40_VL(num+1) = '0' else (not(Diff40_VL(num downto 0))+1);
SAD41_VL    <= Diff41_VL(num downto 0) when Diff41_VL(num+1) = '0' else (not(Diff41_VL(num downto 0))+1);
SAD42_VL    <= Diff42_VL(num downto 0) when Diff42_VL(num+1) = '0' else (not(Diff42_VL(num downto 0))+1);
SAD43_VL    <= Diff43_VL(num downto 0) when Diff43_VL(num+1) = '0' else (not(Diff43_VL(num downto 0))+1);
SAD44_VL    <= Diff44_VL(num downto 0) when Diff44_VL(num+1) = '0' else (not(Diff44_VL(num downto 0))+1);
SAD45_VL    <= Diff45_VL(num downto 0) when Diff45_VL(num+1) = '0' else (not(Diff45_VL(num downto 0))+1);
SAD46_VL    <= Diff46_VL(num downto 0) when Diff46_VL(num+1) = '0' else (not(Diff46_VL(num downto 0))+1);		
SAD47_VL    <= Diff47_VL(num downto 0) when Diff47_VL(num+1) = '0' else (not(Diff47_VL(num downto 0))+1);
SAD48_VL    <= Diff48_VL(num downto 0) when Diff48_VL(num+1) = '0' else (not(Diff48_VL(num downto 0))+1);
SAD49_VL    <= Diff49_VL(num downto 0) when Diff49_VL(num+1) = '0' else (not(Diff49_VL(num downto 0))+1);
SAD50_VL    <= Diff50_VL(num downto 0) when Diff50_VL(num+1) = '0' else (not(Diff50_VL(num downto 0))+1);
SAD51_VL    <= Diff51_VL(num downto 0) when Diff51_VL(num+1) = '0' else (not(Diff51_VL(num downto 0))+1);
SAD52_VL    <= Diff52_VL(num downto 0) when Diff52_VL(num+1) = '0' else (not(Diff52_VL(num downto 0))+1);
SAD53_VL    <= Diff53_VL(num downto 0) when Diff53_VL(num+1) = '0' else (not(Diff53_VL(num downto 0))+1);
SAD54_VL    <= Diff54_VL(num downto 0) when Diff54_VL(num+1) = '0' else (not(Diff54_VL(num downto 0))+1);
SAD55_VL    <= Diff55_VL(num downto 0) when Diff55_VL(num+1) = '0' else (not(Diff55_VL(num downto 0))+1);
SAD56_VL    <= Diff56_VL(num downto 0) when Diff56_VL(num+1) = '0' else (not(Diff56_VL(num downto 0))+1);
SAD57_VL    <= Diff57_VL(num downto 0) when Diff57_VL(num+1) = '0' else (not(Diff57_VL(num downto 0))+1);
SAD58_VL    <= Diff58_VL(num downto 0) when Diff58_VL(num+1) = '0' else (not(Diff58_VL(num downto 0))+1);
SAD59_VL    <= Diff59_VL(num downto 0) when Diff59_VL(num+1) = '0' else (not(Diff59_VL(num downto 0))+1);
SAD60_VL    <= Diff60_VL(num downto 0) when Diff60_VL(num+1) = '0' else (not(Diff60_VL(num downto 0))+1);
SAD61_VL    <= Diff61_VL(num downto 0) when Diff61_VL(num+1) = '0' else (not(Diff61_VL(num downto 0))+1);
SAD62_VL    <= Diff62_VL(num downto 0) when Diff62_VL(num+1) = '0' else (not(Diff62_VL(num downto 0))+1);		
SAD63_VL    <= Diff63_VL(num downto 0) when Diff63_VL(num+1) = '0' else (not(Diff63_VL(num downto 0))+1);
--HU
SAD0_HU     <= Diff0_HU(num downto 0) when Diff0_HU(num+1) = '0' else (not(Diff0_HU(num downto 0))+1);
SAD1_HU     <= Diff1_HU(num downto 0) when Diff1_HU(num+1) = '0' else (not(Diff1_HU(num downto 0))+1);
SAD2_HU     <= Diff2_HU(num downto 0) when Diff2_HU(num+1) = '0' else (not(Diff2_HU(num downto 0))+1);
SAD3_HU     <= Diff3_HU(num downto 0) when Diff3_HU(num+1) = '0' else (not(Diff3_HU(num downto 0))+1);
SAD4_HU     <= Diff4_HU(num downto 0) when Diff4_HU(num+1) = '0' else (not(Diff4_HU(num downto 0))+1);
SAD5_HU     <= Diff5_HU(num downto 0) when Diff5_HU(num+1) = '0' else (not(Diff5_HU(num downto 0))+1);
SAD6_HU     <= Diff6_HU(num downto 0) when Diff6_HU(num+1) = '0' else (not(Diff6_HU(num downto 0))+1);
SAD7_HU     <= Diff7_HU(num downto 0) when Diff7_HU(num+1) = '0' else (not(Diff7_HU(num downto 0))+1);
SAD8_HU     <= Diff8_HU(num downto 0) when Diff8_HU(num+1) = '0' else (not(Diff8_HU(num downto 0))+1);
SAD9_HU     <= Diff9_HU(num downto 0) when Diff9_HU(num+1) = '0' else (not(Diff9_HU(num downto 0))+1);
SAD10_HU    <= Diff10_HU(num downto 0) when Diff10_HU(num+1) = '0' else (not(Diff10_HU(num downto 0))+1);
SAD11_HU    <= Diff11_HU(num downto 0) when Diff11_HU(num+1) = '0' else (not(Diff11_HU(num downto 0))+1);
SAD12_HU    <= Diff12_HU(num downto 0) when Diff12_HU(num+1) = '0' else (not(Diff12_HU(num downto 0))+1);
SAD13_HU    <= Diff13_HU(num downto 0) when Diff13_HU(num+1) = '0' else (not(Diff13_HU(num downto 0))+1);
SAD14_HU    <= Diff14_HU(num downto 0) when Diff14_HU(num+1) = '0' else (not(Diff14_HU(num downto 0))+1);		
SAD15_HU    <= Diff15_HU(num downto 0) when Diff15_HU(num+1) = '0' else (not(Diff15_HU(num downto 0))+1);
SAD16_HU    <= Diff16_HU(num downto 0) when Diff16_HU(num+1) = '0' else (not(Diff16_HU(num downto 0))+1);
SAD17_HU    <= Diff17_HU(num downto 0) when Diff17_HU(num+1) = '0' else (not(Diff17_HU(num downto 0))+1);
SAD18_HU    <= Diff18_HU(num downto 0) when Diff18_HU(num+1) = '0' else (not(Diff18_HU(num downto 0))+1);
SAD19_HU    <= Diff19_HU(num downto 0) when Diff19_HU(num+1) = '0' else (not(Diff19_HU(num downto 0))+1);
SAD20_HU    <= Diff20_HU(num downto 0) when Diff20_HU(num+1) = '0' else (not(Diff20_HU(num downto 0))+1);
SAD21_HU    <= Diff21_HU(num downto 0) when Diff21_HU(num+1) = '0' else (not(Diff21_HU(num downto 0))+1);
SAD22_HU    <= Diff22_HU(num downto 0) when Diff22_HU(num+1) = '0' else (not(Diff22_HU(num downto 0))+1);
SAD23_HU    <= Diff23_HU(num downto 0) when Diff23_HU(num+1) = '0' else (not(Diff23_HU(num downto 0))+1);
SAD24_HU    <= Diff24_HU(num downto 0) when Diff24_HU(num+1) = '0' else (not(Diff24_HU(num downto 0))+1);
SAD25_HU    <= Diff25_HU(num downto 0) when Diff25_HU(num+1) = '0' else (not(Diff25_HU(num downto 0))+1);
SAD26_HU    <= Diff26_HU(num downto 0) when Diff26_HU(num+1) = '0' else (not(Diff26_HU(num downto 0))+1);
SAD27_HU    <= Diff27_HU(num downto 0) when Diff27_HU(num+1) = '0' else (not(Diff27_HU(num downto 0))+1);
SAD28_HU    <= Diff28_HU(num downto 0) when Diff28_HU(num+1) = '0' else (not(Diff28_HU(num downto 0))+1);
SAD29_HU    <= Diff29_HU(num downto 0) when Diff29_HU(num+1) = '0' else (not(Diff29_HU(num downto 0))+1);
SAD30_HU    <= Diff30_HU(num downto 0) when Diff30_HU(num+1) = '0' else (not(Diff30_HU(num downto 0))+1);		
SAD31_HU    <= Diff31_HU(num downto 0) when Diff31_HU(num+1) = '0' else (not(Diff31_HU(num downto 0))+1);
SAD32_HU    <= Diff32_HU(num downto 0) when Diff32_HU(num+1) = '0' else (not(Diff32_HU(num downto 0))+1);
SAD33_HU    <= Diff33_HU(num downto 0) when Diff33_HU(num+1) = '0' else (not(Diff33_HU(num downto 0))+1);
SAD34_HU    <= Diff34_HU(num downto 0) when Diff34_HU(num+1) = '0' else (not(Diff34_HU(num downto 0))+1);
SAD35_HU    <= Diff35_HU(num downto 0) when Diff35_HU(num+1) = '0' else (not(Diff35_HU(num downto 0))+1);
SAD36_HU    <= Diff36_HU(num downto 0) when Diff36_HU(num+1) = '0' else (not(Diff36_HU(num downto 0))+1);
SAD37_HU    <= Diff37_HU(num downto 0) when Diff37_HU(num+1) = '0' else (not(Diff37_HU(num downto 0))+1);
SAD38_HU    <= Diff38_HU(num downto 0) when Diff38_HU(num+1) = '0' else (not(Diff38_HU(num downto 0))+1);
SAD39_HU    <= Diff39_HU(num downto 0) when Diff39_HU(num+1) = '0' else (not(Diff39_HU(num downto 0))+1);
SAD40_HU    <= Diff40_HU(num downto 0) when Diff40_HU(num+1) = '0' else (not(Diff40_HU(num downto 0))+1);
SAD41_HU    <= Diff41_HU(num downto 0) when Diff41_HU(num+1) = '0' else (not(Diff41_HU(num downto 0))+1);
SAD42_HU    <= Diff42_HU(num downto 0) when Diff42_HU(num+1) = '0' else (not(Diff42_HU(num downto 0))+1);
SAD43_HU    <= Diff43_HU(num downto 0) when Diff43_HU(num+1) = '0' else (not(Diff43_HU(num downto 0))+1);
SAD44_HU    <= Diff44_HU(num downto 0) when Diff44_HU(num+1) = '0' else (not(Diff44_HU(num downto 0))+1);
SAD45_HU    <= Diff45_HU(num downto 0) when Diff45_HU(num+1) = '0' else (not(Diff45_HU(num downto 0))+1);
SAD46_HU    <= Diff46_HU(num downto 0) when Diff46_HU(num+1) = '0' else (not(Diff46_HU(num downto 0))+1);		
SAD47_HU    <= Diff47_HU(num downto 0) when Diff47_HU(num+1) = '0' else (not(Diff47_HU(num downto 0))+1);
SAD48_HU    <= Diff48_HU(num downto 0) when Diff48_HU(num+1) = '0' else (not(Diff48_HU(num downto 0))+1);
SAD49_HU    <= Diff49_HU(num downto 0) when Diff49_HU(num+1) = '0' else (not(Diff49_HU(num downto 0))+1);
SAD50_HU    <= Diff50_HU(num downto 0) when Diff50_HU(num+1) = '0' else (not(Diff50_HU(num downto 0))+1);
SAD51_HU    <= Diff51_HU(num downto 0) when Diff51_HU(num+1) = '0' else (not(Diff51_HU(num downto 0))+1);
SAD52_HU    <= Diff52_HU(num downto 0) when Diff52_HU(num+1) = '0' else (not(Diff52_HU(num downto 0))+1);
SAD53_HU    <= Diff53_HU(num downto 0) when Diff53_HU(num+1) = '0' else (not(Diff53_HU(num downto 0))+1);
SAD54_HU    <= Diff54_HU(num downto 0) when Diff54_HU(num+1) = '0' else (not(Diff54_HU(num downto 0))+1);
SAD55_HU    <= Diff55_HU(num downto 0) when Diff55_HU(num+1) = '0' else (not(Diff55_HU(num downto 0))+1);
SAD56_HU    <= Diff56_HU(num downto 0) when Diff56_HU(num+1) = '0' else (not(Diff56_HU(num downto 0))+1);
SAD57_HU    <= Diff57_HU(num downto 0) when Diff57_HU(num+1) = '0' else (not(Diff57_HU(num downto 0))+1);
SAD58_HU    <= Diff58_HU(num downto 0) when Diff58_HU(num+1) = '0' else (not(Diff58_HU(num downto 0))+1);
SAD59_HU    <= Diff59_HU(num downto 0) when Diff59_HU(num+1) = '0' else (not(Diff59_HU(num downto 0))+1);
SAD60_HU    <= Diff60_HU(num downto 0) when Diff60_HU(num+1) = '0' else (not(Diff60_HU(num downto 0))+1);
SAD61_HU    <= Diff61_HU(num downto 0) when Diff61_HU(num+1) = '0' else (not(Diff61_HU(num downto 0))+1);
SAD62_HU    <= Diff62_HU(num downto 0) when Diff62_HU(num+1) = '0' else (not(Diff62_HU(num downto 0))+1);		
SAD63_HU    <= Diff63_HU(num downto 0) when Diff63_HU(num+1) = '0' else (not(Diff63_HU(num downto 0))+1);

-- end SAD stuff


-------------------------------------------------------------------------------------------------------------------------
--#####################################################################################################################--
--# Store SADs in registers
--#####################################################################################################################--
-------------------------------------------------------------------------------------------------------------------------
PROCESS (clk, rst)
BEGIN
  IF (rst = '1') THEN

	SAD0_V_reg  <= (others=>'0'); SAD1_V_reg  <= (others=>'0');	SAD2_V_reg  <= (others=>'0'); SAD3_V_reg  <= (others=>'0');
	SAD4_V_reg  <= (others=>'0'); SAD5_V_reg  <= (others=>'0');	SAD6_V_reg  <= (others=>'0'); SAD7_V_reg  <= (others=>'0');
	SAD8_V_reg  <= (others=>'0'); SAD9_V_reg  <= (others=>'0');	SAD10_V_reg <= (others=>'0'); SAD11_V_reg <= (others=>'0');
	SAD12_V_reg <= (others=>'0'); SAD13_V_reg <= (others=>'0');	SAD14_V_reg <= (others=>'0'); SAD15_V_reg <= (others=>'0');
	SAD16_V_reg <= (others=>'0'); SAD17_V_reg <= (others=>'0');	SAD18_V_reg <= (others=>'0'); SAD19_V_reg <= (others=>'0');
	SAD20_V_reg <= (others=>'0'); SAD21_V_reg <= (others=>'0');	SAD22_V_reg <= (others=>'0'); SAD23_V_reg <= (others=>'0');
	SAD24_V_reg <= (others=>'0'); SAD25_V_reg <= (others=>'0');	SAD26_V_reg <= (others=>'0'); SAD27_V_reg <= (others=>'0');
	SAD28_V_reg <= (others=>'0'); SAD29_V_reg <= (others=>'0');	SAD30_V_reg <= (others=>'0'); SAD31_V_reg <= (others=>'0');
	SAD32_V_reg <= (others=>'0'); SAD33_V_reg <= (others=>'0');	SAD34_V_reg <= (others=>'0'); SAD35_V_reg <= (others=>'0');
	SAD36_V_reg <= (others=>'0'); SAD37_V_reg <= (others=>'0');	SAD38_V_reg <= (others=>'0'); SAD39_V_reg <= (others=>'0');
	SAD40_V_reg <= (others=>'0'); SAD41_V_reg <= (others=>'0');	SAD42_V_reg <= (others=>'0'); SAD43_V_reg <= (others=>'0');
	SAD44_V_reg <= (others=>'0'); SAD45_V_reg <= (others=>'0');	SAD46_V_reg <= (others=>'0'); SAD47_V_reg <= (others=>'0');
	SAD48_V_reg <= (others=>'0'); SAD49_V_reg <= (others=>'0');	SAD50_V_reg <= (others=>'0'); SAD51_V_reg <= (others=>'0');
	SAD52_V_reg <= (others=>'0'); SAD53_V_reg <= (others=>'0');	SAD54_V_reg <= (others=>'0'); SAD55_V_reg <= (others=>'0');
	SAD56_V_reg <= (others=>'0'); SAD57_V_reg <= (others=>'0');	SAD58_V_reg <= (others=>'0'); SAD59_V_reg <= (others=>'0');
	SAD60_V_reg <= (others=>'0'); SAD61_V_reg <= (others=>'0');	SAD62_V_reg <= (others=>'0'); SAD63_V_reg <= (others=>'0');
	
	SAD0_H_reg  <= (others=>'0'); SAD1_H_reg  <= (others=>'0');	SAD2_H_reg  <= (others=>'0'); SAD3_H_reg  <= (others=>'0');
	SAD4_H_reg  <= (others=>'0'); SAD5_H_reg  <= (others=>'0');	SAD6_H_reg  <= (others=>'0'); SAD7_H_reg  <= (others=>'0');
	SAD8_H_reg  <= (others=>'0'); SAD9_H_reg  <= (others=>'0');	SAD10_H_reg <= (others=>'0'); SAD11_H_reg <= (others=>'0');
	SAD12_H_reg <= (others=>'0'); SAD13_H_reg <= (others=>'0');	SAD14_H_reg <= (others=>'0'); SAD15_H_reg <= (others=>'0');
	SAD16_H_reg <= (others=>'0'); SAD17_H_reg <= (others=>'0');	SAD18_H_reg <= (others=>'0'); SAD19_H_reg <= (others=>'0');
	SAD20_H_reg <= (others=>'0'); SAD21_H_reg <= (others=>'0');	SAD22_H_reg <= (others=>'0'); SAD23_H_reg <= (others=>'0');
	SAD24_H_reg <= (others=>'0'); SAD25_H_reg <= (others=>'0');	SAD26_H_reg <= (others=>'0'); SAD27_H_reg <= (others=>'0');
	SAD28_H_reg <= (others=>'0'); SAD29_H_reg <= (others=>'0');	SAD30_H_reg <= (others=>'0'); SAD31_H_reg <= (others=>'0');
	SAD32_H_reg <= (others=>'0'); SAD33_H_reg <= (others=>'0');	SAD34_H_reg <= (others=>'0'); SAD35_H_reg <= (others=>'0');
	SAD36_H_reg <= (others=>'0'); SAD37_H_reg <= (others=>'0');	SAD38_H_reg <= (others=>'0'); SAD39_H_reg <= (others=>'0');
	SAD40_H_reg <= (others=>'0'); SAD41_H_reg <= (others=>'0');	SAD42_H_reg <= (others=>'0'); SAD43_H_reg <= (others=>'0');
	SAD44_H_reg <= (others=>'0'); SAD45_H_reg <= (others=>'0');	SAD46_H_reg <= (others=>'0'); SAD47_H_reg <= (others=>'0');
	SAD48_H_reg <= (others=>'0'); SAD49_H_reg <= (others=>'0');	SAD50_H_reg <= (others=>'0'); SAD51_H_reg <= (others=>'0');
	SAD52_H_reg <= (others=>'0'); SAD53_H_reg <= (others=>'0');	SAD54_H_reg <= (others=>'0'); SAD55_H_reg <= (others=>'0');
	SAD56_H_reg <= (others=>'0'); SAD57_H_reg <= (others=>'0');	SAD58_H_reg <= (others=>'0'); SAD59_H_reg <= (others=>'0');
	SAD60_H_reg <= (others=>'0'); SAD61_H_reg <= (others=>'0');	SAD62_H_reg <= (others=>'0'); SAD63_H_reg <= (others=>'0');

	SAD0_DC_reg  <= (others=>'0'); SAD1_DC_reg  <= (others=>'0');	SAD2_DC_reg  <= (others=>'0'); SAD3_DC_reg  <= (others=>'0');
	SAD4_DC_reg  <= (others=>'0'); SAD5_DC_reg  <= (others=>'0');	SAD6_DC_reg  <= (others=>'0'); SAD7_DC_reg  <= (others=>'0');
	SAD8_DC_reg  <= (others=>'0'); SAD9_DC_reg  <= (others=>'0');	SAD10_DC_reg <= (others=>'0'); SAD11_DC_reg <= (others=>'0');
	SAD12_DC_reg <= (others=>'0'); SAD13_DC_reg <= (others=>'0');	SAD14_DC_reg <= (others=>'0'); SAD15_DC_reg <= (others=>'0');
	SAD16_DC_reg <= (others=>'0'); SAD17_DC_reg <= (others=>'0');	SAD18_DC_reg <= (others=>'0'); SAD19_DC_reg <= (others=>'0');
	SAD20_DC_reg <= (others=>'0'); SAD21_DC_reg <= (others=>'0');	SAD22_DC_reg <= (others=>'0'); SAD23_DC_reg <= (others=>'0');
	SAD24_DC_reg <= (others=>'0'); SAD25_DC_reg <= (others=>'0');	SAD26_DC_reg <= (others=>'0'); SAD27_DC_reg <= (others=>'0');
	SAD28_DC_reg <= (others=>'0'); SAD29_DC_reg <= (others=>'0');	SAD30_DC_reg <= (others=>'0'); SAD31_DC_reg <= (others=>'0');
	SAD32_DC_reg <= (others=>'0'); SAD33_DC_reg <= (others=>'0');	SAD34_DC_reg <= (others=>'0'); SAD35_DC_reg <= (others=>'0');
	SAD36_DC_reg <= (others=>'0'); SAD37_DC_reg <= (others=>'0');	SAD38_DC_reg <= (others=>'0'); SAD39_DC_reg <= (others=>'0');
	SAD40_DC_reg <= (others=>'0'); SAD41_DC_reg <= (others=>'0');	SAD42_DC_reg <= (others=>'0'); SAD43_DC_reg <= (others=>'0');
	SAD44_DC_reg <= (others=>'0'); SAD45_DC_reg <= (others=>'0');	SAD46_DC_reg <= (others=>'0'); SAD47_DC_reg <= (others=>'0');
	SAD48_DC_reg <= (others=>'0'); SAD49_DC_reg <= (others=>'0');	SAD50_DC_reg <= (others=>'0'); SAD51_DC_reg <= (others=>'0');
	SAD52_DC_reg <= (others=>'0'); SAD53_DC_reg <= (others=>'0');	SAD54_DC_reg <= (others=>'0'); SAD55_DC_reg <= (others=>'0');
	SAD56_DC_reg <= (others=>'0'); SAD57_DC_reg <= (others=>'0');	SAD58_DC_reg <= (others=>'0'); SAD59_DC_reg <= (others=>'0');
	SAD60_DC_reg <= (others=>'0'); SAD61_DC_reg <= (others=>'0');	SAD62_DC_reg <= (others=>'0'); SAD63_DC_reg <= (others=>'0');

	SAD0_DDL_reg  <= (others=>'0'); SAD1_DDL_reg  <= (others=>'0');	SAD2_DDL_reg  <= (others=>'0'); SAD3_DDL_reg  <= (others=>'0');
	SAD4_DDL_reg  <= (others=>'0'); SAD5_DDL_reg  <= (others=>'0');	SAD6_DDL_reg  <= (others=>'0'); SAD7_DDL_reg  <= (others=>'0');
	SAD8_DDL_reg  <= (others=>'0'); SAD9_DDL_reg  <= (others=>'0');	SAD10_DDL_reg <= (others=>'0'); SAD11_DDL_reg <= (others=>'0');
	SAD12_DDL_reg <= (others=>'0'); SAD13_DDL_reg <= (others=>'0');	SAD14_DDL_reg <= (others=>'0'); SAD15_DDL_reg <= (others=>'0');
	SAD16_DDL_reg <= (others=>'0'); SAD17_DDL_reg <= (others=>'0');	SAD18_DDL_reg <= (others=>'0'); SAD19_DDL_reg <= (others=>'0');
	SAD20_DDL_reg <= (others=>'0'); SAD21_DDL_reg <= (others=>'0');	SAD22_DDL_reg <= (others=>'0'); SAD23_DDL_reg <= (others=>'0');
	SAD24_DDL_reg <= (others=>'0'); SAD25_DDL_reg <= (others=>'0');	SAD26_DDL_reg <= (others=>'0'); SAD27_DDL_reg <= (others=>'0');
	SAD28_DDL_reg <= (others=>'0'); SAD29_DDL_reg <= (others=>'0');	SAD30_DDL_reg <= (others=>'0'); SAD31_DDL_reg <= (others=>'0');
	SAD32_DDL_reg <= (others=>'0'); SAD33_DDL_reg <= (others=>'0');	SAD34_DDL_reg <= (others=>'0'); SAD35_DDL_reg <= (others=>'0');
	SAD36_DDL_reg <= (others=>'0'); SAD37_DDL_reg <= (others=>'0');	SAD38_DDL_reg <= (others=>'0'); SAD39_DDL_reg <= (others=>'0');
	SAD40_DDL_reg <= (others=>'0'); SAD41_DDL_reg <= (others=>'0');	SAD42_DDL_reg <= (others=>'0'); SAD43_DDL_reg <= (others=>'0');
	SAD44_DDL_reg <= (others=>'0'); SAD45_DDL_reg <= (others=>'0');	SAD46_DDL_reg <= (others=>'0'); SAD47_DDL_reg <= (others=>'0');
	SAD48_DDL_reg <= (others=>'0'); SAD49_DDL_reg <= (others=>'0');	SAD50_DDL_reg <= (others=>'0'); SAD51_DDL_reg <= (others=>'0');
	SAD52_DDL_reg <= (others=>'0'); SAD53_DDL_reg <= (others=>'0');	SAD54_DDL_reg <= (others=>'0'); SAD55_DDL_reg <= (others=>'0');
	SAD56_DDL_reg <= (others=>'0'); SAD57_DDL_reg <= (others=>'0');	SAD58_DDL_reg <= (others=>'0'); SAD59_DDL_reg <= (others=>'0');
	SAD60_DDL_reg <= (others=>'0'); SAD61_DDL_reg <= (others=>'0');	SAD62_DDL_reg <= (others=>'0'); SAD63_DDL_reg <= (others=>'0');

	SAD0_DDR_reg  <= (others=>'0'); SAD1_DDR_reg  <= (others=>'0');	SAD2_DDR_reg  <= (others=>'0'); SAD3_DDR_reg  <= (others=>'0');
	SAD4_DDR_reg  <= (others=>'0'); SAD5_DDR_reg  <= (others=>'0');	SAD6_DDR_reg  <= (others=>'0'); SAD7_DDR_reg  <= (others=>'0');
	SAD8_DDR_reg  <= (others=>'0'); SAD9_DDR_reg  <= (others=>'0');	SAD10_DDR_reg <= (others=>'0'); SAD11_DDR_reg <= (others=>'0');
	SAD12_DDR_reg <= (others=>'0'); SAD13_DDR_reg <= (others=>'0');	SAD14_DDR_reg <= (others=>'0'); SAD15_DDR_reg <= (others=>'0');
	SAD16_DDR_reg <= (others=>'0'); SAD17_DDR_reg <= (others=>'0');	SAD18_DDR_reg <= (others=>'0'); SAD19_DDR_reg <= (others=>'0');
	SAD20_DDR_reg <= (others=>'0'); SAD21_DDR_reg <= (others=>'0');	SAD22_DDR_reg <= (others=>'0'); SAD23_DDR_reg <= (others=>'0');
	SAD24_DDR_reg <= (others=>'0'); SAD25_DDR_reg <= (others=>'0');	SAD26_DDR_reg <= (others=>'0'); SAD27_DDR_reg <= (others=>'0');
	SAD28_DDR_reg <= (others=>'0'); SAD29_DDR_reg <= (others=>'0');	SAD30_DDR_reg <= (others=>'0'); SAD31_DDR_reg <= (others=>'0');
	SAD32_DDR_reg <= (others=>'0'); SAD33_DDR_reg <= (others=>'0');	SAD34_DDR_reg <= (others=>'0'); SAD35_DDR_reg <= (others=>'0');
	SAD36_DDR_reg <= (others=>'0'); SAD37_DDR_reg <= (others=>'0');	SAD38_DDR_reg <= (others=>'0'); SAD39_DDR_reg <= (others=>'0');
	SAD40_DDR_reg <= (others=>'0'); SAD41_DDR_reg <= (others=>'0');	SAD42_DDR_reg <= (others=>'0'); SAD43_DDR_reg <= (others=>'0');
	SAD44_DDR_reg <= (others=>'0'); SAD45_DDR_reg <= (others=>'0');	SAD46_DDR_reg <= (others=>'0'); SAD47_DDR_reg <= (others=>'0');
	SAD48_DDR_reg <= (others=>'0'); SAD49_DDR_reg <= (others=>'0');	SAD50_DDR_reg <= (others=>'0'); SAD51_DDR_reg <= (others=>'0');
	SAD52_DDR_reg <= (others=>'0'); SAD53_DDR_reg <= (others=>'0');	SAD54_DDR_reg <= (others=>'0'); SAD55_DDR_reg <= (others=>'0');
	SAD56_DDR_reg <= (others=>'0'); SAD57_DDR_reg <= (others=>'0');	SAD58_DDR_reg <= (others=>'0'); SAD59_DDR_reg <= (others=>'0');
	SAD60_DDR_reg <= (others=>'0'); SAD61_DDR_reg <= (others=>'0');	SAD62_DDR_reg <= (others=>'0'); SAD63_DDR_reg <= (others=>'0');

	SAD0_VR_reg  <= (others=>'0'); SAD1_VR_reg  <= (others=>'0');	SAD2_VR_reg  <= (others=>'0'); SAD3_VR_reg  <= (others=>'0');
	SAD4_VR_reg  <= (others=>'0'); SAD5_VR_reg  <= (others=>'0');	SAD6_VR_reg  <= (others=>'0'); SAD7_VR_reg  <= (others=>'0');
	SAD8_VR_reg  <= (others=>'0'); SAD9_VR_reg  <= (others=>'0');	SAD10_VR_reg <= (others=>'0'); SAD11_VR_reg <= (others=>'0');
	SAD12_VR_reg <= (others=>'0'); SAD13_VR_reg <= (others=>'0');	SAD14_VR_reg <= (others=>'0'); SAD15_VR_reg <= (others=>'0');
	SAD16_VR_reg <= (others=>'0'); SAD17_VR_reg <= (others=>'0');	SAD18_VR_reg <= (others=>'0'); SAD19_VR_reg <= (others=>'0');
	SAD20_VR_reg <= (others=>'0'); SAD21_VR_reg <= (others=>'0');	SAD22_VR_reg <= (others=>'0'); SAD23_VR_reg <= (others=>'0');
	SAD24_VR_reg <= (others=>'0'); SAD25_VR_reg <= (others=>'0');	SAD26_VR_reg <= (others=>'0'); SAD27_VR_reg <= (others=>'0');
	SAD28_VR_reg <= (others=>'0'); SAD29_VR_reg <= (others=>'0');	SAD30_VR_reg <= (others=>'0'); SAD31_VR_reg <= (others=>'0');
	SAD32_VR_reg <= (others=>'0'); SAD33_VR_reg <= (others=>'0');	SAD34_VR_reg <= (others=>'0'); SAD35_VR_reg <= (others=>'0');
	SAD36_VR_reg <= (others=>'0'); SAD37_VR_reg <= (others=>'0');	SAD38_VR_reg <= (others=>'0'); SAD39_VR_reg <= (others=>'0');
	SAD40_VR_reg <= (others=>'0'); SAD41_VR_reg <= (others=>'0');	SAD42_VR_reg <= (others=>'0'); SAD43_VR_reg <= (others=>'0');
	SAD44_VR_reg <= (others=>'0'); SAD45_VR_reg <= (others=>'0');	SAD46_VR_reg <= (others=>'0'); SAD47_VR_reg <= (others=>'0');
	SAD48_VR_reg <= (others=>'0'); SAD49_VR_reg <= (others=>'0');	SAD50_VR_reg <= (others=>'0'); SAD51_VR_reg <= (others=>'0');
	SAD52_VR_reg <= (others=>'0'); SAD53_VR_reg <= (others=>'0');	SAD54_VR_reg <= (others=>'0'); SAD55_VR_reg <= (others=>'0');
	SAD56_VR_reg <= (others=>'0'); SAD57_VR_reg <= (others=>'0');	SAD58_VR_reg <= (others=>'0'); SAD59_VR_reg <= (others=>'0');
	SAD60_VR_reg <= (others=>'0'); SAD61_VR_reg <= (others=>'0');	SAD62_VR_reg <= (others=>'0'); SAD63_VR_reg <= (others=>'0');

	SAD0_HD_reg  <= (others=>'0'); SAD1_HD_reg  <= (others=>'0');	SAD2_HD_reg  <= (others=>'0'); SAD3_HD_reg  <= (others=>'0');
	SAD4_HD_reg  <= (others=>'0'); SAD5_HD_reg  <= (others=>'0');	SAD6_HD_reg  <= (others=>'0'); SAD7_HD_reg  <= (others=>'0');
	SAD8_HD_reg  <= (others=>'0'); SAD9_HD_reg  <= (others=>'0');	SAD10_HD_reg <= (others=>'0'); SAD11_HD_reg <= (others=>'0');
	SAD12_HD_reg <= (others=>'0'); SAD13_HD_reg <= (others=>'0');	SAD14_HD_reg <= (others=>'0'); SAD15_HD_reg <= (others=>'0');
	SAD16_HD_reg <= (others=>'0'); SAD17_HD_reg <= (others=>'0');	SAD18_HD_reg <= (others=>'0'); SAD19_HD_reg <= (others=>'0');
	SAD20_HD_reg <= (others=>'0'); SAD21_HD_reg <= (others=>'0');	SAD22_HD_reg <= (others=>'0'); SAD23_HD_reg <= (others=>'0');
	SAD24_HD_reg <= (others=>'0'); SAD25_HD_reg <= (others=>'0');	SAD26_HD_reg <= (others=>'0'); SAD27_HD_reg <= (others=>'0');
	SAD28_HD_reg <= (others=>'0'); SAD29_HD_reg <= (others=>'0');	SAD30_HD_reg <= (others=>'0'); SAD31_HD_reg <= (others=>'0');
	SAD32_HD_reg <= (others=>'0'); SAD33_HD_reg <= (others=>'0');	SAD34_HD_reg <= (others=>'0'); SAD35_HD_reg <= (others=>'0');
	SAD36_HD_reg <= (others=>'0'); SAD37_HD_reg <= (others=>'0');	SAD38_HD_reg <= (others=>'0'); SAD39_HD_reg <= (others=>'0');
	SAD40_HD_reg <= (others=>'0'); SAD41_HD_reg <= (others=>'0');	SAD42_HD_reg <= (others=>'0'); SAD43_HD_reg <= (others=>'0');
	SAD44_HD_reg <= (others=>'0'); SAD45_HD_reg <= (others=>'0');	SAD46_HD_reg <= (others=>'0'); SAD47_HD_reg <= (others=>'0');
	SAD48_HD_reg <= (others=>'0'); SAD49_HD_reg <= (others=>'0');	SAD50_HD_reg <= (others=>'0'); SAD51_HD_reg <= (others=>'0');
	SAD52_HD_reg <= (others=>'0'); SAD53_HD_reg <= (others=>'0');	SAD54_HD_reg <= (others=>'0'); SAD55_HD_reg <= (others=>'0');
	SAD56_HD_reg <= (others=>'0'); SAD57_HD_reg <= (others=>'0');	SAD58_HD_reg <= (others=>'0'); SAD59_HD_reg <= (others=>'0');
	SAD60_HD_reg <= (others=>'0'); SAD61_HD_reg <= (others=>'0');	SAD62_HD_reg <= (others=>'0'); SAD63_HD_reg <= (others=>'0');

	SAD0_VL_reg  <= (others=>'0'); SAD1_VL_reg  <= (others=>'0');	SAD2_VL_reg  <= (others=>'0'); SAD3_VL_reg  <= (others=>'0');
	SAD4_VL_reg  <= (others=>'0'); SAD5_VL_reg  <= (others=>'0');	SAD6_VL_reg  <= (others=>'0'); SAD7_VL_reg  <= (others=>'0');
	SAD8_VL_reg  <= (others=>'0'); SAD9_VL_reg  <= (others=>'0');	SAD10_VL_reg <= (others=>'0'); SAD11_VL_reg <= (others=>'0');
	SAD12_VL_reg <= (others=>'0'); SAD13_VL_reg <= (others=>'0');	SAD14_VL_reg <= (others=>'0'); SAD15_VL_reg <= (others=>'0');
	SAD16_VL_reg <= (others=>'0'); SAD17_VL_reg <= (others=>'0');	SAD18_VL_reg <= (others=>'0'); SAD19_VL_reg <= (others=>'0');
	SAD20_VL_reg <= (others=>'0'); SAD21_VL_reg <= (others=>'0');	SAD22_VL_reg <= (others=>'0'); SAD23_VL_reg <= (others=>'0');
	SAD24_VL_reg <= (others=>'0'); SAD25_VL_reg <= (others=>'0');	SAD26_VL_reg <= (others=>'0'); SAD27_VL_reg <= (others=>'0');
	SAD28_VL_reg <= (others=>'0'); SAD29_VL_reg <= (others=>'0');	SAD30_VL_reg <= (others=>'0'); SAD31_VL_reg <= (others=>'0');
	SAD32_VL_reg <= (others=>'0'); SAD33_VL_reg <= (others=>'0');	SAD34_VL_reg <= (others=>'0'); SAD35_VL_reg <= (others=>'0');
	SAD36_VL_reg <= (others=>'0'); SAD37_VL_reg <= (others=>'0');	SAD38_VL_reg <= (others=>'0'); SAD39_VL_reg <= (others=>'0');
	SAD40_VL_reg <= (others=>'0'); SAD41_VL_reg <= (others=>'0');	SAD42_VL_reg <= (others=>'0'); SAD43_VL_reg <= (others=>'0');
	SAD44_VL_reg <= (others=>'0'); SAD45_VL_reg <= (others=>'0');	SAD46_VL_reg <= (others=>'0'); SAD47_VL_reg <= (others=>'0');
	SAD48_VL_reg <= (others=>'0'); SAD49_VL_reg <= (others=>'0');	SAD50_VL_reg <= (others=>'0'); SAD51_VL_reg <= (others=>'0');
	SAD52_VL_reg <= (others=>'0'); SAD53_VL_reg <= (others=>'0');	SAD54_VL_reg <= (others=>'0'); SAD55_VL_reg <= (others=>'0');
	SAD56_VL_reg <= (others=>'0'); SAD57_VL_reg <= (others=>'0');	SAD58_VL_reg <= (others=>'0'); SAD59_VL_reg <= (others=>'0');
	SAD60_VL_reg <= (others=>'0'); SAD61_VL_reg <= (others=>'0');	SAD62_VL_reg <= (others=>'0'); SAD63_VL_reg <= (others=>'0');

	SAD0_HU_reg  <= (others=>'0'); SAD1_HU_reg  <= (others=>'0');	SAD2_HU_reg  <= (others=>'0'); SAD3_HU_reg  <= (others=>'0');
	SAD4_HU_reg  <= (others=>'0'); SAD5_HU_reg  <= (others=>'0');	SAD6_HU_reg  <= (others=>'0'); SAD7_HU_reg  <= (others=>'0');
	SAD8_HU_reg  <= (others=>'0'); SAD9_HU_reg  <= (others=>'0');	SAD10_HU_reg <= (others=>'0'); SAD11_HU_reg <= (others=>'0');
	SAD12_HU_reg <= (others=>'0'); SAD13_HU_reg <= (others=>'0');	SAD14_HU_reg <= (others=>'0'); SAD15_HU_reg <= (others=>'0');
	SAD16_HU_reg <= (others=>'0'); SAD17_HU_reg <= (others=>'0');	SAD18_HU_reg <= (others=>'0'); SAD19_HU_reg <= (others=>'0');
	SAD20_HU_reg <= (others=>'0'); SAD21_HU_reg <= (others=>'0');	SAD22_HU_reg <= (others=>'0'); SAD23_HU_reg <= (others=>'0');
	SAD24_HU_reg <= (others=>'0'); SAD25_HU_reg <= (others=>'0');	SAD26_HU_reg <= (others=>'0'); SAD27_HU_reg <= (others=>'0');
	SAD28_HU_reg <= (others=>'0'); SAD29_HU_reg <= (others=>'0');	SAD30_HU_reg <= (others=>'0'); SAD31_HU_reg <= (others=>'0');
	SAD32_HU_reg <= (others=>'0'); SAD33_HU_reg <= (others=>'0');	SAD34_HU_reg <= (others=>'0'); SAD35_HU_reg <= (others=>'0');
	SAD36_HU_reg <= (others=>'0'); SAD37_HU_reg <= (others=>'0');	SAD38_HU_reg <= (others=>'0'); SAD39_HU_reg <= (others=>'0');
	SAD40_HU_reg <= (others=>'0'); SAD41_HU_reg <= (others=>'0');	SAD42_HU_reg <= (others=>'0'); SAD43_HU_reg <= (others=>'0');
	SAD44_HU_reg <= (others=>'0'); SAD45_HU_reg <= (others=>'0');	SAD46_HU_reg <= (others=>'0'); SAD47_HU_reg <= (others=>'0');
	SAD48_HU_reg <= (others=>'0'); SAD49_HU_reg <= (others=>'0');	SAD50_HU_reg <= (others=>'0'); SAD51_HU_reg <= (others=>'0');
	SAD52_HU_reg <= (others=>'0'); SAD53_HU_reg <= (others=>'0');	SAD54_HU_reg <= (others=>'0'); SAD55_HU_reg <= (others=>'0');
	SAD56_HU_reg <= (others=>'0'); SAD57_HU_reg <= (others=>'0');	SAD58_HU_reg <= (others=>'0'); SAD59_HU_reg <= (others=>'0');
	SAD60_HU_reg <= (others=>'0'); SAD61_HU_reg <= (others=>'0');	SAD62_HU_reg <= (others=>'0'); SAD63_HU_reg <= (others=>'0');	
	           
  ELSIF (clk'EVENT AND clk = '1') THEN	  
	  
	SAD0_V_reg  <= SAD0_V;  SAD1_V_reg  <= SAD1_V;	SAD2_V_reg  <= SAD2_V;	SAD3_V_reg  <= SAD3_V;
	SAD4_V_reg  <= SAD4_V;	SAD5_V_reg  <= SAD5_V;	SAD6_V_reg  <= SAD6_V;	SAD7_V_reg  <= SAD7_V;
	SAD8_V_reg  <= SAD8_V;	SAD9_V_reg  <= SAD9_V;	SAD10_V_reg <= SAD10_V;	SAD11_V_reg <= SAD11_V;
	SAD12_V_reg <= SAD12_V;	SAD13_V_reg <= SAD13_V;	SAD14_V_reg <= SAD14_V;	SAD15_V_reg <= SAD15_V;
	SAD16_V_reg <= SAD16_V; SAD17_V_reg <= SAD17_V;	SAD18_V_reg <= SAD18_V;	SAD19_V_reg <= SAD19_V;
	SAD20_V_reg <= SAD20_V;	SAD21_V_reg <= SAD21_V;	SAD22_V_reg <= SAD22_V;	SAD23_V_reg <= SAD23_V;
	SAD24_V_reg <= SAD24_V;	SAD25_V_reg <= SAD25_V;	SAD26_V_reg <= SAD26_V;	SAD27_V_reg <= SAD27_V;
	SAD28_V_reg <= SAD28_V;	SAD29_V_reg <= SAD29_V;	SAD30_V_reg <= SAD30_V;	SAD31_V_reg <= SAD31_V;
	SAD32_V_reg <= SAD32_V; SAD33_V_reg <= SAD33_V;	SAD34_V_reg <= SAD34_V;	SAD35_V_reg <= SAD35_V;
	SAD36_V_reg <= SAD36_V;	SAD37_V_reg <= SAD37_V;	SAD38_V_reg <= SAD38_V;	SAD39_V_reg <= SAD39_V;
	SAD40_V_reg <= SAD40_V;	SAD41_V_reg <= SAD41_V;	SAD42_V_reg <= SAD42_V;	SAD43_V_reg <= SAD43_V;
	SAD44_V_reg <= SAD44_V;	SAD45_V_reg <= SAD45_V;	SAD46_V_reg <= SAD46_V;	SAD47_V_reg <= SAD47_V;
	SAD48_V_reg <= SAD48_V; SAD49_V_reg <= SAD49_V;	SAD50_V_reg <= SAD50_V;	SAD51_V_reg <= SAD51_V;
	SAD52_V_reg <= SAD52_V;	SAD53_V_reg <= SAD53_V;	SAD54_V_reg <= SAD54_V;	SAD55_V_reg <= SAD55_V;
	SAD56_V_reg <= SAD56_V;	SAD57_V_reg <= SAD57_V;	SAD58_V_reg <= SAD58_V;	SAD59_V_reg <= SAD59_V;
	SAD60_V_reg <= SAD60_V;	SAD61_V_reg <= SAD61_V;	SAD62_V_reg <= SAD62_V;	SAD63_V_reg <= SAD63_V;	
	
	SAD0_H_reg  <= SAD0_H;  SAD1_H_reg  <= SAD1_H;	SAD2_H_reg  <= SAD2_H;	SAD3_H_reg  <= SAD3_H;
	SAD4_H_reg  <= SAD4_H;	SAD5_H_reg  <= SAD5_H;	SAD6_H_reg  <= SAD6_H;	SAD7_H_reg  <= SAD7_H;
	SAD8_H_reg  <= SAD8_H;	SAD9_H_reg  <= SAD9_H;	SAD10_H_reg <= SAD10_H;	SAD11_H_reg <= SAD11_H;
	SAD12_H_reg <= SAD12_H;	SAD13_H_reg <= SAD13_H;	SAD14_H_reg <= SAD14_H;	SAD15_H_reg <= SAD15_H;
	SAD16_H_reg <= SAD16_H; SAD17_H_reg <= SAD17_H;	SAD18_H_reg <= SAD18_H;	SAD19_H_reg <= SAD19_H;
	SAD20_H_reg <= SAD20_H;	SAD21_H_reg <= SAD21_H;	SAD22_H_reg <= SAD22_H;	SAD23_H_reg <= SAD23_H;
	SAD24_H_reg <= SAD24_H;	SAD25_H_reg <= SAD25_H;	SAD26_H_reg <= SAD26_H;	SAD27_H_reg <= SAD27_H;
	SAD28_H_reg <= SAD28_H;	SAD29_H_reg <= SAD29_H;	SAD30_H_reg <= SAD30_H;	SAD31_H_reg <= SAD31_H;
	SAD32_H_reg <= SAD32_H; SAD33_H_reg <= SAD33_H;	SAD34_H_reg <= SAD34_H;	SAD35_H_reg <= SAD35_H;
	SAD36_H_reg <= SAD36_H;	SAD37_H_reg <= SAD37_H;	SAD38_H_reg <= SAD38_H;	SAD39_H_reg <= SAD39_H;
	SAD40_H_reg <= SAD40_H;	SAD41_H_reg <= SAD41_H;	SAD42_H_reg <= SAD42_H;	SAD43_H_reg <= SAD43_H;
	SAD44_H_reg <= SAD44_H;	SAD45_H_reg <= SAD45_H;	SAD46_H_reg <= SAD46_H;	SAD47_H_reg <= SAD47_H;
	SAD48_H_reg <= SAD48_H; SAD49_H_reg <= SAD49_H;	SAD50_H_reg <= SAD50_H;	SAD51_H_reg <= SAD51_H;
	SAD52_H_reg <= SAD52_H;	SAD53_H_reg <= SAD53_H;	SAD54_H_reg <= SAD54_H;	SAD55_H_reg <= SAD55_H;
	SAD56_H_reg <= SAD56_H;	SAD57_H_reg <= SAD57_H;	SAD58_H_reg <= SAD58_H;	SAD59_H_reg <= SAD59_H;
	SAD60_H_reg <= SAD60_H;	SAD61_H_reg <= SAD61_H;	SAD62_H_reg <= SAD62_H;	SAD63_H_reg <= SAD63_H;	

	SAD0_DC_reg  <= SAD0_DC;  SAD1_DC_reg  <= SAD1_DC;	SAD2_DC_reg  <= SAD2_DC;  SAD3_DC_reg  <= SAD3_DC;
	SAD4_DC_reg  <= SAD4_DC;  SAD5_DC_reg  <= SAD5_DC;	SAD6_DC_reg  <= SAD6_DC;  SAD7_DC_reg  <= SAD7_DC;
	SAD8_DC_reg  <= SAD8_DC;  SAD9_DC_reg  <= SAD9_DC;	SAD10_DC_reg <= SAD10_DC; SAD11_DC_reg <= SAD11_DC;
	SAD12_DC_reg <= SAD12_DC; SAD13_DC_reg <= SAD13_DC;	SAD14_DC_reg <= SAD14_DC; SAD15_DC_reg <= SAD15_DC;
	SAD16_DC_reg <= SAD16_DC; SAD17_DC_reg <= SAD17_DC;	SAD18_DC_reg <= SAD18_DC; SAD19_DC_reg <= SAD19_DC;
	SAD20_DC_reg <= SAD20_DC; SAD21_DC_reg <= SAD21_DC;	SAD22_DC_reg <= SAD22_DC; SAD23_DC_reg <= SAD23_DC;
	SAD24_DC_reg <= SAD24_DC; SAD25_DC_reg <= SAD25_DC;	SAD26_DC_reg <= SAD26_DC; SAD27_DC_reg <= SAD27_DC;
	SAD28_DC_reg <= SAD28_DC; SAD29_DC_reg <= SAD29_DC;	SAD30_DC_reg <= SAD30_DC; SAD31_DC_reg <= SAD31_DC;
	SAD32_DC_reg <= SAD32_DC; SAD33_DC_reg <= SAD33_DC;	SAD34_DC_reg <= SAD34_DC; SAD35_DC_reg <= SAD35_DC;
	SAD36_DC_reg <= SAD36_DC; SAD37_DC_reg <= SAD37_DC;	SAD38_DC_reg <= SAD38_DC; SAD39_DC_reg <= SAD39_DC;
	SAD40_DC_reg <= SAD40_DC; SAD41_DC_reg <= SAD41_DC;	SAD42_DC_reg <= SAD42_DC; SAD43_DC_reg <= SAD43_DC;
	SAD44_DC_reg <= SAD44_DC; SAD45_DC_reg <= SAD45_DC;	SAD46_DC_reg <= SAD46_DC; SAD47_DC_reg <= SAD47_DC;
	SAD48_DC_reg <= SAD48_DC; SAD49_DC_reg <= SAD49_DC;	SAD50_DC_reg <= SAD50_DC; SAD51_DC_reg <= SAD51_DC;
	SAD52_DC_reg <= SAD52_DC; SAD53_DC_reg <= SAD53_DC;	SAD54_DC_reg <= SAD54_DC; SAD55_DC_reg <= SAD55_DC;
	SAD56_DC_reg <= SAD56_DC; SAD57_DC_reg <= SAD57_DC;	SAD58_DC_reg <= SAD58_DC; SAD59_DC_reg <= SAD59_DC;
	SAD60_DC_reg <= SAD60_DC; SAD61_DC_reg <= SAD61_DC;	SAD62_DC_reg <= SAD62_DC; SAD63_DC_reg <= SAD63_DC;	

	SAD0_DDL_reg  <= SAD0_DDL;  SAD1_DDL_reg  <= SAD1_DDL;	SAD2_DDL_reg  <= SAD2_DDL;  SAD3_DDL_reg  <= SAD3_DDL;
	SAD4_DDL_reg  <= SAD4_DDL;  SAD5_DDL_reg  <= SAD5_DDL;	SAD6_DDL_reg  <= SAD6_DDL;  SAD7_DDL_reg  <= SAD7_DDL;
	SAD8_DDL_reg  <= SAD8_DDL;  SAD9_DDL_reg  <= SAD9_DDL;	SAD10_DDL_reg <= SAD10_DDL; SAD11_DDL_reg <= SAD11_DDL;
	SAD12_DDL_reg <= SAD12_DDL; SAD13_DDL_reg <= SAD13_DDL;	SAD14_DDL_reg <= SAD14_DDL; SAD15_DDL_reg <= SAD15_DDL;
	SAD16_DDL_reg <= SAD16_DDL; SAD17_DDL_reg <= SAD17_DDL;	SAD18_DDL_reg <= SAD18_DDL; SAD19_DDL_reg <= SAD19_DDL;
	SAD20_DDL_reg <= SAD20_DDL; SAD21_DDL_reg <= SAD21_DDL;	SAD22_DDL_reg <= SAD22_DDL; SAD23_DDL_reg <= SAD23_DDL;
	SAD24_DDL_reg <= SAD24_DDL; SAD25_DDL_reg <= SAD25_DDL;	SAD26_DDL_reg <= SAD26_DDL; SAD27_DDL_reg <= SAD27_DDL;
	SAD28_DDL_reg <= SAD28_DDL; SAD29_DDL_reg <= SAD29_DDL;	SAD30_DDL_reg <= SAD30_DDL; SAD31_DDL_reg <= SAD31_DDL;
	SAD32_DDL_reg <= SAD32_DDL; SAD33_DDL_reg <= SAD33_DDL;	SAD34_DDL_reg <= SAD34_DDL; SAD35_DDL_reg <= SAD35_DDL;
	SAD36_DDL_reg <= SAD36_DDL; SAD37_DDL_reg <= SAD37_DDL;	SAD38_DDL_reg <= SAD38_DDL; SAD39_DDL_reg <= SAD39_DDL;
	SAD40_DDL_reg <= SAD40_DDL; SAD41_DDL_reg <= SAD41_DDL;	SAD42_DDL_reg <= SAD42_DDL; SAD43_DDL_reg <= SAD43_DDL;
	SAD44_DDL_reg <= SAD44_DDL; SAD45_DDL_reg <= SAD45_DDL;	SAD46_DDL_reg <= SAD46_DDL; SAD47_DDL_reg <= SAD47_DDL;
	SAD48_DDL_reg <= SAD48_DDL; SAD49_DDL_reg <= SAD49_DDL;	SAD50_DDL_reg <= SAD50_DDL; SAD51_DDL_reg <= SAD51_DDL;
	SAD52_DDL_reg <= SAD52_DDL; SAD53_DDL_reg <= SAD53_DDL;	SAD54_DDL_reg <= SAD54_DDL; SAD55_DDL_reg <= SAD55_DDL;
	SAD56_DDL_reg <= SAD56_DDL; SAD57_DDL_reg <= SAD57_DDL;	SAD58_DDL_reg <= SAD58_DDL; SAD59_DDL_reg <= SAD59_DDL;
	SAD60_DDL_reg <= SAD60_DDL; SAD61_DDL_reg <= SAD61_DDL;	SAD62_DDL_reg <= SAD62_DDL; SAD63_DDL_reg <= SAD63_DDL;	

	SAD0_DDR_reg  <= SAD0_DDR;  SAD1_DDR_reg  <= SAD1_DDR;	SAD2_DDR_reg  <= SAD2_DDR;  SAD3_DDR_reg  <= SAD3_DDR;
	SAD4_DDR_reg  <= SAD4_DDR;  SAD5_DDR_reg  <= SAD5_DDR;	SAD6_DDR_reg  <= SAD6_DDR;  SAD7_DDR_reg  <= SAD7_DDR;
	SAD8_DDR_reg  <= SAD8_DDR;  SAD9_DDR_reg  <= SAD9_DDR;	SAD10_DDR_reg <= SAD10_DDR; SAD11_DDR_reg <= SAD11_DDR;
	SAD12_DDR_reg <= SAD12_DDR; SAD13_DDR_reg <= SAD13_DDR;	SAD14_DDR_reg <= SAD14_DDR; SAD15_DDR_reg <= SAD15_DDR;
	SAD16_DDR_reg <= SAD16_DDR; SAD17_DDR_reg <= SAD17_DDR;	SAD18_DDR_reg <= SAD18_DDR; SAD19_DDR_reg <= SAD19_DDR;
	SAD20_DDR_reg <= SAD20_DDR; SAD21_DDR_reg <= SAD21_DDR;	SAD22_DDR_reg <= SAD22_DDR; SAD23_DDR_reg <= SAD23_DDR;
	SAD24_DDR_reg <= SAD24_DDR; SAD25_DDR_reg <= SAD25_DDR;	SAD26_DDR_reg <= SAD26_DDR; SAD27_DDR_reg <= SAD27_DDR;
	SAD28_DDR_reg <= SAD28_DDR; SAD29_DDR_reg <= SAD29_DDR;	SAD30_DDR_reg <= SAD30_DDR; SAD31_DDR_reg <= SAD31_DDR;
	SAD32_DDR_reg <= SAD32_DDR; SAD33_DDR_reg <= SAD33_DDR;	SAD34_DDR_reg <= SAD34_DDR; SAD35_DDR_reg <= SAD35_DDR;
	SAD36_DDR_reg <= SAD36_DDR; SAD37_DDR_reg <= SAD37_DDR;	SAD38_DDR_reg <= SAD38_DDR; SAD39_DDR_reg <= SAD39_DDR;
	SAD40_DDR_reg <= SAD40_DDR; SAD41_DDR_reg <= SAD41_DDR;	SAD42_DDR_reg <= SAD42_DDR; SAD43_DDR_reg <= SAD43_DDR;
	SAD44_DDR_reg <= SAD44_DDR; SAD45_DDR_reg <= SAD45_DDR;	SAD46_DDR_reg <= SAD46_DDR; SAD47_DDR_reg <= SAD47_DDR;
	SAD48_DDR_reg <= SAD48_DDR; SAD49_DDR_reg <= SAD49_DDR;	SAD50_DDR_reg <= SAD50_DDR; SAD51_DDR_reg <= SAD51_DDR;
	SAD52_DDR_reg <= SAD52_DDR; SAD53_DDR_reg <= SAD53_DDR;	SAD54_DDR_reg <= SAD54_DDR; SAD55_DDR_reg <= SAD55_DDR;
	SAD56_DDR_reg <= SAD56_DDR; SAD57_DDR_reg <= SAD57_DDR;	SAD58_DDR_reg <= SAD58_DDR; SAD59_DDR_reg <= SAD59_DDR;
	SAD60_DDR_reg <= SAD60_DDR; SAD61_DDR_reg <= SAD61_DDR;	SAD62_DDR_reg <= SAD62_DDR; SAD63_DDR_reg <= SAD63_DDR;	

	SAD0_VR_reg  <= SAD0_VR;  SAD1_VR_reg  <= SAD1_VR;	SAD2_VR_reg  <= SAD2_VR;  SAD3_VR_reg  <= SAD3_VR;
	SAD4_VR_reg  <= SAD4_VR;  SAD5_VR_reg  <= SAD5_VR;	SAD6_VR_reg  <= SAD6_VR;  SAD7_VR_reg  <= SAD7_VR;
	SAD8_VR_reg  <= SAD8_VR;  SAD9_VR_reg  <= SAD9_VR;	SAD10_VR_reg <= SAD10_VR; SAD11_VR_reg <= SAD11_VR;
	SAD12_VR_reg <= SAD12_VR; SAD13_VR_reg <= SAD13_VR;	SAD14_VR_reg <= SAD14_VR; SAD15_VR_reg <= SAD15_VR;
	SAD16_VR_reg <= SAD16_VR; SAD17_VR_reg <= SAD17_VR;	SAD18_VR_reg <= SAD18_VR; SAD19_VR_reg <= SAD19_VR;
	SAD20_VR_reg <= SAD20_VR; SAD21_VR_reg <= SAD21_VR;	SAD22_VR_reg <= SAD22_VR; SAD23_VR_reg <= SAD23_VR;
	SAD24_VR_reg <= SAD24_VR; SAD25_VR_reg <= SAD25_VR;	SAD26_VR_reg <= SAD26_VR; SAD27_VR_reg <= SAD27_VR;
	SAD28_VR_reg <= SAD28_VR; SAD29_VR_reg <= SAD29_VR;	SAD30_VR_reg <= SAD30_VR; SAD31_VR_reg <= SAD31_VR;
	SAD32_VR_reg <= SAD32_VR; SAD33_VR_reg <= SAD33_VR;	SAD34_VR_reg <= SAD34_VR; SAD35_VR_reg <= SAD35_VR;
	SAD36_VR_reg <= SAD36_VR; SAD37_VR_reg <= SAD37_VR;	SAD38_VR_reg <= SAD38_VR; SAD39_VR_reg <= SAD39_VR;
	SAD40_VR_reg <= SAD40_VR; SAD41_VR_reg <= SAD41_VR;	SAD42_VR_reg <= SAD42_VR; SAD43_VR_reg <= SAD43_VR;
	SAD44_VR_reg <= SAD44_VR; SAD45_VR_reg <= SAD45_VR;	SAD46_VR_reg <= SAD46_VR; SAD47_VR_reg <= SAD47_VR;
	SAD48_VR_reg <= SAD48_VR; SAD49_VR_reg <= SAD49_VR;	SAD50_VR_reg <= SAD50_VR; SAD51_VR_reg <= SAD51_VR;
	SAD52_VR_reg <= SAD52_VR; SAD53_VR_reg <= SAD53_VR;	SAD54_VR_reg <= SAD54_VR; SAD55_VR_reg <= SAD55_VR;
	SAD56_VR_reg <= SAD56_VR; SAD57_VR_reg <= SAD57_VR;	SAD58_VR_reg <= SAD58_VR; SAD59_VR_reg <= SAD59_VR;
	SAD60_VR_reg <= SAD60_VR; SAD61_VR_reg <= SAD61_VR;	SAD62_VR_reg <= SAD62_VR; SAD63_VR_reg <= SAD63_VR;	

	SAD0_HD_reg  <= SAD0_HD;  SAD1_HD_reg  <= SAD1_HD;	SAD2_HD_reg  <= SAD2_HD;  SAD3_HD_reg  <= SAD3_HD;
	SAD4_HD_reg  <= SAD4_HD;  SAD5_HD_reg  <= SAD5_HD;	SAD6_HD_reg  <= SAD6_HD;  SAD7_HD_reg  <= SAD7_HD;
	SAD8_HD_reg  <= SAD8_HD;  SAD9_HD_reg  <= SAD9_HD;	SAD10_HD_reg <= SAD10_HD; SAD11_HD_reg <= SAD11_HD;
	SAD12_HD_reg <= SAD12_HD; SAD13_HD_reg <= SAD13_HD;	SAD14_HD_reg <= SAD14_HD; SAD15_HD_reg <= SAD15_HD;
	SAD16_HD_reg <= SAD16_HD; SAD17_HD_reg <= SAD17_HD;	SAD18_HD_reg <= SAD18_HD; SAD19_HD_reg <= SAD19_HD;
	SAD20_HD_reg <= SAD20_HD; SAD21_HD_reg <= SAD21_HD;	SAD22_HD_reg <= SAD22_HD; SAD23_HD_reg <= SAD23_HD;
	SAD24_HD_reg <= SAD24_HD; SAD25_HD_reg <= SAD25_HD;	SAD26_HD_reg <= SAD26_HD; SAD27_HD_reg <= SAD27_HD;
	SAD28_HD_reg <= SAD28_HD; SAD29_HD_reg <= SAD29_HD;	SAD30_HD_reg <= SAD30_HD; SAD31_HD_reg <= SAD31_HD;
	SAD32_HD_reg <= SAD32_HD; SAD33_HD_reg <= SAD33_HD;	SAD34_HD_reg <= SAD34_HD; SAD35_HD_reg <= SAD35_HD;
	SAD36_HD_reg <= SAD36_HD; SAD37_HD_reg <= SAD37_HD;	SAD38_HD_reg <= SAD38_HD; SAD39_HD_reg <= SAD39_HD;
	SAD40_HD_reg <= SAD40_HD; SAD41_HD_reg <= SAD41_HD;	SAD42_HD_reg <= SAD42_HD; SAD43_HD_reg <= SAD43_HD;
	SAD44_HD_reg <= SAD44_HD; SAD45_HD_reg <= SAD45_HD;	SAD46_HD_reg <= SAD46_HD; SAD47_HD_reg <= SAD47_HD;
	SAD48_HD_reg <= SAD48_HD; SAD49_HD_reg <= SAD49_HD;	SAD50_HD_reg <= SAD50_HD; SAD51_HD_reg <= SAD51_HD;
	SAD52_HD_reg <= SAD52_HD; SAD53_HD_reg <= SAD53_HD;	SAD54_HD_reg <= SAD54_HD; SAD55_HD_reg <= SAD55_HD;
	SAD56_HD_reg <= SAD56_HD; SAD57_HD_reg <= SAD57_HD;	SAD58_HD_reg <= SAD58_HD; SAD59_HD_reg <= SAD59_HD;
	SAD60_HD_reg <= SAD60_HD; SAD61_HD_reg <= SAD61_HD;	SAD62_HD_reg <= SAD62_HD; SAD63_HD_reg <= SAD63_HD;	

	SAD0_VL_reg  <= SAD0_VL;  SAD1_VL_reg  <= SAD1_VL;	SAD2_VL_reg  <= SAD2_VL;  SAD3_VL_reg  <= SAD3_VL;
	SAD4_VL_reg  <= SAD4_VL;  SAD5_VL_reg  <= SAD5_VL;	SAD6_VL_reg  <= SAD6_VL;  SAD7_VL_reg  <= SAD7_VL;
	SAD8_VL_reg  <= SAD8_VL;  SAD9_VL_reg  <= SAD9_VL;	SAD10_VL_reg <= SAD10_VL; SAD11_VL_reg <= SAD11_VL;
	SAD12_VL_reg <= SAD12_VL; SAD13_VL_reg <= SAD13_VL;	SAD14_VL_reg <= SAD14_VL; SAD15_VL_reg <= SAD15_VL;
	SAD16_VL_reg <= SAD16_VL; SAD17_VL_reg <= SAD17_VL;	SAD18_VL_reg <= SAD18_VL; SAD19_VL_reg <= SAD19_VL;
	SAD20_VL_reg <= SAD20_VL; SAD21_VL_reg <= SAD21_VL;	SAD22_VL_reg <= SAD22_VL; SAD23_VL_reg <= SAD23_VL;
	SAD24_VL_reg <= SAD24_VL; SAD25_VL_reg <= SAD25_VL;	SAD26_VL_reg <= SAD26_VL; SAD27_VL_reg <= SAD27_VL;
	SAD28_VL_reg <= SAD28_VL; SAD29_VL_reg <= SAD29_VL;	SAD30_VL_reg <= SAD30_VL; SAD31_VL_reg <= SAD31_VL;
	SAD32_VL_reg <= SAD32_VL; SAD33_VL_reg <= SAD33_VL;	SAD34_VL_reg <= SAD34_VL; SAD35_VL_reg <= SAD35_VL;
	SAD36_VL_reg <= SAD36_VL; SAD37_VL_reg <= SAD37_VL;	SAD38_VL_reg <= SAD38_VL; SAD39_VL_reg <= SAD39_VL;
	SAD40_VL_reg <= SAD40_VL; SAD41_VL_reg <= SAD41_VL;	SAD42_VL_reg <= SAD42_VL; SAD43_VL_reg <= SAD43_VL;
	SAD44_VL_reg <= SAD44_VL; SAD45_VL_reg <= SAD45_VL;	SAD46_VL_reg <= SAD46_VL; SAD47_VL_reg <= SAD47_VL;
	SAD48_VL_reg <= SAD48_VL; SAD49_VL_reg <= SAD49_VL;	SAD50_VL_reg <= SAD50_VL; SAD51_VL_reg <= SAD51_VL;
	SAD52_VL_reg <= SAD52_VL; SAD53_VL_reg <= SAD53_VL;	SAD54_VL_reg <= SAD54_VL; SAD55_VL_reg <= SAD55_VL;
	SAD56_VL_reg <= SAD56_VL; SAD57_VL_reg <= SAD57_VL;	SAD58_VL_reg <= SAD58_VL; SAD59_VL_reg <= SAD59_VL;
	SAD60_VL_reg <= SAD60_VL; SAD61_VL_reg <= SAD61_VL;	SAD62_VL_reg <= SAD62_VL; SAD63_VL_reg <= SAD63_VL;	

	SAD0_HU_reg  <= SAD0_HU;  SAD1_HU_reg  <= SAD1_HU;	SAD2_HU_reg  <= SAD2_HU;  SAD3_HU_reg  <= SAD3_HU;
	SAD4_HU_reg  <= SAD4_HU;  SAD5_HU_reg  <= SAD5_HU;	SAD6_HU_reg  <= SAD6_HU;  SAD7_HU_reg  <= SAD7_HU;
	SAD8_HU_reg  <= SAD8_HU;  SAD9_HU_reg  <= SAD9_HU;	SAD10_HU_reg <= SAD10_HU; SAD11_HU_reg <= SAD11_HU;
	SAD12_HU_reg <= SAD12_HU; SAD13_HU_reg <= SAD13_HU;	SAD14_HU_reg <= SAD14_HU; SAD15_HU_reg <= SAD15_HU;
	SAD16_HU_reg <= SAD16_HU; SAD17_HU_reg <= SAD17_HU;	SAD18_HU_reg <= SAD18_HU; SAD19_HU_reg <= SAD19_HU;
	SAD20_HU_reg <= SAD20_HU; SAD21_HU_reg <= SAD21_HU;	SAD22_HU_reg <= SAD22_HU; SAD23_HU_reg <= SAD23_HU;
	SAD24_HU_reg <= SAD24_HU; SAD25_HU_reg <= SAD25_HU;	SAD26_HU_reg <= SAD26_HU; SAD27_HU_reg <= SAD27_HU;
	SAD28_HU_reg <= SAD28_HU; SAD29_HU_reg <= SAD29_HU;	SAD30_HU_reg <= SAD30_HU; SAD31_HU_reg <= SAD31_HU;
	SAD32_HU_reg <= SAD32_HU; SAD33_HU_reg <= SAD33_HU;	SAD34_HU_reg <= SAD34_HU; SAD35_HU_reg <= SAD35_HU;
	SAD36_HU_reg <= SAD36_HU; SAD37_HU_reg <= SAD37_HU;	SAD38_HU_reg <= SAD38_HU; SAD39_HU_reg <= SAD39_HU;
	SAD40_HU_reg <= SAD40_HU; SAD41_HU_reg <= SAD41_HU;	SAD42_HU_reg <= SAD42_HU; SAD43_HU_reg <= SAD43_HU;
	SAD44_HU_reg <= SAD44_HU; SAD45_HU_reg <= SAD45_HU;	SAD46_HU_reg <= SAD46_HU; SAD47_HU_reg <= SAD47_HU;
	SAD48_HU_reg <= SAD48_HU; SAD49_HU_reg <= SAD49_HU;	SAD50_HU_reg <= SAD50_HU; SAD51_HU_reg <= SAD51_HU;
	SAD52_HU_reg <= SAD52_HU; SAD53_HU_reg <= SAD53_HU;	SAD54_HU_reg <= SAD54_HU; SAD55_HU_reg <= SAD55_HU;
	SAD56_HU_reg <= SAD56_HU; SAD57_HU_reg <= SAD57_HU;	SAD58_HU_reg <= SAD58_HU; SAD59_HU_reg <= SAD59_HU;
	SAD60_HU_reg <= SAD60_HU; SAD61_HU_reg <= SAD61_HU;	SAD62_HU_reg <= SAD62_HU; SAD63_HU_reg <= SAD63_HU;		
	  
  end if;
end process;

-- sum each SAD_Reg to SAD_Quart for each clk cycle. If rdy_in4 = 1, then write out...
-- 16 additions in one clk cycle.

--Corner cases: equals infinity (others 1) when modes are not available due to firstrow/firstcol
SAD_V_quart0 <=  (("0000" &  SAD0_V_reg) + ("0000" &  SAD1_V_reg) + ("0000" &  SAD2_V_reg) +  ("0000" &  SAD3_V_reg) +		  
						("0000" &  SAD4_V_reg) + ("0000" &  SAD5_V_reg) + ("0000" &  SAD6_V_reg) +  ("0000" &  SAD7_V_reg) +
						("0000" &  SAD8_V_reg) + ("0000" &  SAD9_V_reg) + ("0000" & SAD10_V_reg) +  ("0000" & SAD11_V_reg) +
						("0000" & SAD12_V_reg) + ("0000" & SAD13_V_reg) + ("0000" & SAD14_V_reg) +  ("0000" & SAD15_V_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_V_quart1 <=  (("0000" & SAD16_V_reg) + ("0000" & SAD17_V_reg) + ("0000" & SAD18_V_reg) +  ("0000" & SAD19_V_reg) +		  
						("0000" & SAD20_V_reg) + ("0000" & SAD21_V_reg) + ("0000" & SAD22_V_reg) +  ("0000" & SAD23_V_reg) +
						("0000" & SAD24_V_reg) + ("0000" & SAD25_V_reg) + ("0000" & SAD26_V_reg) +  ("0000" & SAD27_V_reg) +
						("0000" & SAD28_V_reg) + ("0000" & SAD29_V_reg) + ("0000" & SAD30_V_reg) +  ("0000" & SAD31_V_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_V_quart2 <=  (("0000" & SAD32_V_reg) + ("0000" & SAD33_V_reg) + ("0000" & SAD34_V_reg) +  ("0000" & SAD35_V_reg) +		  
						("0000" & SAD36_V_reg) + ("0000" & SAD37_V_reg) + ("0000" & SAD38_V_reg) +  ("0000" & SAD39_V_reg) +
						("0000" & SAD40_V_reg) + ("0000" & SAD41_V_reg) + ("0000" & SAD42_V_reg) +  ("0000" & SAD43_V_reg) +
						("0000" & SAD44_V_reg) + ("0000" & SAD45_V_reg) + ("0000" & SAD46_V_reg) +  ("0000" & SAD47_V_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_V_quart3 <=  (("0000" & SAD48_V_reg) + ("0000" & SAD49_V_reg) + ("0000" & SAD50_V_reg) +  ("0000" & SAD51_V_reg) +		  
						("0000" & SAD52_V_reg) + ("0000" & SAD53_V_reg) + ("0000" & SAD54_V_reg) +  ("0000" & SAD55_V_reg) +
						("0000" & SAD56_V_reg) + ("0000" & SAD57_V_reg) + ("0000" & SAD58_V_reg) +  ("0000" & SAD59_V_reg) +
						("0000" & SAD60_V_reg) + ("0000" & SAD61_V_reg) + ("0000" & SAD62_V_reg) +  ("0000" & SAD63_V_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');						

SAD_H_quart0 <=  (("0000" &  SAD0_H_reg) + ("0000" &  SAD1_H_reg) + ("0000" &  SAD2_H_reg) +  ("0000" &  SAD3_H_reg) +		  
						("0000" &  SAD4_H_reg) + ("0000" &  SAD5_H_reg) + ("0000" &  SAD6_H_reg) +  ("0000" &  SAD7_H_reg) +
						("0000" &  SAD8_H_reg) + ("0000" &  SAD9_H_reg) + ("0000" & SAD10_H_reg) +  ("0000" & SAD11_H_reg) +
						("0000" & SAD12_H_reg) + ("0000" & SAD13_H_reg) + ("0000" & SAD14_H_reg) +  ("0000" & SAD15_H_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_H_quart1 <=  (("0000" & SAD16_H_reg) + ("0000" & SAD17_H_reg) + ("0000" & SAD18_H_reg) +  ("0000" & SAD19_H_reg) +		  
						("0000" & SAD20_H_reg) + ("0000" & SAD21_H_reg) + ("0000" & SAD22_H_reg) +  ("0000" & SAD23_H_reg) +
						("0000" & SAD24_H_reg) + ("0000" & SAD25_H_reg) + ("0000" & SAD26_H_reg) +  ("0000" & SAD27_H_reg) +
						("0000" & SAD28_H_reg) + ("0000" & SAD29_H_reg) + ("0000" & SAD30_H_reg) +  ("0000" & SAD31_H_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_H_quart2 <=  (("0000" & SAD32_H_reg) + ("0000" & SAD33_H_reg) + ("0000" & SAD34_H_reg) +  ("0000" & SAD35_H_reg) +		  
						("0000" & SAD36_H_reg) + ("0000" & SAD37_H_reg) + ("0000" & SAD38_H_reg) +  ("0000" & SAD39_H_reg) +
						("0000" & SAD40_H_reg) + ("0000" & SAD41_H_reg) + ("0000" & SAD42_H_reg) +  ("0000" & SAD43_H_reg) +
						("0000" & SAD44_H_reg) + ("0000" & SAD45_H_reg) + ("0000" & SAD46_H_reg) +  ("0000" & SAD47_H_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_H_quart3 <=  (("0000" & SAD48_H_reg) + ("0000" & SAD49_H_reg) + ("0000" & SAD50_H_reg) +  ("0000" & SAD51_H_reg) +		  
						("0000" & SAD52_H_reg) + ("0000" & SAD53_H_reg) + ("0000" & SAD54_H_reg) +  ("0000" & SAD55_H_reg) +
						("0000" & SAD56_H_reg) + ("0000" & SAD57_H_reg) + ("0000" & SAD58_H_reg) +  ("0000" & SAD59_H_reg) +
						("0000" & SAD60_H_reg) + ("0000" & SAD61_H_reg) + ("0000" & SAD62_H_reg) +  ("0000" & SAD63_H_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');
				
				
SAD_DC_quart0 <= (("0000" &  SAD0_DC_reg) + ("0000" &  SAD1_DC_reg) + ("0000" &  SAD2_DC_reg) +  ("0000" &  SAD3_DC_reg) +		  
						("0000" &  SAD4_DC_reg) + ("0000" &  SAD5_DC_reg) + ("0000" &  SAD6_DC_reg) +  ("0000" &  SAD7_DC_reg) +
						("0000" &  SAD8_DC_reg) + ("0000" &  SAD9_DC_reg) + ("0000" & SAD10_DC_reg) +  ("0000" & SAD11_DC_reg) +
						("0000" & SAD12_DC_reg) + ("0000" & SAD13_DC_reg) + ("0000" & SAD14_DC_reg) +  ("0000" & SAD15_DC_reg))
						when (rdy_in_4='1' ) else (others=>'0');

SAD_DC_quart1 <= (("0000" & SAD16_DC_reg) + ("0000" & SAD17_DC_reg) + ("0000" & SAD18_DC_reg) +  ("0000" & SAD19_DC_reg) +		  
						("0000" & SAD20_DC_reg) + ("0000" & SAD21_DC_reg) + ("0000" & SAD22_DC_reg) +  ("0000" & SAD23_DC_reg) +
						("0000" & SAD24_DC_reg) + ("0000" & SAD25_DC_reg) + ("0000" & SAD26_DC_reg) +  ("0000" & SAD27_DC_reg) +
						("0000" & SAD28_DC_reg) + ("0000" & SAD29_DC_reg) + ("0000" & SAD30_DC_reg) +  ("0000" & SAD31_DC_reg))
						when (rdy_in_4='1' ) else (others=>'0');

SAD_DC_quart2 <= (("0000" & SAD32_DC_reg) + ("0000" & SAD33_DC_reg) + ("0000" & SAD34_DC_reg) +  ("0000" & SAD35_DC_reg) +		  
						("0000" & SAD36_DC_reg) + ("0000" & SAD37_DC_reg) + ("0000" & SAD38_DC_reg) +  ("0000" & SAD39_DC_reg) +
						("0000" & SAD40_DC_reg) + ("0000" & SAD41_DC_reg) + ("0000" & SAD42_DC_reg) +  ("0000" & SAD43_DC_reg) +
						("0000" & SAD44_DC_reg) + ("0000" & SAD45_DC_reg) + ("0000" & SAD46_DC_reg) +  ("0000" & SAD47_DC_reg))
						when (rdy_in_4='1' ) else (others=>'0');

SAD_DC_quart3 <= (("0000" & SAD48_DC_reg) + ("0000" & SAD49_DC_reg) + ("0000" & SAD50_DC_reg) +  ("0000" & SAD51_DC_reg) +		  
						("0000" & SAD52_DC_reg) + ("0000" & SAD53_DC_reg) + ("0000" & SAD54_DC_reg) +  ("0000" & SAD55_DC_reg) +
						("0000" & SAD56_DC_reg) + ("0000" & SAD57_DC_reg) + ("0000" & SAD58_DC_reg) +  ("0000" & SAD59_DC_reg) +
						("0000" & SAD60_DC_reg) + ("0000" & SAD61_DC_reg) + ("0000" & SAD62_DC_reg) +  ("0000" & SAD63_DC_reg))
						when (rdy_in_4='1' ) else (others=>'0');				


SAD_DDL_quart0 <=(("0000" &  SAD0_DDL_reg) + ("0000" &  SAD1_DDL_reg) + ("0000" &  SAD2_DDL_reg) +  ("0000" &  SAD3_DDL_reg) +		  
						("0000" &  SAD4_DDL_reg) + ("0000" &  SAD5_DDL_reg) + ("0000" &  SAD6_DDL_reg) +  ("0000" &  SAD7_DDL_reg) +
						("0000" &  SAD8_DDL_reg) + ("0000" &  SAD9_DDL_reg) + ("0000" & SAD10_DDL_reg) +  ("0000" & SAD11_DDL_reg) +
						("0000" & SAD12_DDL_reg) + ("0000" & SAD13_DDL_reg) + ("0000" & SAD14_DDL_reg) +  ("0000" & SAD15_DDL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_DDL_quart1 <=(("0000" & SAD16_DDL_reg) + ("0000" & SAD17_DDL_reg) + ("0000" & SAD18_DDL_reg) +  ("0000" & SAD19_DDL_reg) +		  
						("0000" & SAD20_DDL_reg) + ("0000" & SAD21_DDL_reg) + ("0000" & SAD22_DDL_reg) +  ("0000" & SAD23_DDL_reg) +
						("0000" & SAD24_DDL_reg) + ("0000" & SAD25_DDL_reg) + ("0000" & SAD26_DDL_reg) +  ("0000" & SAD27_DDL_reg) +
						("0000" & SAD28_DDL_reg) + ("0000" & SAD29_DDL_reg) + ("0000" & SAD30_DDL_reg) +  ("0000" & SAD31_DDL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_DDL_quart2 <=(("0000" & SAD32_DDL_reg) + ("0000" & SAD33_DDL_reg) + ("0000" & SAD34_DDL_reg) +  ("0000" & SAD35_DDL_reg) +		  
						("0000" & SAD36_DDL_reg) + ("0000" & SAD37_DDL_reg) + ("0000" & SAD38_DDL_reg) +  ("0000" & SAD39_DDL_reg) +
						("0000" & SAD40_DDL_reg) + ("0000" & SAD41_DDL_reg) + ("0000" & SAD42_DDL_reg) +  ("0000" & SAD43_DDL_reg) +
						("0000" & SAD44_DDL_reg) + ("0000" & SAD45_DDL_reg) + ("0000" & SAD46_DDL_reg) +  ("0000" & SAD47_DDL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_DDL_quart3 <=(("0000" & SAD48_DDL_reg) + ("0000" & SAD49_DDL_reg) + ("0000" & SAD50_DDL_reg) +  ("0000" & SAD51_DDL_reg) +		  
						("0000" & SAD52_DDL_reg) + ("0000" & SAD53_DDL_reg) + ("0000" & SAD54_DDL_reg) +  ("0000" & SAD55_DDL_reg) +
						("0000" & SAD56_DDL_reg) + ("0000" & SAD57_DDL_reg) + ("0000" & SAD58_DDL_reg) +  ("0000" & SAD59_DDL_reg) +
						("0000" & SAD60_DDL_reg) + ("0000" & SAD61_DDL_reg) + ("0000" & SAD62_DDL_reg) +  ("0000" & SAD63_DDL_reg))				
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');


SAD_DDR_quart0 <=(("0000" &  SAD0_DDR_reg) + ("0000" &  SAD1_DDR_reg) + ("0000" &  SAD2_DDR_reg) +  ("0000" &  SAD3_DDR_reg) +		  
						("0000" &  SAD4_DDR_reg) + ("0000" &  SAD5_DDR_reg) + ("0000" &  SAD6_DDR_reg) +  ("0000" &  SAD7_DDR_reg) +
						("0000" &  SAD8_DDR_reg) + ("0000" &  SAD9_DDR_reg) + ("0000" & SAD10_DDR_reg) +  ("0000" & SAD11_DDR_reg) +
						("0000" & SAD12_DDR_reg) + ("0000" & SAD13_DDR_reg) + ("0000" & SAD14_DDR_reg) +  ("0000" & SAD15_DDR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_DDR_quart1 <=(("0000" & SAD16_DDR_reg) + ("0000" & SAD17_DDR_reg) + ("0000" & SAD18_DDR_reg) +  ("0000" & SAD19_DDR_reg) +		  
						("0000" & SAD20_DDR_reg) + ("0000" & SAD21_DDR_reg) + ("0000" & SAD22_DDR_reg) +  ("0000" & SAD23_DDR_reg) +
						("0000" & SAD24_DDR_reg) + ("0000" & SAD25_DDR_reg) + ("0000" & SAD26_DDR_reg) +  ("0000" & SAD27_DDR_reg) +
						("0000" & SAD28_DDR_reg) + ("0000" & SAD29_DDR_reg) + ("0000" & SAD30_DDR_reg) +  ("0000" & SAD31_DDR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_DDR_quart2 <=(("0000" & SAD32_DDR_reg) + ("0000" & SAD33_DDR_reg) + ("0000" & SAD34_DDR_reg) +  ("0000" & SAD35_DDR_reg) +		  
						("0000" & SAD36_DDR_reg) + ("0000" & SAD37_DDR_reg) + ("0000" & SAD38_DDR_reg) +  ("0000" & SAD39_DDR_reg) +
						("0000" & SAD40_DDR_reg) + ("0000" & SAD41_DDR_reg) + ("0000" & SAD42_DDR_reg) +  ("0000" & SAD43_DDR_reg) +
						("0000" & SAD44_DDR_reg) + ("0000" & SAD45_DDR_reg) + ("0000" & SAD46_DDR_reg) +  ("0000" & SAD47_DDR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_DDR_quart3 <=(("0000" & SAD48_DDR_reg) + ("0000" & SAD49_DDR_reg) + ("0000" & SAD50_DDR_reg) +  ("0000" & SAD51_DDR_reg) +		  
						("0000" & SAD52_DDR_reg) + ("0000" & SAD53_DDR_reg) + ("0000" & SAD54_DDR_reg) +  ("0000" & SAD55_DDR_reg) +
						("0000" & SAD56_DDR_reg) + ("0000" & SAD57_DDR_reg) + ("0000" & SAD58_DDR_reg) +  ("0000" & SAD59_DDR_reg) +
						("0000" & SAD60_DDR_reg) + ("0000" & SAD61_DDR_reg) + ("0000" & SAD62_DDR_reg) +  ("0000" & SAD63_DDR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');
						
SAD_VR_quart0 <= (("0000" &  SAD0_VR_reg) + ("0000" &  SAD1_VR_reg) + ("0000" &  SAD2_VR_reg) +  ("0000" &  SAD3_VR_reg) +		  
						("0000" &  SAD4_VR_reg) + ("0000" &  SAD5_VR_reg) + ("0000" &  SAD6_VR_reg) +  ("0000" &  SAD7_VR_reg) +
						("0000" &  SAD8_VR_reg) + ("0000" &  SAD9_VR_reg) + ("0000" & SAD10_VR_reg) +  ("0000" & SAD11_VR_reg) +
						("0000" & SAD12_VR_reg) + ("0000" & SAD13_VR_reg) + ("0000" & SAD14_VR_reg) +  ("0000" & SAD15_VR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_VR_quart1 <= (("0000" & SAD16_VR_reg) + ("0000" & SAD17_VR_reg) + ("0000" & SAD18_VR_reg) +  ("0000" & SAD19_VR_reg) +		  
						("0000" & SAD20_VR_reg) + ("0000" & SAD21_VR_reg) + ("0000" & SAD22_VR_reg) +  ("0000" & SAD23_VR_reg) +
						("0000" & SAD24_VR_reg) + ("0000" & SAD25_VR_reg) + ("0000" & SAD26_VR_reg) +  ("0000" & SAD27_VR_reg) +
						("0000" & SAD28_VR_reg) + ("0000" & SAD29_VR_reg) + ("0000" & SAD30_VR_reg) +  ("0000" & SAD31_VR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_VR_quart2 <= (("0000" & SAD32_VR_reg) + ("0000" & SAD33_VR_reg) + ("0000" & SAD34_VR_reg) +  ("0000" & SAD35_VR_reg) +		  
						("0000" & SAD36_VR_reg) + ("0000" & SAD37_VR_reg) + ("0000" & SAD38_VR_reg) +  ("0000" & SAD39_VR_reg) +
						("0000" & SAD40_VR_reg) + ("0000" & SAD41_VR_reg) + ("0000" & SAD42_VR_reg) +  ("0000" & SAD43_VR_reg) +
						("0000" & SAD44_VR_reg) + ("0000" & SAD45_VR_reg) + ("0000" & SAD46_VR_reg) +  ("0000" & SAD47_VR_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_VR_quart3 <= (("0000" & SAD48_VR_reg) + ("0000" & SAD49_VR_reg) + ("0000" & SAD50_VR_reg) +  ("0000" & SAD51_VR_reg) +		  
						("0000" & SAD52_VR_reg) + ("0000" & SAD53_VR_reg) + ("0000" & SAD54_VR_reg) +  ("0000" & SAD55_VR_reg) +
						("0000" & SAD56_VR_reg) + ("0000" & SAD57_VR_reg) + ("0000" & SAD58_VR_reg) +  ("0000" & SAD59_VR_reg) +
						("0000" & SAD60_VR_reg) + ("0000" & SAD61_VR_reg) + ("0000" & SAD62_VR_reg) +  ("0000" & SAD63_VR_reg))						
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1') else (others=>'0');

SAD_HD_quart0 <= (("0000" &  SAD0_HD_reg) + ("0000" &  SAD1_HD_reg) + ("0000" &  SAD2_HD_reg) +  ("0000" &  SAD3_HD_reg) +		  
						("0000" &  SAD4_HD_reg) + ("0000" &  SAD5_HD_reg) + ("0000" &  SAD6_HD_reg) +  ("0000" &  SAD7_HD_reg) +
						("0000" &  SAD8_HD_reg) + ("0000" &  SAD9_HD_reg) + ("0000" & SAD10_HD_reg) +  ("0000" & SAD11_HD_reg) +
						("0000" & SAD12_HD_reg) + ("0000" & SAD13_HD_reg) + ("0000" & SAD14_HD_reg) +  ("0000" & SAD15_HD_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1' ) else (others=>'0');

SAD_HD_quart1 <= (("0000" & SAD16_HD_reg) + ("0000" & SAD17_HD_reg) + ("0000" & SAD18_HD_reg) +  ("0000" & SAD19_HD_reg) +		  
						("0000" & SAD20_HD_reg) + ("0000" & SAD21_HD_reg) + ("0000" & SAD22_HD_reg) +  ("0000" & SAD23_HD_reg) +
						("0000" & SAD24_HD_reg) + ("0000" & SAD25_HD_reg) + ("0000" & SAD26_HD_reg) +  ("0000" & SAD27_HD_reg) +
						("0000" & SAD28_HD_reg) + ("0000" & SAD29_HD_reg) + ("0000" & SAD30_HD_reg) +  ("0000" & SAD31_HD_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1' ) else (others=>'0');

SAD_HD_quart2 <= (("0000" & SAD32_HD_reg) + ("0000" & SAD33_HD_reg) + ("0000" & SAD34_HD_reg) +  ("0000" & SAD35_HD_reg) +		  
						("0000" & SAD36_HD_reg) + ("0000" & SAD37_HD_reg) + ("0000" & SAD38_HD_reg) +  ("0000" & SAD39_HD_reg) +
						("0000" & SAD40_HD_reg) + ("0000" & SAD41_HD_reg) + ("0000" & SAD42_HD_reg) +  ("0000" & SAD43_HD_reg) +
						("0000" & SAD44_HD_reg) + ("0000" & SAD45_HD_reg) + ("0000" & SAD46_HD_reg) +  ("0000" & SAD47_HD_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1' ) else (others=>'0');

SAD_HD_quart3 <= (("0000" & SAD48_HD_reg) + ("0000" & SAD49_HD_reg) + ("0000" & SAD50_HD_reg) +  ("0000" & SAD51_HD_reg) +		  
						("0000" & SAD52_HD_reg) + ("0000" & SAD53_HD_reg) + ("0000" & SAD54_HD_reg) +  ("0000" & SAD55_HD_reg) +
						("0000" & SAD56_HD_reg) + ("0000" & SAD57_HD_reg) + ("0000" & SAD58_HD_reg) +  ("0000" & SAD59_HD_reg) +
						("0000" & SAD60_HD_reg) + ("0000" & SAD61_HD_reg) + ("0000" & SAD62_HD_reg) +  ("0000" & SAD63_HD_reg))
						when (rdy_in_4='1' AND (first_col_reg2 = '0' and first_row_reg2='0')) else  (others=>'1') when (rdy_in_4='1' ) else (others=>'0');

SAD_VL_quart0 <= (("0000" &  SAD0_VL_reg) + ("0000" &  SAD1_VL_reg) + ("0000" &  SAD2_VL_reg) +  ("0000" &  SAD3_VL_reg) +		  
						("0000" &  SAD4_VL_reg) + ("0000" &  SAD5_VL_reg) + ("0000" &  SAD6_VL_reg) +  ("0000" &  SAD7_VL_reg) +
						("0000" &  SAD8_VL_reg) + ("0000" &  SAD9_VL_reg) + ("0000" & SAD10_VL_reg) +  ("0000" & SAD11_VL_reg) +
						("0000" & SAD12_VL_reg) + ("0000" & SAD13_VL_reg) + ("0000" & SAD14_VL_reg) +  ("0000" & SAD15_VL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_VL_quart1 <= (("0000" & SAD16_VL_reg) + ("0000" & SAD17_VL_reg) + ("0000" & SAD18_VL_reg) +  ("0000" & SAD19_VL_reg) +		  
						("0000" & SAD20_VL_reg) + ("0000" & SAD21_VL_reg) + ("0000" & SAD22_VL_reg) +  ("0000" & SAD23_VL_reg) +
						("0000" & SAD24_VL_reg) + ("0000" & SAD25_VL_reg) + ("0000" & SAD26_VL_reg) +  ("0000" & SAD27_VL_reg) +
						("0000" & SAD28_VL_reg) + ("0000" & SAD29_VL_reg) + ("0000" & SAD30_VL_reg) +  ("0000" & SAD31_VL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_VL_quart2 <= (("0000" & SAD32_VL_reg) + ("0000" & SAD33_VL_reg) + ("0000" & SAD34_VL_reg) +  ("0000" & SAD35_VL_reg) +		  
						("0000" & SAD36_VL_reg) + ("0000" & SAD37_VL_reg) + ("0000" & SAD38_VL_reg) +  ("0000" & SAD39_VL_reg) +
						("0000" & SAD40_VL_reg) + ("0000" & SAD41_VL_reg) + ("0000" & SAD42_VL_reg) +  ("0000" & SAD43_VL_reg) +
						("0000" & SAD44_VL_reg) + ("0000" & SAD45_VL_reg) + ("0000" & SAD46_VL_reg) +  ("0000" & SAD47_VL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_VL_quart3 <= (("0000" & SAD48_VL_reg) + ("0000" & SAD49_VL_reg) + ("0000" & SAD50_VL_reg) +  ("0000" & SAD51_VL_reg) +		  
						("0000" & SAD52_VL_reg) + ("0000" & SAD53_VL_reg) + ("0000" & SAD54_VL_reg) +  ("0000" & SAD55_VL_reg) +
						("0000" & SAD56_VL_reg) + ("0000" & SAD57_VL_reg) + ("0000" & SAD58_VL_reg) +  ("0000" & SAD59_VL_reg) +
						("0000" & SAD60_VL_reg) + ("0000" & SAD61_VL_reg) + ("0000" & SAD62_VL_reg) +  ("0000" & SAD63_VL_reg))
						when (rdy_in_4='1' AND first_row_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_row_reg2 = '1') else (others=>'0');

SAD_HU_quart0 <= (("0000" &  SAD0_HU_reg) + ("0000" &  SAD1_HU_reg) + ("0000" &  SAD2_HU_reg) +  ("0000" &  SAD3_HU_reg) +		  
						("0000" &  SAD4_HU_reg) + ("0000" &  SAD5_HU_reg) + ("0000" &  SAD6_HU_reg) +  ("0000" &  SAD7_HU_reg) +
						("0000" &  SAD8_HU_reg) + ("0000" &  SAD9_HU_reg) + ("0000" & SAD10_HU_reg) +  ("0000" & SAD11_HU_reg) +
						("0000" & SAD12_HU_reg) + ("0000" & SAD13_HU_reg) + ("0000" & SAD14_HU_reg) +  ("0000" & SAD15_HU_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_HU_quart1 <= (("0000" & SAD16_HU_reg) + ("0000" & SAD17_HU_reg) + ("0000" & SAD18_HU_reg) +  ("0000" & SAD19_HU_reg) +		  
						("0000" & SAD20_HU_reg) + ("0000" & SAD21_HU_reg) + ("0000" & SAD22_HU_reg) +  ("0000" & SAD23_HU_reg) +
						("0000" & SAD24_HU_reg) + ("0000" & SAD25_HU_reg) + ("0000" & SAD26_HU_reg) +  ("0000" & SAD27_HU_reg) +
						("0000" & SAD28_HU_reg) + ("0000" & SAD29_HU_reg) + ("0000" & SAD30_HU_reg) +  ("0000" & SAD31_HU_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_HU_quart2 <= (("0000" & SAD32_HU_reg) + ("0000" & SAD33_HU_reg) + ("0000" & SAD34_HU_reg) +  ("0000" & SAD35_HU_reg) +		  
						("0000" & SAD36_HU_reg) + ("0000" & SAD37_HU_reg) + ("0000" & SAD38_HU_reg) +  ("0000" & SAD39_HU_reg) +
						("0000" & SAD40_HU_reg) + ("0000" & SAD41_HU_reg) + ("0000" & SAD42_HU_reg) +  ("0000" & SAD43_HU_reg) +
						("0000" & SAD44_HU_reg) + ("0000" & SAD45_HU_reg) + ("0000" & SAD46_HU_reg) +  ("0000" & SAD47_HU_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');

SAD_HU_quart3 <= (("0000" & SAD48_HU_reg) + ("0000" & SAD49_HU_reg) + ("0000" & SAD50_HU_reg) +  ("0000" & SAD51_HU_reg) +		  
						("0000" & SAD52_HU_reg) + ("0000" & SAD53_HU_reg) + ("0000" & SAD54_HU_reg) +  ("0000" & SAD55_HU_reg) +
						("0000" & SAD56_HU_reg) + ("0000" & SAD57_HU_reg) + ("0000" & SAD58_HU_reg) +  ("0000" & SAD59_HU_reg) +
						("0000" & SAD60_HU_reg) + ("0000" & SAD61_HU_reg) + ("0000" & SAD62_HU_reg) +  ("0000" & SAD63_HU_reg))
						when (rdy_in_4='1' AND first_col_reg2 = '0') else  (others=>'1') when (rdy_in_4='1' AND first_col_reg2 = '1') else (others=>'0');
					

-- Store SAD_quart values in registers						
PROCESS (clk, rst)
BEGIN
  IF (rst = '1') THEN
	SAD_V_quart_reg0   <= (others=>'0'); SAD_V_quart_reg1   <= (others=>'0'); SAD_V_quart_reg2   <= (others=>'0'); SAD_V_quart_reg3   <= (others=>'0');  
	SAD_H_quart_reg0   <= (others=>'0'); SAD_H_quart_reg1   <= (others=>'0'); SAD_H_quart_reg2   <= (others=>'0'); SAD_H_quart_reg3   <= (others=>'0');    		 
	SAD_DC_quart_reg0  <= (others=>'0'); SAD_DC_quart_reg1  <= (others=>'0'); SAD_DC_quart_reg2  <= (others=>'0'); SAD_DC_quart_reg3  <= (others=>'0');  
	SAD_DDL_quart_reg0 <= (others=>'0'); SAD_DDL_quart_reg1 <= (others=>'0'); SAD_DDL_quart_reg2 <= (others=>'0'); SAD_DDL_quart_reg3 <= (others=>'0'); 
	SAD_DDR_quart_reg0 <= (others=>'0'); SAD_DDR_quart_reg1 <= (others=>'0'); SAD_DDR_quart_reg2 <= (others=>'0'); SAD_DDR_quart_reg3 <= (others=>'0'); 
	SAD_VR_quart_reg0  <= (others=>'0'); SAD_VR_quart_reg1  <= (others=>'0'); SAD_VR_quart_reg2  <= (others=>'0'); SAD_VR_quart_reg3  <= (others=>'0');  
	SAD_HD_quart_reg0  <= (others=>'0'); SAD_HD_quart_reg1  <= (others=>'0'); SAD_HD_quart_reg2  <= (others=>'0'); SAD_HD_quart_reg3  <= (others=>'0');
	SAD_VL_quart_reg0  <= (others=>'0'); SAD_VL_quart_reg1  <= (others=>'0'); SAD_VL_quart_reg2  <= (others=>'0'); SAD_VL_quart_reg3  <= (others=>'0'); 
	SAD_HU_quart_reg0  <= (others=>'0'); SAD_HU_quart_reg1  <= (others=>'0'); SAD_HU_quart_reg2  <= (others=>'0'); SAD_HU_quart_reg3  <= (others=>'0');

	
  ELSIF (clk'EVENT AND clk = '1') THEN	
	  
	 SAD_V_quart_reg0   <= SAD_V_quart0;   SAD_V_quart_reg1   <= SAD_V_quart1;   SAD_V_quart_reg2   <=	SAD_V_quart2;   SAD_V_quart_reg3   <= SAD_V_quart3;
    SAD_H_quart_reg0   <= SAD_H_quart0;   SAD_H_quart_reg1   <= SAD_H_quart1;   SAD_H_quart_reg2   <=	SAD_H_quart2;   SAD_H_quart_reg3   <= SAD_H_quart3;
	 SAD_DC_quart_reg0  <= SAD_DC_quart0;  SAD_DC_quart_reg1  <= SAD_DC_quart1;  SAD_DC_quart_reg2  <=	SAD_DC_quart2;  SAD_DC_quart_reg3  <= SAD_DC_quart3;
	 SAD_DDL_quart_reg0 <= SAD_DDL_quart0; SAD_DDL_quart_reg1 <= SAD_DDL_quart1; SAD_DDL_quart_reg2 <=	SAD_DDL_quart2; SAD_DDL_quart_reg3 <= SAD_DDL_quart3;
	 SAD_DDR_quart_reg0 <= SAD_DDR_quart0; SAD_DDR_quart_reg1 <= SAD_DDR_quart1; SAD_DDR_quart_reg2 <=	SAD_DDR_quart2; SAD_DDR_quart_reg3 <= SAD_DDR_quart3;
	 SAD_VR_quart_reg0  <= SAD_VR_quart0;  SAD_VR_quart_reg1  <= SAD_VR_quart1;  SAD_VR_quart_reg2  <=	SAD_VR_quart2;  SAD_VR_quart_reg3  <= SAD_VR_quart3;
	 SAD_HD_quart_reg0  <= SAD_HD_quart0;  SAD_HD_quart_reg1  <= SAD_HD_quart1;  SAD_HD_quart_reg2  <=	SAD_HD_quart2;  SAD_HD_quart_reg3  <= SAD_HD_quart3; 
    SAD_VL_quart_reg0  <= SAD_VL_quart0;  SAD_VL_quart_reg1  <= SAD_VL_quart1;  SAD_VL_quart_reg2  <= SAD_VL_quart2;  SAD_VL_quart_reg3  <= SAD_VL_quart3;
	 SAD_HU_quart_reg0  <= SAD_HU_quart0;  SAD_HU_quart_reg1  <= SAD_HU_quart1;  SAD_HU_quart_reg2  <= SAD_HU_quart2;  SAD_HU_quart_reg3  <= SAD_HU_quart3;
	 
  END IF;
END PROCESS;

--Assign final sums of absolute differences (4 additions/clk cycle)
SAD_V <=   ("00" & SAD_V_quart_reg0)  + ("00" & SAD_V_quart_reg1)  + ("00" & SAD_V_quart_reg2)  + ("00" & SAD_V_quart_reg3);
SAD_H <=   ("00" & SAD_H_quart_reg0)  + ("00" & SAD_H_quart_reg1)  + ("00" & SAD_H_quart_reg2)  + ("00" & SAD_H_quart_reg3);
SAD_DC <=  ("00" & SAD_DC_quart_reg0) + ("00" & SAD_DC_quart_reg1) + ("00" & SAD_DC_quart_reg2) + ("00" & SAD_DC_quart_reg3);
SAD_DDL <= ("00" & SAD_DDL_quart_reg0)+ ("00" & SAD_DDL_quart_reg1)+ ("00" & SAD_DDL_quart_reg2)+ ("00" & SAD_DDL_quart_reg3);
SAD_DDR <= ("00" & SAD_DDR_quart_reg0)+ ("00" & SAD_DDR_quart_reg1)+ ("00" & SAD_DDR_quart_reg2)+ ("00" & SAD_DDR_quart_reg3);
SAD_VR <=  ("00" & SAD_VR_quart_reg0) + ("00" & SAD_VR_quart_reg1) + ("00" & SAD_VR_quart_reg2) + ("00" & SAD_VR_quart_reg3);
SAD_HD <=  ("00" & SAD_HD_quart_reg0) + ("00" & SAD_HD_quart_reg1) + ("00" & SAD_HD_quart_reg2) + ("00" & SAD_HD_quart_reg3);
SAD_VL <=  ("00" & SAD_VL_quart_reg0) + ("00" & SAD_VL_quart_reg1) + ("00" & SAD_VL_quart_reg2) + ("00" & SAD_VL_quart_reg3);
SAD_HU <=  ("00" & SAD_HU_quart_reg0) + ("00" & SAD_HU_quart_reg1) + ("00" & SAD_HU_quart_reg2) + ("00" & SAD_HU_quart_reg3);
			 
PROCESS (clk, rst)
BEGIN
  IF (rst = '1') THEN
	SAD_V_reg    <= (others=>'0');   
	SAD_H_reg    <= (others=>'0');   		 
	SAD_DC_reg   <= (others=>'0');   
	SAD_DDL_reg  <= (others=>'0');  
	SAD_DDR_reg  <= (others=>'0');  
	SAD_VR_reg   <= (others=>'0');   
	SAD_HD_reg   <= (others=>'0'); 
	SAD_VL_reg   <= (others=>'0');  
	SAD_HU_reg   <= (others=>'0'); 

	
  ELSIF (clk'EVENT AND clk = '1') THEN	
	  
	 SAD_V_reg    <=	SAD_V;
    SAD_H_reg    <=	SAD_H;
	 SAD_DC_reg   <=	SAD_DC;
	 SAD_DDL_reg  <=	SAD_DDL;
	 SAD_DDR_reg  <=	SAD_DDR;
	 SAD_VR_reg   <=	SAD_VR;
	 SAD_HD_reg   <=	SAD_HD;
    SAD_VL_reg   <=  SAD_VL;
	 SAD_HU_reg   <=  SAD_HU;
	 
  END IF;
END PROCESS;						

--compute MINIMUM prediction mode
--this part is a copy from Milica Orlandic's 4x4 code
PROCESS (clk, rst)
BEGIN
  IF (rst = '1') THEN    
	 COMP4 <= (others=>'0'); 	
  ELSIF (clk'EVENT AND clk = '1') THEN	
	 if (COMP0 <= COMP1) then 
		COMP4 <= COMP0;     
	 else
		COMP4 <= COMP1;
	 end if;
  END IF;
END PROCESS; 

PROCESS (clk, rst)
BEGIN
  IF (rst = '1') THEN    
	 COMP5 <= (others=>'0'); 
  ELSIF (clk'EVENT AND clk = '1') THEN	
	 if  (COMP2 <= COMP3)  then 
	   COMP5 <= COMP2;     
	 else
	   COMP5 <= COMP3 ;
	 end if;
  END IF;
END PROCESS; 

COMP0   <= SAD_V   when (SAD_V   <= SAD_H)      else SAD_H;
COMP1   <= SAD_DDL when (SAD_DDL <= SAD_DDR)    else SAD_DDR;
COMP2   <= SAD_VR  when (SAD_VR  <= SAD_VL)     else SAD_VL;
COMP3   <= SAD_HD  when (SAD_HD  <= SAD_HU)     else SAD_HU;
COMP6   <= COMP4   when (COMP4   <= COMP5)      else COMP5;
MINIMUM <= COMP6   when (COMP6   <= SAD_DC_reg) else SAD_DC_reg;	
--compute MINIMUM end

-- PREDICTION OUTPUT
process (clk, rst)
begin
	if (rst = '1') then 
		V_reg		<= (others=> (others=>'0'));
		H_reg		<= (others=> (others=>'0'));
		DC_reg	<= (others=>'0');
		DDL_reg	<= (others=> (others=>'0'));
		DDR_reg	<= (others=> (others=>'0'));
		VL_reg	<= (others=> (others=>'0'));
		HD_reg	<= (others=> (others=>'0'));
		VR_reg	<= (others=> (others=>'0'));
		HU_reg	<= (others=> (others=>'0'));
	elsif (clk'event and clk = '1') then
		if (rdy_in_3 = '1') then
			V_reg(0)  <= V_0;  V_reg(1)  <= V_1;  V_reg(2)  <= V_2;  V_reg(3) <= V_3; 
			V_reg(4)  <= V_4;  V_reg(5)  <= V_5;  V_reg(6)  <= V_6;  V_reg(7) <= V_7; 
			V_reg(8)  <= V_8;  V_reg(9)  <= V_9;  V_reg(10) <= V_10; V_reg(11) <= V_11; 
			V_reg(12) <= V_12; V_reg(13) <= V_13; V_reg(14) <= V_14; V_reg(15) <= V_15; 
			V_reg(16) <= V_16; V_reg(17) <= V_17; V_reg(18) <= V_18; V_reg(19) <= V_19; 
			V_reg(20) <= V_20; V_reg(21) <= V_21; V_reg(22) <= V_22; V_reg(23) <= V_23; 
			V_reg(24) <= V_24; V_reg(25) <= V_25; V_reg(26) <= V_26; V_reg(27) <= V_27; 
			V_reg(28) <= V_28; V_reg(29) <= V_29; V_reg(30) <= V_30; V_reg(31) <= V_31; 
			V_reg(32) <= V_32; V_reg(33) <= V_33; V_reg(34) <= V_34; V_reg(35) <= V_35; 
			V_reg(36) <= V_36; V_reg(37) <= V_37; V_reg(38) <= V_38; V_reg(39) <= V_39; 
			V_reg(40) <= V_40; V_reg(41) <= V_41; V_reg(42) <= V_42; V_reg(43) <= V_43; 
			V_reg(44) <= V_44; V_reg(45) <= V_45; V_reg(46) <= V_46; V_reg(47) <= V_47; 
			V_reg(48) <= V_48; V_reg(49) <= V_49; V_reg(50) <= V_50; V_reg(51) <= V_51; 
			V_reg(52) <= V_52; V_reg(53) <= V_53; V_reg(54) <= V_54; V_reg(55) <= V_55; 
			V_reg(56) <= V_56; V_reg(57) <= V_57; V_reg(58) <= V_58; V_reg(59) <= V_59; 
			V_reg(60) <= V_60; V_reg(61) <= V_61; V_reg(62) <= V_62; V_reg(63) <= V_63; 

			H_reg(0)  <= H_0;  H_reg(1)  <= H_1;  H_reg(2)  <= H_2;  H_reg(3) <= H_3; 
			H_reg(4)  <= H_4;  H_reg(5)  <= H_5;  H_reg(6)  <= H_6;  H_reg(7) <= H_7; 
			H_reg(8)  <= H_8;  H_reg(9)  <= H_9;  H_reg(10) <= H_10; H_reg(11) <= H_11; 
			H_reg(12) <= H_12; H_reg(13) <= H_13; H_reg(14) <= H_14; H_reg(15) <= H_15; 
			H_reg(16) <= H_16; H_reg(17) <= H_17; H_reg(18) <= H_18; H_reg(19) <= H_19; 
			H_reg(20) <= H_20; H_reg(21) <= H_21; H_reg(22) <= H_22; H_reg(23) <= H_23; 
			H_reg(24) <= H_24; H_reg(25) <= H_25; H_reg(26) <= H_26; H_reg(27) <= H_27; 
			H_reg(28) <= H_28; H_reg(29) <= H_29; H_reg(30) <= H_30; H_reg(31) <= H_31; 
			H_reg(32) <= H_32; H_reg(33) <= H_33; H_reg(34) <= H_34; H_reg(35) <= H_35; 
			H_reg(36) <= H_36; H_reg(37) <= H_37; H_reg(38) <= H_38; H_reg(39) <= H_39; 
			H_reg(40) <= H_40; H_reg(41) <= H_41; H_reg(42) <= H_42; H_reg(43) <= H_43; 
			H_reg(44) <= H_44; H_reg(45) <= H_45; H_reg(46) <= H_46; H_reg(47) <= H_47; 
			H_reg(48) <= H_48; H_reg(49) <= H_49; H_reg(50) <= H_50; H_reg(51) <= H_51; 
			H_reg(52) <= H_52; H_reg(53) <= H_53; H_reg(54) <= H_54; H_reg(55) <= H_55; 
			H_reg(56) <= H_56; H_reg(57) <= H_57; H_reg(58) <= H_58; H_reg(59) <= H_59; 
			H_reg(60) <= H_60; H_reg(61) <= H_61; H_reg(62) <= H_62; H_reg(63) <= H_63; 

			DC_reg <= DC;

			DDL_reg(0)  <= DDL_0;  DDL_reg(1)  <= DDL_1;  DDL_reg(2)  <= DDL_2;  DDL_reg(3) <= DDL_3; 
			DDL_reg(4)  <= DDL_4;  DDL_reg(5)  <= DDL_5;  DDL_reg(6)  <= DDL_6;  DDL_reg(7) <= DDL_7; 
			DDL_reg(8)  <= DDL_8;  DDL_reg(9)  <= DDL_9;  DDL_reg(10) <= DDL_10; DDL_reg(11) <= DDL_11; 
			DDL_reg(12) <= DDL_12; DDL_reg(13) <= DDL_13; DDL_reg(14) <= DDL_14; DDL_reg(15) <= DDL_15; 
			DDL_reg(16) <= DDL_16; DDL_reg(17) <= DDL_17; DDL_reg(18) <= DDL_18; DDL_reg(19) <= DDL_19; 
			DDL_reg(20) <= DDL_20; DDL_reg(21) <= DDL_21; DDL_reg(22) <= DDL_22; DDL_reg(23) <= DDL_23; 
			DDL_reg(24) <= DDL_24; DDL_reg(25) <= DDL_25; DDL_reg(26) <= DDL_26; DDL_reg(27) <= DDL_27; 
			DDL_reg(28) <= DDL_28; DDL_reg(29) <= DDL_29; DDL_reg(30) <= DDL_30; DDL_reg(31) <= DDL_31; 
			DDL_reg(32) <= DDL_32; DDL_reg(33) <= DDL_33; DDL_reg(34) <= DDL_34; DDL_reg(35) <= DDL_35; 
			DDL_reg(36) <= DDL_36; DDL_reg(37) <= DDL_37; DDL_reg(38) <= DDL_38; DDL_reg(39) <= DDL_39; 
			DDL_reg(40) <= DDL_40; DDL_reg(41) <= DDL_41; DDL_reg(42) <= DDL_42; DDL_reg(43) <= DDL_43; 
			DDL_reg(44) <= DDL_44; DDL_reg(45) <= DDL_45; DDL_reg(46) <= DDL_46; DDL_reg(47) <= DDL_47; 
			DDL_reg(48) <= DDL_48; DDL_reg(49) <= DDL_49; DDL_reg(50) <= DDL_50; DDL_reg(51) <= DDL_51; 
			DDL_reg(52) <= DDL_52; DDL_reg(53) <= DDL_53; DDL_reg(54) <= DDL_54; DDL_reg(55) <= DDL_55; 
			DDL_reg(56) <= DDL_56; DDL_reg(57) <= DDL_57; DDL_reg(58) <= DDL_58; DDL_reg(59) <= DDL_59; 
			DDL_reg(60) <= DDL_60; DDL_reg(61) <= DDL_61; DDL_reg(62) <= DDL_62; DDL_reg(63) <= DDL_63; 

			DDR_reg(0)  <= DDR_0;  DDR_reg(1)  <= DDR_1;  DDR_reg(2)  <= DDR_2;  DDR_reg(3) <= DDR_3; 
			DDR_reg(4)  <= DDR_4;  DDR_reg(5)  <= DDR_5;  DDR_reg(6)  <= DDR_6;  DDR_reg(7) <= DDR_7; 
			DDR_reg(8)  <= DDR_8;  DDR_reg(9)  <= DDR_9;  DDR_reg(10) <= DDR_10; DDR_reg(11) <= DDR_11; 
			DDR_reg(12) <= DDR_12; DDR_reg(13) <= DDR_13; DDR_reg(14) <= DDR_14; DDR_reg(15) <= DDR_15; 
			DDR_reg(16) <= DDR_16; DDR_reg(17) <= DDR_17; DDR_reg(18) <= DDR_18; DDR_reg(19) <= DDR_19; 
			DDR_reg(20) <= DDR_20; DDR_reg(21) <= DDR_21; DDR_reg(22) <= DDR_22; DDR_reg(23) <= DDR_23; 
			DDR_reg(24) <= DDR_24; DDR_reg(25) <= DDR_25; DDR_reg(26) <= DDR_26; DDR_reg(27) <= DDR_27; 
			DDR_reg(28) <= DDR_28; DDR_reg(29) <= DDR_29; DDR_reg(30) <= DDR_30; DDR_reg(31) <= DDR_31; 
			DDR_reg(32) <= DDR_32; DDR_reg(33) <= DDR_33; DDR_reg(34) <= DDR_34; DDR_reg(35) <= DDR_35; 
			DDR_reg(36) <= DDR_36; DDR_reg(37) <= DDR_37; DDR_reg(38) <= DDR_38; DDR_reg(39) <= DDR_39; 
			DDR_reg(40) <= DDR_40; DDR_reg(41) <= DDR_41; DDR_reg(42) <= DDR_42; DDR_reg(43) <= DDR_43; 
			DDR_reg(44) <= DDR_44; DDR_reg(45) <= DDR_45; DDR_reg(46) <= DDR_46; DDR_reg(47) <= DDR_47; 
			DDR_reg(48) <= DDR_48; DDR_reg(49) <= DDR_49; DDR_reg(50) <= DDR_50; DDR_reg(51) <= DDR_51; 
			DDR_reg(52) <= DDR_52; DDR_reg(53) <= DDR_53; DDR_reg(54) <= DDR_54; DDR_reg(55) <= DDR_55; 
			DDR_reg(56) <= DDR_56; DDR_reg(57) <= DDR_57; DDR_reg(58) <= DDR_58; DDR_reg(59) <= DDR_59; 
			DDR_reg(60) <= DDR_60; DDR_reg(61) <= DDR_61; DDR_reg(62) <= DDR_62; DDR_reg(63) <= DDR_63; 

			VL_reg(0)  <= VL_0;  VL_reg(1)  <= VL_1;  VL_reg(2)  <= VL_2;  VL_reg(3) <= VL_3; 
			VL_reg(4)  <= VL_4;  VL_reg(5)  <= VL_5;  VL_reg(6)  <= VL_6;  VL_reg(7) <= VL_7; 
			VL_reg(8)  <= VL_8;  VL_reg(9)  <= VL_9;  VL_reg(10) <= VL_10; VL_reg(11) <= VL_11; 
			VL_reg(12) <= VL_12; VL_reg(13) <= VL_13; VL_reg(14) <= VL_14; VL_reg(15) <= VL_15; 
			VL_reg(16) <= VL_16; VL_reg(17) <= VL_17; VL_reg(18) <= VL_18; VL_reg(19) <= VL_19; 
			VL_reg(20) <= VL_20; VL_reg(21) <= VL_21; VL_reg(22) <= VL_22; VL_reg(23) <= VL_23; 
			VL_reg(24) <= VL_24; VL_reg(25) <= VL_25; VL_reg(26) <= VL_26; VL_reg(27) <= VL_27; 
			VL_reg(28) <= VL_28; VL_reg(29) <= VL_29; VL_reg(30) <= VL_30; VL_reg(31) <= VL_31; 
			VL_reg(32) <= VL_32; VL_reg(33) <= VL_33; VL_reg(34) <= VL_34; VL_reg(35) <= VL_35; 
			VL_reg(36) <= VL_36; VL_reg(37) <= VL_37; VL_reg(38) <= VL_38; VL_reg(39) <= VL_39; 
			VL_reg(40) <= VL_40; VL_reg(41) <= VL_41; VL_reg(42) <= VL_42; VL_reg(43) <= VL_43; 
			VL_reg(44) <= VL_44; VL_reg(45) <= VL_45; VL_reg(46) <= VL_46; VL_reg(47) <= VL_47; 
			VL_reg(48) <= VL_48; VL_reg(49) <= VL_49; VL_reg(50) <= VL_50; VL_reg(51) <= VL_51; 
			VL_reg(52) <= VL_52; VL_reg(53) <= VL_53; VL_reg(54) <= VL_54; VL_reg(55) <= VL_55; 
			VL_reg(56) <= VL_56; VL_reg(57) <= VL_57; VL_reg(58) <= VL_58; VL_reg(59) <= VL_59; 
			VL_reg(60) <= VL_60; VL_reg(61) <= VL_61; VL_reg(62) <= VL_62; VL_reg(63) <= VL_63; 

			HD_reg(0)  <= HD_0;  HD_reg(1)  <= HD_1;  HD_reg(2)  <= HD_2;  HD_reg(3) <= HD_3; 
			HD_reg(4)  <= HD_4;  HD_reg(5)  <= HD_5;  HD_reg(6)  <= HD_6;  HD_reg(7) <= HD_7; 
			HD_reg(8)  <= HD_8;  HD_reg(9)  <= HD_9;  HD_reg(10) <= HD_10; HD_reg(11) <= HD_11; 
			HD_reg(12) <= HD_12; HD_reg(13) <= HD_13; HD_reg(14) <= HD_14; HD_reg(15) <= HD_15; 
			HD_reg(16) <= HD_16; HD_reg(17) <= HD_17; HD_reg(18) <= HD_18; HD_reg(19) <= HD_19; 
			HD_reg(20) <= HD_20; HD_reg(21) <= HD_21; HD_reg(22) <= HD_22; HD_reg(23) <= HD_23; 
			HD_reg(24) <= HD_24; HD_reg(25) <= HD_25; HD_reg(26) <= HD_26; HD_reg(27) <= HD_27; 
			HD_reg(28) <= HD_28; HD_reg(29) <= HD_29; HD_reg(30) <= HD_30; HD_reg(31) <= HD_31; 
			HD_reg(32) <= HD_32; HD_reg(33) <= HD_33; HD_reg(34) <= HD_34; HD_reg(35) <= HD_35; 
			HD_reg(36) <= HD_36; HD_reg(37) <= HD_37; HD_reg(38) <= HD_38; HD_reg(39) <= HD_39; 
			HD_reg(40) <= HD_40; HD_reg(41) <= HD_41; HD_reg(42) <= HD_42; HD_reg(43) <= HD_43; 
			HD_reg(44) <= HD_44; HD_reg(45) <= HD_45; HD_reg(46) <= HD_46; HD_reg(47) <= HD_47; 
			HD_reg(48) <= HD_48; HD_reg(49) <= HD_49; HD_reg(50) <= HD_50; HD_reg(51) <= HD_51; 
			HD_reg(52) <= HD_52; HD_reg(53) <= HD_53; HD_reg(54) <= HD_54; HD_reg(55) <= HD_55; 
			HD_reg(56) <= HD_56; HD_reg(57) <= HD_57; HD_reg(58) <= HD_58; HD_reg(59) <= HD_59; 
			HD_reg(60) <= HD_60; HD_reg(61) <= HD_61; HD_reg(62) <= HD_62; HD_reg(63) <= HD_63; 

			VR_reg(0)  <= VR_0;  VR_reg(1)  <= VR_1;  VR_reg(2)  <= VR_2;  VR_reg(3) <= VR_3; 
			VR_reg(4)  <= VR_4;  VR_reg(5)  <= VR_5;  VR_reg(6)  <= VR_6;  VR_reg(7) <= VR_7; 
			VR_reg(8)  <= VR_8;  VR_reg(9)  <= VR_9;  VR_reg(10) <= VR_10; VR_reg(11) <= VR_11; 
			VR_reg(12) <= VR_12; VR_reg(13) <= VR_13; VR_reg(14) <= VR_14; VR_reg(15) <= VR_15; 
			VR_reg(16) <= VR_16; VR_reg(17) <= VR_17; VR_reg(18) <= VR_18; VR_reg(19) <= VR_19; 
			VR_reg(20) <= VR_20; VR_reg(21) <= VR_21; VR_reg(22) <= VR_22; VR_reg(23) <= VR_23; 
			VR_reg(24) <= VR_24; VR_reg(25) <= VR_25; VR_reg(26) <= VR_26; VR_reg(27) <= VR_27; 
			VR_reg(28) <= VR_28; VR_reg(29) <= VR_29; VR_reg(30) <= VR_30; VR_reg(31) <= VR_31; 
			VR_reg(32) <= VR_32; VR_reg(33) <= VR_33; VR_reg(34) <= VR_34; VR_reg(35) <= VR_35; 
			VR_reg(36) <= VR_36; VR_reg(37) <= VR_37; VR_reg(38) <= VR_38; VR_reg(39) <= VR_39; 
			VR_reg(40) <= VR_40; VR_reg(41) <= VR_41; VR_reg(42) <= VR_42; VR_reg(43) <= VR_43; 
			VR_reg(44) <= VR_44; VR_reg(45) <= VR_45; VR_reg(46) <= VR_46; VR_reg(47) <= VR_47; 
			VR_reg(48) <= VR_48; VR_reg(49) <= VR_49; VR_reg(50) <= VR_50; VR_reg(51) <= VR_51; 
			VR_reg(52) <= VR_52; VR_reg(53) <= VR_53; VR_reg(54) <= VR_54; VR_reg(55) <= VR_55; 
			VR_reg(56) <= VR_56; VR_reg(57) <= VR_57; VR_reg(58) <= VR_58; VR_reg(59) <= VR_59; 
			VR_reg(60) <= VR_60; VR_reg(61) <= VR_61; VR_reg(62) <= VR_62; VR_reg(63) <= VR_63; 

			HU_reg(0)  <= HU_0;  HU_reg(1)  <= HU_1;  HU_reg(2)  <= HU_2;  HU_reg(3) <= HU_3; 
			HU_reg(4)  <= HU_4;  HU_reg(5)  <= HU_5;  HU_reg(6)  <= HU_6;  HU_reg(7) <= HU_7; 
			HU_reg(8)  <= HU_8;  HU_reg(9)  <= HU_9;  HU_reg(10) <= HU_10; HU_reg(11) <= HU_11; 
			HU_reg(12) <= HU_12; HU_reg(13) <= HU_13; HU_reg(14) <= HU_14; HU_reg(15) <= HU_15; 
			HU_reg(16) <= HU_16; HU_reg(17) <= HU_17; HU_reg(18) <= HU_18; HU_reg(19) <= HU_19; 
			HU_reg(20) <= HU_20; HU_reg(21) <= HU_21; HU_reg(22) <= HU_22; HU_reg(23) <= HU_23; 
			HU_reg(24) <= HU_24; HU_reg(25) <= HU_25; HU_reg(26) <= HU_26; HU_reg(27) <= HU_27; 
			HU_reg(28) <= HU_28; HU_reg(29) <= HU_29; HU_reg(30) <= HU_30; HU_reg(31) <= HU_31; 
			HU_reg(32) <= HU_32; HU_reg(33) <= HU_33; HU_reg(34) <= HU_34; HU_reg(35) <= HU_35; 
			HU_reg(36) <= HU_36; HU_reg(37) <= HU_37; HU_reg(38) <= HU_38; HU_reg(39) <= HU_39; 
			HU_reg(40) <= HU_40; HU_reg(41) <= HU_41; HU_reg(42) <= HU_42; HU_reg(43) <= HU_43; 
			HU_reg(44) <= HU_44; HU_reg(45) <= HU_45; HU_reg(46) <= HU_46; HU_reg(47) <= HU_47; 
			HU_reg(48) <= HU_48; HU_reg(49) <= HU_49; HU_reg(50) <= HU_50; HU_reg(51) <= HU_51; 
			HU_reg(52) <= HU_52; HU_reg(53) <= HU_53; HU_reg(54) <= HU_54; HU_reg(55) <= HU_55; 
			HU_reg(56) <= HU_56; HU_reg(57) <= HU_57; HU_reg(58) <= HU_58; HU_reg(59) <= HU_59; 
			HU_reg(60) <= HU_60; HU_reg(61) <= HU_61; HU_reg(62) <= HU_62; HU_reg(63) <= HU_63; 
		end if;
	end if;
end process;

-- RESIDUALs stored in registers
process (clk, rst)
begin

	IF (rst = '1') THEN
		Diff_regV   <= (others=> (others =>'0'));
		Diff_regH   <= (others=> (others =>'0'));
		Diff_regDC  <= (others=> (others =>'0'));
		Diff_regDDL <= (others=> (others =>'0'));
		Diff_regDDR <= (others=> (others =>'0'));
		Diff_regVL  <= (others=> (others =>'0'));
		Diff_regHD  <= (others=> (others =>'0'));
		Diff_regVR  <= (others=> (others =>'0'));
		Diff_regHU  <= (others=> (others =>'0'));		
		
	ELSIF (clk'EVENT AND clk = '1') THEN
		if (rdy_in_3 = '1') then
	Diff_regV(0)  <= Diff0_V;  Diff_regV(1)  <= Diff1_V;	Diff_regV(2)  <= Diff2_V;  Diff_regV(3)  <= Diff3_V;
	Diff_regV(4)  <= Diff4_V;  Diff_regV(5)  <= Diff5_V;	Diff_regV(6)  <= Diff6_V;  Diff_regV(7)  <= Diff7_V;
	Diff_regV(8)  <= Diff8_V;  Diff_regV(9)  <= Diff9_V;	Diff_regV(10) <= Diff10_V; Diff_regV(11) <= Diff11_V;
	Diff_regV(12) <= Diff12_V; Diff_regV(13) <= Diff13_V;	Diff_regV(14) <= Diff14_V; Diff_regV(15) <= Diff15_V;
	Diff_regV(16) <= Diff16_V; Diff_regV(17) <= Diff17_V;	Diff_regV(18) <= Diff18_V; Diff_regV(19) <= Diff19_V;
	Diff_regV(20) <= Diff20_V; Diff_regV(21) <= Diff21_V;	Diff_regV(22) <= Diff22_V; Diff_regV(23) <= Diff23_V;
	Diff_regV(24) <= Diff24_V; Diff_regV(25) <= Diff25_V;	Diff_regV(26) <= Diff26_V; Diff_regV(27) <= Diff27_V;
	Diff_regV(28) <= Diff28_V; Diff_regV(29) <= Diff29_V;	Diff_regV(30) <= Diff30_V; Diff_regV(31) <= Diff31_V;
	Diff_regV(32) <= Diff32_V; Diff_regV(33) <= Diff33_V;	Diff_regV(34) <= Diff34_V; Diff_regV(35) <= Diff35_V;
	Diff_regV(36) <= Diff36_V; Diff_regV(37) <= Diff37_V;	Diff_regV(38) <= Diff38_V; Diff_regV(39) <= Diff39_V;
	Diff_regV(40) <= Diff40_V; Diff_regV(41) <= Diff41_V;	Diff_regV(42) <= Diff42_V; Diff_regV(43) <= Diff43_V;
	Diff_regV(44) <= Diff44_V; Diff_regV(45) <= Diff45_V;	Diff_regV(46) <= Diff46_V; Diff_regV(47) <= Diff47_V;
	Diff_regV(48) <= Diff48_V; Diff_regV(49) <= Diff49_V;	Diff_regV(50) <= Diff50_V; Diff_regV(51) <= Diff51_V;
	Diff_regV(52) <= Diff52_V; Diff_regV(53) <= Diff53_V;	Diff_regV(54) <= Diff54_V; Diff_regV(55) <= Diff55_V;
	Diff_regV(56) <= Diff56_V; Diff_regV(57) <= Diff57_V;	Diff_regV(58) <= Diff58_V; Diff_regV(59) <= Diff59_V;
	Diff_regV(60) <= Diff60_V; Diff_regV(61) <= Diff61_V;	Diff_regV(62) <= Diff62_V; Diff_regV(63) <= Diff63_V;

	Diff_regH(0)  <= Diff0_H;  Diff_regH(1)  <= Diff1_H;	Diff_regH(2)  <= Diff2_H;  Diff_regH(3)  <= Diff3_H;
	Diff_regH(4)  <= Diff4_H;  Diff_regH(5)  <= Diff5_H;	Diff_regH(6)  <= Diff6_H;  Diff_regH(7)  <= Diff7_H;
	Diff_regH(8)  <= Diff8_H;  Diff_regH(9)  <= Diff9_H;	Diff_regH(10) <= Diff10_H; Diff_regH(11) <= Diff11_H;
	Diff_regH(12) <= Diff12_H; Diff_regH(13) <= Diff13_H;	Diff_regH(14) <= Diff14_H; Diff_regH(15) <= Diff15_H;
	Diff_regH(16) <= Diff16_H; Diff_regH(17) <= Diff17_H;	Diff_regH(18) <= Diff18_H; Diff_regH(19) <= Diff19_H;
	Diff_regH(20) <= Diff20_H; Diff_regH(21) <= Diff21_H;	Diff_regH(22) <= Diff22_H; Diff_regH(23) <= Diff23_H;
	Diff_regH(24) <= Diff24_H; Diff_regH(25) <= Diff25_H;	Diff_regH(26) <= Diff26_H; Diff_regH(27) <= Diff27_H;
	Diff_regH(28) <= Diff28_H; Diff_regH(29) <= Diff29_H;	Diff_regH(30) <= Diff30_H; Diff_regH(31) <= Diff31_H;
	Diff_regH(32) <= Diff32_H; Diff_regH(33) <= Diff33_H;	Diff_regH(34) <= Diff34_H; Diff_regH(35) <= Diff35_H;
	Diff_regH(36) <= Diff36_H; Diff_regH(37) <= Diff37_H;	Diff_regH(38) <= Diff38_H; Diff_regH(39) <= Diff39_H;
	Diff_regH(40) <= Diff40_H; Diff_regH(41) <= Diff41_H;	Diff_regH(42) <= Diff42_H; Diff_regH(43) <= Diff43_H;
	Diff_regH(44) <= Diff44_H; Diff_regH(45) <= Diff45_H;	Diff_regH(46) <= Diff46_H; Diff_regH(47) <= Diff47_H;
	Diff_regH(48) <= Diff48_H; Diff_regH(49) <= Diff49_H;	Diff_regH(50) <= Diff50_H; Diff_regH(51) <= Diff51_H;
	Diff_regH(52) <= Diff52_H; Diff_regH(53) <= Diff53_H;	Diff_regH(54) <= Diff54_H; Diff_regH(55) <= Diff55_H;
	Diff_regH(56) <= Diff56_H; Diff_regH(57) <= Diff57_H;	Diff_regH(58) <= Diff58_H; Diff_regH(59) <= Diff59_H;
	Diff_regH(60) <= Diff60_H; Diff_regH(61) <= Diff61_H;	Diff_regH(62) <= Diff62_H; Diff_regH(63) <= Diff63_H;

	Diff_regDC(0)  <= Diff0_DC;  Diff_regDC(1)  <= Diff1_DC;	Diff_regDC(2)  <= Diff2_DC;  Diff_regDC(3)  <= Diff3_DC;
	Diff_regDC(4)  <= Diff4_DC;  Diff_regDC(5)  <= Diff5_DC;	Diff_regDC(6)  <= Diff6_DC;  Diff_regDC(7)  <= Diff7_DC;
	Diff_regDC(8)  <= Diff8_DC;  Diff_regDC(9)  <= Diff9_DC;	Diff_regDC(10) <= Diff10_DC; Diff_regDC(11) <= Diff11_DC;
	Diff_regDC(12) <= Diff12_DC; Diff_regDC(13) <= Diff13_DC;	Diff_regDC(14) <= Diff14_DC; Diff_regDC(15) <= Diff15_DC;
	Diff_regDC(16) <= Diff16_DC; Diff_regDC(17) <= Diff17_DC;	Diff_regDC(18) <= Diff18_DC; Diff_regDC(19) <= Diff19_DC;
	Diff_regDC(20) <= Diff20_DC; Diff_regDC(21) <= Diff21_DC;	Diff_regDC(22) <= Diff22_DC; Diff_regDC(23) <= Diff23_DC;
	Diff_regDC(24) <= Diff24_DC; Diff_regDC(25) <= Diff25_DC;	Diff_regDC(26) <= Diff26_DC; Diff_regDC(27) <= Diff27_DC;
	Diff_regDC(28) <= Diff28_DC; Diff_regDC(29) <= Diff29_DC;	Diff_regDC(30) <= Diff30_DC; Diff_regDC(31) <= Diff31_DC;
	Diff_regDC(32) <= Diff32_DC; Diff_regDC(33) <= Diff33_DC;	Diff_regDC(34) <= Diff34_DC; Diff_regDC(35) <= Diff35_DC;
	Diff_regDC(36) <= Diff36_DC; Diff_regDC(37) <= Diff37_DC;	Diff_regDC(38) <= Diff38_DC; Diff_regDC(39) <= Diff39_DC;
	Diff_regDC(40) <= Diff40_DC; Diff_regDC(41) <= Diff41_DC;	Diff_regDC(42) <= Diff42_DC; Diff_regDC(43) <= Diff43_DC;
	Diff_regDC(44) <= Diff44_DC; Diff_regDC(45) <= Diff45_DC;	Diff_regDC(46) <= Diff46_DC; Diff_regDC(47) <= Diff47_DC;
	Diff_regDC(48) <= Diff48_DC; Diff_regDC(49) <= Diff49_DC;	Diff_regDC(50) <= Diff50_DC; Diff_regDC(51) <= Diff51_DC;
	Diff_regDC(52) <= Diff52_DC; Diff_regDC(53) <= Diff53_DC;	Diff_regDC(54) <= Diff54_DC; Diff_regDC(55) <= Diff55_DC;
	Diff_regDC(56) <= Diff56_DC; Diff_regDC(57) <= Diff57_DC;	Diff_regDC(58) <= Diff58_DC; Diff_regDC(59) <= Diff59_DC;
	Diff_regDC(60) <= Diff60_DC; Diff_regDC(61) <= Diff61_DC;	Diff_regDC(62) <= Diff62_DC; Diff_regDC(63) <= Diff63_DC;

	Diff_regDDL(0)  <= Diff0_DDL;  Diff_regDDL(1)  <= Diff1_DDL;	Diff_regDDL(2)  <= Diff2_DDL;  Diff_regDDL(3)  <= Diff3_DDL;
	Diff_regDDL(4)  <= Diff4_DDL;  Diff_regDDL(5)  <= Diff5_DDL;	Diff_regDDL(6)  <= Diff6_DDL;  Diff_regDDL(7)  <= Diff7_DDL;
	Diff_regDDL(8)  <= Diff8_DDL;  Diff_regDDL(9)  <= Diff9_DDL;	Diff_regDDL(10) <= Diff10_DDL; Diff_regDDL(11) <= Diff11_DDL;
	Diff_regDDL(12) <= Diff12_DDL; Diff_regDDL(13) <= Diff13_DDL;	Diff_regDDL(14) <= Diff14_DDL; Diff_regDDL(15) <= Diff15_DDL;
	Diff_regDDL(16) <= Diff16_DDL; Diff_regDDL(17) <= Diff17_DDL;	Diff_regDDL(18) <= Diff18_DDL; Diff_regDDL(19) <= Diff19_DDL;
	Diff_regDDL(20) <= Diff20_DDL; Diff_regDDL(21) <= Diff21_DDL;	Diff_regDDL(22) <= Diff22_DDL; Diff_regDDL(23) <= Diff23_DDL;
	Diff_regDDL(24) <= Diff24_DDL; Diff_regDDL(25) <= Diff25_DDL;	Diff_regDDL(26) <= Diff26_DDL; Diff_regDDL(27) <= Diff27_DDL;
	Diff_regDDL(28) <= Diff28_DDL; Diff_regDDL(29) <= Diff29_DDL;	Diff_regDDL(30) <= Diff30_DDL; Diff_regDDL(31) <= Diff31_DDL;
	Diff_regDDL(32) <= Diff32_DDL; Diff_regDDL(33) <= Diff33_DDL;	Diff_regDDL(34) <= Diff34_DDL; Diff_regDDL(35) <= Diff35_DDL;
	Diff_regDDL(36) <= Diff36_DDL; Diff_regDDL(37) <= Diff37_DDL;	Diff_regDDL(38) <= Diff38_DDL; Diff_regDDL(39) <= Diff39_DDL;
	Diff_regDDL(40) <= Diff40_DDL; Diff_regDDL(41) <= Diff41_DDL;	Diff_regDDL(42) <= Diff42_DDL; Diff_regDDL(43) <= Diff43_DDL;
	Diff_regDDL(44) <= Diff44_DDL; Diff_regDDL(45) <= Diff45_DDL;	Diff_regDDL(46) <= Diff46_DDL; Diff_regDDL(47) <= Diff47_DDL;
	Diff_regDDL(48) <= Diff48_DDL; Diff_regDDL(49) <= Diff49_DDL;	Diff_regDDL(50) <= Diff50_DDL; Diff_regDDL(51) <= Diff51_DDL;
	Diff_regDDL(52) <= Diff52_DDL; Diff_regDDL(53) <= Diff53_DDL;	Diff_regDDL(54) <= Diff54_DDL; Diff_regDDL(55) <= Diff55_DDL;
	Diff_regDDL(56) <= Diff56_DDL; Diff_regDDL(57) <= Diff57_DDL;	Diff_regDDL(58) <= Diff58_DDL; Diff_regDDL(59) <= Diff59_DDL;
	Diff_regDDL(60) <= Diff60_DDL; Diff_regDDL(61) <= Diff61_DDL;	Diff_regDDL(62) <= Diff62_DDL; Diff_regDDL(63) <= Diff63_DDL;

	Diff_regDDR(0)  <= Diff0_DDR;  Diff_regDDR(1)  <= Diff1_DDR;	Diff_regDDR(2)  <= Diff2_DDR;  Diff_regDDR(3)  <= Diff3_DDR;
	Diff_regDDR(4)  <= Diff4_DDR;  Diff_regDDR(5)  <= Diff5_DDR;	Diff_regDDR(6)  <= Diff6_DDR;  Diff_regDDR(7)  <= Diff7_DDR;
	Diff_regDDR(8)  <= Diff8_DDR;  Diff_regDDR(9)  <= Diff9_DDR;	Diff_regDDR(10) <= Diff10_DDR; Diff_regDDR(11) <= Diff11_DDR;
	Diff_regDDR(12) <= Diff12_DDR; Diff_regDDR(13) <= Diff13_DDR;	Diff_regDDR(14) <= Diff14_DDR; Diff_regDDR(15) <= Diff15_DDR;
	Diff_regDDR(16) <= Diff16_DDR; Diff_regDDR(17) <= Diff17_DDR;	Diff_regDDR(18) <= Diff18_DDR; Diff_regDDR(19) <= Diff19_DDR;
	Diff_regDDR(20) <= Diff20_DDR; Diff_regDDR(21) <= Diff21_DDR;	Diff_regDDR(22) <= Diff22_DDR; Diff_regDDR(23) <= Diff23_DDR;
	Diff_regDDR(24) <= Diff24_DDR; Diff_regDDR(25) <= Diff25_DDR;	Diff_regDDR(26) <= Diff26_DDR; Diff_regDDR(27) <= Diff27_DDR;
	Diff_regDDR(28) <= Diff28_DDR; Diff_regDDR(29) <= Diff29_DDR;	Diff_regDDR(30) <= Diff30_DDR; Diff_regDDR(31) <= Diff31_DDR;
	Diff_regDDR(32) <= Diff32_DDR; Diff_regDDR(33) <= Diff33_DDR;	Diff_regDDR(34) <= Diff34_DDR; Diff_regDDR(35) <= Diff35_DDR;
	Diff_regDDR(36) <= Diff36_DDR; Diff_regDDR(37) <= Diff37_DDR;	Diff_regDDR(38) <= Diff38_DDR; Diff_regDDR(39) <= Diff39_DDR;
	Diff_regDDR(40) <= Diff40_DDR; Diff_regDDR(41) <= Diff41_DDR;	Diff_regDDR(42) <= Diff42_DDR; Diff_regDDR(43) <= Diff43_DDR;
	Diff_regDDR(44) <= Diff44_DDR; Diff_regDDR(45) <= Diff45_DDR;	Diff_regDDR(46) <= Diff46_DDR; Diff_regDDR(47) <= Diff47_DDR;
	Diff_regDDR(48) <= Diff48_DDR; Diff_regDDR(49) <= Diff49_DDR;	Diff_regDDR(50) <= Diff50_DDR; Diff_regDDR(51) <= Diff51_DDR;
	Diff_regDDR(52) <= Diff52_DDR; Diff_regDDR(53) <= Diff53_DDR;	Diff_regDDR(54) <= Diff54_DDR; Diff_regDDR(55) <= Diff55_DDR;
	Diff_regDDR(56) <= Diff56_DDR; Diff_regDDR(57) <= Diff57_DDR;	Diff_regDDR(58) <= Diff58_DDR; Diff_regDDR(59) <= Diff59_DDR;
	Diff_regDDR(60) <= Diff60_DDR; Diff_regDDR(61) <= Diff61_DDR;	Diff_regDDR(62) <= Diff62_DDR; Diff_regDDR(63) <= Diff63_DDR;

	Diff_regVL(0)  <= Diff0_VL;  Diff_regVL(1)  <= Diff1_VL;	Diff_regVL(2)  <= Diff2_VL;  Diff_regVL(3)  <= Diff3_VL;
	Diff_regVL(4)  <= Diff4_VL;  Diff_regVL(5)  <= Diff5_VL;	Diff_regVL(6)  <= Diff6_VL;  Diff_regVL(7)  <= Diff7_VL;
	Diff_regVL(8)  <= Diff8_VL;  Diff_regVL(9)  <= Diff9_VL;	Diff_regVL(10) <= Diff10_VL; Diff_regVL(11) <= Diff11_VL;
	Diff_regVL(12) <= Diff12_VL; Diff_regVL(13) <= Diff13_VL;	Diff_regVL(14) <= Diff14_VL; Diff_regVL(15) <= Diff15_VL;
	Diff_regVL(16) <= Diff16_VL; Diff_regVL(17) <= Diff17_VL;	Diff_regVL(18) <= Diff18_VL; Diff_regVL(19) <= Diff19_VL;
	Diff_regVL(20) <= Diff20_VL; Diff_regVL(21) <= Diff21_VL;	Diff_regVL(22) <= Diff22_VL; Diff_regVL(23) <= Diff23_VL;
	Diff_regVL(24) <= Diff24_VL; Diff_regVL(25) <= Diff25_VL;	Diff_regVL(26) <= Diff26_VL; Diff_regVL(27) <= Diff27_VL;
	Diff_regVL(28) <= Diff28_VL; Diff_regVL(29) <= Diff29_VL;	Diff_regVL(30) <= Diff30_VL; Diff_regVL(31) <= Diff31_VL;
	Diff_regVL(32) <= Diff32_VL; Diff_regVL(33) <= Diff33_VL;	Diff_regVL(34) <= Diff34_VL; Diff_regVL(35) <= Diff35_VL;
	Diff_regVL(36) <= Diff36_VL; Diff_regVL(37) <= Diff37_VL;	Diff_regVL(38) <= Diff38_VL; Diff_regVL(39) <= Diff39_VL;
	Diff_regVL(40) <= Diff40_VL; Diff_regVL(41) <= Diff41_VL;	Diff_regVL(42) <= Diff42_VL; Diff_regVL(43) <= Diff43_VL;
	Diff_regVL(44) <= Diff44_VL; Diff_regVL(45) <= Diff45_VL;	Diff_regVL(46) <= Diff46_VL; Diff_regVL(47) <= Diff47_VL;
	Diff_regVL(48) <= Diff48_VL; Diff_regVL(49) <= Diff49_VL;	Diff_regVL(50) <= Diff50_VL; Diff_regVL(51) <= Diff51_VL;
	Diff_regVL(52) <= Diff52_VL; Diff_regVL(53) <= Diff53_VL;	Diff_regVL(54) <= Diff54_VL; Diff_regVL(55) <= Diff55_VL;
	Diff_regVL(56) <= Diff56_VL; Diff_regVL(57) <= Diff57_VL;	Diff_regVL(58) <= Diff58_VL; Diff_regVL(59) <= Diff59_VL;
	Diff_regVL(60) <= Diff60_VL; Diff_regVL(61) <= Diff61_VL;	Diff_regVL(62) <= Diff62_VL; Diff_regVL(63) <= Diff63_VL;

	Diff_regHD(0)  <= Diff0_HD;  Diff_regHD(1)  <= Diff1_HD;	Diff_regHD(2)  <= Diff2_HD;  Diff_regHD(3)  <= Diff3_HD;
	Diff_regHD(4)  <= Diff4_HD;  Diff_regHD(5)  <= Diff5_HD;	Diff_regHD(6)  <= Diff6_HD;  Diff_regHD(7)  <= Diff7_HD;
	Diff_regHD(8)  <= Diff8_HD;  Diff_regHD(9)  <= Diff9_HD;	Diff_regHD(10) <= Diff10_HD; Diff_regHD(11) <= Diff11_HD;
	Diff_regHD(12) <= Diff12_HD; Diff_regHD(13) <= Diff13_HD;	Diff_regHD(14) <= Diff14_HD; Diff_regHD(15) <= Diff15_HD;
	Diff_regHD(16) <= Diff16_HD; Diff_regHD(17) <= Diff17_HD;	Diff_regHD(18) <= Diff18_HD; Diff_regHD(19) <= Diff19_HD;
	Diff_regHD(20) <= Diff20_HD; Diff_regHD(21) <= Diff21_HD;	Diff_regHD(22) <= Diff22_HD; Diff_regHD(23) <= Diff23_HD;
	Diff_regHD(24) <= Diff24_HD; Diff_regHD(25) <= Diff25_HD;	Diff_regHD(26) <= Diff26_HD; Diff_regHD(27) <= Diff27_HD;
	Diff_regHD(28) <= Diff28_HD; Diff_regHD(29) <= Diff29_HD;	Diff_regHD(30) <= Diff30_HD; Diff_regHD(31) <= Diff31_HD;
	Diff_regHD(32) <= Diff32_HD; Diff_regHD(33) <= Diff33_HD;	Diff_regHD(34) <= Diff34_HD; Diff_regHD(35) <= Diff35_HD;
	Diff_regHD(36) <= Diff36_HD; Diff_regHD(37) <= Diff37_HD;	Diff_regHD(38) <= Diff38_HD; Diff_regHD(39) <= Diff39_HD;
	Diff_regHD(40) <= Diff40_HD; Diff_regHD(41) <= Diff41_HD;	Diff_regHD(42) <= Diff42_HD; Diff_regHD(43) <= Diff43_HD;
	Diff_regHD(44) <= Diff44_HD; Diff_regHD(45) <= Diff45_HD;	Diff_regHD(46) <= Diff46_HD; Diff_regHD(47) <= Diff47_HD;
	Diff_regHD(48) <= Diff48_HD; Diff_regHD(49) <= Diff49_HD;	Diff_regHD(50) <= Diff50_HD; Diff_regHD(51) <= Diff51_HD;
	Diff_regHD(52) <= Diff52_HD; Diff_regHD(53) <= Diff53_HD;	Diff_regHD(54) <= Diff54_HD; Diff_regHD(55) <= Diff55_HD;
	Diff_regHD(56) <= Diff56_HD; Diff_regHD(57) <= Diff57_HD;	Diff_regHD(58) <= Diff58_HD; Diff_regHD(59) <= Diff59_HD;
	Diff_regHD(60) <= Diff60_HD; Diff_regHD(61) <= Diff61_HD;	Diff_regHD(62) <= Diff62_HD; Diff_regHD(63) <= Diff63_HD;

	Diff_regVR(0)  <= Diff0_VR;  Diff_regVR(1)  <= Diff1_VR;	Diff_regVR(2)  <= Diff2_VR;  Diff_regVR(3)  <= Diff3_VR;
	Diff_regVR(4)  <= Diff4_VR;  Diff_regVR(5)  <= Diff5_VR;	Diff_regVR(6)  <= Diff6_VR;  Diff_regVR(7)  <= Diff7_VR;
	Diff_regVR(8)  <= Diff8_VR;  Diff_regVR(9)  <= Diff9_VR;	Diff_regVR(10) <= Diff10_VR; Diff_regVR(11) <= Diff11_VR;
	Diff_regVR(12) <= Diff12_VR; Diff_regVR(13) <= Diff13_VR;	Diff_regVR(14) <= Diff14_VR; Diff_regVR(15) <= Diff15_VR;
	Diff_regVR(16) <= Diff16_VR; Diff_regVR(17) <= Diff17_VR;	Diff_regVR(18) <= Diff18_VR; Diff_regVR(19) <= Diff19_VR;
	Diff_regVR(20) <= Diff20_VR; Diff_regVR(21) <= Diff21_VR;	Diff_regVR(22) <= Diff22_VR; Diff_regVR(23) <= Diff23_VR;
	Diff_regVR(24) <= Diff24_VR; Diff_regVR(25) <= Diff25_VR;	Diff_regVR(26) <= Diff26_VR; Diff_regVR(27) <= Diff27_VR;
	Diff_regVR(28) <= Diff28_VR; Diff_regVR(29) <= Diff29_VR;	Diff_regVR(30) <= Diff30_VR; Diff_regVR(31) <= Diff31_VR;
	Diff_regVR(32) <= Diff32_VR; Diff_regVR(33) <= Diff33_VR;	Diff_regVR(34) <= Diff34_VR; Diff_regVR(35) <= Diff35_VR;
	Diff_regVR(36) <= Diff36_VR; Diff_regVR(37) <= Diff37_VR;	Diff_regVR(38) <= Diff38_VR; Diff_regVR(39) <= Diff39_VR;
	Diff_regVR(40) <= Diff40_VR; Diff_regVR(41) <= Diff41_VR;	Diff_regVR(42) <= Diff42_VR; Diff_regVR(43) <= Diff43_VR;
	Diff_regVR(44) <= Diff44_VR; Diff_regVR(45) <= Diff45_VR;	Diff_regVR(46) <= Diff46_VR; Diff_regVR(47) <= Diff47_VR;
	Diff_regVR(48) <= Diff48_VR; Diff_regVR(49) <= Diff49_VR;	Diff_regVR(50) <= Diff50_VR; Diff_regVR(51) <= Diff51_VR;
	Diff_regVR(52) <= Diff52_VR; Diff_regVR(53) <= Diff53_VR;	Diff_regVR(54) <= Diff54_VR; Diff_regVR(55) <= Diff55_VR;
	Diff_regVR(56) <= Diff56_VR; Diff_regVR(57) <= Diff57_VR;	Diff_regVR(58) <= Diff58_VR; Diff_regVR(59) <= Diff59_VR;
	Diff_regVR(60) <= Diff60_VR; Diff_regVR(61) <= Diff61_VR;	Diff_regVR(62) <= Diff62_VR; Diff_regVR(63) <= Diff63_VR;

	Diff_regHU(0)  <= Diff0_HU;  Diff_regHU(1)  <= Diff1_HU;	Diff_regHU(2)  <= Diff2_HU;  Diff_regHU(3)  <= Diff3_HU;
	Diff_regHU(4)  <= Diff4_HU;  Diff_regHU(5)  <= Diff5_HU;	Diff_regHU(6)  <= Diff6_HU;  Diff_regHU(7)  <= Diff7_HU;
	Diff_regHU(8)  <= Diff8_HU;  Diff_regHU(9)  <= Diff9_HU;	Diff_regHU(10) <= Diff10_HU; Diff_regHU(11) <= Diff11_HU;
	Diff_regHU(12) <= Diff12_HU; Diff_regHU(13) <= Diff13_HU;	Diff_regHU(14) <= Diff14_HU; Diff_regHU(15) <= Diff15_HU;
	Diff_regHU(16) <= Diff16_HU; Diff_regHU(17) <= Diff17_HU;	Diff_regHU(18) <= Diff18_HU; Diff_regHU(19) <= Diff19_HU;
	Diff_regHU(20) <= Diff20_HU; Diff_regHU(21) <= Diff21_HU;	Diff_regHU(22) <= Diff22_HU; Diff_regHU(23) <= Diff23_HU;
	Diff_regHU(24) <= Diff24_HU; Diff_regHU(25) <= Diff25_HU;	Diff_regHU(26) <= Diff26_HU; Diff_regHU(27) <= Diff27_HU;
	Diff_regHU(28) <= Diff28_HU; Diff_regHU(29) <= Diff29_HU;	Diff_regHU(30) <= Diff30_HU; Diff_regHU(31) <= Diff31_HU;
	Diff_regHU(32) <= Diff32_HU; Diff_regHU(33) <= Diff33_HU;	Diff_regHU(34) <= Diff34_HU; Diff_regHU(35) <= Diff35_HU;
	Diff_regHU(36) <= Diff36_HU; Diff_regHU(37) <= Diff37_HU;	Diff_regHU(38) <= Diff38_HU; Diff_regHU(39) <= Diff39_HU;
	Diff_regHU(40) <= Diff40_HU; Diff_regHU(41) <= Diff41_HU;	Diff_regHU(42) <= Diff42_HU; Diff_regHU(43) <= Diff43_HU;
	Diff_regHU(44) <= Diff44_HU; Diff_regHU(45) <= Diff45_HU;	Diff_regHU(46) <= Diff46_HU; Diff_regHU(47) <= Diff47_HU;
	Diff_regHU(48) <= Diff48_HU; Diff_regHU(49) <= Diff49_HU;	Diff_regHU(50) <= Diff50_HU; Diff_regHU(51) <= Diff51_HU;
	Diff_regHU(52) <= Diff52_HU; Diff_regHU(53) <= Diff53_HU;	Diff_regHU(54) <= Diff54_HU; Diff_regHU(55) <= Diff55_HU;
	Diff_regHU(56) <= Diff56_HU; Diff_regHU(57) <= Diff57_HU;	Diff_regHU(58) <= Diff58_HU; Diff_regHU(59) <= Diff59_HU;
	Diff_regHU(60) <= Diff60_HU; Diff_regHU(61) <= Diff61_HU;	Diff_regHU(62) <= Diff62_HU; Diff_regHU(63) <= Diff63_HU;
				
		end if;
	end if;	
end process;

process (clk, rst)
begin
	if (rst = '1') then
		Intra_out_reg <= (others=> (others =>'0'));
		Diff_out_reg  <= (others=> (others =>'0'));
	elsif (clk'event and clk = '1') then

		if (rdy_in_6 = '1') then
			if    (MINIMUM=SAD_V_reg)   then
				Intra_out_reg <= V_reg;
				Diff_out_reg <= Diff_regV;
			elsif (MINIMUM=SAD_H_reg)   then
				Intra_out_reg <= H_reg;
				Diff_out_reg <= Diff_regH;			
			elsif (MINIMUM=SAD_DC_reg)  then
				Intra_out_reg <= (others=> DC_reg);
				Diff_out_reg <= Diff_regDC;			
			elsif (MINIMUM=SAD_DDL_reg) then
				Intra_out_reg <= DDL_reg;
				Diff_out_reg <= Diff_regDDL;			
			elsif (MINIMUM=SAD_DDR_reg) then
				Intra_out_reg <= DDR_reg;
				Diff_out_reg <= Diff_regDDR;			
			elsif (MINIMUM=SAD_VR_reg)  then
				Intra_out_reg <= VR_reg;
				Diff_out_reg <= Diff_regVR;			
			elsif (MINIMUM=SAD_HD_reg)  then
				Intra_out_reg <= HD_reg;
				Diff_out_reg <= Diff_regHD;			
			elsif (MINIMUM=SAD_VL_reg)  then
				Intra_out_reg <= VL_reg;
				Diff_out_reg <= Diff_regVL;			
			elsif (MINIMUM=SAD_HU_reg)  then
				Intra_out_reg <= HU_reg;
				Diff_out_reg <= Diff_regHU;
			else
				Intra_out_reg <= (others=> (others =>'0'));
				Diff_out_reg <=  (others=> (others =>'0'));
			end if;	
		end if;
	end if;
end process;

--Outputs
Diff0  <= Diff_out_reg(0)  when (rdy_in_7='1') else
          Diff_out_reg(16) when (rdy_in_8='1') else
          Diff_out_reg(32) when (rdy_in_9='1') else
          Diff_out_reg(48) when (rdy_in_10='1') else (others=>'0');
		  
Diff1  <= Diff_out_reg(1)  when (rdy_in_7='1') else
          Diff_out_reg(17) when (rdy_in_8='1') else
          Diff_out_reg(33) when (rdy_in_9='1') else
          Diff_out_reg(49) when (rdy_in_10='1') else (others=>'0');
		  
Diff2  <= Diff_out_reg(2)  when (rdy_in_7='1') else
          Diff_out_reg(18) when (rdy_in_8='1') else
          Diff_out_reg(34) when (rdy_in_9='1') else
          Diff_out_reg(50) when (rdy_in_10='1') else (others=>'0');
		  
Diff3  <= Diff_out_reg(3)  when (rdy_in_7='1') else
          Diff_out_reg(19) when (rdy_in_8='1') else
          Diff_out_reg(35) when (rdy_in_9='1') else
          Diff_out_reg(51) when (rdy_in_10='1') else (others=>'0');
		  
Diff4  <= Diff_out_reg(4)  when (rdy_in_7='1') else
          Diff_out_reg(20) when (rdy_in_8='1') else
          Diff_out_reg(36) when (rdy_in_9='1') else
          Diff_out_reg(52) when (rdy_in_10='1') else (others=>'0');
		  
Diff5  <= Diff_out_reg(5)  when (rdy_in_7='1') else
          Diff_out_reg(21) when (rdy_in_8='1') else
          Diff_out_reg(37) when (rdy_in_9='1') else
          Diff_out_reg(53) when (rdy_in_10='1') else (others=>'0');
		  
Diff6  <= Diff_out_reg(6)  when (rdy_in_7='1') else
          Diff_out_reg(22) when (rdy_in_8='1') else
          Diff_out_reg(38) when (rdy_in_9='1') else
          Diff_out_reg(54) when (rdy_in_10='1') else (others=>'0');
		  
Diff7  <= Diff_out_reg(7)  when (rdy_in_7='1') else
          Diff_out_reg(23) when (rdy_in_8='1') else
          Diff_out_reg(39) when (rdy_in_9='1') else
          Diff_out_reg(55) when (rdy_in_10='1') else (others=>'0');
		  
Diff8  <= Diff_out_reg(8)  when (rdy_in_7='1') else
          Diff_out_reg(24) when (rdy_in_8='1') else
          Diff_out_reg(40) when (rdy_in_9='1') else
          Diff_out_reg(56) when (rdy_in_10='1') else (others=>'0');
		  
Diff9  <= Diff_out_reg(9)  when (rdy_in_7='1') else
          Diff_out_reg(25) when (rdy_in_8='1') else
          Diff_out_reg(41) when (rdy_in_9='1') else
          Diff_out_reg(57) when (rdy_in_10='1') else (others=>'0');
		  
Diff10 <= Diff_out_reg(10) when (rdy_in_7='1') else
          Diff_out_reg(26) when (rdy_in_8='1') else
          Diff_out_reg(42) when (rdy_in_9='1') else
          Diff_out_reg(58) when (rdy_in_10='1') else (others=>'0');
		  
Diff11 <= Diff_out_reg(11) when (rdy_in_7='1') else
          Diff_out_reg(27) when (rdy_in_8='1') else
          Diff_out_reg(43) when (rdy_in_9='1') else
          Diff_out_reg(59) when (rdy_in_10='1') else (others=>'0');
		  
Diff12 <= Diff_out_reg(12) when (rdy_in_7='1') else
          Diff_out_reg(28) when (rdy_in_8='1') else
          Diff_out_reg(44) when (rdy_in_9='1') else
          Diff_out_reg(60) when (rdy_in_10='1') else (others=>'0');
		  
Diff13 <= Diff_out_reg(13) when (rdy_in_7='1') else
          Diff_out_reg(29) when (rdy_in_8='1') else
          Diff_out_reg(45) when (rdy_in_9='1') else
          Diff_out_reg(61) when (rdy_in_10='1') else (others=>'0');
		  
Diff14 <= Diff_out_reg(14) when (rdy_in_7='1') else
          Diff_out_reg(30) when (rdy_in_8='1') else
          Diff_out_reg(46) when (rdy_in_9='1') else
          Diff_out_reg(62) when (rdy_in_10='1') else (others=>'0');
		  
Diff15 <= Diff_out_reg(15) when (rdy_in_7='1') else
          Diff_out_reg(31) when (rdy_in_8='1') else
          Diff_out_reg(47) when (rdy_in_9='1') else
          Diff_out_reg(63) when (rdy_in_10='1') else (others=>'0');		  		  
			
--Intra0  <= Intra_out_reg(0) when (rdy_in_7='1') else
--          Intra_out_reg(16) when (rdy_in_8='1') else
--          Intra_out_reg(32) when (rdy_in_9='1') else
--          Intra_out_reg(48) when (rdy_in_10='1') else (others=>'0');
--		  
--Intra1  <= Intra_out_reg(1) when (rdy_in_7='1') else
--          Intra_out_reg(17) when (rdy_in_8='1') else
--         Intra_out_reg(33) when (rdy_in_9='1') else
--         Intra_out_reg(49) when (rdy_in_10='1') else (others=>'0');
--		  
--Intra2  <= Intra_out_reg(2) when (rdy_in_7='1') else
--          Intra_out_reg(18) when (rdy_in_8='1') else
--          Intra_out_reg(34) when (rdy_in_9='1') else
--          Intra_out_reg(50) when (rdy_in_10='1') else (others=>'0');
--		  
--Intra3  <= Intra_out_reg(3) when (rdy_in_7='1') else
--          Intra_out_reg(19) when (rdy_in_8='1') else
--          Intra_out_reg(35) when (rdy_in_9='1') else
--          Intra_out_reg(51) when (rdy_in_10='1') else (others=>'0');
--
--Intra4  <= Intra_out_reg(4) when (rdy_in_7='1') else
--          Intra_out_reg(20) when (rdy_in_8='1') else
--          Intra_out_reg(36) when (rdy_in_9='1') else
--          Intra_out_reg(52) when (rdy_in_10='1') else (others=>'0');
--		  
--Intra5  <= Intra_out_reg(5) when (rdy_in_7='1') else
--          Intra_out_reg(21) when (rdy_in_8='1') else
--          Intra_out_reg(37) when (rdy_in_9='1') else
--          Intra_out_reg(53) when (rdy_in_10='1') else (others=>'0');
--		  
--Intra6  <= Intra_out_reg(6) when (rdy_in_7='1') else
--          Intra_out_reg(22) when (rdy_in_8='1') else
--          Intra_out_reg(38) when (rdy_in_9='1') else
--          Intra_out_reg(54) when (rdy_in_10='1') else (others=>'0');
--		  
Intra7  <= Intra_out_reg(7) when (rdy_in_7='1') else
          Intra_out_reg(23) when (rdy_in_8='1') else
          Intra_out_reg(39) when (rdy_in_9='1') else
          Intra_out_reg(55) when (rdy_in_10='1') else (others=>'0');
		  
Intra8  <= Intra_out_reg(8) when (rdy_in_7='1') else
          Intra_out_reg(24) when (rdy_in_8='1') else
          Intra_out_reg(40) when (rdy_in_9='1') else
          Intra_out_reg(56) when (rdy_in_10='1') else (others=>'0');
		  
Intra9  <= Intra_out_reg(9) when (rdy_in_7='1') else
          Intra_out_reg(25) when (rdy_in_8='1') else
          Intra_out_reg(41) when (rdy_in_9='1') else
          Intra_out_reg(57) when (rdy_in_10='1') else (others=>'0');
		  
Intra10 <= Intra_out_reg(10) when (rdy_in_7='1') else
          Intra_out_reg(26) when (rdy_in_8='1') else
          Intra_out_reg(42) when (rdy_in_9='1') else
          Intra_out_reg(58) when (rdy_in_10='1') else (others=>'0');
		  
Intra11 <= Intra_out_reg(11) when (rdy_in_7='1') else
          Intra_out_reg(27) when (rdy_in_8='1') else
          Intra_out_reg(43) when (rdy_in_9='1') else
          Intra_out_reg(59) when (rdy_in_10='1') else (others=>'0');
		  
Intra12 <= Intra_out_reg(12) when (rdy_in_7='1') else
          Intra_out_reg(28) when (rdy_in_8='1') else
          Intra_out_reg(44) when (rdy_in_9='1') else
          Intra_out_reg(60) when (rdy_in_10='1') else (others=>'0');
		  
Intra13 <= Intra_out_reg(13) when (rdy_in_7='1') else
          Intra_out_reg(29) when (rdy_in_8='1') else
          Intra_out_reg(45) when (rdy_in_9='1') else
          Intra_out_reg(61) when (rdy_in_10='1') else (others=>'0');
		  
Intra14 <= Intra_out_reg(14) when (rdy_in_7='1') else
          Intra_out_reg(30) when (rdy_in_8='1') else
          Intra_out_reg(46) when (rdy_in_9='1') else
          Intra_out_reg(62) when (rdy_in_10='1') else (others=>'0');
		  
Intra15 <= Intra_out_reg(15) when (rdy_in_7='1') else
          Intra_out_reg(31) when (rdy_in_8='1') else
          Intra_out_reg(47) when (rdy_in_9='1') else
          Intra_out_reg(63) when (rdy_in_10='1') else (others=>'0');		  		  
						
rdy_out <= '1' when (rdy_in_7 = '1') or (rdy_in_8 = '1') or (rdy_in_9 = '1') or (rdy_in_10 = '1') else '0';


end Behavioral;

